import { ChangeDetectorRef, Component, OnInit, EventEmitter } from '@angular/core';
import { TranslateService } from '@ngx-translate/core';
import { VitisCheck, VitisRequest, VitisClone } from 'src/app/vitis/core/services/utils/vitis-utils';
import { FunctionService } from 'src/app/vitis/core/services/utils/function/function.service';
import { environment } from 'src/environments/environment';
import { DatasourceTreeviewDatasourceTypes } from './models/DatasourceTreeviewDatasourceTypes';
import { AdminDatasourceService } from './services/admin-datasource.service';
import { DatasourceTree } from './models/DatasourceTree.model';
import { ActivatedRoute, NavigationEnd, Router } from '@angular/router';
import { Injector } from '@angular/core';

declare var $: any;

@Component({
  selector: 'app-admin-datasource',
  templateUrl: './admin-datasource.component.html',
  styleUrls: ['./admin-datasource.component.scss']
})
export class AdminDatasourceComponent implements OnInit {

  public aTree: DatasourceTree;
  public leftPosition: string;
  public topPosition: string;
  public showRightComponent: boolean = true;
  public routeSubscription: any;

  constructor(
    private translate: TranslateService,
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private functionService: FunctionService,
    public adminDatasourceService: AdminDatasourceService,
    private route: ActivatedRoute,
    private router: Router,
    private cdRef: ChangeDetectorRef,
    private vitisClone: VitisClone,
    private injector: Injector,
  ) { }

  ngOnInit() {
    // Traductions (obligatoire sinon ne fonctionne pas dans le module en lazy loading)
    this.translate.addLangs(['en', 'fr']);
    this.translate.setDefaultLang('en');
    this.translate.use(environment.properties.language ? environment.properties.language : 'en');

    // Si la valeur de l'observable est actuellement nulle charge la liste de 0, autrement l'observable utilisara la valeurs en cours
    if (this.adminDatasourceService.getTreeObs().getValue() === null) {
      const datasourceTreeviewDatasourceType = new DatasourceTreeviewDatasourceTypes(this.vitisCheck, this.vitisRequest, this.injector);
      datasourceTreeviewDatasourceType.loadChildrens().then((datasourceTypes) => {
        const aTree = this.adminDatasourceService.setChildrensMetadata(datasourceTypes);
        this.adminDatasourceService.setTree(aTree);
      });
    }

    // Abonnement à l'objservable tree
    this.adminDatasourceService.getTreeObs().subscribe((aTree) => {
      this.aTree = aTree;
    });

    // Detecte les changements sur un objet pour lancer detectChanges
    this.adminDatasourceService.getTreeObjectUpdatedObs().subscribe(() => {
      // Détecte les changements sur l'interface, pour éviter l'erreur ExpressionChangedAfterItHasBeenCheckedError
      this.cdRef.detectChanges();
    });


    this.router.navigate(['.'], { relativeTo: this.route, queryParamsHandling: 'merge' });

    // Récupère les paramètres de la route (pour en cas de refresh, pour ne rafraichir que si nécessaire)
    this.routeSubscription = this.router.events.subscribe(e => {
      if (e instanceof NavigationEnd) {
        if (this.vitisCheck.isDefAndNotNull(this.route.firstChild)) {
          if (this.vitisCheck.isDefAndNotNull(this.route.firstChild.firstChild)) {
            this.adminDatasourceService.childrenParams = this.vitisClone.clone(this.route.firstChild.firstChild.snapshot.params);
          } else {
            this.adminDatasourceService.childrenParams = this.vitisClone.clone(this.route.firstChild.snapshot.params);
          }
        } else {
          this.adminDatasourceService.childrenParams = null;
        }
      }
    });

  }

  /**
   * Récupère les objets enfants, et si besoin les charge
   * @param {any} object
   * @param {boolean} loadIfCantGet Si vrai essaye de charger les données si le getchildrens ne donne pas de résultat.
   */
  public getChildrens(object: any, loadIfCantGet: boolean = true) {
    let aChildrens = null;
    if (this.isCollapsableObject(object)) {
      if (this.vitisCheck.isFunction(object.getChildrens) && this.vitisCheck.isFunction(object.loadChildrens)) {
        aChildrens = object.getChildrens();
        if (!this.vitisCheck.isDefAndNotNull(aChildrens) || aChildrens.length == 0 && loadIfCantGet) {
          object.loadChildrens().then((childrens: Array<any>) => {
            aChildrens = this.adminDatasourceService.setChildrensMetadata(childrens, object);
            this.adminDatasourceService.setTree(this.aTree);
          });
        }
      }
    }
    return aChildrens;
  }

  /**
   * Ouvre le menu du clique droit si il y en a un
   * @param {any} object
   * @param {string} sGlobalIndexI sGlobalIndex + i
   */
  public openRightClickMenu(event: MouseEvent, object: any, sGlobalIndexI: string) {

    if (object.aRightClickMenu && object.aRightClickMenu.length > 0) {
      // preventDefault évite d'afficher le menu par défaut du clique gauche du navigateur
      event.preventDefault();

      // attend un peu pour éviter le déclenchement de la fermeture juste après l'ouverture
      setTimeout(() => {
        this.leftPosition = event.clientX + 'px';
        this.topPosition = event.clientY + 'px';

        object.isOpenSubMenu = true;
      }, 200);
    }
  }

  /**
   * Si c'est possible, lance une fonction
   * @param oObject
   * @param event Nom de la fonction a lancer
   */
  public clickEvent(oObject: any, eventName: string, sCondition?: string) {
    if (!this.isValidCondition(oObject, sCondition)) {
      return;
    }
    this.functionService.callFunction(eventName, this, oObject);
    this.closeRightClickMenu(oObject);
  }

  /**
   * Ferme le menu du clique droit si il y en a un
   * @param {any} object
   */
  public closeRightClickMenu(object: any) {
    if (object.aRightClickMenu && object.aRightClickMenu.length > 0) {
      if (object.isOpenSubMenu && object.isOpenSubMenu === true) {
        object.isOpenSubMenu = false;
      }
    }
  }

  /**
   * Vérifie si l'objet peut être ouvert (si il a potentiellement des enfants)
   * @param object
   */
  public isCollapsableObject(object: any) {
    return this.vitisCheck.isDefAndNotNull(object.collapsed);
  }

  /**
   * Si l'objet peut être ouvert alors on change le status (ouvert/fermé) et on charge les enfants
   * @param object
   */
  public updateCollapse(object: any) {
    if (this.isCollapsableObject(object)) {
      object.collapsed = !object.collapsed;
      this.getChildrens(object);
    }
  }

  /**
   * Ouvre le formulaire sur l'espace de droite
   * @param object
   * @param sFormDefinitionName
   */
  public openRoute(object: any) {
    if (object.route) {
      this.router.navigate(Object.keys(object).map(key => object[key]), { relativeTo: this.route, queryParamsHandling: 'merge' });
    } else {
      this.router.navigate(['.'], { relativeTo: this.route, queryParamsHandling: 'merge' });
    }
  }

  /**
   * Formate l'objet pour lancer la fonction openForm
   * @param object
   */
  public openFormDisplay(object: any) {
    this.openRoute({
      route: object.route,
      formName: 'display',
      uId: object.uId
    });
  }

  /**
   * Si il y a une condition vérifie qu'elle soit valide
   * @param oObject
   * @param sCondition
   */
  public isValidCondition(oObject: any, sCondition?: string) {
    if (this.vitisCheck.isDefAndNotNull(sCondition) && !this.adminDatasourceService.isFulfilledCondition(sCondition, oObject)) {
      return false;
    }
    return true;
  }
}
