import { Injectable } from '@angular/core';
import { ActivatedRoute, Router } from '@angular/router';
import { TranslateService } from '@ngx-translate/core';
import { vitisAlert, VitisCheck, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { FormreaderService } from 'src/app/vitis/formreader/services/formreader.service';
import { environment } from 'src/environments/environment';

@Injectable({
  providedIn: 'root'
})
export class AdminDatasourceFormService {

  oOptions: any;
  oFormReader: any;
  oParams: any;
  adminDatasourceService: any;
  route: ActivatedRoute;

  constructor(
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private translate: TranslateService,
    private router: Router,
    private formreaderSrvc: FormreaderService,
  ) {
  }

  /**
   * Initialise les params qui doivent provenir du composant
   * @param route
   * @param adminDatasourceService
   */
  public init(route: ActivatedRoute, adminDatasourceService: any) {
    this.route = route;
    this.adminDatasourceService = adminDatasourceService;
  }

  /**
   * Ferme le formulaire en cours
   * @param route
   */
  public cancelForm() {
    this.router.navigate(['/dashboard/datasource']);
  }

  /**
   * Valide l'envoi d'un formulaire
   * @param  {any} oValues
   * @return {false|Promise<any>}
   */
  public sendForm(oValues: any): any {
    if (this.oFormReader.isFormValid()) {
      if (this.vitisCheck.isDefAndNotNull(oValues)) {
        if (this.vitisCheck.isDefAndNotNull(this.oOptions.ressource)) {

          // Chemin de la ressource REST
          let sRessourceURL = environment.properties.app_api_url;
          sRessourceURL += '/' + this.oOptions.ressource;

          if (this.oParams.formName === 'update') {
            return this.updateRessource(sRessourceURL, oValues);
          }

          if (this.oParams.formName === 'insert') {
            return this.insertRessource(sRessourceURL, oValues);
          }
        } else {
          console.error('option ressource not defined');
        }
      }
    }
    return false;
  }

  /**
   * Met à jour l'enregistrement
   * @param  {any} oValues
   * @return {false|Promise<any>}
   */
  private updateRessource(sRessourceURL: string, oValues: any): any {

    // Attribut identifiant
    if (this.vitisCheck.isDefAndNotNull(this.oOptions.id_field)) {
      const sId = oValues[this.oOptions.id_field];

      if (this.vitisCheck.isDefAndNotNull(sId)) {
        sRessourceURL += '/' + sId;
      } else {
        console.error('Id field not valid');

        this.translate.get(['VITIS_ERROR', 'VITIS_SIMPLEFORM_FIELD_NOT_VALID']).subscribe((translations) => {
          translations['VITIS_SIMPLEFORM_FIELD_NOT_VALID'] = translations['VITIS_SIMPLEFORM_FIELD_NOT_VALID']
            .replace('{{::id_field}}', this.oOptions.id_field);

          vitisAlert.alert({
            type: 'error',
            title: translations['VITIS_ERROR'],
            text: translations['VITIS_SIMPLEFORM_FIELD_NOT_VALID']
          });
        });
        return false;
      }
    }

    return new Promise<void>((resolve, reject) => {

      // Lancement de la requête PUT
      this.vitisRequest.ajaxRequestPromise({
        method: 'PUT',
        url: sRessourceURL,
        data: this.formreaderSrvc.transformToDataForm(oValues, true)
      }).then((resp) => {
        if (resp.data) {
          if (resp.data.status === 1) {
            this.translate.get('LIST_EDIT_OK').subscribe((translation) => {
              vitisAlert.fireToast({
                icon: 'success',
                title: translation
              });

              const oObject = this.adminDatasourceService.getObjectById(this.oParams.treeid);

              // rafraichit l'objet
              if (this.vitisCheck.isDefAndNotNull(oObject) && this.vitisCheck.isFunction(oObject.refreshData)) {
                oObject.refreshData();
              } else {
                console.error('L\'objet n\'a pas de fonction refreshData', oObject);
              }

              resolve();
            });
          } else {
            reject();
          }
        }
      });
    });
  }

  /**
   * Ajoute l'enregistrement
   * @param  {any} oValues
   * @return {false|Promise<any>}
   */
  private insertRessource(sRessourceURL: string, oValues: any): any {
    return new Promise<void>((resolve, reject) => {

      // Lancement de la requête PUT
      this.vitisRequest.ajaxRequestPromise({
        method: 'POST',
        url: sRessourceURL,
        data: this.formreaderSrvc.transformToDataForm(oValues, true)
      }).then((resp) => {
        if (resp.data) {
          if (resp.data.status === 1) {

            // Message succès
            this.translate.get('LIST_ADD_OK').subscribe((translation) => {
              vitisAlert.fireToast({
                icon: 'success',
                title: translation
              });

              // Dans le cas de l'insert, l'objet oObject est l'objet parent de celui qui vient d'être créé
              // Par exemple si on ajoute un DatasourceTreeviewPgConnection, l'objet oObject est de type DatasourceTreeviewPostgis
              const oObject = this.adminDatasourceService.getObjectById(this.oParams.treeid);
              if (this.vitisCheck.isDefAndNotNull(oObject)) {
                this.refreshChildrens(oObject).then(e => {
                  if (resp.data.data) {
                    const idsRessource = {
                      layer_id: 'layerId',
                      publication_id: 'publicationId',
                      datasource_id: 'datasourceId'
                    };
                    for (const idRessourceName of Object.keys(idsRessource)) {
                      if (resp.data.data[idRessourceName]) {
                        for (let i = 0; i < oObject.aChildrenObjects.length; i++) {
                          if ((resp.data.data[idRessourceName] === oObject.aChildrenObjects[i][idsRessource[idRessourceName]]) ||
                            (resp.data.data[idRessourceName] === oObject.aChildrenObjects[i][idRessourceName])) {
                            // redirige l'utilisateur vers son formulaire d'édition
                            this.router.navigate([`/dashboard/datasource/${oObject.aChildrenObjects[i].route}/update/${oObject.aChildrenObjects[i].uId}`]);
                          }
                        }
                      }
                    }
                  }
                });
              }
              resolve();
            });
          } else {
            reject();
          }
        }
      });
    });
  }

  /**
   * Rafraichit l'abre au niveau de l'objet
   * Function qui retourne un promise pour pouvoir utiliser les modifications apportées à l'arbre
   * @param oObject
   */
  private refreshChildrens(oObject: any): Promise<void> {
    return new Promise((resolve, reject) => {
      oObject.loadChildrens().then((aChildrens: Array<any>) => {
        for (let i = 0; i < aChildrens.length; i++) {
          this.adminDatasourceService.setObjectMetadata(aChildrens[i], oObject);
        }
        this.adminDatasourceService.setTree(this.adminDatasourceService.getTreeObs().getValue());
        resolve();
      });
    });
  }


  /**
   * Prépare un select
   * @param aArray Tableau à parcourrir
   * @param oField Champ select a préparer
   * @param aSubElLabel Label, exemple ['Identifier', ' (', 'SupportedCRS', ')'] => PM (EPSG:3857)
   * @param sSubElValue Value
   */
  public prepareSelectForm(aArray: Array<any>, oField: any, aSubElLabel?: Array<string>, sSubElValue?: string) {
    if (this.vitisCheck.isDefAndNotNull(oField)) {
      if (this.vitisCheck.isDefAndNotNull(aArray) && aArray.length > 0) {
        const sFieldValue = oField.getValue();
        let isValueInOptions = false;
        const oOptions = [];
        for (let i = 0; i < aArray.length; i++) {
          // Label
          let sLabel: string = aArray[i];
          // Construit le label à partir du tableau de chaine de caractère
          if (this.vitisCheck.isDefAndNotNull(aSubElLabel) && this.vitisCheck.isObject(aArray[i])) {
            sLabel = '';
            for (let ii = 0; ii < aSubElLabel.length; ii++) {
              // Par défaut le texte a ajouter est la chaine
              let sLabelValue = aSubElLabel[ii];

              const aLabel = aSubElLabel[ii].split('|');
              // Si la chaine est un identifiant de l'objet on récupère la valeur de l'objet.
              for (let j = 0; j < aLabel.length; j++) {
                const label = aLabel[j];
                if (this.vitisCheck.isDefAndNotNull(aArray[i][label])) {
                  sLabelValue = aArray[i][label];
                  break;
                }
              }

              sLabel += sLabelValue;
            }
          }
          // Value
          const sCurrentValue = this.vitisCheck.isDefAndNotNull(sSubElValue) ? aArray[i][sSubElValue] : aArray[i];
          // Vérifie si la valeur du champ est bien dans les options
          if (sFieldValue === sCurrentValue) {
            isValueInOptions = true;
          }
          // Ajoute une option
          oOptions.push({
            label: sLabel,
            value: sCurrentValue
          });
        }
        // Ajoute les options au champ
        oField.options.choises = oOptions;

        // Si besoin séléctionne la première option
        this.selectFirstOption_(sFieldValue, oField, isValueInOptions);
      } else {
        // Si il n'y a pas de valeur a insert dans les options on vide la valeur du champ.
        oField.options.choises = [];
        oField.setValue('');
      }
    }
  }

  /**
   * Set first option
   * @param oValue
   * @param oField
   * @param isValueInOptions
   */
  private selectFirstOption_(oValue: any, oField: any, isValueInOptions: boolean) {
    // Si il y a au moins 1 option dans la liste déroulante et pas encore de valeur on selectionne le premier element.
    if ((!this.vitisCheck.isDefAndNotNull(oValue) || oValue == '' || !isValueInOptions) &&
      this.vitisCheck.isDefAndNotNull(oField.options.choises[0])) {
      oField.setValue(oField.options.choises[0].value);
    }
  }
}
