import 'src/app/vitis/core/services/utils/string.extensions';
import { Injectable } from '@angular/core';
import { TranslateService } from '@ngx-translate/core';
import { vitisAlert, VitisCheck, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { environment } from 'src/environments/environment';
import WMSCapabilities from 'ol/format/WMSCapabilities';
declare var $: any;

@Injectable({
  providedIn: 'root'
})
export class WmsService {

  constructor(
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private translate: TranslateService,
    ) { }

  /**
   * Make a WMS getCapabilities
   * @param {any} oOptions
   * @param {string} oOptions.flux_url
   * @param {string} oOptions.flux_version
   * @param {string} oOptions.showErrorMessage
   * @param {string} oOptions.flux_login
   * @param {string} oOptions.flux_password
   *
   */
  public getAjaxCapabilitiesWMS(oOptions: any): Promise<boolean | any> {
    return new Promise((resolve, reject) => {
      if (!this.vitisCheck.isDefAndNotNull(oOptions.flux_url)) {
        this.printErrorMsg('DATASOURCE_FORM_WMS_EVENT_ERROR_MISSING_LINK');
        return null;
      }

      if (!this.vitisCheck.isDefAndNotNull(oOptions.flux_version)) {
        oOptions.flux_version = '1.3.0';
      }
      if (!this.vitisCheck.isDefAndNotNull(oOptions.showErrorMessage)) {
        oOptions.showErrorMessage = true;
      }

      const fluxUrl: string = oOptions.flux_url;
      const fluxVersion: string = oOptions.flux_version;

      let sCapabilitiesUrl: string;

      // Enlève certains attributs au cas où l'utilisateur est ait écrits dans l'url
      sCapabilitiesUrl = fluxUrl.removeURLParams(['request', 'service', 'version']);

      // Création de l'url vers le service.
      if (sCapabilitiesUrl.indexOf('?') === -1) {
        sCapabilitiesUrl += '?';
      } else {
        sCapabilitiesUrl += '&';
      }
      sCapabilitiesUrl += 'service=WMS&version=' + fluxVersion + '&request=GetCapabilities';

      const oHeaders: any = {};
      if (this.vitisCheck.isDefAndNotNull(oOptions.flux_login) && this.vitisCheck.isDefAndNotNull(oOptions.flux_password)) {
        if (oOptions.flux_login.length > 0 && oOptions.flux_password.length > 0) {
          oHeaders.Authorization = 'Basic ' + btoa(oOptions.flux_login + ':' + oOptions.flux_password);
        }
      }

      if (this.vitisCheck.isDefAndNotNull(sCapabilitiesUrl)) {
        this.vitisRequest.ajaxRequest({
          method: 'GET',
          url: environment.properties.proxy_url,
          params: {
            url: sCapabilitiesUrl,
            headers: oHeaders
          },
          httpMethodOverride: true,
          success: (response: any) => {
            let oGetCapabilities: any = {};
            let sErrorMessage: string = '';

            if (this.vitisCheck.isDefAndNotNull(response.data) && response.data.status === 1) {
              const sText = response.data.data;

              if (sText == '') {
                this.printErrorMsg('DATASOURCE_WMS_GETCAPABILITES_ERROR_NO_DATA');
              } else {
                // Sauve le résultat du GetCapabilities
                oGetCapabilities = {
                  'xml': sText
                };

                // Vérification du résultat.
                const parser = new DOMParser();
                const xmlDoc = parser.parseFromString(sText, 'text/xml');

                if (xmlDoc.querySelector('parsererror') !== null) {
                  oGetCapabilities = null;
                  sErrorMessage = $(xmlDoc).find('parsererror').html();
                } else if (xmlDoc.querySelector('Service') === null) {

                  oGetCapabilities = null;
                  if (!this.vitisCheck.isDefAndNotNull(sErrorMessage)) {
                    if (this.vitisCheck.isDefAndNotNull(xmlDoc.querySelector('BODY'))) {
                      if (this.vitisCheck.isDefAndNotNull(xmlDoc.querySelector('BODY').textContent)) {
                        sErrorMessage = xmlDoc.querySelector('BODY').textContent;
                      } else {
                        sErrorMessage = $(xmlDoc).find('body').html();
                      }
                    } else {
                      sErrorMessage = $(xmlDoc).find('body').html();
                    }
                  }
                } else {
                  // Parse le xml retourné par le getCapabilities du service.
                  const oParser = new WMSCapabilities();
                  oGetCapabilities.json = oParser.read(sText);
                }
              }

            } else {
              if (this.vitisCheck.isDefAndNotNull(response.data.sMessage)) {
                oGetCapabilities = null;
                sErrorMessage = '<a href="' + fluxUrl + '" target="_blank">' + fluxUrl + '</a><br><br>';
                sErrorMessage += response.data.sMessage;
              }
            }

            // Affiche l'erreur si il y en a une
            if (this.vitisCheck.isDefAndNotNull(sErrorMessage) && sErrorMessage != '') {
              if (oOptions.showErrorMessage) {
                vitisAlert.fireToast({
                  icon: 'error',
                  html: sErrorMessage
                });
              } else {
                console.error(sErrorMessage);
              }
            }

            // Retourne les données du getCapabilities.
            resolve(oGetCapabilities);
          },
          'error': (response: any) => {
            // Certificat ssl invalide.
            if (environment.properties.proxy_check_ssl === true) {
              this.translate.get([
                'DATASOURCE_WMS_GETCAPABILITES_ERROR_SSL',
                'DATASOURCE_WMS_GETCAPABILITES_ERROR'
              ]).subscribe((translations: string) => {
                vitisAlert.fireToast({
                  icon: 'error',
                  title: translations['DATASOURCE_WMS_GETCAPABILITES_ERROR'],
                  text: translations['DATASOURCE_WMS_GETCAPABILITES_ERROR_SSL']
                });
                resolve(false);
              });
              console.error('error: ', response);
            } else {
              resolve(false);
            }
          }
        });
      }
    });

  }

  /**
   * Affiche l'erreur dans un toast
   * @param {string} msg
   * @param {any} oParams
   */
  private printErrorMsg(msg: string, oParams?: any) {
    this.translate.get(msg, oParams).subscribe((translation: string) => {
      vitisAlert.fireToast({
        icon: 'error',
        title: translation
      });
    });
  }

  /**
   * Browse all the <layer> tags to fill aLayers
   * @param {object} layer node
   */
  public searchLayers(oLayer: any) {
    if (!oLayer.Layer) {
      return [oLayer];
    } else {
      let aLayers: Array<any> = [];
      for (let i = 0; i < oLayer.Layer.length; i++) {
        aLayers = aLayers.concat(this.searchLayers(oLayer.Layer[i]));
      }
      return aLayers;
    }
  }
}
