import 'src/app/vitis/core/services/utils/string.extensions';
import { Injectable } from '@angular/core';
import { TranslateService } from '@ngx-translate/core';
import { vitisAlert, VitisCheck, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { environment } from 'src/environments/environment';
import WMTSCapabilities from 'ol/format/WMTSCapabilities';
declare var $: any;

@Injectable({
  providedIn: 'root'
})
export class WmtsService {

  constructor(
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private translate: TranslateService,
  ) { }

  /**
   * Make a WMTS getCapabilities
   * @param {any} opt_options
   * @param {string} opt_options.flux_url
   * @param {string} opt_options.flux_version
   * @param {string} opt_options.flux_type
   */
  public getAjaxCapabilitiesWMTS(opt_options: any): Promise<boolean | any> {
    return new Promise((resolve, reject) => {
      if (!this.vitisCheck.isDefAndNotNull(opt_options.flux_url)) {
        this.printErrorMsg('DATASOURCE_FORM_WMTS_EVENT_ERROR_MISSING_LINK');
        return null;
      }
      if (!this.vitisCheck.isDefAndNotNull(opt_options.flux_version)) {
        this.printErrorMsg('DATASOURCE_FORM_WMTS_EVENT_ERROR_MISSING_VERSION');
        return null;
      }
      if (!this.vitisCheck.isDefAndNotNull(opt_options.flux_type)) {
        this.printErrorMsg('DATASOURCE_FORM_WMTS_EVENT_ERROR_MISSING_TYPE');
        return null;
      }

      let fluxUrl: string = opt_options.flux_url;
      const fluxVersion: string = opt_options.flux_version;
      const WMTSType: string = opt_options.flux_type;

      let sCapabilitiesUrl: string;
      let bError: boolean = true;
      let sError: string = '';

      // Enlève certains attributs au cas où l'utilisateur est ait écrits dans l'url
      fluxUrl = fluxUrl.removeURLParams(['request', 'service', 'version']);

      if (WMTSType === 'KVP') {
        if (fluxUrl.indexOf('?') === -1) {
          sCapabilitiesUrl = fluxUrl + '?SERVICE=WMTS&VERSION=' + fluxVersion + '&REQUEST=GetCapabilities';
        } else {
          sCapabilitiesUrl = fluxUrl + '&SERVICE=WMTS&VERSION=' + fluxVersion + '&REQUEST=GetCapabilities';
        }
      } else if (WMTSType === 'REST') {
        sCapabilitiesUrl = fluxUrl + '/' + fluxVersion + '/WMTSCapabilities.xml';
      }

      if (this.vitisCheck.isDefAndNotNull(sCapabilitiesUrl)) {
        this.vitisRequest.ajaxRequest({
          method: 'GET',
          url: environment.properties.proxy_url,
          params: {
            url: sCapabilitiesUrl
          },
          httpMethodOverride: true,
          success: (response: any) => {
            const text = response.data.data;
            let result: any = {};

            if (!this.vitisCheck.isDefAndNotNull(response.data) || response.data.status !== 1) {
              sError = 'missing response data or status 0';
            } else if (this.vitisCheck.isString(text)) {
              if (text.length > 0) {
                try {
                  const parser = new WMTSCapabilities;
                  result = parser.read(text);
                } catch (e) {
                  sError = text;
                }

                if (this.vitisCheck.isDefAndNotNull(result.Contents)) {
                  if (this.vitisCheck.isArray(result.Contents.Layer)) {
                    if (this.vitisCheck.isDefAndNotNull(result.Contents.Layer[0])) {
                      if (this.vitisCheck.isArray(result.Contents.Layer[0].TileMatrixSetLink)) {
                        if (this.vitisCheck.isDefAndNotNull(result.Contents.Layer[0].TileMatrixSetLink[0])) {
                          if (this.vitisCheck.isString(result.Contents.Layer[0].TileMatrixSetLink[0].TileMatrixSet)) {
                            bError = false;
                          } else {
                            sError = 'Contents.Layer[0].TileMatrixSetLink[0].TileMatrixSet undefined';
                          }
                        } else {
                          sError = 'Contents.Layer[0].TileMatrixSetLink[0] undefined';
                        }
                      } else {
                        sError = 'Contents.Layer[0].TileMatrixSetLink undefined';
                      }
                    } else {
                      sError = 'Contents.Layer[0] undefined';
                    }
                  } else {
                    sError = 'Contents.Layer undefined';
                  }
                } else {
                  sError = 'Contents undefined';
                }
              }
            }

            if (bError) {
              this.translate.get(['CANNOT_GET_URL', 'DATASOURCE_FORM_WMTS_EVENT_ERROR_FLUX'], {
                url: sCapabilitiesUrl.replace(/</g, '&lt;').replace(/>/g, '&gt;')
              }).subscribe((translations: string) => {
                if (text.length < 500) {
                  translations['CANNOT_GET_URL'] += '<br><br>' + text;
                }
                vitisAlert.fireToast({
                  icon: 'error',
                  title: translations['DATASOURCE_FORM_WMTS_EVENT_ERROR_FLUX'],
                  html: translations['CANNOT_GET_URL']
                });
                console.error('proxy error :>> ', sError);
                resolve(false);
              });
            } else {
              resolve(result);
            }
          },
          'error': (response: any) => {
            this.translate.get(['CANNOT_GET_URL', 'DATASOURCE_FORM_WMTS_EVENT_ERROR_FLUX'], {
              url: sCapabilitiesUrl.replace(/</g, '&lt;').replace(/>/g, '&gt;')
            }).subscribe((translations: string) => {
              vitisAlert.fireToast({
                icon: 'error',
                title: translations['DATASOURCE_FORM_WMTS_EVENT_ERROR_FLUX'],
                text: translations['CANNOT_GET_URL']
              });
              resolve(false);
            });
            console.error('error: ', response);
          }
        });
      }
    });

  }

  /**
   * Affiche l'erreur dans un toast
   * @param {string} msg
   * @param {any} oParams
   */
  public printErrorMsg(msg: string, oParams?: any) {
    this.translate.get(msg, oParams).subscribe((translation: string) => {
      vitisAlert.fireToast({
        icon: 'error',
        title: translation
      });
    });
  }

  /**
   * Browse all the <layer> tags to fill aLayers
   * @param {object} layer node
   */
  public searchLayers(oLayer: any) {
    return oLayer;
  }

  /**
   * getTileMatrixSets
   * @param oSelectedLayer
   * @param oCapabilities
   */
  public getTileMatrixSets(oSelectedLayer: any, oCapabilities: any) {
    const aTileMatrixSets = [];
    const aTileMatrixIdentifiers = [];

    // Rempli aTileMatrixIdentifiers
    if (this.vitisCheck.isDefAndNotNull(oSelectedLayer.TileMatrixSetLink)) {
      for (let i = 0; i < oSelectedLayer.TileMatrixSetLink.length; i++) {
        if (this.vitisCheck.isString(oSelectedLayer.TileMatrixSetLink[i].TileMatrixSet)) {
          aTileMatrixIdentifiers.push(oSelectedLayer.TileMatrixSetLink[i].TileMatrixSet);
        }
      }
    }

    // Rempli aTileMatrixSets
    if (this.vitisCheck.isDefAndNotNull(oCapabilities) && this.vitisCheck.isDefAndNotNull(oCapabilities.Contents) &&
      this.vitisCheck.isDefAndNotNull(oCapabilities.Contents.TileMatrixSet) && aTileMatrixIdentifiers.length > 0) {
      for (let i = 0; i < oCapabilities.Contents.TileMatrixSet.length; i++) {
        if (aTileMatrixIdentifiers.indexOf(oCapabilities.Contents.TileMatrixSet[i].Identifier) !== -1) {
          aTileMatrixSets.push(oCapabilities.Contents.TileMatrixSet[i]);
        }
      }
    }

    return aTileMatrixSets;
  }

  /**
   * Retourne l'index d'une matrice
   * @param oSelectedLayer
   * @param oCapabilities
   * @param oProjection
   */
  public getTileMatrixSetsIndex(oSelectedLayer: any, oCapabilities: any, oProjection: any) {
    const aTileMatrixSets = this.getTileMatrixSets(oSelectedLayer, oCapabilities);

    return aTileMatrixSets.findIndex((oTileMatrixSet) => {
      if (this.vitisCheck.isArray(oProjection.srid)) {
        for (let j = 0; j < oProjection.srid.length; j++) {
          if (oTileMatrixSet.SupportedCRS.includes(oProjection.srid[j])) {
            return true;
          }
        }
        return false;
      } else {
        return oTileMatrixSet.SupportedCRS.includes(oProjection.srid);
      }
    });

  }
}
