import { Component, Input, OnDestroy, OnInit } from '@angular/core';
import { Map, Overlay } from 'ol';
import { MapToolDef } from '../../map-tools.model';
import { MapCoreManagerService } from 'src/app/vitis/map-core/services/map-core-manager/map-core-manager.service';
import { unByKey } from 'ol/Observable';
import { transform, transformExtent } from 'node_modules/ol/proj.js';
import { MapCoreUtilsService } from 'src/app/vitis/map-core/services/map-core-utils/map-core-utils.service';
import { Subscription } from 'rxjs';

declare var $: any;

@Component({
  selector: 'app-vmap-tools-compare-map',
  templateUrl: './vmap-tools-compare-map.component.html',
  styleUrls: ['./vmap-tools-compare-map.component.scss']
})
export class VmapToolsCompareMapComponent implements OnInit, OnDestroy {

  @Input() olMap: Map;
  @Input() oTool: MapToolDef;
  // Id de la div OL3
  @Input() targetCompareId: string;
  @Input() tileSize: Array<number>;

  public bToolOpen;
  public active: boolean = false;

  public mapId: number;
  // Carte de comparaison en cours
  public mapCompareId: number;
  private olMapCompare: Map;
  private targetOverlay;

  private keyChangeMap;
  private keyPointerMoveMap;
  private keyPointerMoveMapCompare;
  private keyPointerDragMap;
  private keyPointerDragMapCompare;
  private keyChangeResolutionMap;
  private keyChangeResolutionMapCompare;
  private keyChangeMouseEnterMap;
  private keyChangeMouseLeaveMap;
  private keyChangeMouseEnterMapCompare;
  private keyChangeMouseLeaveMapCompare;

  // Mémorise toutes les souscriptions de manière à les desouscrire à la destruction du composant
  private subscriptions: Subscription[] = [];

  constructor(
    private mapCoreManagerService: MapCoreManagerService,
    private mapUtils: MapCoreUtilsService,
  ) { }

  ngOnInit(): void {
    this.mapId = this.olMap.get('map_id');
    // Ouverture de l'outil
    this.subscriptions.push(this.oTool.onOpen.subscribe(() => {
      this.bToolOpen = true;
    }));

    // Fermeture de l'outil
    this.subscriptions.push(this.oTool.onClose.subscribe(() => {
      this.bToolOpen = false;
    }));

    this.keyChangeMap = this.olMap.on('change', (event: any) => {
      setTimeout(() => {
        if (this.mapId != event.target.get('map_id')) {
          this.mapId = event.target.get('map_id');
          this.active = false;
          this.onChangeCompareMode();
          document.getElementsByClassName('mapCompare')[0].innerHTML = '';
          unByKey(this.keyChangeMap);
        }
      });
    });
  }

  /**
   * Fonction de destruction
   */
  ngOnDestroy(): void {
    this.subscriptions.forEach(subscription => subscription.unsubscribe());
  }

  /**
   * Active / Désactive le mode de comparaison d'une carte
   */
  onChangeCompareMode() {
    const mapContainer = document.getElementsByClassName('mapContainer');
    const mapCompare = document.getElementsByClassName('mapCompare');
    const controls = document.getElementsByClassName('controls-container');
    const controlsTopCenter = controls[0].getElementsByClassName('top-center');
    const controlsBottomRight = controls[0].getElementsByClassName('bottom-right');
    if (this.active === true) { // Active le mode de comparaison
      mapContainer[0]['style']['display'] = 'flex';
      mapCompare[0]['style']['display'] = 'block';
    } else { // Désactive le mode de comparaison
      mapContainer[0]['style']['display'] = 'block';
      mapCompare[0]['style']['display'] = 'none';
      controlsTopCenter[0].classList.remove('map-compare');
      controlsBottomRight[0].classList.remove('map-compare');
      if (this.olMap.get('map_compare')) {
        this.olMap.set('map_compare', null);
        this.olMap.changed();
      }

      // Supprime le pointeur carto
      if (this.targetOverlay) {
        this.targetOverlay.setPosition(null);
        this.olMap.removeOverlay(this.targetOverlay);
        this.olMapCompare.removeOverlay(this.targetOverlay);
        this.targetOverlay = null;
      }
    }
    this.resetKeyEvent();
    this.olMap.updateSize();
  }

  /**
   * Détruit les events des cartes
   */
  private resetKeyEvent() {
    if (this.keyPointerMoveMap) {
      unByKey(this.keyPointerMoveMap);
    }
    if (this.keyPointerMoveMapCompare) {
      unByKey(this.keyPointerMoveMapCompare);
    }
    if (this.keyPointerDragMap) {
      unByKey(this.keyPointerDragMap);
    }
    if (this.keyPointerDragMapCompare) {
      unByKey(this.keyPointerDragMapCompare);
    }
    if (this.keyChangeResolutionMap) {
      unByKey(this.keyChangeResolutionMap);
    }
    if (this.keyChangeResolutionMapCompare) {
      unByKey(this.keyChangeResolutionMapCompare);
    }
    if (this.keyChangeMouseEnterMap) {
      $(this.olMap.getTargetElement()).off('mouseenter');
    }
    if (this.keyChangeMouseLeaveMap) {
      $(this.olMap.getTargetElement()).off('mouseleave');
    }
    if (this.keyChangeMouseEnterMapCompare) {
      $(this.olMapCompare.getTargetElement()).off('mouseenter');
    }
    if (this.keyChangeMouseLeaveMapCompare) {
      $(this.olMapCompare.getTargetElement()).off('mouseleave');
    }
  }

  /**
   * Sélection de la carte de comparaison
   * @param mapId
   */
  public selectedCompareMap(mapId) {
    this.resetKeyEvent();

    const controls = document.getElementsByClassName('controls-container');
    const controlsTopCenter = controls[0].getElementsByClassName('top-center');
    const controlsBottomRight = controls[0].getElementsByClassName('bottom-right');
    controlsTopCenter[0].classList.add('map-compare');
    controlsBottomRight[0].classList.add('map-compare');

    this.mapCompareId = mapId;

    this.initCompareMapId(mapId).then((olMapCompare: Map) => {

      this.olMap.set('map_compare', olMapCompare);
      this.olMap.changed();

      // Synchronisation de l'extent des views
      this.keyChangeMouseEnterMap = $(this.olMap.getTargetElement()).on('mouseenter', (e) => {
        this.keyChangeResolutionMap = this.olMap.on('moveend', (event: any) => {
          this.setMapCompareCenter(this.olMap.getView());
          this.setMapCompareScale();
        });
      });

      this.keyChangeMouseLeaveMap = $(this.olMap.getTargetElement()).on('mouseleave', (e) => {
        if (this.keyChangeResolutionMap) {
          unByKey(this.keyChangeResolutionMap);
        }
      });

      this.keyChangeMouseEnterMapCompare = $(this.olMapCompare.getTargetElement()).on('mouseenter', (e) => {
        this.keyChangeResolutionMapCompare = this.olMapCompare.on('moveend', (event: any) => {
          this.setMapCenter(this.olMapCompare.getView());
          this.setMapScale();
        });
      });

      this.keyChangeMouseLeaveMapCompare = $(this.olMapCompare.getTargetElement()).on('mouseleave', (e) => {
        if (this.keyChangeResolutionMapCompare) {
          unByKey(this.keyChangeResolutionMapCompare);
        }
      });

      // Synchronisation du déplacement sur les carte
      this.keyPointerDragMap = this.olMap.on('pointerdrag', (e) => {
        this.setMapCompareCenter(this.olMap.getView());
      });

      this.keyPointerDragMapCompare = this.olMapCompare.on('pointerdrag', (e) => {
        this.setMapCenter(this.olMapCompare.getView());
      });

      // Synchronisation des pointers sur les cartes
      this.keyPointerMoveMap = this.olMap.on('pointermove', (e) => {
        if (document.getElementsByClassName('over-map-compare').length > 0) {
          this.targetOverlay.setPosition(null);
          this.olMap.removeOverlay(this.targetOverlay);
          this.targetOverlay = null;
        }
        if (!this.targetOverlay) {
          const element = document.createElement('div');
          element.className = 'ol-target';
          this.targetOverlay = new Overlay({
            element: element,
            position: [0, 0],
            positioning: 'center-center',
          });
          this.olMapCompare.addOverlay(this.targetOverlay);
          element.parentElement.classList.add('ol-target-overlay');
          element.parentElement.classList.add('over-map');
        }
        const coordinate = transform(e.coordinate, this.olMap.getView().getProjection(), this.olMapCompare.getView().getProjection());
        this.targetOverlay.setPosition(coordinate);
      });

      this.keyPointerMoveMapCompare = olMapCompare.on('pointermove', (e) => {
        if (document.getElementsByClassName('over-map').length > 0) {
          this.targetOverlay.setPosition(null);
          this.olMapCompare.removeOverlay(this.targetOverlay);
          this.targetOverlay = null;
        }
        if (!this.targetOverlay) {
          const element = document.createElement('div');
          element.className = 'ol-target';
          this.targetOverlay = new Overlay({
            element: element,
            position: [0, 0],
            positioning: 'center-center',
          });
          this.olMap.addOverlay(this.targetOverlay);
          element.parentElement.classList.add('ol-target-overlay');
          element.parentElement.classList.add('over-map-compare');
        }
        const coordinate = transform(e.coordinate, this.olMapCompare.getView().getProjection(), this.olMap.getView().getProjection());
        this.targetOverlay.setPosition(coordinate);
      });
    });
  }

  /**
   * Met à jour le centre de la carte de comparaison
   * @param olMapView
   */
  private setMapCompareCenter(olMapView) {
    const center = olMapView.getCenter();
    const centerMapCompare = transform(center, olMapView.getProjection(), this.olMapCompare.getView().getProjection());
    this.olMapCompare.getView().setCenter(centerMapCompare);
  }

  /**
   * Met à jour le centre de la carte principale
   * @param olMapCompareView
   */
  private setMapCenter(olMapCompareView) {
    const center = olMapCompareView.getCenter();
    const centerMap = transform(center, olMapCompareView.getProjection(), this.olMap.getView().getProjection());
    this.olMap.getView().setCenter(centerMap);
  }

  /**
   * Met à jour l'étendue de la carte de comparaison
   */
  private setMapCompareScale() {
    this.mapUtils.setScale(this.olMapCompare, this.mapUtils.getScale(this.olMap));
  }

  /**
   * Met à jour l'étendue de la carte principale
   */
  private setMapScale() {
    this.mapUtils.setScale(this.olMap, this.mapUtils.getScale(this.olMapCompare));
  }

  /**
   * Initialise la carte de comparaison
   * @param mapId
   * @returns
   */
  private initCompareMapId(mapId): Promise<Map> {
    return new Promise((resolve, reject) => {
      this.subscriptions.push(this.mapCoreManagerService.initOlMap(mapId, this.targetCompareId, this.tileSize, this.olMapCompare).subscribe((olMap: any) => {

        //  // Enregistre l'objet Map
        this.olMapCompare = olMap;

        // Enregistre l'identifiant de carte utilisé
        this.olMapCompare.set('map_id', mapId);

        // Enregistre l'identifiant de l'élement utilisé
        this.olMapCompare.set('target_id', this.targetCompareId);

        // Fit l'étendue
        this.setMapCompareScale();
        this.setMapCompareCenter(this.olMap.getView());

        // Enregistre le nom de la carte utilisé
        this.olMapCompare.set('name', this.olMapCompare.get('map_definition')['name']);

        resolve(this.olMapCompare);
      }));
    });
  }
}
