import { Component, Input, OnDestroy, OnInit } from '@angular/core';
import { Map } from 'ol';
import { Subscription } from 'rxjs';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';
import { MapCoreManagerService } from 'src/app/vitis/map-core/services/map-core-manager/map-core-manager.service';
import { VmapLayer } from 'src/app/vitis/map-core/services/map-models.model';
import { MapToolDef } from 'src/modules/vmap/components/carto/components/vmap-tools/map-tools.model';
import { VmapToolsService } from 'src/modules/vmap/components/carto/services/vmap-tools/vmap-tools.service';

class LayerFilterOperator {
  column: string;
  value: string;
  compare_operator: string;
  sql_operator: string;
  oSelectedColumn?: LayerColumn;

  constructor(obj: any = {}) {
    this.column = obj.column || null;
    this.value = obj.value || null;
    this.compare_operator = obj.compare_operator || null;
    this.sql_operator = obj.sql_operator || null;
  }
}

interface LayerColumn {
  column: string;
  filter: boolean;
  index: string;
  name: string;
  requestor: boolean;
  type: string;
}

@Component({
  selector: 'app-vmap-tools-filter',
  templateUrl: './vmap-tools-filter.component.html',
  styleUrls: ['./vmap-tools-filter.component.scss']
})
export class VmapToolsFilterComponent implements OnInit, OnDestroy {

  @Input() olMap: Map;
  @Input() oTool: MapToolDef;

  public aAvailableLayers: Array<VmapLayer> = [];
  public oSelectedLayer: VmapLayer;
  public bToolOpen: boolean;
  public aActiveFilter: Array<LayerFilterOperator> = [new LayerFilterOperator()];
  public aLayerColumns: Array<LayerColumn> | null;
  public isOpenDropDown: boolean = false;

  // Mémorise toutes les souscriptions de manière à les desouscrire à la destruction du composant
  private subscriptions: Subscription[] = [];

  constructor(
    private vitisCheck: VitisCheck,
    private mapManagerService: MapCoreManagerService,
    private vmapToolsService: VmapToolsService,
  ) { }

  ngOnInit(): void {

    // Ouverture de l'outil
    this.subscriptions.push(this.oTool.onOpen.subscribe(() => {
      this.initTool();
      this.bToolOpen = true;
    }));

    // Fermeture de l'outil
    this.subscriptions.push(this.oTool.onClose.subscribe(() => {
      this.bToolOpen = false;
    }));
  }

  /**
   * Fonction de destruction
   */
  ngOnDestroy(): void {
    this.subscriptions.forEach(subscription => subscription.unsubscribe());
  }

  /**
   * Initialise l'outil
   */
  private initTool() {
    this.aAvailableLayers = this.getAvailableLayers_(this.olMap);
    const layerExists = this.aAvailableLayers.find(layer => {
      if (this.oSelectedLayer &&
        layer.layer_id === this.oSelectedLayer.layer_id) {
        return true;
      } else {
        return false;
      }
    });
    if (!layerExists) {
      this.oSelectedLayer = null;
      this.aActiveFilter = [new LayerFilterOperator()];
    }
  }

  /**
   * Appelé au changement de couche
   * @param oLayer
   */
  public onLayerChange(oLayer: VmapLayer) {
    this.aLayerColumns = this.getAvailableColumns_(this.oSelectedLayer);
    this.aActiveFilter = [new LayerFilterOperator()];
  }

  /**
   * Appelé à la sélection d'une colonne
   * @param sColumn
   */
  public onSelectColumn(index: number, sColumn: string) {
    if (this.vitisCheck.isArray(this.aLayerColumns) && this.aLayerColumns.length > 0) {
      const oSelectedColumn = this.aLayerColumns.find(col => {
        return col.column === sColumn;
      });
      this.aActiveFilter[index].oSelectedColumn = oSelectedColumn;
    }
  }

  /**
   * Affiche le bouton permettant d'ajouter une ligne
   * @param index
   */
  public openAddExpression(index: number) {
    if (this.aActiveFilter.length <= 1) {
      document.getElementById(`vmap_tool_filter_dropdown_${index}`).classList.toggle('show');
      this.isOpenDropDown = !this.isOpenDropDown;
    } else {
      this.addExpression(this.aActiveFilter[0].sql_operator, index);
    }
  }

  /**
   * Ajoute une ligne au filtre
   * @param sql_operator
   * @param index
   */
  public addExpression(sql_operator: string, index: number) {
    this.aActiveFilter[index].sql_operator = sql_operator;
    this.aActiveFilter = [
      ...this.aActiveFilter,
      new LayerFilterOperator()
    ];
    const dropElem = document.getElementById(`vmap_tool_filter_dropdown_${index}`);
    if (dropElem) {
      dropElem.classList.remove('show');
    }
  }

  /**
   * Supprime une ligne
   * @param index
   */
  public deleteExpression(index: number) {
    if (this.aActiveFilter.length === 1) {
      this.aActiveFilter = [new LayerFilterOperator()];
    } else {
      if (index === this.aActiveFilter.length - 1) {
        this.aActiveFilter[index - 1].sql_operator = null;
      }
      this.aActiveFilter = [...this.aActiveFilter.slice(0, index), ...this.aActiveFilter.slice(Number(index) + 1)];
    }
  }

  /**
   * Retourne les opérateurs en fonction du type de colonne
   * @param index
   * @returns
   */
  public getAvailableOperators(index: number): Array<string> {
    const aReturn = [];
    if (this.aActiveFilter &&
      this.aActiveFilter[index] &&
      this.aActiveFilter[index].oSelectedColumn &&
      this.aActiveFilter[index].oSelectedColumn.type) {

      const type = this.aActiveFilter[index].oSelectedColumn.type;

      switch (type) {
        case 'string':
          return ['=', '~', '!=', 'IS NOT NULL', 'IS NULL'];
          break;
        case 'number':
          return ['=', '!=', '<', '<=', '>', '>=', 'IS NOT NULL', 'IS NULL'];
          break;
        case 'bool':
          return ['IS TRUE', 'IS FALSE', 'IS NOT NULL', 'IS NULL'];
          break;
        default:
          return ['=', '!=', 'IS NOT NULL', 'IS NULL'];
          break;
      }
    }

    return aReturn;
  }

  /**
   * Lance la requête de filtre
   */
  public sendFilter() {
    const oAPIFilter = this.getAPIFilter_(this.aActiveFilter);
    this.vmapToolsService.refreshMap(this.olMap, false, null, oAPIFilter);
  }

  /**
   * Retourne les couches disponibles pour filtrage
   * @returns
   */
  private getAvailableLayers_(olMap: Map): Array<VmapLayer> {
    const aAvailableLayers = [];
    const aLayers = this.mapManagerService.getLayersArray(olMap);

    for (let i = 0; i < aLayers.length; i++) {
      const oLayer = aLayers[i];
      if (oLayer.type === 'postgres' &&
        oLayer.is_filter === true) {
        const columns = this.getAvailableColumns_(oLayer);
        if (this.vitisCheck.isArray(columns) && columns.length > 0) {
          aAvailableLayers.push(oLayer);
        }
      }
    }

    return aAvailableLayers.sort((a, b) => a.name > b.name && 1 || -1);
  }

  /**
   * Retourne la définition des colonnes de la couche en se basant sur json_columns
   * @param oLayer
   * @returns
   */
  private getAvailableColumns_(oLayer: VmapLayer): Array<LayerColumn> | null {
    if (this.vitisCheck.isString(oLayer.json_columns)) {
      const columns = JSON.parse(oLayer.json_columns);
      if (this.vitisCheck.isArray(columns) && columns.length > 0) {
        return columns.filter((column: any) => {
          return column.filter === true;
        });
      }
    } else {
      return null;
    }
  }

  /**
   * Transforme le filtre en filtre compréhensible par l'API
   * @param aFilter
   * @returns
   */
  private getAPIFilter_(aFilter: Array<LayerFilterOperator>) {
    if (this.vitisCheck.isArray(aFilter) && aFilter.length > 0) {

      const aAPIFilter = [];
      const oFilter = {
        relation: aFilter[0].sql_operator || 'AND',
        operators: [],
      };

      for (let i = 0; i < aFilter.length; i++) {
        const operator = aFilter[i];

        if (operator.column &&
          operator.column.length > 0 &&
          operator.compare_operator &&
          operator.compare_operator.length > 0) {

          if (operator.value === null) {
            operator.value = '';
          }

          if (operator.compare_operator === '~') {
            oFilter.operators.push({
              column: operator.column,
              compare_operator: 'LIKE',
              value: `%${operator.value}%`,
              compare_operator_options: {
                column_type: operator.oSelectedColumn.type,
                case_insensitive: true,
              }
            });
          } else if (operator.compare_operator === 'IS TRUE' || operator.compare_operator === 'IS FALSE') {
            oFilter.operators.push({
              column: operator.column,
              compare_operator: '=',
              value: operator.compare_operator === 'IS TRUE' ? true : false,
              compare_operator_options: {
                column_type: operator.oSelectedColumn.type,
              }
            });
          } else {
            oFilter.operators.push({
              column: operator.column,
              compare_operator: operator.compare_operator,
              value: operator.value,
              compare_operator_options: {
                column_type: operator.oSelectedColumn.type,
              }
            });
          }
        }
      }

      if (oFilter.operators.length > 0) {
        aAPIFilter.push({
          layer_id: this.oSelectedLayer.layer_id,
          filter: oFilter,
        });
      }

      return aAPIFilter;
    }
  }
}
