import { Component, Input, Output, OnChanges, OnInit, EventEmitter, ViewChild, OnDestroy } from '@angular/core';
import { environment } from 'src/environments/environment';
import { VitisCheck, VitisClone, vitisAlert, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { MapCoreUtilsService } from 'src/app/vitis/map-core/services/map-core-utils/map-core-utils.service';
import { MapSelection, VmapSelectService } from 'src/modules/vmap/components/carto/services/vmap-select/vmap-select.service';
import { ActivatedRoute, Router } from '@angular/router';
import { TranslateService } from '@ngx-translate/core';
import { VmapToolsService } from 'src/modules/vmap/components/carto/services/vmap-tools/vmap-tools.service';
import { VmapToolsFormComponent } from 'src/app/modules/vmap/carto/components/vmap-tools/components/vmap-tools-select-right-panel/vmap-tools-form/vmap-tools-form.component';

import { Geometry } from 'ol/geom';
import { Feature } from 'ol';
import { FormreaderComponent } from 'src/app/vitis/formreader/formreader.component';
import { Subscription } from 'rxjs';
import { FormreaderService } from 'src/app/vitis/formreader/services/formreader.service';

@Component({
  selector: 'app-vmap-tools-modify-right-panel',
  templateUrl: './vmap-tools-modify-right-panel.component.html',
  styleUrls: ['./vmap-tools-modify-right-panel.component.scss']
})
export class VmapToolsModifyRightPanelComponent implements OnInit, OnChanges, OnDestroy {

  @ViewChild(VmapToolsFormComponent) oFormComponent: VmapToolsFormComponent;

  @Input() map;
  @Input() selections: Array<MapSelection>;
  @Input() indexSelection: number;

  @Input() isSmallScreen: boolean;

  @Output() subscribeFormName = new EventEmitter();
  @Output() updatePopup = new EventEmitter();
  @Output() subscribeFormSize = new EventEmitter();


  public oLayer;

  private oParamsId;
  private oParamsLayerId;

  public bOpenActions = false;

  public sFormName = 'display';
  public sDefinition;
  public sDatasForm;

  private feature: Feature<Geometry>;
  public geomEventValue: string = '';
  private oFormReader: FormreaderComponent;

  // Affichage des boutons d'actions.
  public oActions: any = {
    google_street_view: true
  };

  // Mémorise toutes les souscriptions de manière à les desouscrire à la destruction du composant
  private subscriptions: Subscription[] = [];

  constructor(
    private vitisClone: VitisClone,
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private mapSelectService: VmapSelectService,
    private mapUtils: MapCoreUtilsService,
    private mapToolsService: VmapToolsService,
    private translate: TranslateService,
    private route: ActivatedRoute,
    private router: Router,
    private formreaderSrvc: FormreaderService,
  ) { }

  ngOnInit(): void {
    this.subscriptions.push(this.route.data.subscribe(mode => {
      this.subscriptions.push(this.route.queryParams.subscribe(oParams => {
        this.oParamsId = oParams.id;
        this.oParamsLayerId = oParams.layer_id;
      }));
    }));


    if (this.selections && this.vitisCheck.isNumber(this.indexSelection) && this.selections[this.indexSelection]) {
      const aSelection = this.selections[this.indexSelection];
      this.oLayer = aSelection.oLayer;
      this.initFeature(aSelection);
      const def = this.vitisCheck.isString(aSelection.oLayer.published_form_json_def) ?
        JSON.parse(aSelection.oLayer.published_form_json_def) : aSelection.oLayer.published_form_json_def;

      this.sDefinition = this.vitisClone.clone(def);
      this.sDatasForm = aSelection.data;
    }
  }

  /**
   * Fonction de destruction
   */
  ngOnDestroy(): void {
    this.subscriptions.forEach(subscription => subscription.unsubscribe());
  }

  ngOnChanges(changes): void {
    let aSelections = this.selections;
    let indexSelection = this.indexSelection;
    if (changes) {
      if (changes.selections && changes.selections.currentValue) {
        aSelections = changes.selections.currentValue;
      }
      if (changes.indexSelection && changes.indexSelection.currentValue) {
        indexSelection = changes.indexSelection.currentValue;
      }
      const aSelection = aSelections[indexSelection];
      if (aSelection) {
        this.oLayer = aSelection.oLayer;
        this.initFeature(aSelection);
        const def = this.vitisCheck.isString(aSelection.oLayer.published_form_json_def) ?
          JSON.parse(aSelection.oLayer.published_form_json_def) : aSelection.oLayer.published_form_json_def;

        this.sDefinition = this.vitisClone.clone(def);
        this.sDatasForm = aSelection.data;

        // Affichage du bouton "Voir sur Google Street View".
        if (this.vitisCheck.isDefAndNotNull(this.oLayer)) {
          if (this.vitisCheck.isDefAndNotNull(this.oLayer.google_street_view)) {
            this.oActions.google_street_view = this.oLayer.google_street_view;
          }
        }
      }
    }
  }

  /**
   * Initialise la feature pour sa modification
   * @param aSelection
   */
  private initFeature(aSelection): void {
    const EWKT = aSelection.data[aSelection.oLayer.geomfield];
    const geometry = this.mapUtils.getGeomFromEWKT(EWKT, this.map);
    this.feature = new Feature({ geometry });
  }

  /**
   * Permet de récupérer l'update de la geométrie de la feature
   * @param feature
   */
  public subscribeFeature(feature) {
    this.feature = feature;
  }

  /**
   * Ouvre le menu des actions
   */
  public openActions() {
    this.bOpenActions = !this.bOpenActions;
  }

  /**
   * Ouvre la fenêtre google street view
   */
  public googleStreetView(): void {
    const clickedMapCoordinates = this.mapSelectService.getClickedCoordinate();
    const EWKT = this.selections[this.indexSelection].data[this.selections[this.indexSelection].oLayer.geomfield];
    const url = this.mapUtils.getOpenStreetViewUrlFromEWKT(EWKT, this.map, clickedMapCoordinates);
    window.open(url);
  }

  /**
   * Retourne si geom est de type multi
   * @returns
   */
  public isGeometryTypeMulti(): boolean {
    if (this.feature) {
      const geom = this.feature.getGeometry();
      if (geom) {
        const geometryType = geom.getType();
        return geometryType === 'MultiPoint' || geometryType === 'MultiLineString' || geometryType === 'MultiPolygon';
      }
    }
    return false;
  }

  /**
  * Copie du lien vers cet objet
  */
  public copyPressObjectLink(): void {
    let link = this.router.url;
    if (!this.oParamsLayerId) {
      link += `&layer_id=${this.selections[this.indexSelection].oLayer.layer_id}`;
    }
    if (!this.oParamsId) {
      link += `&id=${this.selections[this.indexSelection].data[this.selections[this.indexSelection].oLayer.idfield]}`;
    }
    const txt = document.createElement('textarea');
    document.body.appendChild(txt);
    txt.value = `${environment.properties.app_client_url}${link}`; // CHROME
    txt.textContent = `${environment.properties.app_client_url}${link}`; // FIREFOX
    const sel = getSelection();
    const range = document.createRange();
    range.selectNode(txt);
    sel.removeAllRanges();
    sel.addRange(range);
    if (document.execCommand('copy')) {
      this.translate.get('URL_COPY_PRESS_OK').subscribe((sTranslate: string) => {
        vitisAlert.fireToast({
          type: 'success',
          title: `${sTranslate} :\n${environment.properties.app_client_url}${link}`
        });
      });
    }
    document.body.removeChild(txt);
  }

  /**
   * Ouvre le formulaire d'édition
   */
  public editForm(): void {
    this.sFormName = 'update';
    this.subscribeFormName.next(this.sFormName);
    this.subscribeFormSize.next(this.selections[this.indexSelection].oLayer.form_update_size_id ?? 1);
    // Permet de déclencher la recharge du formulaire par FormReader
    this.sDefinition = this.vitisClone.clone(
      this.vitisCheck.isString(this.selections[this.indexSelection].oLayer.published_form_json_def) ?
        JSON.parse(this.selections[this.indexSelection].oLayer.published_form_json_def) : this.selections[this.indexSelection].oLayer.published_form_json_def);
    this.sDatasForm = this.vitisClone.clone(this.sDatasForm);
  }

  /**
 * Event appelé si on redéfinit la géométrie et aux actions présent dans la redéfinition de la géométrie
 */
  public updateGeometry(): void {
    this.sFormName = 'geometry';
    this.subscribeFormName.next(this.sFormName);
    this.sDatasForm = this.selections[this.indexSelection].data;
  }

  /**
   * Sauvegarde la nouvelle géométrie de l'object
   */
  public saveGeometryModifications(): void {
    const oLayer = this.selections[this.indexSelection].oLayer;
    const projLayer = oLayer.coordsys_id;

    const sEWKTGeom = this.mapUtils.getEWKTFromGeomAndProj(this.feature.getGeometry(), this.map, projLayer);

    this.sendUpdateForm({
      [oLayer.geomfield]: sEWKTGeom,
      [oLayer.idfield]: this.sDatasForm[oLayer.idfield]
    }, true);
  }

  /**
   * Annule la redéfinition de la géométrie et renvoie sur le formulaire de description
   */
  public cancelGeometryModifications(): void {
    this.sendCloseForm();
  }

  /**
   * Envoi de la requête pour le formulaire d'édition
   * @param event
   */
  public sendUpdateForm(event: any, updateGeometry?: any): void {
    const data = this.vitisCheck.isDefAndNotNull(event) ? event : this.oFormComponent.getFormValues();
    const sendFunction = (resData: any) => {
      this.formreaderSrvc.manageLargesFiles(resData).then((managedData: any) => {
        return new Promise<void>((resolve, reject) => {
          this.vitisRequest.ajaxRequestPromise({
            method: 'PUT',
            url: `${environment.properties.app_api_url}/vmap/layers/${this.oLayer.layer_id}/query/${this.sDatasForm[this.oLayer.idfield]}`,
            data: this.formreaderSrvc.transformToDataForm(managedData),
          }).then((resp: any) => {
            if (resp && resp.data && resp.data.data) {
              this.refreshLayer(this.oLayer.layer_id);
              this.sendCloseForm(resp.data.data);
              this.updatePopup.next(resp.data.data);
              // Add Toast
              vitisAlert.fireToast({
                icon: 'success',
                title: this.translate.instant('COMPONENT_MAP_ENABLE_UTILS_SUCCESS_TOAST')
              });
              resolve();
            }
          }).catch(e => {
            console.error(e);
            this.translate.get(['ERROR_UPDATE_LAYER_OBJECT', 'ERROR_UPDATE_LAYER_OBJECT_DETAIL']).subscribe(async (translation) => {
              vitisAlert.alert({
                type: 'error',
                title: translation['ERROR_UPDATE_LAYER_OBJECT'],
                text: translation['ERROR_UPDATE_LAYER_OBJECT_DETAIL'],
              });
            });
            reject();
          });
        });
      });
    };

    if (updateGeometry) {
      sendFunction(data);
    } else {
      this.oFormComponent.oFormReader['updateObjectLayerForm'] = sendFunction;
      this.sDefinition[this.sFormName].event = 'updateObjectLayerForm()';
      this.oFormComponent.oFormReader.submitForm(this.sDefinition, this.sFormName, data);
    }
  }

  /**
   * Ferme le formulaire actuel
   */
  public sendCloseForm(oNewData?: any): void {
    this.sFormName = 'display';
    this.subscribeFormSize.next(this.selections[this.indexSelection].oLayer.form_display_size_id ?? 1);
    this.subscribeFormName.next(this.sFormName);
    this.sDefinition = this.vitisClone.clone(this.vitisCheck.isString(this.selections[this.indexSelection].oLayer.published_form_json_def) ?
      JSON.parse(this.selections[this.indexSelection].oLayer.published_form_json_def) : this.selections[this.indexSelection].oLayer.published_form_json_def);
    if (oNewData) {
      this.sDatasForm = this.vitisClone.clone(oNewData);
    } else {
      this.sDatasForm = this.vitisClone.clone(this.sDatasForm);
    }
    this.bOpenActions = false;
  }

  /**
   * Rafraîchit la couche
   * @param iLayerId
   */
  private refreshLayer(iLayerId: number): void {
    this.mapToolsService.refreshLayer(this.map, iLayerId, true);
  }

  /**
   * Envoi la requête de suppression d'un object
   */
  public deleteLayerObject(): void {
    this.displayDeleteWarning().then((bResponse: boolean) => {
      if (bResponse === true) {
        const oLayer = this.selections[this.indexSelection].oLayer;
        this.vitisRequest.ajaxRequestPromise({
          method: 'DELETE',
          url: `${environment.properties.app_api_url}/vmap/layers/${oLayer.layer_id}/query/${this.sDatasForm[oLayer.idfield]}`,
        }).then((resp: any) => {
          if (resp && resp.data && resp.data.data) {
            this.refreshLayer(oLayer.layer_id);
            this.mapToolsService.closeAllPanels();
            this.updatePopup.next(null);
          }
        }).catch(e => {
          console.error(e);
          this.translate.get('ERROR_DELETE_LAYER_OBJECT').subscribe(async (translation) => {
            vitisAlert.alert({
              text: translation,
              showCancelButton: true
            });
          });
        });
      }
    });
  }

  /**
   * Affiche le message de warning à la suppression
   * @returns
   */
  private displayDeleteWarning(): Promise<boolean> {
    return new Promise((resolve, reject) => {
      this.translate.get(['FORM_DELETE_WARNING',
        'CONFIRM_CANCEL',
        'CONFIRM_CONFIRM']
      ).subscribe((translations) => {
        vitisAlert.alert({
          title: translations['FORM_DELETE_WARNING'],
          type: 'warning',
          showCancelButton: true,
          confirmButtonText: translations['CONFIRM_CONFIRM'],
          cancelButtonText: translations['CONFIRM_CANCEL'],
        }).then((result: any) => {
          if (result &&
            result.isConfirmed === true) {
            resolve(true);
          } else {
            resolve(false);
          }
        });
      });
    });
  }

  /**
   * Sélectionne le mode insert avec la couche qui va avec
   */
  public selectInsertMode() {
    this.mapSelectService.closeSelections();
    this.mapSelectService.setSelectMode('add');
    this.mapSelectService.setSelection(this.oLayer);
  }

  /**
   * Recharge le javascript lorsque le formreader est modifié
   * @param oFormReader
   */
  public updateFormReader(oFormReader: FormreaderComponent) {
    const aSelection = this.selections[this.indexSelection];

    // Charge le JS du formulaire
    if (this.vitisCheck.isDefAndNotNull(aSelection.oLayer.published_form_js_def)) {
      if (!this.vitisCheck.isSameObject(this.oFormReader, oFormReader)) {
        const oScope = {
          map: this.map,
          ...aSelection,
        };
        this.mapToolsService.loadJs(aSelection.oLayer.published_form_js_def, aSelection.oLayer.layer_id, oFormReader, oScope);
        this.oFormReader = this.vitisClone.clone(oFormReader);
      }
    }
  }
}
