import { Injectable, EventEmitter } from '@angular/core';
import { Map } from 'ol';
import VectorLayer from 'ol/layer/Vector';
import VectorSource from 'ol/source/Vector';
import Collection from 'ol/Collection';
import { unByKey } from 'ol/Observable';
import Feature from 'ol/Feature';
import Translate from 'ol/interaction/Translate';
import Polygon from 'ol/geom/Polygon';
import { fromExtent } from 'ol/geom/Polygon';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';
import { MapCoreUtilsService } from 'src/app/vitis/map-core/services/map-core-utils/map-core-utils.service';
import { Source } from 'ol-mapbox-style';
import { Geometry } from 'ol/geom';

@Injectable({
  providedIn: 'root'
})
export class PrintBoxService {

  private oPrintBoxFeature;
  private aPrintBoxSize: Array<number>;
  private aOriginalSize: Array<number>;
  private iOffset: number = 0;
  private oPrintBoxEvent;

  private oOpenLayersOverlay: VectorLayer<VectorSource<Source>>;
  private oOpenLayersOverlayFeatures: Collection<Feature<Geometry>>;
  private oTranslateInteraction: Translate;
  private oTranslateInteractionEvent;

  constructor(
    private vitisCheck: VitisCheck,
    private mapUtils: MapCoreUtilsService,
  ) { }

  /**
   * Initialise ou réinitialise la boite d'impression
   * @param olMap
   * @param boxSize
   * @param iOffset permet de décaler la fenètre vers la gauche (compense le bandeau de l'outil d'impression)
   */
  public initPrintBox(olMap: Map, boxSize: Array<number>, iOffset = 0) {

    // Initialisation des tailles de la printBox
    this.aPrintBoxSize = boxSize;
    this.aOriginalSize = [...this.aPrintBoxSize];
    this.iOffset = iOffset;

    // Vérifie que les objets du service ne soient pas déjà instanciés
    if (this.vitisCheck.isDefAndNotNull(this.oOpenLayersOverlay)) {
      olMap.removeLayer(this.oOpenLayersOverlay);
    }
    this.oOpenLayersOverlay = new VectorLayer({
      map: olMap,
      source: new VectorSource({
        useSpatialIndex: false
      }),
      updateWhileAnimating: true,
      updateWhileInteracting: true,
      zIndex: 1000,
    });

    this.oOpenLayersOverlayFeatures = this.oOpenLayersOverlay.getSource().getFeaturesCollection();
    this.oTranslateInteraction = new Translate({
      features: this.oOpenLayersOverlayFeatures
    });
  }

  /**
   * Affiche la printBox sur la carte en cours
   * @param bListenMovements
   */
  public showPrintBox(olMap: Map, bListenMovements: boolean = true) {
    // Crée la première printBox
    this.drawPrintBox(olMap);

    // Re-dessine une printBox à chaque mouvement de la carte si bListenMovements
    if (bListenMovements) {
      this.listenMapMovements(olMap);
    }
  }

  /**
   * Masque la printBox de la carte
   */
  public hidePrintBox() {
    // Arrête d'écouter les mouvements de la carte
    this.unlistenMapMovements();

    // Supprime la print box en cours
    this.oOpenLayersOverlayFeatures.clear();
  }

  /**
   * Renvoie true si la printBox est masquée
   * @returns boolean
   */
  public isPrintBoxHidden() {
    return this.oOpenLayersOverlayFeatures.get('values').length === 0;
  }

  /**
   * Get the printBox size (warning: this only works when the printBox zone is managed)
   * @returns
   */
  public getSize() {
    return this.aPrintBoxSize;
  }

  /**
   * Set the printBox size
   * @param printBoxSize
   */
  public setSize(aPrintBoxSize: Array<number>) {
    this.aPrintBoxSize = aPrintBoxSize;
  }

  /**
   * Get the printBox extent
   * @export
   * @returns {array<number>}
   */
  public getExtent() {
    return this.oPrintBoxFeature.getGeometry().getExtent();
  }

  /**
   * Draw a new print box on each map precompose
   * @private
   */
  private listenMapMovements(olMap: Map) {
    // Supprime les anciens évènements et interractions
    unByKey(this.oPrintBoxEvent);

    // Ajoute le nouvel évènement
    this.oPrintBoxEvent = olMap.on('precompose', () => ((map) => {
      try {
        this.drawPrintBox(map);
      } catch (error) { }
    })(olMap));
  }

  /**
   * Stop drawing print boxes on each map precompose
   * @private
   */
  private unlistenMapMovements = function () {
    unByKey(this.oPrintBoxEvent);
  };

  /**
   * Draw the printBox descrived by this.oPrintBoxFeature
   * @returns {undefined}
   * @private
   */
  private drawPrintBox(olMap: Map) {
    if (!this.vitisCheck.isDefAndNotNull(this.aPrintBoxSize)) {
      console.error('Print box size undefined');
      return 0;
    }

    // Supprime la zone d'impression
    this.oOpenLayersOverlayFeatures.clear();

    // Dessine la zone d'impression
    this.oPrintBoxFeature = this.createPrintBox(olMap, this.aPrintBoxSize);
    this.oOpenLayersOverlay.getSource().addFeature(this.oPrintBoxFeature);
  }

  /**
   * Create a printBox feature according to this.printedMapSize
   * @param olMap
   * @param aPrintBoxSize
   * @returns {Feature}
   */
  private createPrintBox(olMap: Map, aPrintBoxSize: Array<number>) {

    const boxWidth = aPrintBoxSize[0];
    const boxHeight = aPrintBoxSize[1];

    const oMapElement = olMap.getTargetElement();
    const oComputedStyle = window.getComputedStyle(oMapElement, null);

    const x1 = parseFloat(oComputedStyle.width.replace('px', '')) / 2 - boxWidth / 2 - this.iOffset;
    const x2 = parseFloat(oComputedStyle.width.replace('px', '')) / 2 + boxWidth / 2 - this.iOffset;
    const y1 = parseFloat(oComputedStyle.height.replace('px', '')) / 2 - boxHeight / 2;
    const y2 = parseFloat(oComputedStyle.height.replace('px', '')) / 2 + boxHeight / 2;

    const point1 = [x1, y2];
    const point3 = [x2, y1];

    const coord1 = olMap.getCoordinateFromPixel(point1);
    const coord3 = olMap.getCoordinateFromPixel(point3);

    const box = fromExtent([coord1[0], coord3[1], coord3[0], coord1[1]]);

    const boxFeature = new Feature({
      geometry: box
    });

    boxFeature.set('pixelExtent', [x1, y1, x2, y2]);

    return boxFeature;
  }

}
