import { Injectable, EventEmitter } from '@angular/core';
import { Map } from 'ol';
import { Layer } from 'ol/layer';
import { Source } from 'ol/source';
import ImageWMS from 'ol/source/ImageWMS';
import TileWMS from 'ol/source/TileWMS';
import OSM from 'ol/source/OSM';
import XYZ from 'ol/source/XYZ';
import WMTS from 'ol/source/WMTS';
import VectorTile from 'ol/source/VectorTile';
import { vitisAlert, VitisCheck, VitisClone, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { MapCoreManagerService } from 'src/app/vitis/map-core/services/map-core-manager/map-core-manager.service';
import { MapCoreJSONService } from 'src/app/vitis/map-core/services/map-core-json/map-core-json.service';
import { FormreaderComponent } from 'src/app/vitis/formreader/formreader.component';
import { environment } from 'src/environments/environment';
import { TranslateService } from '@ngx-translate/core';
import { FormreaderService } from 'src/app/vitis/formreader/services/formreader.service';
import { MapCoreDrawService } from 'src/app/vitis/map-core/services/map-core-draw/map-core-draw.service';
import { MapCorePopupService } from 'src/app/vitis/map-core/services/map-core-popup/map-core-popup.service';
import { MapCorePopupOverlayService } from 'src/app/vitis/map-core/services/map-core-popup-overlay/map-core-popup-overlay.service';
import { MapCoreUtilsService } from 'src/app/vitis/map-core/services/map-core-utils/map-core-utils.service';
import { VmapSelectService } from '../vmap-select/vmap-select.service';
import { VmapLayer } from 'src/app/vitis/map-core/services/map-models.model';
import { LegendObject } from 'src/modules/vmap/components/carto/components/vmap-layers-panel/components/vmap-layers-legend/vmap-layers-legend.model';
import { nativeMapToolsDef } from 'src/app/modules/vmap/carto/components/vmap-tools/map-tools-buttons';
import { MapToolDef } from 'src/app/modules/vmap/carto/components/vmap-tools/map-tools.model';

interface CloseToolsOptions {
  bExceptBottom?: boolean; // Ne ferme pas les outils de type bottom
  bOnlyTypeMenu?: boolean; // ne ferme que les outils de type menu
}

@Injectable({
  providedIn: 'root'
})
export class VmapToolsService {

  // Gère les tailles
  private oLeftPanelEmiter: EventEmitter<number> = new EventEmitter();
  private oRightPanelEmiter: EventEmitter<number> = new EventEmitter();
  private oBottomPanelEmiter: EventEmitter<number> = new EventEmitter();

  // État actuel des panneaux en cours
  private iLeftPanel: number;
  private iRightPanel: number;
  private iBottomPanel: number;

  // Gère l'outil en cours
  private oActiveToolEmiter: EventEmitter<string> = new EventEmitter();

  // Gère les panneaux en cours
  private oActiveLeftPanelEmiter: EventEmitter<string> = new EventEmitter();
  private oActiveRightPanelEmiter: EventEmitter<string> = new EventEmitter();
  private oActiveBottomPanelEmiter: EventEmitter<string> = new EventEmitter();

  // État actuel des panneaux en cours
  private sActiveLeftPanel: string = '';
  private sActiveRightPanel: string = '';
  private sActiveBottomPanel: string = '';

  // Renvoie les panneaux qui sont ouverts
  private oCurrentOpenedLeftPanel: EventEmitter<string> = new EventEmitter();
  private oCurrentOpenedRightPanel: EventEmitter<string> = new EventEmitter();
  private oCurrentOpenedBottomPanel: EventEmitter<string> = new EventEmitter();

  // Renvoie l'outil qui est ouvert
  private currentOpenedTool: EventEmitter<string> = new EventEmitter();

  // Event de rafraîchissement de la carte
  private oRefreshMapEmitter: EventEmitter<number> = new EventEmitter();

  // Outils disponible
  private oToolsDef: any;

  constructor(
    private mapCoreManagerService: MapCoreManagerService,
    private mapCoreJSONService: MapCoreJSONService,
    private mapCoreDrawService: MapCoreDrawService,
    private mapCorePopupService: MapCorePopupService,
    private mapCorePopupOverlayService: MapCorePopupOverlayService,
    private mapCoreUtilsService: MapCoreUtilsService,
    private mapSelectService: VmapSelectService,
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private vitisClone: VitisClone,
    private translate: TranslateService,
    private formreaderSrvc: FormreaderService,
  ) {
    this.getActiveLeftPanelEmiter().subscribe((next: string) => {
      this.sActiveLeftPanel = next;
      this.oCurrentOpenedLeftPanel.next(next);
    });

    this.getActiveRightPanelEmiter().subscribe((next: string) => {
      this.sActiveRightPanel = next;
      this.oCurrentOpenedRightPanel.next(next);
    });

    this.getActiveBottomPanelEmiter().subscribe((next: string) => {
      this.sActiveBottomPanel = next;
      this.oCurrentOpenedBottomPanel.next(next);
    });

    this.getLeftPanelEmiter().subscribe((next: number) => {
      this.iLeftPanel = next;
    });

    this.getRightPanelEmiter().subscribe((next: number) => {
      this.iRightPanel = next;
    });

    this.getBottomPanelEmiter().subscribe((next: number) => {
      this.iBottomPanel = next;
    });

    // Charge la liste des outils disponibles
    this.loadToolsDef();
  }

  /**
   * Ferme tous les panneaux ouverts
   * @returns
   */
  public closeAllPanels(): void {
    // Fermeture du panneau de gauche
    this.setLeftPanel(0, '');

    // Fermeture du panneau de droite
    this.setRightPanel(0, '');

    // Fermeture du panneau du bas
    this.setBottomPanel(0, '');
  }

  /**
   * Définit un outil à une position et une taille donnée
   * @param position
   * @param cols
   * @param tool
   */
  public setPanel(position: string, cols: number, tool: string) {
    if (position === 'left') {
      this.setLeftPanel(cols, tool);
    } else if (position === 'right') {
      this.setRightPanel(cols, tool);
    } else if (position === 'bottom') {
      this.setBottomPanel(cols, tool);
    }
  }

  /**
   * Change l'état du panneau de gauche
   */
  public setLeftPanel(cols: number, tool: string): void {
    this.getLeftPanelEmiter().next(cols);

    // Permet de ne pas fermer / ouvrir l'outil si la taille change
    if (this.sActiveLeftPanel !== tool) {
      this.getActiveLeftPanelEmiter().next(tool);
    }
  }

  /**
   * Change l'état du panneau de droite
   */
  public setRightPanel(cols: number, tool: string): void {
    this.getRightPanelEmiter().next(cols);

    // Permet de ne pas fermer / ouvrir l'outil si la taille change
    if (this.sActiveRightPanel !== tool) {
      this.getActiveRightPanelEmiter().next(tool);
    }
  }

  /**
   * Change l'état du panneau du bas
   */
  public setBottomPanel(cols: number, tool: string): void {
    this.getBottomPanelEmiter().next(cols);

    // Permet de ne pas fermer / ouvrir l'outil si la taille change
    if (this.sActiveBottomPanel !== tool) {
      this.getActiveBottomPanelEmiter().next(tool);
    }
  }

  /**
   * Emiter appelé lors de l'ouverture d'un des menus de gauche
   * la valeur vaut 0, 1, 2, 3 en fonction de la largeur de la fenêtre
   * @returns
   */
  public getLeftPanelEmiter(): EventEmitter<number> {
    return this.oLeftPanelEmiter;
  }

  /**
   * Emiter appelé lors de l'ouverture d'un des menus de droite
   * la valeur vaut 0, 1, 2, 3 en fonction de la largeur de la fenêtre
   * @returns
   */
  public getRightPanelEmiter(): EventEmitter<number> {
    return this.oRightPanelEmiter;
  }

  /**
   * Emiter appelé lors de l'ouverture d'un des menus du bas
   * la valeur vaut 0, 1, 2, 3 en fonction de la largeur de la fenêtre
   * @returns
   */
  public getBottomPanelEmiter(): EventEmitter<number> {
    return this.oBottomPanelEmiter;
  }

  /**
   * Emiter appelé lors de l'ouverture d'un des menus
   * la valeur vaut le nom de l'outil
   * @returns
   */
  public getActiveToolEmiter(): EventEmitter<string> {
    return this.oActiveToolEmiter;
  }

  /**
   * Emiter appelé lors de l'ouverture d'un des panneaux de gauche
   * la valeur vaut le nom de l'outil
   * @returns
   */
  public getActiveLeftPanelEmiter(): EventEmitter<string> {
    return this.oActiveLeftPanelEmiter;
  }

  /**
   * Emiter appelé lors de l'ouverture d'un des panneaux de droite
   * la valeur vaut le nom de l'outil
   * @returns
   */
  public getActiveRightPanelEmiter(): EventEmitter<string> {
    return this.oActiveRightPanelEmiter;
  }

  /**
   * Emiter appelé lors de l'ouverture d'un des panneaux du bas
   * la valeur vaut le nom de l'outil
   * @returns
   */
  public getActiveBottomPanelEmiter(): EventEmitter<string> {
    return this.oActiveBottomPanelEmiter;
  }

  /**
   * Emiter appelé lors du changement de l'outil qui est ouvert
   * sur le panneau de gauche
   * @returns
   */
  public getCurrentOpenedLeftPanel(): EventEmitter<string> {
    return this.oCurrentOpenedLeftPanel;
  }

  /**
   * Emiter appelé lors du changement de l'outil qui est ouvert
   * sur le panneau de droite
   * @returns
   */
  public getCurrentOpenedRightPanel(): EventEmitter<string> {
    return this.oCurrentOpenedRightPanel;
  }

  /**
   * Emiter appelé lors du changement de l'outil qui est ouvert
   * sur le panneau du bas
   * @returns
   */
  public getCurrentOpenedBottomPanel(): EventEmitter<string> {
    return this.oCurrentOpenedBottomPanel;
  }

  /**
   * Emiter appelé lors de l'ouverture d'un des menus
   * la valeur vaut le nom de l'outil ouvert
   * @returns
   */
  public getCurrentOpenedTool(): EventEmitter<string> {
    return this.currentOpenedTool;
  }

  /**
   * Retourne un emitter appelé au rafraichissement de la carte
   * @returns
   */
  public getRefreshMapEmitter(): EventEmitter<number> {
    return this.oRefreshMapEmitter;
  }

  /**
   * Rafraîchit la map en ne conservant rien (comme si on faisait F5)
   * @param olMap
   */
  public reloadMap(olMap: Map): void {
    const reloadMapTrigger = olMap.get('reloadMapTrigger');
    if (reloadMapTrigger) {
      reloadMapTrigger.next();
    }
  }

  /**
   * Rafraîchit la carte en cours
   *
   * @param olMap
   * @param keepLoadedTiles si true alors garde les tuiles tant qu'elles n'ont pas été chargées
   * @param aLayersIds permet de définir les couches à recharger
   * @param oAPIFilter permet d'effectuer un filtre sur les couches postgres
   */
  public refreshMap(olMap: Map, keepLoadedTiles = false, aLayersIds = null, oAPIFilter = null): void {
    const iMapId = olMap.get('map_id');
    const checkLayerId = this.vitisCheck.isArray(aLayersIds) && aLayersIds.length > 0;

    if (iMapId) {
      // Permet de rafraichir le mapfile
      this.mapCoreManagerService.ajaxGetMap(iMapId, oAPIFilter).then((oMapDef: any) => {
        const aLayers = olMap.getLayers().getArray();
        for (let i = 0; i < aLayers.length; i++) {
          const olLayer: any = aLayers[i];
          const olSource = olLayer.getSource();

          // Test aLayersIds
          if (checkLayerId && !this.isLayerInIds_(olLayer, aLayersIds)) {
            continue;
          }

          if (
            olSource instanceof ImageWMS ||
            olSource instanceof TileWMS
          ) {
            const params = olSource.getParams();
            params['TIMESTAMP'] = Date.now();
            olSource.updateParams(params);
          }

          if (
            olSource instanceof ImageWMS ||
            olSource instanceof TileWMS ||
            olSource instanceof OSM ||
            olSource instanceof XYZ ||
            olSource instanceof WMTS ||
            olSource instanceof VectorTile
          ) {
            if (keepLoadedTiles !== true) {
              olSource.refresh();
            }
          }

          if (oAPIFilter) {
            olMap.set('layers_filters', oAPIFilter);
          } else {
            olMap.set('layers_filters', null);
          }
        }

        this.oRefreshMapEmitter.next(Date.now());
      });
    } else {
      console.error('cannot found map_id', olMap);
    }
  }

  /**
   * Rafraîchit une couche de la carte
   * @param olMap
   * @param iLayerId
   * @param keepLoadedTiles
   */
  public refreshLayer(olMap: Map, iLayerId: number, keepLoadedTiles = false): void {
    this.refreshMap(olMap, keepLoadedTiles, [iLayerId]);
  }

  /**
   * Centre sur la position initiale de la carte
   * @param olMap
   */
  public centerHome(olMap: Map): void {

    // Vue d'après la définition
    const view = this.mapCoreJSONService.getViewFromDef(olMap.get('map_definition'), {
      size: olMap.getSize()
    });
    const mapCenter = view.getCenter();
    const mapResolution = view.getResolution();

    if (mapCenter && mapResolution) {
      olMap.getView().setCenter(mapCenter);
      olMap.getView().setResolution(mapResolution);
    } else {
      console.error('Cannot get map original center and resolution');
    }
  }

  /**
   * Retourne true si la couche fait partie des aIds
   * @param olLayer
   * @param aIds
   * @returns
   */
  private isLayerInIds_(olLayer: Layer<Source>, aIds: Array<number>): boolean {
    if (this.vitisCheck.isArray(aIds) && aIds.length > 0) {
      if (this.vitisCheck.isArray(olLayer.get('fusion_layers'))) {
        for (let i = 0; i < olLayer.get('fusion_layers').length; i++) {
          const olChildLayer = olLayer.get('fusion_layers')[i];
          const isInIds = this.isLayerInIds_(olChildLayer, aIds);
          if (isInIds === true) {
            return true;
          }
        }
      } else {
        if (olLayer.get('layer_id') && aIds.indexOf(olLayer.get('layer_id')) !== -1) {
          return true;
        }
      }
    }
    return false;
  }

  /**
   * Charge le js du formulaire
   * @param published_form_js_def
   * @param iObjectId
   * @param oFormReader
   */
  public loadJs(published_form_js_def: string, iObjectId: string | number, oFormReader: FormreaderComponent, oScope: any) {

    // Supprime les js précédement chargés
    const oldDiv = document.getElementsByClassName('vmap-form-js');
    for (let i = 0; i < oldDiv.length; i++) {
      oldDiv[i].remove();
    }

    if (published_form_js_def) {
      const oScript = document.createElement('script');
      oScript.textContent = published_form_js_def;
      oScript.setAttribute('class', 'vmap-form-js');
      oScript.setAttribute('type', 'text/javascript');
      document.getElementsByTagName('head')[0].appendChild(oScript);
    }
    this.runJS(oFormReader, oScope);
  }

  /**
   * Execute le javascript
   * @param oFormReader
   */
  private runJS(oFormReader: FormreaderComponent, oScope: any) {
    // Lance le JS associé
    if ((<any>window).constructor_form) {
      (<any>window).constructor_form(oScope, environment, {
        vitisRequest: this.vitisRequest,
        vitisCheck: this.vitisCheck,
        vitisAlert: vitisAlert,
        vitisClone: this.vitisClone,
        translate: this.translate,
        formreaderSrvc: this.formreaderSrvc,
        mapCoreManagerService: this.mapCoreManagerService,
        mapCoreJSONService: this.mapCoreJSONService,
        mapCoreDrawService: this.mapCoreDrawService,
        mapCorePopupService: this.mapCorePopupService,
        mapCorePopupOverlayService: this.mapCorePopupOverlayService,
        mapCoreUtilsService: this.mapCoreUtilsService,
        mapSelectService: this.mapSelectService,
        maptoolsService: this,
        oFormReader: oFormReader,
      });
    }
  }

  /**
   * Retourne le tableau de légendes
   * @param olMap
   * @param scale
   * @returns
   */
  public getLayersLegends(olMap: Map, scale?: number) {
    const aLayersLegends = [];
    const aLayers = this.mapCoreManagerService.getLayersArray(olMap);
    let scaleParam = '';

    for (let i = 0; i < aLayers.length; i++) {
      const olLayer = aLayers[i].olLayer;

      if (olLayer.get('legend') === true && olLayer.get('layerType') === 'postgres') {
        if (olLayer.get('visible') === true) {
          scaleParam = '';

          if (olLayer.get('is_dynamic_legend') && this.vitisCheck.isDefAndNotNull(scale)) {
            scaleParam = '&SCALE=' + scale;
          }

          if (olLayer.get('type') === 'tilewms' || olLayer.get('type') === 'imagewms') {
            const aWMSLegendObjects = this.getWMSLegendObjects_(aLayers[i], scaleParam);
            if (aWMSLegendObjects) {
              aLayersLegends.push({
                collapsed: false,
                layerName: olLayer.get('name'),
                legendObjects: aWMSLegendObjects,
              });
            }
          }
        }
      }
    }

    // Tri par ordre alpha
    aLayersLegends.sort((a, b) => {
      if (a.layerName < b.layerName) {
        return -1;
      }
      if (a.layerName > b.layerName) {
        return 1;
      }
      return 0;
    });

    return aLayersLegends;
  }

  /**
   * Parse la couche et retourne le tableau de LegendObject
   * @param oLayer
   * @param scaleParam
   * @returns
   */
  private getWMSLegendObjects_(oLayer: VmapLayer, scaleParam: any): Array<LegendObject> {
    const olLayer = oLayer.olLayer;
    const sLayers = olLayer.getSource().getParams()['LAYERS'];
    let aLayerParam = [];

    if (this.vitisCheck.isDefAndNotNull(sLayers)) {
      aLayerParam = olLayer.getSource().getParams()['LAYERS'].split(',');
    }

    // Permet de ne pas visualiser les couches non visibles
    const aActiveSublayers = olLayer.get('activeSublayers');
    if (this.vitisCheck.isDefAndNotNull(aActiveSublayers)) {
      aLayerParam = aActiveSublayers;
    }

    const sSourceUrl = olLayer.get('type') === 'imagewms' ? olLayer.getSource().getUrl() : olLayer.get('type') === 'tilewms' ? olLayer.getSource().getUrls()[0] : null;
    if (!sSourceUrl) {
      return;
    }

    const aLegendsArray = [];
    for (let iii = 0; iii < aLayerParam.length; iii++) {
      const separator = sSourceUrl.indexOf('?') !== -1 ? '&' : '?';

      let legendURL = sSourceUrl + separator + 'request=GetLegendGraphic&LAYER=' + aLayerParam[iii] + '&FORMAT=image/png&VERSION=1.1.0&service=WMS' + scaleParam;
      let legendName: any;
      let legendType = 'image';

      if (this.vitisCheck.isDefAndNotNull(oLayer.definition_sym_json)) {
        const oSymDef = this.vitisCheck.isString(oLayer.definition_sym_json) ? JSON.parse(oLayer.definition_sym_json) : oLayer.definition_sym_json;

        // Type singleclass
        if (oSymDef &&
          oSymDef.type === 'singleclass') {
          legendName = oSymDef.legendName ? oSymDef.legendName : oLayer.name;
          legendURL += `&RULE=${aLayerParam[iii]}`;
          legendType = 'vmap';
        }
      }

      aLegendsArray.push({
        url: sSourceUrl,
        layerName: olLayer.get('name'),
        service_login: this.vitisCheck.isDefAndNotNull(olLayer.get('service_login')) ? olLayer.get('service_login') : null,
        service_password: this.vitisCheck.isDefAndNotNull(olLayer.get('service_password')) ? olLayer.get('service_password') : null,
        legendURL,
        legendName,
        legendType,
      });
    }

    return aLegendsArray;
  }

  /**
   * Calcul du padding à donner lors de l'appel à centerMapOnGeom en fonction des menus ouverts
   * @returns
   */
  public getCenterPadding(): Array<number> {
    return [
      200, // Haut
      this.getRightPadding_(), // Droite
      this.getBottomPadding_(), // Bas
      this.getLeftPadding_(), // Gauche
    ];
  }

  /**
   * Calcul du padding du bas
   * @returns
   */
  private getBottomPadding_() {
    switch (this.iBottomPanel) {
      case 1:
        return 400;
        break;
      default:
        return 200;
        break;
    }
  }

  /**
   * Calcul du padding de droite
   * @returns
   */
  private getRightPadding_() {
    switch (this.iRightPanel) {
      case 1:
        return 400;
        break;
      case 2:
        return 600;
        break;
      default:
        return 200;
        break;
    }
  }

  /**
   * Calcul du padding de gauche
   * @returns
   */
  private getLeftPadding_() {
    switch (this.iLeftPanel) {
      case 1:
        return 400;
        break;
      case 2:
        return 600;
        break;
      default:
        return 200;
        break;
    }
  }

  /**
   * Charge les outils disponibles en concaténant ceux du code avec les modules complémentaires
   * @returns
   */
  public loadToolsDef() {
    return new Promise((resolve, reject) => {

      if (this.vitisCheck.isDefAndNotNull(this.oToolsDef)) {
        resolve(this.oToolsDef);
      } else {

        // Charge les modules complémentaires depuis la base
        this.vitisRequest.ajaxRequestPromise({
          url: `${environment.properties.app_api_url}/vmap/toolmap`,
          method: 'GET',
        }).then((resp: any) => {
          const oCustomTool = {};
          if (resp.data && this.vitisCheck.isArray(resp.data.data)) {
            for (let i = 0; i < resp.data.data.length; i++) {
              const oTool = resp.data.data[i];
              if (oTool &&
                oTool.toolmap_id &&
                oTool.name &&
                oTool.icon &&
                oTool.title &&
                oTool.type &&
                oTool.panel_position &&
                oTool.panel_size &&
                oTool.component &&
                oTool.html_content &&
                oTool.js_content) {
                oCustomTool[oTool.name] = {
                  name: oTool.name,
                  icon: oTool.icon,
                  title: oTool.title,
                  type: oTool.type,
                  panelPosition: oTool.panel_position,
                  panelSize: oTool.panel_size,
                  component: oTool.component,
                  htmlContent: oTool.html_content,
                  jsContent: oTool.js_content,
                };
              }
            }
          }

          this.oToolsDef = { ...nativeMapToolsDef, ...oCustomTool };
          resolve(this.oToolsDef);

        }).catch(() => {
          console.error('error while loading toolmap');
          this.oToolsDef = { ...nativeMapToolsDef };
          resolve(this.oToolsDef);
        });
      }
    });
  }

  /**
   * Retourne les outils disponibles
   * @param olMap
   * @returns
   */
  public getMapToolsDef(olMap: Map) {
    return this.oToolsDef;
  }

  /**
   * Ferme l'ensemble des outils sauf exceptions (voir options)
   */
  public closeTools(olMap: Map, options?: CloseToolsOptions) {
    const oMapToolsDef = this.getMapToolsDef(olMap);
    options = this.vitisCheck.isObject(options) ? options : {};
    for (const key in oMapToolsDef) {
      if (Object.prototype.hasOwnProperty.call(oMapToolsDef, key)) {
        const oTool = oMapToolsDef[key];

        let close = true;
        if (options.bOnlyTypeMenu && oTool.type !== 'submenu') {
          close = false;
        }
        if (options.bExceptBottom && oTool.panelPosition === 'bottom') {
          close = false;
        }

        if (close) {
          this.openTool_(oTool, false);
        }
      }
    }
  }

  /**
   * Ouvre un outil
   * @param olMap
   * @param sToolName
   * @param bHandleMenus si true alors gère l'affichage en fonction des autres outils
   */
  public openTool(olMap: Map, sToolName: string, bHandleMenus = true) {
    const oTool = this.getToolByName(olMap, sToolName);
    if (bHandleMenus) {
      this.openToolWithMenus_(olMap, oTool, true);
    } else {
      this.openTool_(oTool, true);
    }
  }

  /**
   * Ferme un outil
   * @param olMap
   * @param sToolName
   * @param bHandleMenus si true alors gère l'affichage en fonction des autres outils
   */
  public closeTool(olMap: Map, sToolName: string, bHandleMenus = true) {
    const oTool = this.getToolByName(olMap, sToolName);
    if (bHandleMenus) {
      this.openToolWithMenus_(olMap, oTool, false);
    } else {
      this.openTool_(oTool, false);
    }
  }

  /**
   * Retourne un outil par son nom
   * @param olMap
   * @param sToolName
   * @returns
   */
  public getToolByName(olMap: Map, sToolName: string) {
    const oMapToolsDef = this.getMapToolsDef(olMap);

    let oTool: any;
    for (const key in oMapToolsDef) {
      if (Object.prototype.hasOwnProperty.call(oMapToolsDef, key)) {
        if (oMapToolsDef[key].name === sToolName) {
          oTool = oMapToolsDef[key];
        }
      }
    }

    if (this.vitisCheck.isDefAndNotNull(oTool)) {
      return oTool;
    } else {
      console.error('tool not defined', sToolName, oMapToolsDef);
    }
  }

  /**
   * Ferme un outil de type panneau à la position donnée
   */
  public closePanelAt(olMap: Map, position: string) {
    const oMapToolsDef = this.getMapToolsDef(olMap);
    for (const key in oMapToolsDef) {
      if (Object.prototype.hasOwnProperty.call(oMapToolsDef, key)) {
        const oTool: MapToolDef = oMapToolsDef[key];

        if (oTool.type === 'panel' && oTool.panelPosition === position) {
          if (oTool.onClose && oTool.collapsed_ === false) {
            oTool.onClose.next(true);
          }
          oTool.collapsed_ = true;
        }
      }
    }
  }

  /**
   * Ouvre ou ferme l'outil
   * @param bOpen
   */
  private openTool_(oTool: MapToolDef, bOpen: boolean) {
    if (oTool.onOpen) {
      if (bOpen === true && oTool.collapsed_ !== false) {
        oTool.onOpen.next(true);
      }
    }
    if (oTool.onClose && oTool.collapsed_ === false) {
      if (bOpen === false) {
        oTool.onClose.next(true);
      }
    }
    oTool.collapsed_ = !bOpen;
  }

  /**
   * Ouvre l'outil et gère les différents menus
   * @param olMap
   * @param oTool
   * @param bOpen
   */
  private openToolWithMenus_(olMap: Map, oTool: MapToolDef, bOpen: boolean) {

    if (oTool.type === 'submenu') {

      // Ferme les panneaux de gauche et droite
      this.setLeftPanel(0, '');
      this.setRightPanel(0, '');

      if (bOpen) {
        // Ferme les outils de gauche et de droite
        this.closeTools(olMap, {
          bExceptBottom: true,
        });
        this.openTool_(oTool, true);
      } else {
        this.openTool_(oTool, false);
      }
    } else if (oTool.type === 'panel') {

      if (oTool.collapsed_ === false) {
        if (oTool.panelPosition === 'left') {
          this.setLeftPanel(0, '');

        } else if (oTool.panelPosition === 'right') {
          this.setRightPanel(0, '');

        } else if (oTool.panelPosition === 'bottom') {
          this.setBottomPanel(0, '');
        }
        this.openTool_(oTool, false);

      } else {
        if (oTool.panelPosition === 'left') {
          this.setLeftPanel(oTool.panelSize, oTool.name);

        } else if (oTool.panelPosition === 'right') {
          this.setRightPanel(oTool.panelSize, oTool.name);

        } else if (oTool.panelPosition === 'bottom') {
          this.setBottomPanel(oTool.panelSize, oTool.name);
        }

        this.closeTools(olMap, {
          bOnlyTypeMenu: true,
        });
        this.closePanelAt(olMap, oTool.panelPosition);
        this.openTool_(oTool, true);
      }
    }
  }

  /**
   * Outils de la carte avec les positions
   * @param olMap
   * @returns
   */
  public getActiveTools(olMap: Map) {
    const oMapDef = olMap.get('map_definition');
    if (oMapDef && oMapDef.maptools && oMapDef.maptools.tools) {
      return oMapDef.maptools.tools;
    }
  }

  /**
   * Outils de la carte au format d'un tableau
   * @param olMap
   * @returns
   */
  public getActiveToolsAsArray(olMap: Map) {
    const aMapTools = [];
    const oMapTools = this.getActiveTools(olMap);

    for (const sPosition in oMapTools) {
      if (Object.prototype.hasOwnProperty.call(oMapTools, sPosition)) {
        if (this.vitisCheck.isArray(oMapTools[sPosition])) {
          for (let i = 0; i < oMapTools[sPosition].length; i++) {
            const sTool = oMapTools[sPosition][i];
            if (aMapTools.indexOf(sTool) === -1) {
              aMapTools.push(sTool);
            }
          }
        }
      }
    }

    return aMapTools;
  }

  /**
   * Retourne true/false si un outil est disponible pour cette carte
   * @param olMap
   * @param sTool
   * @returns
   */
  public isToolActive(olMap: Map, sTool: string) {
    const aMapTools = this.getActiveToolsAsArray(olMap);
    if (this.vitisCheck.isArray(aMapTools) && aMapTools.indexOf(sTool) !== -1) {
      return true;
    } else {
      return false;
    }
  }
}
