import { Injectable, EventEmitter } from '@angular/core';
import { VitisRequest, VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';
import { environment } from 'src/environments/environment';
import { BehaviorSubject, Observable, Subject } from 'rxjs';
import { StyleSubFields } from 'src/modules/vmap/components/models/json-symbology/json-symbology-style.model';

@Injectable({
  providedIn: 'root'
})
export class LayerSymbologyService {

  public columnsAvancedInput = new BehaviorSubject([]);
  public attributName: string;

  public symbols = new BehaviorSubject([]);
  public fonts = new BehaviorSubject([]);
  public positions = new BehaviorSubject({});
  public aligns = new BehaviorSubject({});
  public types = new BehaviorSubject({});
  public forces = new BehaviorSubject({});
  public lineCap = new BehaviorSubject({});
  public linejoin = new BehaviorSubject({});
  public obsRunTest = new Subject(); // Appelé pour lancer le test sur la couche

  public bddDatas = {
    layer_id: null,
    schema: null,
    database: null,
    table: null,
    datasource_id: null
  };

  constructor(
    private vitisRequest: VitisRequest,
    private vitisCheck: VitisCheck,
  ) {
    this.getSymbols();
    this.getFonts();
    this.getPositions();
    this.getAligns();
    this.getTypes();
    this.getForces();
    this.getLineCap();
    this.getLinejoin();
  }

  /**
   * Requête pour répupérer les colonnes de la table
   * @returns
   */
  public getColumns(): Promise<boolean> {
    const { layer_id, ...params } = this.bddDatas;
    return new Promise((resolve, reject) => {
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: `${environment.properties.app_api_url}/vmap/pglayers/columns`,
        params,
      }).then((resp) => {
        if (resp.data) {
          if (resp.data.status === 1) {
            if (this.vitisCheck.isArray(resp.data.data) && resp.data.data.length > 0) {
              this.columnsAvancedInput.next(resp.data.data);
              resolve(true);
            }
          }
        }
        reject(false);
      });
    });
  }

  /**
   * Requête pour récupérer les symboles dusponibles
   * @returns
   */
  public getSymbols(): Promise<boolean> {
    return new Promise((resolve, reject) => {
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: environment.properties.app_api_url + '/vmap/symbols',
        params: {
          attributs: 'name|label'
        }
      }).then((resp) => {
        if (resp.data) {
          if (resp.data.status === 1) {
            if (resp.data.data) {
              const arr = resp.data.data;
              const formatOptions = arr.reduce((acc, curr) => {
                return { ...acc, [curr.name]: curr.label || curr.name };
              }, {});
              this.symbols.next(formatOptions);
              resolve(true);
            }
          }
        }
        reject(false);
      });
    });
  }

  /**
   * Requête pour récupérer les polices dusponibles
   * @returns
   */
  public getFonts(): Promise<boolean> {
    return new Promise((resolve, reject) => {
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: environment.properties.app_api_url + '/vmap/fonts',
      }).then((resp) => {
        if (resp.data) {
          if (resp.data.status === 1) {
            if (resp.data.data) {
              const arr = resp.data.data;
              const formatOptions = arr.reduce((acc, curr) => {
                return { ...acc, [curr.name]: curr.name };
              }, {});
              this.fonts.next(formatOptions);
              resolve(true);
            }
          }
        }
        reject(false);
      });
    });
  }

  /**
   * Retourne les positions disponibles
   * @returns
   */
  public getPositions(): Promise<boolean> {
    return new Promise((resolve, reject) => {

      this.positions.next({
        ul: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_POSITION_UL',
        uc: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_POSITION_UC',
        ur: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_POSITION_UR',
        cl: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_POSITION_CL',
        cc: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_POSITION_CC',
        cr: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_POSITION_CR',
        ll: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_POSITION_LL',
        lc: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_POSITION_LC',
        lr: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_POSITION_LR',
        auto: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_POSITION_AUTO'
      });
      resolve(true);
    });
  }

  /**
   * Retourne les alignements disponibles
   * @returns
   */
  public getAligns(): Promise<boolean> {
    return new Promise((resolve, reject) => {

      this.aligns.next({
        left: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_ALIGN_LEFT',
        center: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_ALIGN_CENTER',
        right: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_ALIGN_RIGHT'
      });
      resolve(true);
    });
  }

  /**
   * Retourne les types disponibles
   * @returns
   */
  public getTypes(): Promise<boolean> {
    return new Promise((resolve, reject) => {
      this.types.next({
        bitmap: 'bitmap',
        truetype: 'truetype',
      });
      resolve(true);
    });
  }

  /**
   * Retourne les forces disponibles
   * @returns
   */
  public getForces(): Promise<boolean> {
    return new Promise((resolve, reject) => {
      this.forces.next({
        true: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_FORCE_TRUE',
        false: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_FORCE_FALSE',
        group: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_FORCE_GROUP'
      });
      resolve(true);
    });
  }

  /** Retourne les types de cap disponible */
  public getLineCap(): Promise<boolean> {
    return new Promise((resolve, reject) => {
      this.lineCap.next({
        BUTT: 'COMPONENT_MAP_EDIT_SYMBOLOGY_BUTT',
        ROUND: 'COMPONENT_MAP_EDIT_SYMBOLOGY_ROUND',
        SQUARE: 'COMPONENT_MAP_EDIT_SYMBOLOGY_SQUARE'
      });
      resolve(true);
    });
  }

  /** Retourne les types de linejoin disponible */
  public getLinejoin(): Promise<boolean> {
    return new Promise((resolve, reject) => {
      this.linejoin.next({
        ROUND: 'COMPONENT_MAP_EDIT_SYMBOLOGY_ROUND',
        MITER: 'COMPONENT_MAP_EDIT_SYMBOLOGY_MITER',
        BEVEL: 'COMPONENT_MAP_EDIT_SYMBOLOGY_BEVEL',
        NONE: 'COMPONENT_MAP_EDIT_SYMBOLOGY_NONE',
      });
      resolve(true);
    });
  }

  /**
   * Traduit une expression en objet
   * @param expression
   * @returns
   */
  public expressionTaductionArrToString(expression): string {
    return typeof expression == 'object' ? expression.reduce((acc, curr) =>
      `${acc} [${curr.table}] ${curr.operator} ${curr.value} ${curr.sql_operator || ''}`, '') : '';
  }

  /**
   * Retourne un pourcentage arrondi
   * @param iVal
   * @returns
   */
  public getPrettyPercent(iVal: number): string {
    return `${iVal ? Math.round(iVal * 100) : 100}%`;
  }

  /**
   * Retourne l'observable obsRunTest
   * @returns
   */
  public getObsRunTest() {
    return this.obsRunTest;
  }

  /**
   * Lance le test de couche en envoyant une nouvelle valeur sur l'observable obsRunTest
   * @param oJsonSymbologyExpression
   */
  public runTest(oJsonSymbologyExpression: any) {
    this.obsRunTest.next(oJsonSymbologyExpression);
  }

  /**
   * Récupère la liste des types qui peuvent être ajoutéavec leur libellé
   */
  public getLabelsList(currGeomType: string, isLabel: boolean = false) {
    const sVisibilityLabel = 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_PARAMS_VISIBILITY';
    if (!isLabel) {
      const oLabels: any = {
        visibility: sVisibilityLabel,
        opacity: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_PARAMS_OPACITY',
        symbolOffset: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_PARAMS_SYMBOL_OFFSET',
        symbolSizeVariation: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_PARAMS_SYMBOL_SIZE_VARIATION',
      };

      if (currGeomType === 'point') {
        oLabels.symbolOutline = 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_PARAMS_SYMBOL_OUTLINE';
      }

      if (currGeomType === 'polygon') {
        oLabels.polygonSymbolRepeat = 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_PARAMS_POLYGON_SYMBOL_REPEAT';
        oLabels.polygonSymbolCenter = 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_PARAMS_POLYGON_SYMBOL_CENTER';
      }

      if (currGeomType === 'line') {
        oLabels.lineDash = 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_PARAMS_LINE_DASH';
        oLabels.lineSymbol = 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_PARAMS_LINE_SYMBOL';
      }

      return oLabels;
    } else {
      return {
        visibility: sVisibilityLabel,
        shadow: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABELS_SHADOW',
        shadowBackground: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABELS_SHADOW_BACKGROUND',
        wrap: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABELS_WRAP_WORD',
        sizeAndPosition: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABELS_SAS',
        outline: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABELS_OUTLINE',
        other: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABELS_OTHER'
      };
    }
  }

  /**
   * réucpère la liste des icons pour les containers
   */
  public getIcons() {
    return {
      visibility: '',
      opacity: '',
      symbolOutline: '',
      symbolOffset: '',
      symbolSizeVariation: '',
      shadow: '',
      shadowBackground: '',
      wrap: '',
      sizeAndPosition: '',
      outline: '',
      other: ''
    };
  }

  /**
   * Set la valeur dans otherFields en fonction du chemin dans event
   */
  public setOtherFieldValue(event: StyleSubFields, object: any) {
    const sEventFirstKey = Object.keys(event)[0];
    const sEventSecondKey = Object.keys(event[sEventFirstKey])[0];
    const sEventValue = event[sEventFirstKey][sEventSecondKey];
    for (const key in object.otherFields) {
      if (Object.prototype.hasOwnProperty.call(object.otherFields, key)) {
        if (key === sEventFirstKey) {
          object.otherFields[key][sEventSecondKey] = sEventValue;
        }
      }
    }
    return object;
  }

  /**
   * Créér un géométrie par défaut
   * @param currGeomType
   */
  public setDefaultValue(currGeomType: string) {
    if (currGeomType == 'point') {
      return {
        symbol: 'circle',
        fillcolor: '#000000'
      };
    }
    if (currGeomType == 'line') {
      return {
        width: '10',
        fillcolor: '#000000'
      };
    }
    if (currGeomType == 'polygon') {
      return {
        width: '2',
        fillcolor: '#FF0000',
        outlineColor: '#000000'
      };
    }
    return {};
  }

  /**
   * Vérifie si les conditions sont remplis
   * @param key
   * @param oConditions
   */
  public isReadOnly(key: string, oConditions: any) {
    for (const k in oConditions) {
      if (Object.prototype.hasOwnProperty.call(oConditions, k)) {
        const condition = oConditions[k];
        if (k === key && condition) {
          return true;
        }
      }
    }
    return false;
  }
}