<?php

namespace App\Module\Vmap\WebService\Controller;

use App\Module\Vitis\WebService\Controller\GenericController;
use App\Module\Vitis\WebService\Repository\GenericRepository;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\DataAccessor\DatabaseConnection;
use App\Module\Vitis\WebService\Service\FileAccessor\FileManager;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\VitisRequest;
use App\Module\Vmap\WebService\Entity\AbstractLayer;
use App\Module\Vmap\WebService\Entity\LayerFile;
use App\Module\Vmap\WebService\Entity\LayerOSM;
use App\Module\Vmap\WebService\Entity\LayerPostgres;
use App\Module\Vmap\WebService\Entity\LayerVectorTiles;
use App\Module\Vmap\WebService\Entity\LayerWMS;
use App\Module\Vmap\WebService\Entity\LayerWMTS;
use App\Module\Vmap\WebService\Entity\LayerXYZ;
use App\Module\Vmap\WebService\Entity\Map;
use App\Module\Vmap\WebService\Service\MapServerService;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller pour le MapJSON de vMap.
 *
 */
class MapJSONController extends GenericController {
  // Exemple de JSON
  //
  // {
  //   type: string;
  //   tree?: any; // arbre de définition de la carte
  //   map_id?: number;
  //   center?: any;
  //   extent?: any;
  //   source?: any; // utilisé pour cartes Bing
  //   circle_radius?: number;
  //   contour_color?: string;
  //   contour_size?: number;
  //   controls: {
  //     map_projection?: boolean,
  //     mobile_gps_zoom?: boolean,
  //     mouse_position?: boolean,
  //     scale_line?: boolean,
  //     zoom_controls?: boolean,
  //     location_controls?: boolean,
  //     attribution?: boolean, // True par défaut
  //   };
  //   coord_accuracy?: number;
  //   draw_color?: string;
  //   interactions: {
  //     business_objects?: boolean,
  //     draw_point?: boolean,
  //     draw_line?: boolean,
  //     draw_polygon?: boolean,
  //     full_screen?: boolean,
  //     modify?: boolean,
  //     multi_geometry?: boolean,
  //     remove?: boolean,
  //     remove_all?: boolean,
  //   };
  //   output_format?: string;
  //   proj?: string;
  //   map_geocoders?: any;
  //   map_default_geocoders?: string;
  // }

  /**
   * @var MapServerService
   */
  private $oMapServerService;

  /**
   * @var Properties
   */
  private $oPropertiesSrvc;

  /**
   * @var LocalAccessor
   */
  private $oLocalFileAccessor;

  /**
   * Initialise le controler et instancie les services requis
   * @param Properties $oPropertiesSrvc Service d'accès aux properties
   * @param VitisRequest $oVitisRequest Service d'abstraction de la requête Symfony
   * @param Route $oRouteConfig Service d'accès à la configuration de la route
   * @param LoggerManager $oLoggerManager Service d'utilsation des logs
   * @param DatabaseConnection $oDatabaseConnection Service d'accès à la base de données
   * @param FileManager $oFileManager Service de gestion des fichiers
   * @param MapServerService $oMapServerService
   */
  public function __construct(Properties $oPropertiesSrvc, VitisRequest $oVitisRequest, Route $oRouteConfig, LoggerManager $oLoggerManager, DatabaseConnection $oDatabaseConnection, FileManager $oFileManager, MapServerService $oMapServerService) {
    parent::__construct($oPropertiesSrvc, $oVitisRequest, $oRouteConfig, $oLoggerManager, $oDatabaseConnection, $oFileManager);
    $this->oMapServerService = $oMapServerService;
    $this->oPropertiesSrvc = $oPropertiesSrvc;
    $this->oLocalFileAccessor = $oFileManager->getLocalFileAccessor();
  }

  /**
   * Récupère une carte au format MapJSON
   * @param mixed $id id de la carte à trouver
   * @return Response réponse à retourner par l'API
   */
  public function getData($id): Response {
    $aData = $this->requestToArray();

    // chargement de la connexion à la base de donnée
    $oDataAccessor = $this->getDataAccessor();
    $oEntity = $oDataAccessor->getEntity($id, $aData);

    if ($oEntity['status'] === 0) {
      return $this->oResponse;
    }

    // Filtre à appliquer sur les couches
    $aFilter = null;
    if (array_key_exists('filter', $aData)) {
      $tmpFilter = json_decode($aData['filter'], true);
      if (is_array($tmpFilter) && count($tmpFilter) > 0) {
        $aFilter = $tmpFilter;
      }
    }

    $oMapTree = $this->getMapTree($oEntity['data'], $aFilter);

    if ($oMapTree === null) {
      return $this->formatResponse([
        'status' => 0,
        'message' => 'Error during get',
        'errorMessage' => 'ERROR_GENERIC_CONTROLLER_GET',
      ], Response::HTTP_UNAUTHORIZED);
    }

    $oMap = [
      'name' => $oEntity['data']['name'],
      'children' => $oMapTree,
      'maptools' => $this->getMapTools($oEntity['data']),
      'options' => [
        'optimized' => true,
        'default_layer_location_id' => $oEntity['data']['default_layer_location_id'],
      ],
    ];

    $oEntity['data'] = $oMap;

    $this->formatResponse($oEntity, Response::HTTP_OK);
    return $this->oResponse;
  }

  /**
   * Retourne maptools parsé
   * @param array $aMap
   * @return array
   */
  private function getMapTools(array $aMap): array{
    try {
      return json_decode($aMap['maptools'], true);
    } catch (\Throwable $th) {
      return [];
    }
  }

  /**
   * Retourne l'arbre définition de la carte
   * @param array $aMap
   * @param array $aFilter filtre à appliquer sur les couches
   * @return array|null
   */
  private function getMapTree(array $aMap, ?array $aFilter = null): ?array{
    $aTree = [[
      'view' => [
        'extent' => explode('|', $aMap['extent']),
        'projection' => $aMap['coordsys_epsg'],
      ],
    ]];

    $mPublishedLayers = $this->publishLayers($aMap, $aFilter);

    if ($mPublishedLayers === null) {
      return null;
    }

    $aLayers = $this->orderLayersByTheme($mPublishedLayers);

    // Sort les valeurs par le champ 'name'
    usort($aLayers, function ($a, $b) {
      return strcmp($a['name'], $b['name']);
    });

    foreach ($aLayers as $aLayer) {
      $aTree[] = $aLayer;
    }

    return $aTree;
  }

  /**
   * Retourne la définition des couches (avec héritage) d'une carte
   * @param array $aMap
   * @return array|null
   */
  private function getMapLayers(array $aMap): ?array{
    // Récupération du contenu de layer_map
    try {
      $oMapRepository = $this->oDatabaseConnection->getAdminEntityManager()->getRepository(Map::class);

      if (!$oMapRepository instanceof GenericRepository) {
        $this->oLoggerManager->get('API_ERROR')->error('Bad type repository');
        return null;
      }

      $oMap = $oMapRepository->findById($aMap['map_id'])[0];
    } catch (\Exception $e) {
      $this->oLoggerManager->get('API_ERROR')->error('Cannot find map with id : ' . $aMap['map_id']);
      return null;
    }

    $aLayers = $oMap->getLayersWithEntity();
    $oLayerRepository = $this->oDatabaseConnection->getAdminEntityManager()->getRepository(AbstractLayer::class);

    if (!$oLayerRepository instanceof GenericRepository) {
      $this->oLoggerManager->get('API_ERROR')->error('Bad type repository');
      return null;
    }

    // Remplit le reste des infos de la couche
    foreach ($aLayers as $key => $aLayerMapDef) {
      if (isset($aLayerMapDef['layer_id'])) {
        $oLayer = $oLayerRepository->findById($aLayerMapDef['layer']->getLayerId())[0];
        unset($aLayerMapDef['layer']);

        // Attributs de AbstractLayer
        $oLayerRoute = $this->oRouteConfig->getRequestParameters('vmap_layer');
        $aLayerDef = $oLayer->jsonSerialize($oLayerRoute);

        // is_location : si dans layer vaut true alors dépend de layer_map (true=true, false=false)
        if (array_key_exists('is_location', $aLayerDef) && $aLayerDef['is_location'] === true) {
          if (array_key_exists('is_location', $aLayerMapDef) && $aLayerMapDef['is_location'] === true) {
            $aLayerDef['is_location'] = true;
          } else {
            $aLayerDef['is_location'] = false;
          }
        } else {
          $aLayerDef['is_location'] = false;
        }

        // is_map_selection : si dans layer vaut true alors dépend de layer_map (true=true, false=false)
        if (array_key_exists('is_map_selection', $aLayerDef) && $aLayerDef['is_map_selection'] === true) {
          if (array_key_exists('is_map_selection', $aLayerMapDef) && $aLayerMapDef['is_map_selection'] === true) {
            $aLayerDef['is_map_selection'] = true;
          } else {
            $aLayerDef['is_map_selection'] = false;
          }
        } else {
          $aLayerDef['is_map_selection'] = false;
        }

        // is_data_grid : si dans layer vaut true alors dépend de layer_map (true=true, false=false)
        if (array_key_exists('is_data_grid', $aLayerDef) && $aLayerDef['is_data_grid'] === true) {
          if (array_key_exists('is_data_grid', $aLayerMapDef) && $aLayerMapDef['is_data_grid'] === true) {
            $aLayerDef['is_data_grid'] = true;
          } else {
            $aLayerDef['is_data_grid'] = false;
          }
        } else {
          $aLayerDef['is_data_grid'] = false;
        }

        // gestion definition_sym_json
        if (array_key_exists('definition_sym_json', $aLayerDef) && array_key_exists('definition_sym_json_map', $aLayerMapDef)) {
          if (!empty($aLayerMapDef['definition_sym_json_map'])) {
            $aLayerDef['definition_sym_json'] = $aLayerMapDef['definition_sym_json_map'];
          }
        }

        // Ajoute les droits d'utilisateur sur les couches postgres
        if ($aLayerDef['layertype'] === LayerPostgres::TYPE) {
          $oEm = $this->oDatabaseConnection->getUserEntityManager();

          if (array_key_exists('datasource', $aLayerDef)) {
            // création de l'entity manager si datasource non privée
            if ($aLayerDef['datasource']['pg_public'] == true) {
              $oEm = $this->oDatabaseConnection->createExternalEntityManager($aLayerDef['database'], $aLayerDef['datasource']['pg_login'], $aLayerDef['datasource']['pg_password'], DatabaseConnection::SGBD, $aLayerDef['datasource']['pg_server'], $aLayerDef['datasource']['pg_port']);
            } else if ($aLayerDef['database'] !== $this->oProperties->get('db_name')) {
              $oEm = $this->oDatabaseConnection->createExternalEntityManager($aLayerDef['database']);
            }
          }

          $aPrivilege = $this->oDatabaseConnection->getQueryAccessor($oEm)->getTableRights($aLayerDef['schema'], $aLayerDef['pg_table']);
          if (is_array($aPrivilege) && !empty($aPrivilege)) {
            $aLayerDef = array_merge($aLayerDef, [
              'have_select_right' => $aPrivilege['SELECT'],
              'have_insert_right' => $aPrivilege['INSERT'],
              'have_update_right' => $aPrivilege['UPDATE'],
              'have_delete_right' => $aPrivilege['DELETE'],
            ]);
          } else {
             $aLayerDef = array_merge($aLayerDef, [
              'have_select_right' => false,
              'have_insert_right' => false,
              'have_update_right' => false,
              'have_delete_right' => false,
            ]);
          }
        }

        $aLayers[$key] = array_merge($aLayerMapDef, $aLayerDef);
      }
    }

    return $aLayers;
  }

  /**
   * Transforme les couches en calques
   * @param array $aMap
   * @param array $aFilter filtre à appliquer sur les couches
   * @return array|null
   */
  private function publishLayers(array $aMap, ?array $aFilter = null): ?array{
    $aLayers = $this->getMapLayers($aMap);
    if ($aLayers === null) {
      return null;
    }

    // Publication du flux
    $bHasLayerPostgres = false;
    $bHasLayerFile = false;
    foreach ($aLayers as $key => $aLayer) {
      if ($aLayer['layertype'] === LayerPostgres::TYPE) {
        $bHasLayerPostgres = true;
      } else if ($aLayer['layertype'] === LayerFile::TYPE) {
        $bHasLayerFile = true;
      }
    }

    // Générer le fichier fonts.list avec les données en bdd
    $this->oMapServerService->generateFontList();

    $sMapfileName = '';
    if ($bHasLayerPostgres || $bHasLayerFile) {
      $sMapfileName = $this->oMapServerService->createMapfile($aMap, $aLayers, $aFilter, false);
    }

    $aLayers = array_filter($aLayers, function ($k) {
      return !is_string($k);
    }, ARRAY_FILTER_USE_KEY);

    // Couches au format OpenLayers
    foreach ($aLayers as $key => $aLayer) {
      switch ($aLayer['layertype']) {
        case LayerWMS::TYPE:
          $aLayers[$key] = LayerWMS::getLayerParams($aLayer);
          break;
        case LayerWMTS::TYPE:
          $aLayers[$key] = LayerWMTS::getLayerParams($aLayer);
          break;
        case LayerOSM::TYPE:
          $aLayers[$key] = LayerOSM::getLayerParams($aLayer);
          break;
        case LayerXYZ::TYPE:
          $aLayers[$key] = LayerXYZ::getLayerParams($aLayer);
          break;
        case LayerVectorTiles::TYPE:
          $aLayers[$key] = LayerVectorTiles::getLayerParams($aLayer);
          break;
        case LayerFile::TYPE:
          $aLayers[$key] = $this->getFileLayerParams($aLayer, $sMapfileName);
          break;
        case LayerPostgres::TYPE:
          $aLayers[$key] = $this->getPostgresLayerParams($aLayer, $sMapfileName);
          if (empty($aLayer['definition_sym_json']) || $aLayers[$key]['is_geometry_layer'] === false) {
            $aLayers[$key]['is_map_hidden'] = true;
            $aLayers[$key]['is_geometry_layer'] = false;
          }
          break;
        default:
          break;
      }
    }

    return $aLayers;
  }

  /**
   * Renvoie les paramètres spécifiques a OSM
   * @param array $aLayer
   * @param string $sMapfileName
   * @return array
   */
  private function getFileLayerParams(array $aLayer, string $sMapfileName): array{
    $sOnlineResourceUrl = str_replace('@file', $sMapfileName, MapServerService::PRIVATE_WMS_MAIN_URL);
    $sOnlineResourceUrl = $this->oPropertiesSrvc->replacePropertiesInString($sOnlineResourceUrl);
    $aLayer['url'] = $sOnlineResourceUrl;
    return LayerFile::getLayerParams($aLayer);
  }

  /**
   * Renvoie les paramètres spécifiques a OSM
   * @param array $aLayer
   * @param string $sMapfileName
   * @return array
   */
  private function getPostgresLayerParams(array $aLayer, string $sMapfileName): array{
    $sOnlineResourceUrl = str_replace('@file', $sMapfileName, MapServerService::PRIVATE_WMS_MAIN_URL);
    $sOnlineResourceUrl = $this->oPropertiesSrvc->replacePropertiesInString($sOnlineResourceUrl);
    $aLayer['url'] = $sOnlineResourceUrl;
    return LayerPostgres::getLayerParams($aLayer);
  }

  /**
   * Ordonne les couches sous forme d'arbre en fonction de leur thème
   * @param array $aLayers
   * @return array
   */
  private function orderLayersByTheme(array $aLayers): array{
    $aReturn = [];
    $aThemes = [];

    $aThemes = [];
    foreach ($aLayers as $aLayer) {
      $sTheme = 'null';
      if (!empty($aLayer['theme']) && !empty($aLayer['theme']['label'])) {
        $sTheme = $aLayer['theme']['label'];
      }
      $aTheme = explode('/', $sTheme);
      if (!in_array($aTheme, $aThemes)) {
        $aThemes[] = $aTheme;
      }
    }

    foreach ($aThemes as $aTheme) {
      if (count($aTheme) > 1) {
        $aReturn[] = $this->recursiveGetThemeLayers($aTheme, $aTheme, $aLayers);
      } else {
        $sTheme = $aTheme[0];
        if ($sTheme !== 'null') {
          $groupedLayers = $this->findLayersByTheme($aTheme, $aLayers);
          // Sort les valeurs par le champ 'name'
          usort($groupedLayers, function ($a, $b) {
            return strcmp($a['name'], $b['name']);
          });

          $aReturn[] = [
            'name' => $sTheme,
            'children' => $groupedLayers,
          ];
        } else {
          $aWithoutThemeLayers = $this->findLayersByTheme($aTheme, $aLayers);
          foreach ($aWithoutThemeLayers as $aWithoutThemeLayer) {
            $aReturn[] = $aWithoutThemeLayer;
          }
        }
      }
    }

    return $aReturn;
  }

  /**
   * Permet la récupération des couches dans des sous-themes
   * @param array $aTheme
   * @param array $aCompleteTheme
   * @param array $aLayers
   * @return array
   */
  private function recursiveGetThemeLayers(array $aTheme, array $aCompleteTheme, array $aLayers): array{
    $sTheme = $aTheme[0];
    array_shift($aTheme);

    $aTree = [
      'name' => $sTheme,
      'children' => [],
    ];

    if (count($aTheme) > 0) {
      $aTree['children'] = [$this->recursiveGetThemeLayers($aTheme, $aCompleteTheme, $aLayers)];
    } else {
      $aTree['children'] = $this->findLayersByTheme($aCompleteTheme, $aLayers);
    }

    return $aTree;
  }

  /**
   * Trouve les layers en fonction du thème passé en paramètre
   * @param array $aTheme
   * @param array $aLayers
   * @return array
   */
  private function findLayersByTheme(array $aTheme, array $aLayers): array{
    $aResult = [];

    foreach ($aLayers as $aLayer) {
      $sLayerTheme = 'null';
      if (!empty($aLayer['theme']) && !empty($aLayer['theme']['label'])) {
        $sLayerTheme = $aLayer['theme']['label'];
      }

      if (join('/', $aTheme) === $sLayerTheme) {
        $aResult[] = $aLayer;
      }
    }

    return $aResult;
  }
}
