<?php

namespace App\Module\Vmap\WebService\Controller;

use App\Module\Vitis\WebService\Controller\GenericController;
use App\Module\Vitis\WebService\Repository\GenericRepository;
use App\Module\Vitis\WebService\Service\Serializer\VitisSerializer;
use App\Module\Vmap\WebService\Entity\AbstractPublication;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Serializer\Normalizer\AbstractNormalizer;

/**
 * Controller pour le datasource treeview de vMap.
 *
 * @author Laurie Ambroise <laurie.ambroise@veremes.com>
 */
class PublicationController extends GenericController {
  /**
   * Créé plusieurs entités pour une route de type POST Module/Objet.
   * @return Response réponse à retourner par l'API
   */
  public function postDatas(): Response {
    $aData = $this->requestToArray();
    if (empty($aData)) {
      $aReturn = ['status' => 0];
      $aReturn['message'] = 'data parameter is missing';
      $aReturn['errorMessage'] = 'ERROR_MISSING_DATA_PARAMETER';
      return $this->formatResponse($aReturn, Response::HTTP_NOT_ACCEPTABLE);
    }

    // chargement de la connexion à la base de donnée
    $oDataAccessor = $this->getDataAccessor();
    // Sauvegarde en base de donnée
    $oEntity = $oDataAccessor->insertEntities($aData['data']); // Uniquement en ORM
    if ($oEntity['status'] == 0) {
      $oEntity['message'] = 'Error during update';
      $oEntity['errorMessage'] = 'ERROR_GENERIC_CONTROLLER_INSERT';
      return $this->formatResponse($oEntity, Response::HTTP_UNAUTHORIZED);
    }
    $this->formatResponse($oEntity);

    return $this->oResponse;
  }

  /**
   * Récupère une entité AbstractPublication
   * @return Response réponse à retourner par l'API
   */
  public function getData($id): Response {
    $aData = $this->requestToArray();

    // Attribution de l'entité oPublication grâce à son publicationtype
    $oPublication = $this->returnPublicationClass($id)['oPublication'];

    if (empty($oPublication)) {
      $this->oLoggerManager->get('API_ERROR')->error('Cannot find publication with id : ' . $id);
      $oResponse = [
        'status' => 0,
        'message' => 'Error during get',
        'errorMessage' => 'ERROR_GENERIC_CONTROLLER_GET',
      ];
      return $this->formatResponse($oResponse, Response::HTTP_UNAUTHORIZED);
    }

    $aReturn = [
      'data' => $oPublication,
      'status' => 1,
    ];

    $aSerializerContext = [];
    if (isset($aData['attributs'])) {
      $aSerializerContext[AbstractNormalizer::ATTRIBUTES] = explode('|', $aData['attributs']);
    }

    return $this->formatResponse($aReturn, Response::HTTP_OK, $aSerializerContext);
  }

  /**
   * Met à jour une entité AbstractPublication
   * @return Response réponse à retourner par l'API
   */
  public function updateData($id): Response {
    $aData = $this->requestToArray();
    $oPublication = $this->returnPublicationClass($id);
    $oEntity = $oPublication['oPublication'];

    $oEntity->setDatemaj(new \DateTime());

    if (empty($oEntity)) {
      $this->oLoggerManager->get('API_ERROR')->error('Cannot find publication with id : ' . $id);
      $oResponse = [
        'status' => 0,
        'message' => 'Error during get',
        'errorMessage' => 'ERROR_GENERIC_CONTROLLER_GET',
      ];
      return $this->formatResponse($oResponse, Response::HTTP_UNAUTHORIZED);
    }

    $parameters = $this->getRequestParameters();
    $oDataAccessor = $this->getDataAccessor();
    $parameters['options']['entity'] = $oPublication['class'];
    $oDataAccessor->setOptions($parameters['options']);

    // Patch booléens
    foreach ($aData as $key => $value) {
      if ($value == 'false') {
        $aData[$key] = false;
      }
      if ($value == 'true') {
        $aData[$key] = true;
      }
    }

    $oSerializer = new VitisSerializer();
    $oSerializer->deserialize(json_encode($aData), $oPublication['class'], ['object_to_populate' => $oEntity]);

    // Sauvegarde en base de donnée
    $this->oDatabaseConnection->getAdminEntityManager()->persist($oEntity);
    $this->oDatabaseConnection->getAdminEntityManager()->flush();
    // rafraichissement en cas de modifications interne à la base
    $this->oDatabaseConnection->getAdminEntityManager()->refresh($oEntity);

    $aReturn = [
      'data' => $oEntity,
      'status' => 1,
    ];

    return $this->formatResponse($aReturn, Response::HTTP_OK);
  }

  /**
   * Ajoute une entité AbstractPublication
   * @return Response réponse à retourner par l'API
   */
  public function postData(): Response {
    $aData = $this->requestToJson();
    $publicationtype = $this->requestToArray()['publicationtype'];
    if (empty($publicationtype)) {
      $this->oLoggerManager->get('API_ERROR')->error('Cannot find publicationtype');
      $oResponse = [
        'status' => 0,
        'message' => 'Error during post',
        'errorMessage' => 'ERROR_GENERIC_CONTROLLER_POST',
      ];
      return $this->formatResponse($oResponse, Response::HTTP_UNAUTHORIZED);
    }

    $oPublication = $this->returnPublicationClass(null, $publicationtype);

    $parameters = $this->getRequestParameters();
    $oDataAccessor = $this->getDataAccessor();
    $parameters['options']['entity'] = $oPublication['class'];
    $oDataAccessor->setOptions($parameters['options']);

    $oPublicationEntity = $oDataAccessor->insertEntity($aData);

    $aReturn = $oPublicationEntity;

    return $this->formatResponse($aReturn, Response::HTTP_OK);
  }

  /**
   * Récupère le repository associé au publicationtype de la couche puis retourne l'entité
   * @param mixed|null $id
   * @param string|null $publicationtype
   * @return array Tableau avec l'entité AbstractPublication et sa class
   */
  private function returnPublicationClass($id = null, ?string $publicationtype = null): array{
    // Récupération du repository parent AbstractPublication
    $oPublicationRepository = $this->oDatabaseConnection->getAdminEntityManager()->getRepository(AbstractPublication::class);
    if (!$oPublicationRepository instanceof GenericRepository) {
      throw new \Exception("Repository is not GenericRepository", 1);
    }
    if ($id !== null) {
      // Récupération de l'entité avec le parent
      $oPublication = $oPublicationRepository->findById($id)[0];

      // Attribution de la class associé au publicationtype
      $publicationtype = $oPublication->getPublicationtype();
    }

    $class = AbstractPublication::getPublicationChildrenPublicationClassName($publicationtype);
    $aReturn = [
      'class' => $class,
    ];

    if ($id !== null) {
      // Récupération de l'entité global avec les champs spécifique à son type
      $oPublicationChildRepository = $this->oDatabaseConnection->getAdminEntityManager()->getRepository($class);
      if (!$oPublicationChildRepository instanceof GenericRepository) {
        throw new \Exception("Repository is not GenericRepository", 1);
      }
      $aReturn['oPublication'] = $oPublicationChildRepository->findById($id)[0];
    }

    // Return de l'entité et de la class du publicationtype si $id existe
    return $aReturn;
  }
}
