import { Component, OnInit, EventEmitter, Compiler, Injector, ViewContainerRef, ViewChild, HostListener } from '@angular/core';
import { Router, UrlSegment, NavigationEnd, NavigationCancel, ActivatedRoute } from '@angular/router';
import { Location, LocationStrategy, PathLocationStrategy } from '@angular/common';
declare var $: any;

// Vitis
import { VitisCheck, LocalStorage, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { environment } from '../../../../../environments/environment';
import { AuthService } from '../../services/auth/auth.service';
import { AppConfig } from 'src/app/vitis/app.model';
import { MapCoreSmallScreenService } from 'src/app/vitis/map-core/services/map-core-small-screen/map-core-small-screen.service';

@Component({
  selector: 'app-dashboard',
  templateUrl: './dashboard.component.html',
  styleUrls: ['./dashboard.component.scss'],
  providers: [Location, { provide: LocationStrategy, useClass: PathLocationStrategy }],
})
export class DashboardComponent implements OnInit {

  constructor(
    private auth: AuthService,
    private route: ActivatedRoute,
    private router: Router,
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private localStorage: LocalStorage,
    private compiler: Compiler,
    private injector: Injector,
    private mapCoreSmallScreenService: MapCoreSmallScreenService,
  ) { }

  @ViewChild('container', { read: ViewContainerRef }) container: ViewContainerRef;

  public appConfig: AppConfig;
  public modes: Array<any>;
  public headers: Array<any>;
  public footers: Array<any>;
  public appName: string = 'Vitis';
  public appLogo: any = {
    type: 'icon',
    class: 'icon-vitis-vitis'
  };
  public sidenavOpen: boolean;
  public currentMode: any;
  public currentObject: any;
  public currentEntity: any;
  public currentSections: Array<any>;
  public userPrivileges: Array<any>;
  public sidenavOpenChange: EventEmitter<any> = new EventEmitter();
  public modalFormMode: any = {};
  public userInfos: any;
  public enableSignUp: boolean = false;
  public showConnectionButtons: boolean = false;

  public isSmallScreen;

  ngOnInit() {

    this.isSmallScreen = window.innerWidth <= 768;
    this.mapCoreSmallScreenService.smallScreenObservable.subscribe(isSmallScreen => {
      this.isSmallScreen = isSmallScreen;

      if (isSmallScreen == true) {
        this.setSideNavValue(false);
      }

      this.setNavigationBar(isSmallScreen);
    });

    // Structure app.json
    this.appConfig = environment.appStruct;

    // Récuperation des infos utilisateur
    this.userInfos = this.auth.getUserInfos();
    this.auth.getUserInfosObs().subscribe((userInfos: any) => {
      this.userInfos = userInfos;
    });

    // Nom de l'application
    if (environment.appStruct.name) {
      this.appName = environment.appStruct.name;
    }

    // Logo de l'application
    if (environment.appStruct.logo) {
      if (environment.appStruct.logo.type) {
        this.appLogo = environment.appStruct.logo;
      }
    }

    // Boutons de connexion / inscription
    if (environment.properties.public_connect_available === true &&
      environment.properties.public_connect_only !== true) {
      this.showConnectionButtons = true;

      // Bouton signup
      if (environment.properties.sign_up === true) {
        this.enableSignUp = true;
      }
    }

    // Menus des différents modes
    this.modes = environment.appStruct.modes;

    // Boutons d'entête
    this.headers = environment.appStruct.headers;

    // Footer elements
    this.footers = environment.appStruct.footers;

    // Mode actuellement selectionné
    this.currentMode = this.getCurrentMode();

    // Objet actuellement selectionné
    this.currentObject = this.getCurrentObject();

    // Enregistrement actuellement selectionné
    this.currentEntity = this.getCurrentEntity();

    // Sections actuellement utilisées
    this.loadCurrentSections();

    // Écoute les changements de routes
    this.router.events.subscribe(event => {
      if (event instanceof NavigationEnd ||
        event instanceof NavigationCancel) {
        if (event instanceof NavigationCancel) {
          console.warn('Navigation cancelled', event);
        }
        this.currentMode = this.getCurrentMode();
        this.currentObject = this.getCurrentObject();
        this.currentEntity = this.getCurrentEntity();
        this.loadCurrentSections();
        this.collapseModes(this.modes);
      }
    });

    // Collapse tous les modes à l'exception de celui en cours
    this.collapseModes(this.modes);

    // Permet de gérer les dorpdown menus des headers
    this.handleHeaderDorpdownMenus_();

    // Permet de gérer la modale formulaire
    this.handleModalForm_();

    this.setNavigationBar(this.isSmallScreen);
  }

  /**
   * Set la valeur de la sidenav
   * @param isSmallScreen
   */
  private setNavigationBar(isSmallScreen) {
    if (isSmallScreen !== true) {
      // Survéillance des paramètres
      this.route.queryParams.subscribe(oParams => {

        // Affichage par défaut de la barre de navigation
        const changedSideNavOpen = this.getSideNavValue(oParams);
        setTimeout(() => {
          if (changedSideNavOpen !== this.sidenavOpen) {
            this.setSideNavValue(changedSideNavOpen);
          }
        });
      });

      // Affichage par défaut de la barre de navigation
      const sideNavOpen = this.getSideNavValue();
      if (sideNavOpen !== this.sidenavOpen) {
        this.setSideNavValue(sideNavOpen);
      }
    }
  }

  /**
   * Affiche / cache la barre latérale
   */
  public toggleSideNav() {
    if (this.isSmallScreen !== true) {
      this.setSideNavValue(!this.sidenavOpen);
    }
  }

  /**
   * Affiche la barre latérale
   */
  public openSideNav() {
    this.setSideNavValue(!this.isSmallScreen);
  }

  /**
   * Cache la barre latérale
   */
  public closeSideNav() {
    this.setSideNavValue(false);
  }

  /**
   * sidenavOpen setter
   * Permet de déclencher l'événement sidenavOpenChange
   * @param  {boolean} newVal
   */
  public setSideNavValue(newVal: boolean) {

    // Set la valeur
    this.sidenavOpen = newVal;

    // émet l'event sidenavOpenChange
    this.sidenavOpenChange.emit(this.sidenavOpen);

    // Mémorise dans localstorage
    this.localStorage.add(this.appName, {
      sidenavOpen: newVal
    });
  }

  /**
   * Retourne la valeur de sideNavOpen à utiliser
   * @param oParams
   */
  public getSideNavValue(oParams?: any): boolean {
    let sideNavopen = true;

    // Paramètres URL
    if (oParams &&
      this.vitisCheck.isDefAndNotNull(oParams.sidenav_open)) {
      sideNavopen = oParams.sidenav_open === 'true' ? true : false;
    } else {

      // Properties
      if (this.vitisCheck.isDefAndNotNull(environment.properties.default_show_sidenav)) {
        sideNavopen = environment.properties.default_show_sidenav === true ? true : false;
      }

      // LocalStorage
      if (this.appName) {
        const localStorageVal = this.localStorage.get(this.appName);
        if (this.vitisCheck.isDefAndNotNull(localStorageVal && localStorageVal.sidenavOpen)) {
          sideNavopen = localStorageVal.sidenavOpen == true ? true : false;
        }
      }

      // Cache la navbar s'il n'y a qu'un seul mode
      const aActiveModes = this.getActiveModes_(this.modes);
      if (aActiveModes.length === 1) {
        sideNavopen = false;
      }
    }

    return sideNavopen;
  }

  /**
   * Ouvre un mode
   * @param  {any} mode
   */
  public openMode(mode: any) {
    if (this.vitisCheck.isDefAndNotNull(mode)) {
      if (this.vitisCheck.isDefAndNotNull(mode.route)) {
        this.router.navigate(['dashboard', mode.route], { queryParamsHandling: 'merge' });
      }
    }
  }

  /**
   * Retourne true/false si l'utilisateur a les droits pour le mode
   * @param  {any} mode
   * @returns boolean
   */
  public hasRights(mode: any): boolean {

    const aUserPrivileges = this.auth.getPrivileges();
    let bHasRight: boolean = false;

    if (this.vitisCheck.isArray(aUserPrivileges)) {
      if (mode.privilege) {
        // conversion de l'ancien en tableau pour uniformisé
        if (this.vitisCheck.isString(mode.privilege)) {
          mode.privilege = [mode.privilege];
        }
        // cherche le match de privilège entre les deux tableaux
        const aMatch: string[] = aUserPrivileges.filter(item => mode.privilege.indexOf(item) > -1);
        // si un privilege a match l'objet est visible
        bHasRight = aMatch.length > 0;
      } else {
        bHasRight = true;
      }
    }
    // si les droits sont déjà bloquant ça ne sert à rien de vérifier la visibilités
    if (!bHasRight) {
      return false;
    }

    if (mode.visible) {
      if (this.vitisCheck.isString(mode.visible)) {
        // remplacement des properties
        const sVisible: string = environment.replacePropertiesInString(mode.visible);
        // évaluation de l'expression
        bHasRight = eval(sVisible);
      }
    }

    return bHasRight;
  }

  /**
   * Retourne le mode en cours à partir de l'URL
   * @returns any
   */
  public getCurrentMode(): any {

    const modeRoute = this.getCurrentModeRoute_();
    if (this.vitisCheck.isDefAndNotNull(modeRoute)) {
      for (let i = 0; i < this.modes.length; i++) {
        if (this.modes[i].route === modeRoute) {
          return this.modes[i];
        }
      }
    }
  }

  /**
   * Fonction récursive qui permet de récupérer l'objet en cours
   * @param  {any} oObject?
   * @returns any
   */
  public getCurrentObject(oObject?: any): any {

    oObject = oObject ? oObject : this.getCurrentMode();

    if (oObject) {
      if (this.vitisCheck.isArray(oObject.objects)) {
        for (let i = 0; i < oObject.objects.length; i++) {
          const currentObject = this.getCurrentObject(oObject.objects[i]);
          if (this.vitisCheck.isDefAndNotNull(currentObject)) {
            return currentObject;
          }
        }
      } else {
        if (this.isRouteActive(oObject.route)) {
          return oObject;
        }
      }
    }
  }

  /**
   * Retourne la section en cours
   * @returns any
   */
  public getCurrentSection(): any {

    let oSection: any;
    const aSections = this.getCurrentSections();

    if (this.vitisCheck.isArray(aSections)) {
      for (let i = 0; i < aSections.length; i++) {
        if (this.isRouteActive(aSections[i].route)) {
          oSection = aSections[i];
        }

        if (aSections[i].options &&
          aSections[i].options.display === true ||
          aSections[i].options.update === true ||
          aSections[i].options.insert === true) {
          if (this.isRouteActive(aSections[i].route + '/:id/:form') ||
            this.isRouteActive(aSections[i].route + '/:form')) {
            oSection = aSections[i];
          }
        }
      }
    }

    return oSection;
  }

  /**
   * Récupère les sections en cours d'utilisation
   * @returns Array
   */
  public getCurrentSections(): Array<any> {

    let aSections = [];

    const oCurrentObject = this.getCurrentObject();

    // Vérifie que au moins une des sections soit active
    if (oCurrentObject && oCurrentObject.sections) {
      const getObjectsKeys: Array<string> = Object.keys(oCurrentObject.sections);
      let sections = [];
      if (getObjectsKeys[0] != '0') {
        sections = getObjectsKeys.reduce((acc, curr) => {
          return [...new Set([...acc, ...oCurrentObject.sections[curr]])];
        }, []);
        if (oCurrentObject.options &&
          oCurrentObject.options.values &&
          oCurrentObject.options.map_field &&
          oCurrentObject.options.values[oCurrentObject.options.map_field]) {
          if (oCurrentObject.sections &&
            oCurrentObject.sections[oCurrentObject.options.values[oCurrentObject.options.map_field]]) {
            sections = oCurrentObject.sections[oCurrentObject.options.values[oCurrentObject.options.map_field]];
          }
        }
      }
      sections = sections.length ? sections : oCurrentObject.sections;
      if (this.vitisCheck.isArray(sections)) {

        // Vérifie si au moins une section est active
        let bIsSectionActive: boolean = false;
        for (let i = 0; i < sections.length; i++) {
          const oSection = sections[i];
          if (this.isRouteActive(oSection.route)) {
            bIsSectionActive = true;
          }

          if (oSection.options &&
            (oSection.options.display === true ||
              oSection.options.update === true ||
              oSection.options.insert === true)) {
            if (this.isRouteActive(oSection.route + '/:id/:form') ||
              this.isRouteActive(oSection.route + '/:form')) {
              bIsSectionActive = true;
            }
          }
        }

        if (bIsSectionActive) {
          aSections = sections;
        }
      }
    }

    return aSections;
  }

  /**
   * Charge la liste des sections en cours d utilisation
   */
  public loadCurrentSections() {
    return new Promise((resolve, reject) => {

      this.currentSections = [];
      const oCurrentObject = this.getCurrentObject();
      if (oCurrentObject && oCurrentObject.sections) {
        const getObjectsKeys: Array<string> = Object.keys(oCurrentObject.sections);

        // Sections dynamiques
        if (getObjectsKeys[0] != '0') {
          if (oCurrentObject.options &&
            oCurrentObject.options.values) {
            this.currentSections = this.getCurrentSections();
          } else {
            if (this.currentEntity && this.currentEntity.id) {

              // Chargement des valeurs
              this.vitisRequest.ajaxRequestPromise({
                method: 'GET',
                url: this.vitisRequest.getApiRessourceUrl(oCurrentObject.options.ressource, this.currentEntity.id)
              }).then((resp) => {
                if (resp.data) {
                  if (resp.data.status === 1) {
                    if (resp.data.data) {
                      oCurrentObject.options.values = resp.data.data;

                      // Chargement des sections
                      this.currentSections = this.getCurrentSections();
                    }
                  }
                }
              });
            }
          }
        } else {
          // Sections normales
          this.currentSections = this.getCurrentSections();
        }
      }
    });
  }

  /**
   * Retourne l'identifiant de l'enregistrement en cours
   * @returns any
   */
  public getCurrentEntity(): any {

    const oEntity: any = {};

    const oSection = this.getCurrentSection();
    const aCurrentRoute = this.router.url.split('/');

    // Utilisation dans une section
    if (this.vitisCheck.isDefAndNotNull(oSection)) {
      if (this.vitisCheck.isDefAndNotNull(oSection.route)) {

        const aSectionRoute = ('/dashboard/' + oSection.route).split('/');
        const isSectionRouteActive = this.isSectionRouteActive(oSection.route);

        if (isSectionRouteActive) {
          for (let i = aSectionRoute.length - 1; i >= 0; i--) {
            if (aSectionRoute[i] === ':id') {
              oEntity.id = aCurrentRoute[i];
            }
            if (aSectionRoute[i] === ':form') {
              oEntity.form = aCurrentRoute[i];
            }
          }
        }
      }
    }

    return oEntity;
  }

  /**
   * Ouvre le mode par défaut
   */
  public openDefaultMode() {
    this.router.navigate(['**'], { queryParamsHandling: 'merge' });
  }

  /**
   * Retourne true/false si la route passée en argument est active (pour sections avec potentiellement sous-formulaires)
   * @param sRoute
   */
  public isSectionRouteActive(sRoute: string): boolean {
    return (this.isRouteActive(sRoute) || this.isRouteActive(sRoute + '/:id/:form') || this.isRouteActive(sRoute + '/:form'));
  }

  /**
   * Retourne true/false si la route passée en argument est active
   * @param  {string} sRoute
   * @returns boolean
   */
  public isRouteActive(sRoute: string): boolean {

    if (!this.vitisCheck.isString(sRoute)) {
      return false;
    }

    const isActive = sRoute.indexOf('/:') !== -1 ?
      this.isRouteWithParamsActive(sRoute) :
      this.router.isActive('dashboard/' + sRoute, false);
    return isActive;
  }

  /**
   * Retourne true/false si la route passée en argument est active (en ignorant les paramètres)
   * @param  {string} sRoute
   * @returns boolean
   */
  public isRouteWithParamsActive(sRoute: string): boolean {

    if (!this.vitisCheck.isString(sRoute)) {
      return false;
    }

    let isActive = true;
    const aRoute = ('/dashboard/' + sRoute).split('/');
    const aCurrentRoute = this.router.url.split('?')[0].split('/');

    if (aRoute.length !== aCurrentRoute.length) {
      return false;
    }

    for (let i = 0; i < aRoute.length; i++) {
      if (aRoute[i].substr(0, 1) === ':') {
        continue;
      }
      if (aRoute[i] !== aCurrentRoute[i]) {
        isActive = false;
      }
    }

    return isActive;
  }

  /**
   * Collapse tous les modes à l'exception de celui en cours
   * @param  {Array<any>} aModes
   */
  public collapseModes(aModes: Array<any>) {

    for (let i = 0; i < aModes.length; i++) {
      const oMode = aModes[i];

      if (this.vitisCheck.isArray(oMode.objects)) {

        oMode.collapsed = true;

        if (oMode.route) {
          if (this.isRouteActive(oMode.route)) {
            oMode.collapsed = false;
          }
        }

        this.collapseModes(oMode.objects);
      }
    }
  }

  /**
   * Affiche / cache le menu du header
   * @param  {any} oHeader
   */
  public toggleHeader(oHeader: any) {
    if (oHeader.collapsed !== false) {
      this.collapseHeaders();
      setTimeout(() => {
        this.displayHeader(oHeader);
      });
    } else {
      this.collapseHeaders();
    }
  }

  /**
   * Affiche le menu du header
   * @param  {any} oHeader
   */
  public displayHeader(oHeader: any) {
    if (oHeader.type === 'dropdownMenu') {
      oHeader.collapsed = false;
    }
  }

  /**
   * Événement lors du clic sur le menu d'un header
   * @param  {any} oHeader
   */
  public headerOnclick(oHeader: any) {

    const sType = oHeader.type;

    // Affichage d'un formulaire dans une modale
    if (sType === 'modalform') {
      if (this.vitisCheck.isDefAndNotNull(oHeader.options)) {
        this.displayModalForm(oHeader);
      }
    }

    // Appel à une fonction
    if (sType === 'function') {
      if (this.vitisCheck.isDefAndNotNull(oHeader.options)) {
        if (this.vitisCheck.isDefAndNotNull(oHeader.options.function)
          && this.vitisCheck.isDefAndNotNull(oHeader.options.functionContainer)) {

          // Container de la fonction
          const oFunctionContainer = oHeader.options.functionContainer;

          // Nom de la fonction
          let sFunction = oHeader.options.function;
          if (sFunction.indexOf('(') !== -1) {
            sFunction = sFunction.substr(0, sFunction.indexOf('('));
          }

          if (this.vitisCheck.isDefAndNotNull(this[oFunctionContainer])) {
            if (this.vitisCheck.isFunction(this[oFunctionContainer][sFunction])) {
              return this[oFunctionContainer][sFunction].call(
                this[oFunctionContainer]
              );
            }
          }
        }
      }
    }
  }

  /**
   * Affiche un simple-form dans une modale
   * @param  {any} oObject
   */
  public displayModalForm(oObject: any) {

    this.modalFormMode = oObject;

    // Import SimpleFormModule
    import('../../../simpleform/simpleform.module').then(({ SimpleformModule }) => {

      // Compile le module
      this.compiler.compileModuleAsync(SimpleformModule).then(moduleFactory => {

        // Crée moduleRef, resolve entry component, crée le composant
        const moduleRef = moduleFactory.create(this.injector);
        const componentFactory = moduleRef.instance.resolveComponent();
        const { instance } = this.container.createComponent(componentFactory, null, moduleRef.injector);

        // set oInputMode du formreader
        setTimeout(() => {
          instance.oInputMode = this.modalFormMode;
        }, 100);

        // Affiche la modale
        $('#dashboard-simpleform-modal').modal('show');
      });
    });
  }

  /**
   * Collapse tous les headers
   */
  public collapseHeaders() {
    if (this.vitisCheck.isArray(this.headers)) {
      for (let i = 0; i < this.headers.length; i++) {
        if (this.headers[i].type === 'dropdownMenu') {
          this.headers[i].collapsed = true;
        }
      }
    }
  }

  /**
   * Affiche la modale de connexion
   */
  public showModalConnection() {
    this.displayModalForm({
      options: {
        form_path: 'assets/forms/publicuser/signin.json',
        js_path: 'assets/forms/publicuser/signin.js',
        module: 'vitis',
        ressource: 'vitis/users'
      }
    });
  }

  /**
   * Affiche la modale de création de compte
   */
  public showModalSignUp() {
    this.displayModalForm({
      options: {
        form_path: 'assets/forms/publicuser/signup.json',
        js_path: 'assets/forms/publicuser/signup.js',
        module: 'vitis',
        ressource: 'vitis/users'
      }
    });
  }

  /**
   * Retourne la route du mode en cours à partie de l'URL
   * @returns any
   */
  private getCurrentModeRoute_(): any {

    const segments = this.getCurrentRouteSegments_();

    if (this.vitisCheck.isArray(segments)) {
      for (let i = 0; i < segments.length; i++) {
        const segment = segments[i];
        if (segment.path === 'dashboard') {

          // Retourne le segment suivant
          if (this.vitisCheck.isDefAndNotNull(segments[i + 1])) {
            if (this.vitisCheck.isDefAndNotNull(segments[i + 1].path)) {
              return segments[i + 1].path;
            }
          }
        }
      }
    }
  }

  /**
   * Retourne les segments de route actuellement utilisés
   * @returns Array
   */
  private getCurrentRouteSegments_(): Array<UrlSegment> {
    const tree = this.router.parseUrl(this.router.url);
    const f = tree.fragment;
    const q = tree.queryParams;
    const g = tree.root.children['primary'];
    return g.segments;
  }

  /**
   * Collapse tous les headers quand on clique n'importe où sur l'application
   */
  private handleHeaderDorpdownMenus_() {

    $('html').click((e) => {
      this.collapseHeaders();
    });
  }

  /**
   * Supprime le formulaire en cours lorsque la modale se ferme
   */
  private handleModalForm_() {

    const emptyForm = {
      'display': {
        'name': 'vitis_dashboard_modal_display_form',
        'title': '',
        'input_size': 'xs',
        'nb_cols': 8,
        'rows': []
      },
      'insert': {
        'name': 'vitis_dashboard_modal_insert_form',
        'title': '',
        'input_size': 'xs',
        'nb_cols': 8,
        'rows': []
      },
      'update': {
        'name': 'vitis_dashboard_modal_update_form',
        'title': '',
        'input_size': 'xs',
        'nb_cols': 8,
        'rows': []
      }
    };
    this.modalFormMode = emptyForm;

    $('#dashboard-simpleform-modal').on('hide.bs.modal', () => {
      this.modalFormMode = emptyForm;

      // Vide le container
      for (let i = 0; i < this.container.length; i++) {
        this.container.remove(i);
      }
    });
  }

  /**
   * Retourne la liste des modes actifs en ignorant les séparateurs
   * @param aModes
   * @returns
   */
  private getActiveModes_(aModes: Array<any>) {
    const aActiveModes = [];
    for (let i = 0; i < aModes.length; i++) {
      const oMode = aModes[i];
      if (oMode &&
        oMode.type !== 'separator' &&
        this.hasRights(oMode)) {
        aActiveModes.push(oMode);
      }
    }
    return aActiveModes;
  }
}
