import { Injectable, Injector, EventEmitter } from '@angular/core';
import { Router } from '@angular/router';
import { TranslateService } from '@ngx-translate/core';
import { Observable, Subject } from 'rxjs';
import { environment } from 'src/environments/environment';
import { CheckService as VitisCheck } from '../utils/check/check.service';
import { vitisAlert } from '../utils/alert/vitisalert';

@Injectable({
  providedIn: 'root'
})
export class AuthService {
  private token: string;
  private user: string;
  private user_id: number;
  private validity_date: string;
  private privileges: Array<string>;
  private userInfos: any;
  private tokenValid: boolean = false;
  private rememberSession: boolean = true;
  private loadedFromStorage: boolean = false;
  private loadedFromUrl: boolean = false;
  private RequestsService: any;
  private isPublicUser: boolean = false;

  public isConnecting: boolean = false;
  public onLogin = new EventEmitter<boolean>();

  // Observables
  public obsUserInfos = new Subject<any>();

  constructor(
    private translate: TranslateService,
    private vitisCheck: VitisCheck,
    private injector: Injector,
    private router: Router
  ) {

    // Récupère les paramètres depuis le localStorage
    this.retrieveLocalStorageParams();

    // Récupère les paramètres depuis l'URL
    this.retrieveUrlParams();

    import('../utils/requests/requests.service').then((oRequests) => {
      this.RequestsService = oRequests.RequestsService;

      // Connexion publique
      if (!this.token && environment.properties.public_connect_available === true) {
        this.retrievePublicConnection();
      }
    });
  }

  /**
   * Fonction retournant si l'utilisateur est authentifié
   * @return {boolean}
   */
  public isAuthenticated(bAllowLocalStorage: boolean = true): boolean {

    if (!bAllowLocalStorage && this.loadedFromStorage) {
      return false;
    }

    // Si token non vérifié, alors le vérifie
    if (!this.tokenValid && this.vitisCheck.isDefAndNotNull(this.token)) {
      // Requete synchrone qui vérifie le token et remplit les variables d'authentification
      this.testTokenSync_(this.token);

      // Si le token a été chagé depuis l'URL alors récupère les properties avec le token long
      if (this.tokenValid && this.loadedFromUrl) {
        // Récupère à nouveau les properties car cette fois l'utilisateur possède un token
        this.getPropertiesFromServer_().then((serverProperties: any) => {

          // Merge les properties server avec les properties client
          environment.properties = { ...serverProperties, ...environment.properties };
        });
      }
    }
    // S'il n'y pas d'id utilisateur est déconnecté ou si le token après vérification n'est toujours pas valide
    if (!this.user_id || !this.tokenValid) {
      return false;
    }
    // Récupère les infos de l'utilisateur
    if (!this.userInfos) {
      this.getUserInfosSync_(this.user_id).then(oResponse => {

        // Infos de la réponse
        this.userInfos = oResponse;

        // Définit si la connexion se fait par l'utilisateur public
        this.userInfos.isPublicUser = this.isPublicUser;

      }).catch(oResponse => {
        console.error('Cannot get user infos', oResponse);
        if (this.vitisCheck.isDefAndNotNull(this.token)) {
          this.disconnect();
        } else {
          this.router.navigate(['login'], { queryParamsHandling: 'merge' });
        }
      });
    }

    if (this.token && this.tokenValid) {
      return true;
    } else {
      return false;
    }
  }

  /**
   * Tocken getter
   * @return {object} token
   */
  public getToken(): string {
    return this.token === 'undefined' ? null : this.token;
  }

  /**
   * Privileges getter
   * @return {Array<string>} privileges
   */
  public getPrivileges(): Array<string> {
    return this.privileges;
  }

  /**
   * Privileges setter
   * @param {string|array} privileges
   */
  public setPrivileges(privileges: any): void {
    if (this.vitisCheck.isString(privileges)) {
      privileges = privileges.split('|');
    }
    if (this.vitisCheck.isArray(privileges)) {
      this.privileges = privileges;
    } else {
      console.error('geven privileges not an array: ', privileges);
    }
  }

  /**
   * UserInfos getter
   * @return {any} userInfos
   */
  public getUserInfos(): any {
    return this.userInfos;
  }

  /**
   * UserInfos observable getter
   * @return {Subject} userInfos
   */
  public getUserInfosObs(): any {
    return this.obsUserInfos;
  }

  /**
   * UserInfos setter
   * @param {any} userInfos
   */
  public setUserInfos(userInfos: any): any {

    // Met à jour this.userInfos
    this.userInfos = {
      ...this.userInfos,
      ...userInfos
    };

    // Met à jour les privilèges si besoin
    if (this.vitisCheck.isDefAndNotNull(userInfos.privileges)) {
      this.setPrivileges(userInfos.privileges);
    }

    // Met à jour la valeur de l'observable
    this.obsUserInfos.next(this.userInfos);
  }

  /**
   * Tente l'identification d'un utilisateur,
   * retourne un observable true/false si l'authentification a réussie
   * @param  {string} user
   * @param  {string} password
   * @param  {boolean|undefined} remember
   * @returns Observable
   */
  public login(user: string, password: string, remember: boolean = false): Observable<any> {
    return new Observable(observer => {

      // mémorise remember
      if (remember) {
        this.rememberSession = true;
      } else {
        this.rememberSession = false;
      }

      const throwError = (error: any) => {
        console.error(error);
        observer.next(error);
      };

      // Token API V1
      this.ajaxGetToken_(user, password).then((oResult) => {

        if (oResult.status === 200) {
          if (oResult.data.status == 1 &&
            oResult.data.data &&
            oResult.data.data.token) {

            this.token = oResult.data.data.token;
            this.user = oResult.data.data.login;
            this.user_id = oResult.data.data.user_id;
            this.validity_date = oResult.data.data.exp;
            this.privileges = oResult.data.data.privileges;

            // Token desormais vérifié
            this.tokenValid = true;

            // Stocke les paramètres dans le localStorage
            if (remember) {
              this.storeLocalStorageParams();
            }

            // Récupère à nouveau les properties car cette fois l'utilisateur possède un token
            this.getPropertiesFromServer_().then((serverProperties: any) => {

              // Merge les properties server avec les properties client
              environment.properties = { ...serverProperties, ...environment.properties };

              // Valide l'autentification
              observer.next(true);

              this.onLogin.emit(true);
            });

          } else {
            throwError(oResult);
          }
        } else {
          throwError(oResult);
        }
      }).catch((oResult) => {
        throwError(oResult);
      });
    });
  }

  /**
   * Déconnecte l'utilisateur
   */
  public disconnect() {
    this.removeLocalStorageParams();
    window.location.href = document.location.href.split('?')[0];
  }

  /**
   * Récupère les paramètres depuis le localStorage
   */
  public retrieveLocalStorageParams() {

    const prefix = this.getLocalStoragePrefix_();
    const tmp_token = localStorage.getItem(prefix + '_token');
    const tmp_user = localStorage.getItem(prefix + '_user');
    const tmp_user_id = localStorage.getItem(prefix + '_user_id');
    const tmp_validity_date = localStorage.getItem(prefix + '_validity_date');
    const tmp_privileges = localStorage.getItem(prefix + '_privileges');

    if (tmp_token) {
      this.token = tmp_token;
      this.loadedFromStorage = true;
    }
    if (tmp_user) {
      this.user = tmp_user;
    }
    if (tmp_user_id) {
      this.user_id = Number(tmp_user_id);
    }
    if (tmp_validity_date) {
      this.validity_date = tmp_validity_date;
    }
    if (tmp_privileges) {
      this.privileges = tmp_privileges.split(',');
    }

  }

  /**
   * Récupère les paramètres de connexion depuis l'url
   */
  public retrieveUrlParams() {
    const token = this.getParamFromUrl_('token');
    const login = this.getParamFromUrl_('login');
    const password = this.getParamFromUrl_('password');
    if (this.vitisCheck.isDefAndNotNull(token)) {
      this.token = token;
      this.rememberSession = false;
      this.loadedFromStorage = false;
      this.loadedFromUrl = true;
    }
    if (this.vitisCheck.isDefAndNotNull(login)
      && this.vitisCheck.isDefAndNotNull(password)) {
      this.isConnecting = true;
      this.login(login, password, false).subscribe((oResult) => {
        this.isConnecting = false;
        if (oResult) {
          this.router.navigate(['**'], { queryParamsHandling: 'merge' });
        } else {
          if (this.vitisCheck.isDefAndNotNull(this.token)) {
            this.disconnect();
          } else {
            this.router.navigate(['login'], { queryParamsHandling: 'merge' });
          }
        }
      });
    }
  }

  /**
   * Utilisation de connexion privée
   */
  public retrievePublicConnection() {

    // liste des routes à ignorer pour les connexion public
    const aIgnoreRoutes: string[] = [
      '/login/forgotten-pwd',
      '/login/sign-up',
      '/widget'
    ];

    const subscription = this.router.events.subscribe(oEvent => {
      // si la route n'est pas encore défini on attend
      if ('url' in oEvent) {
        if (oEvent.url == '' || oEvent.url == '/') {
          return;
        }
      } else {
        return;
      }

      // récupération de la route actuel
      const snapRoute: string = oEvent.url;
      // la route est analysable on coupe pour éviter une boucle infinie
      subscription.unsubscribe();
      // si la route actuel commence par une des routes à ignorer alors on ne teste pas la connexion public
      for (const sIgnoredRoute of aIgnoreRoutes) {
        if (snapRoute.indexOf(sIgnoredRoute) == 0) {
          return;
        }
      }

      this.isConnecting = true;
      // Récupération des properties sans token
      this.getPropertiesFromServer_().then((serverProperties: any) => {

        // Merge les properties server avec les properties client
        environment.properties = { ...serverProperties, ...environment.properties };

        if (environment.properties.public_connect_enabled === true &&
          this.vitisCheck.isDefAndNotNull(environment.properties.public_connect_token)) {
          this.token = environment.properties.public_connect_token;
          this.rememberSession = false;

          // Retrouve les infos de l'utilisateur et marque le token comme valide
          this.testTokenSync_(this.token);

          // Définit la connexion par l'utilisateur public
          this.isPublicUser = true;

          // Récupération des properties avec token
          this.getPropertiesFromServer_().then((serverProperties2: any) => {

            // Merge les properties server avec les properties client
            environment.properties = { ...serverProperties2, ...environment.properties, };

            setTimeout(() => {
              this.onLogin.emit(true);
              this.isConnecting = false;
              setTimeout(() => {
                this.router.navigate(['**'], { queryParamsHandling: 'merge' });
              });
            });
          });
        } else {
          this.isConnecting = false;
          this.router.navigate(['login'], { queryParamsHandling: 'merge' });
        }
      });
    });
  }

  /**
   * Stocke les paramètres dans le localStorage
   */
  public storeLocalStorageParams() {

    const prefix = this.getLocalStoragePrefix_();
    localStorage.setItem(prefix + '_token', this.token);
    localStorage.setItem(prefix + '_user', this.user);
    localStorage.setItem(prefix + '_user_id', String(this.user_id));
    localStorage.setItem(prefix + '_validity_date', this.validity_date);
    localStorage.setItem(prefix + '_privileges', this.privileges.join(','));

  }

  /**
   * Supprime les paramètres du localStorage
   */
  public removeLocalStorageParams() {

    const prefix = this.getLocalStoragePrefix_();
    localStorage.removeItem(prefix + '_token');
    localStorage.removeItem(prefix + '_user');
    localStorage.removeItem(prefix + '_user_id');
    localStorage.removeItem(prefix + '_validity_date');
    localStorage.removeItem(prefix + '_privileges');

  }

  /**
   * Refait la requête et rafraîchit les infos utilisateur
   */
  public refreshUserInfos(): Promise<void> {
    return new Promise((resolve, reject) => {
      this.getUserInfosSync_(this.user_id).then(oResponse => {
        this.setUserInfos(oResponse);
        resolve();
      }).catch(oResponse => {
        console.error('Cannot get user infos', oResponse);
        if (this.vitisCheck.isDefAndNotNull(this.token)) {
          reject();
          this.disconnect();
        } else {
          this.router.navigate(['login'], { queryParamsHandling: 'merge' });
        }
      });
    });
  }

  /**
   * Demande un token et renvoi une promesse
   * @param  {string} user
   * @param  {string} password
   * @returns Promise
   */
  private ajaxGetToken_(user: string, password: string): Promise<any> {

    const vitisRequest = this.injector.get(this.RequestsService);
    const sUrl = `${environment.properties.app_api_url}/vitis/privatetoken`;
    const duration = 600;

    return vitisRequest.ajaxRequestPromise({
      url: sUrl,
      method: 'POST',
      data: {
        user,
        password,
        duration
      }
    });
  }

  /**
   * Retourne les infos de l'utilisateur de manière synchrone
   * @param  {number} userId
   * @returns Promise
   */
  private getUserInfosSync_(userId: number): Promise<any> {

    const vitisRequest = this.injector.get(this.RequestsService);
    const sUrl = `${environment.properties.app_api_url}/vitis/users/${userId}`;

    return new Promise((resolve, reject) => {
      vitisRequest.ajaxRequestPromise({
        url: sUrl,
        method: 'GET',
        async: false,
        responseType: '',
        headers: {
          Authorization: this.getToken()
        },
      }).then((resp: any) => {
        if (this.vitisCheck.isString(resp.data)) {
          const oResult = JSON.parse(resp.data);
          if (oResult.status == 1 && oResult.data) {
            if (oResult.data) {
              if (oResult.data.user_id) {
                resolve(oResult.data);
              } else {
                reject(oResult);
              }
            } else {
              reject(oResult);
            }
          } else {
            reject(oResult);
          }
        } else {
          reject(resp);
        }
      });
    });
  }

  /**
   * Retourne la chaine utilisée comme préfixe dans localStorage
   * Permet l'utilisation de plusieurs environnements
   * @returns string
   */
  private getLocalStoragePrefix_(): string {
    let prefix = environment.properties.application;
    if (environment.properties.environment) {
      if (environment.properties.environment.length > 0) {
        prefix += '_' + environment.properties.environment;
      }
    }
    return prefix;
  }

  /**
   * Requete synchrone qui vérifie le token et remplit les variables suivantes :
   * this.token
   * this.user
   * this.user_id
   * this.validity_date
   * this.privileges
   * @param  {string} sToken
   */
  private testTokenSync_(sToken: string) {

    const vitisRequest = this.injector.get(this.RequestsService);
    const sUrl = `${environment.properties.app_api_url}/vitis/privatetoken`;

    vitisRequest.ajaxRequest({
      url: sUrl,
      method: 'GET',
      async: false,
      responseType: '',
      displayErrorMessage: false,
      success: (resp: any) => {
        // par défaut on considère que le token est non valide
        this.tokenValid = false;
        if (this.vitisCheck.isString(resp.data)) {
          const oResult = JSON.parse(resp.data);
          if (oResult.status === 1) {
            if (!this.vitisCheck.isUndefined(oResult.data.token)) {
              this.token = oResult.data.token;
            }

            this.user = oResult.data.user;
            this.user_id = oResult.data.user_id;
            this.validity_date = oResult.data.exp;
            this.privileges = oResult.data.privileges;

            // Token desormais vérifié
            this.tokenValid = true;

            // Stocke les paramètres dans le localStorage
            if (this.rememberSession) {
              this.storeLocalStorageParams();
            }
          }
        }
      },
      error: (resp: any) => {
        console.error('error', resp);
        this.removeLocalStorageParams();
      }
    });
  }

  /**
   * Recupère un paramètre depuis l'URL
   * @param  {string} sParam
   * @returns string
   */
  private getParamFromUrl_(sParam: string): string {
    const url_string = window.location.href;
    const url = new URL(url_string);
    const value = url.searchParams.get(sParam);
    return value;
  }

  /**
   * Récupère les properties depuis le serveur
   * @returns Promise
   */
  private getPropertiesFromServer_(): Promise<any> {
    return new Promise((resolve, reject) => {

      const vitisRequest = this.injector.get(this.RequestsService);
      let properties = {};

      vitisRequest.ajaxRequest({
        method: 'GET',
        url: `${environment.properties.app_api_url}/vitis/properties`,
        success: (resp: any) => {
          if (this.vitisCheck.isDefAndNotNull(resp.data)) {
            const oResult = resp.data;
            if (oResult.status = 1) {
              properties = oResult.data;
            }
          }
          resolve(properties);
        },
        error: (resp: any) => {
          if (this.vitisCheck.isString(resp.data)) {
            const oResult = JSON.parse(resp.data);
            this.translate.get(oResult.errorMessage).subscribe((translation) => {
              const options: any = {
                type: 'error',
                title: translation
              };
              vitisAlert.alert(options);
            });
          }
          reject();
        }
      });
    });
  }
}