import { Injectable, Injector } from '@angular/core';
import { Observable } from 'rxjs';
import { TranslateService } from '@ngx-translate/core';

import { environment } from 'src/environments/environment';
import { CheckService as VitisCheck } from '../check/check.service';

import { AuthService } from '../../auth/auth.service';
import { vitisAlert } from '../alert/vitisalert';
import { CloneService as VitisClone } from '../clone/clone.service';
import { SessionStorageService } from '../sessionstorage/session-storage.service';
import { Router } from '@angular/router';

@Injectable({
  providedIn: 'root'
})
export class RequestsService {

  private ajaxLoading_: boolean = false;
  private ajaxLoadingCounter_: number = 0;

  constructor(
    private vitisClone: VitisClone,
    private vitisCheck: VitisCheck,
    private injector: Injector,
    private translateService: TranslateService,
    private router: Router,
    private sessionStorageService: SessionStorageService,
  ) { }

  /**
   * ajaxRequest function
   * Envoi d'une requête ajax.
   * @param {object} Options Paramètres de la requête.
   * @param {boolean|undefined} Options.async Définit si la requête doit être lancée de manière asynchrone
   * @param {object|undefined} Options.headers Headers à rajouter à la requette
   * @param {string|undefined} Options.responseType Type de réponse ("", "text", "json", "blob", "arraybuffer", "document")
   * @param {string} Options.method Type de méthode ("GET", "PUT", "POST", "DELETE")
   * @param {string} Options.url URL cible
   * @param {boolean|undefined} Options.ajaxLoader false to not show the ajaxLoader
   * @param {function} Options.success Fonction appellée quand la requête se termine correctement
   * @param {function|undefined} Options.error Fonction appellée quand la requête se termine par une erreur
   * @param {object|undefined} Options.params Paramètres à ajouter dans l'URL
   * @param {object|undefined} Options.data Paramètres à rajouter dans le body
   * @param {boolean|undefined} Options.displayErrorMessage Si false alors n'affiche pas de message en cas d'erreur
   * @param {boolean|undefined} Options.httpMethodOverride Si true alors utilise la méthode POST pour passer un GET tout en utilisant le header X-HTTP-Method-Override
   * @param {boolean|undefined} Options.enableCache Si true alors stoque les résutats dans le session_storage et les re-utilise si besoin
  //  * @param {function|undefined} Options.aborded Fonction appellée quand la requête est annulée -> TO DO
  //  * @param {number|undefined} Options.timeout Timeout en milisecondes permettant d'annuler la requête -> TO DO
  //  * @param {Promise|undefined} Options.abord Promesse angular permettant d'annuler la requête -> TO DO
   * @export
   */
  public ajaxRequest(oOptions: any) {

    const xhr = new XMLHttpRequest();

    oOptions = this.ajaxRequestParseOptions_(oOptions);

    // Paramètres 'method' et 'url' obligatoires.
    if (this.vitisCheck.isDefAndNotNull(oOptions.method) && this.vitisCheck.isDefAndNotNull(oOptions.url)) {

      // Gestion du cache (lecture)
      if ((oOptions.method === 'GET' || oOptions.original_method_ === 'GET') &&
        oOptions.enableCache === true) {
        const oGetFromCache = this.ajaxRequestGetFromCache_(oOptions);
        if (this.vitisCheck.isDefAndNotNull(oGetFromCache)) {
          this.ajaxRequestHandleSuccessResponse_(null, oGetFromCache, oOptions, true);
          return;
        }
      }

      // Traitement à la fin de la requête.
      const this_ = this;
      xhr.onreadystatechange = function (event) {
        if (this.readyState === XMLHttpRequest.DONE) {

          // Requête terminée.
          this_.ajaxRequestHandleResponse_(this, xhr, oOptions);
        }
      };

      // ToDo
      this.ajaxRequestHandleProgress_(xhr, oOptions);

      // Paramètre "responseType" (sauf pour IE).
      if (navigator.userAgent.indexOf('Trident') === -1) {
        xhr.responseType = oOptions.responseType;
      }

      // Préparation de la requête.
      xhr.open(oOptions.method, oOptions.url, oOptions.async);

      // BUG IE : paramètre "responseType" après open (sinon exception avec internet explorer).
      if (navigator.userAgent.indexOf('Trident') !== -1) {
        xhr.responseType = oOptions.responseType;
      }

      // Paramètres de l'entête.
      for (const sHeaderKey in oOptions.headers) {
        if (oOptions.headers.hasOwnProperty(sHeaderKey)) {
          xhr.setRequestHeader(sHeaderKey, oOptions.headers[sHeaderKey]);
        }
      }

      // Création de requestData
      let requestData = null;
      if (typeof (oOptions.data) !== 'undefined') {
        if (this.vitisCheck.isObject(oOptions.data)) {
          if (oOptions.data instanceof FormData) {
            requestData = oOptions.data;
          } else {
            requestData = JSON.stringify(this.cleanParams_(oOptions.data, oOptions.url));
          }
        }
      }

      // Envoi de la requête.
      xhr.send(requestData);

      // Todo
      this.ajaxRequestHandleLoader_(xhr, oOptions);

      // Todo
      this.ajaxRequestHandleTimeout_(xhr, oOptions);

      // Todo
      this.ajaxRequestHandleAbord_(xhr, oOptions);

    } else {
      console.error('Parameter method or url not founded');
    }

    return xhr;
  }

  /**
   * Lance la fonction ajaxRequest et retourne un Observable
   * @param  {any} oOptions
   * @returns Observable
   */
  public ajaxRequestObservable(oOptions: any): Observable<any> {
    return new Observable((observer) => {

      oOptions.success = (resp: any) => {
        observer.next(resp);
        observer.complete();
      };

      oOptions.aborded = (resp: any) => {
        observer.next(resp);
        observer.complete();
      };

      oOptions.error = (resp: any) => {
        observer.error(resp);
        observer.complete();
      };

      // Lance ajaxRequest
      this.ajaxRequest(oOptions);
    });
  }

  /**
   * Lance la fonction ajaxRequest et retourne une promesse
   * @param  {any} oOptions
   * @returns Promise
   */
  public ajaxRequestPromise(oOptions: any): Promise<any> {
    return new Promise((resolve, reject) => {

      oOptions.success = (resp: any) => {
        resolve(resp);
      };

      oOptions.aborded = (resp: any) => {
        reject(resp);
      };

      oOptions.error = (resp: any) => {
        reject(resp);
      };

      // Lance ajaxRequest
      this.ajaxRequest(oOptions);
    });
  }

  /**
   * Télécharge un fichier
   * @param sUrl
   * @param sFileName
   */
  public ajaxDownloadFile(sUrl: string, sFileName: string): Promise<any> {
    return new Promise((resolve, reject) => {

      const throwError = () => {
        // Message d'erreur
        this.translateService.get('ERROR_DOWNLOAD_FILE').subscribe((translation) => {
          const options: any = {
            type: 'error',
            title: translation
          };
          vitisAlert.alert(options);
        });
      };

      this.ajaxRequestPromise({
        method: 'GET',
        url: sUrl,
        responseType: 'blob',
        headers: {
          Accept: 'application/octet-stream'
        }
      }).then((resp: any) => {
        if (resp.data &&
          resp.data instanceof Blob) {
          if (typeof (window.navigator.msSaveOrOpenBlob) == 'undefined') {
            // Crée un lien et simule un click dessus.
            const link = document.createElement('a');
            link.href = window.URL.createObjectURL(resp.data);
            link.download = sFileName;
            link.style.display = 'none';
            document.body.appendChild(link);
            link.click();
            document.body.removeChild(link);
          } else {
            // Pour Edge et IE.
            window.navigator.msSaveOrOpenBlob(resp.data, sFileName);
          }
          resolve();
        } else {
          throwError();
          reject();
        }
      }).catch((err) => {
        throwError();
        reject();
      });
    });



  }

  /**
   * Chage un JS externe
   * @param  {string} sUrl
   * @param  {any} oOptions?
   */
  public loadExternalJs(sUrl: string, oOptions?: any) {
    return new Promise((resolve, reject) => {

      // Options par défaut
      let bAsync: boolean = true;
      let bScriptInBody: boolean = false;

      // Numéro de build (pour éviter caches)
      let sBuild: number = Date.now();
      if (this.vitisCheck.isDefAndNotNull(environment.properties)) {
        sBuild = environment.properties.build;
      }

      // Évite les caches
      if (this.vitisCheck.isString(sUrl)) {
        sUrl += '?version=' + sBuild;
      } else {
        reject();
      }

      if (this.vitisCheck.isDefAndNotNull(oOptions)) {
        if (this.vitisCheck.isDefAndNotNull(oOptions.async)) {
          bAsync = oOptions.async;
        }
        if (this.vitisCheck.isDefAndNotNull(oOptions.scriptInBody)) {
          bScriptInBody = oOptions.scriptInBody;
        }
      }

      const oScript = document.createElement('script');
      oScript.setAttribute('src', sUrl + '?version=' + sBuild);
      oScript.async = bAsync;
      oScript.setAttribute('type', 'text/javascript');
      oScript.onload = () => {
        resolve(true);
      };
      if (bScriptInBody) {
        document.getElementsByTagName('body')[0].appendChild(oScript);
      } else {
        document.getElementsByTagName('head')[0].appendChild(oScript);
      }
    });
  }

  /**
   * Affiche le ajax loader
   */
  public showAjaxLoader() {

    // Incrémente l'ajaxCounter
    this.ajaxLoadingCounter_++;

    // Si un ajaxLoader est déjà en cours
    // Incrémente uniquement le loadingcounter
    // Sinon affiche l'ajaxLoader
    if (!this.ajaxLoading_) {

      this.ajaxLoading_ = true;

      const ajaxLoaderBackdrop = document.createElement('div');
      ajaxLoaderBackdrop.classList.add('modal-backdrop');
      ajaxLoaderBackdrop.classList.add('fade');
      ajaxLoaderBackdrop.classList.add('show');
      ajaxLoaderBackdrop.classList.add('ajaxLoader-backdrop');

      const ajaxLoaderSpinner = document.createElement('div');
      ajaxLoaderSpinner.classList.add('spinner-border');
      ajaxLoaderSpinner.classList.add('ajaxLoader-spinner');

      document.body.appendChild(ajaxLoaderBackdrop);
      document.body.appendChild(ajaxLoaderSpinner);
    }
  }

  /**
   * Cache le ajax loader
   */
  public hideAjaxLoader() {

    // Décrémente ajaxCounter
    this.ajaxLoadingCounter_--;

    // Patch erreurs
    if (this.ajaxLoadingCounter_ < 0) {
      this.ajaxLoadingCounter_ = 0;
    }

    // Si plus aucune requête n'est en cours
    if (this.ajaxLoadingCounter_ === 0) {

      this.ajaxLoading_ = false;

      // Supprime l'ajaxLoader
      const aLoaderClasses = ['ajaxLoader-backdrop', 'ajaxLoader-spinner'];
      for (let i = 0; i < aLoaderClasses.length; i++) {
        const className = aLoaderClasses[i];
        const elems = document.getElementsByClassName(className);

        for (let ii = 0; ii < elems.length; ii++) {
          const element = elems[ii];
          document.body.removeChild(element);
        }
      }
    }
  }

  /**
   * Return l'url de l'api
   * @param sRessource
   * @param sId
   * @returns string
   */
  public getApiRessourceUrl(sRessource: string, sId: any = null): string {

    let sRouteId: any;
    // si un id est passé de force on le garde
    if (this.vitisCheck.isDefAndNotNull(sId)) {
      sRouteId = sId;
    }
    // si un id est défini on essai de bind ou on le rajoute à la fin
    if (sRouteId) {
      if (sRessource.indexOf(':id') > -1) {
        sRessource = sRessource.replace(':id', sRouteId);
      } else {
        sRessource += '/' + sRouteId;
      }
    }

    return environment.properties.app_api_url + '/' + sRessource;
  }

  /**
   * Parse les options de la fonction ajaxRequest
   * @param  {any} oOptions
   * @returns any
   */
  private ajaxRequestParseOptions_(oOptions: any): any {

    // ajaxLoader
    oOptions.ajaxLoader = this.vitisCheck.isDefAndNotNull(oOptions.ajaxLoader) ? oOptions.ajaxLoader : true;

    // URL de l'api (sans le port)
    let sServiceUrl: string = this.vitisClone.clone(environment.properties.app_api_url);
    if (sServiceUrl.indexOf(':', 6) !== -1) {
      sServiceUrl = sServiceUrl.substr(0, sServiceUrl.indexOf(':', 6));
    }

    // Booléen qui définit si une URL vient du serveur Vitis pour savoir si il faut renseigner le token
    let isSelfUrl: boolean = false;
    if (oOptions.url.substr(0, sServiceUrl.length) === sServiceUrl) {
      isSelfUrl = true;
    }

    // Si l'url est celle du proxy, il ne faudra surtout pas renseigner le token
    if (this.vitisCheck.isString(environment.properties.proxy_url) &&
      oOptions.url.substr(0, environment.properties.proxy_url.length) === environment.properties.proxy_url) {
      isSelfUrl = false;
    }

    // Paramètres de l'entête par défaut.
    const oHeaders: any = {
      'Accept': 'application/json',
    };

    // Service d'authentification
    const authService = this.injector.get(AuthService);

    // Token renseigné uniquement si l'URL cible est du même serveur
    if (isSelfUrl && oOptions.sendToken !== false) {
      // Token de l'utilisateur
      if (this.vitisCheck.isDefAndNotNull(authService) &&
        this.vitisCheck.isDefAndNotNull(authService.getToken())) {
        oHeaders.Authorization = authService.getToken();
      }
    }

    // Gestion X-HTTP-Method-Override
    if (isSelfUrl &&
      oOptions.httpMethodOverride === true &&
      oOptions.method === 'GET') {
      oHeaders['X-HTTP-Method-Override'] = 'GET';
      oOptions.method = 'POST';
      oOptions.original_method_ = 'GET';
      if (oOptions.params) {
        oOptions.data = oOptions.params;
        delete oOptions.params;
      }
    }

    // Requête asynchrone.
    if (typeof (oOptions.async) === 'undefined') {
      oOptions.async = true;
    }
    // Paramètres de l'entête.
    if (typeof (oOptions.headers) !== 'undefined') {
      this.mergeObject_(oHeaders, oOptions.headers);
    }

    oOptions.headers = oHeaders;

    if (!this.vitisCheck.isDefAndNotNull(oOptions.timeout)) {
      if (this.vitisCheck.isDefAndNotNull(environment.properties.max_request_time)) {
        oOptions.timeout = environment.properties.max_request_time * 1000;
      }
    }

    // Type de réponse
    switch (oOptions.responseType) {
      case '':
      case 'text':
      case 'json':
      case 'blob':
      case 'arraybuffer':
      case 'document':
        break;
      default:
        oOptions.responseType = 'json';
        break;
    }

    // Paramètres à passer dans l'url.
    if (oOptions.url.indexOf('?') === -1) {
      oOptions.url += '?vitis_version=' + environment.properties.build;
    } else {
      oOptions.url += '&vitis_version=' + environment.properties.build;
    }

    // Traîte les paramètres
    if (this.vitisCheck.isDefAndNotNull(oOptions.params)) {
      oOptions.params = this.cleanParams_(oOptions.params, oOptions.url);
      oOptions.url = this.concatParamsWithURL_(oOptions.params, oOptions.url);
    }

    return oOptions;
  }

  /**
   * Gère la réponse lors d'un appel à la fonction ajaxRequest
   * @param  {any} oRequest
   * @param  {any} xhr
   * @param  {any} oOptions
   */
  private ajaxRequestHandleResponse_(oRequest: any, xhr: any, oOptions: any) {

    // Get tout le header dans une chaine de charactère
    const headers = xhr.getAllResponseHeaders();

    // Split sur CRLF
    const arr = headers.trim().split(/[\r\n]+/);

    // Objet contenant tout le header
    const oHeader = {};
    arr.forEach(function (line) {
      // split sur :  (espace important)
      const parts = line.split(': ');
      // supprime la clé du tableau et la récupére dans une variable
      const header = parts.shift();
      // retransforme le tableau en string
      const value = parts.join(': ');
      oHeader[header] = value;
    });

    // Objet réponse
    let oResponse: any;
    switch (oOptions.responseType) {
      case '':
      case 'text':
        oResponse = {
          'data': oRequest.responseText,
          'headers': oHeader,
          'status': oRequest.status,
          'statusText': oRequest.statusText
        };
        break;
      default:
        oResponse = {
          'data': oRequest.response,
          'headers': oHeader,
          'status': oRequest.status,
          'statusText': oRequest.statusText
        };
        break;
    }

    // BUG IE : retourne une réponse json au format texte.
    if (this.vitisCheck.isDefAndNotNull(oResponse.data) && oOptions.responseType === 'json' && this.vitisCheck.isString(oResponse.data)) {
      oResponse.data = JSON.parse(oResponse.data);
    }

    // Succès de la requête.
    if (oRequest.status >= 200 && oRequest.status < 300) {
      this.ajaxRequestHandleSuccessResponse_(oRequest, oResponse, oOptions);
    } else if (oRequest.status >= 400 && oRequest.status < 600) {
      // Erreur interne serveur ou provoqué
      this.ajaxRequestHandleErrorResponse_(oRequest, oResponse, oOptions);
    } else if (xhr['aborded'] === true) {
      // Requête annulée
      this.ajaxRequestHandleAbordedResponse_(oRequest, oResponse, oOptions);
    } else {
      // Retour erreur

      // Log
      console.error(oOptions.url, oRequest);

      // Fonction callback.
      if (this.vitisCheck.isFunction(oOptions.error)) {
        oOptions.error(oResponse);
      }
    }

    // Timeout car si les données sont en cache iLoadingCounter-- se fait avant iLoadingCounter++
    setTimeout(() => {
      this.hideAjaxLoader();
    });
  }

  /**
   * Gère la réponse lors d'un appel à la fonction ajaxRequest en cas de succès
   * @param  {any} oRequest
   * @param  {any} oResponse
   * @param  {any} oOptions
   * @param  {boolean} bFromCache true si la réponse vient du cache
   */
  private ajaxRequestHandleSuccessResponse_(oRequest: any, oResponse: any, oOptions: any, bFromCache = false) {
    if (bFromCache !== true) {
      // controle du statut de la requête
      if (oResponse.data) {
        if (this.vitisCheck.isDefAndNotNull(oResponse.data.status)) {
          if (oResponse.data.status != 1) {
            this.ajaxRequestHandleErrorResponse_(oRequest, oResponse, oOptions);
            return;
          }
        }
      }

      // Écriture dans le cache
      if ((oOptions.method === 'GET' || oOptions.original_method_ === 'GET') &&
        oOptions.enableCache === true) {
        this.ajaxRequestWriteToCache_(oOptions, oResponse);
      }
    }

    // Fonction callback.
    if (this.vitisCheck.isFunction(oOptions.success)) {
      oOptions.success(oResponse);
    }
  }

  /**
   * Gère la réponse lors d'un appel à la fonction ajaxRequest en cas d'erreur
   * @param  {any} oRequest
   * @param  {any} oResponse
   * @param  {any} oOptions
   */
  private ajaxRequestHandleErrorResponse_(oRequest: any, oResponse: any, oOptions: any) {
    // structures par défaut
    const oModalOptions: any = {
      type: 'error',
    };

    const aTranslationsKey: string[] = ['REQUEST_ERROR'];
    let sTitle = 'REQUEST_ERROR';
    let sMessage = null;

    // gestion des erreurs serveurs
    if (oRequest.status >= 500 && oRequest.status < 600) {
      aTranslationsKey.push('ERROR_SERVER_REQUEST_500');
      sMessage = 'ERROR_SERVER_REQUEST_500';
    } else if (this.vitisCheck.isDefAndNotNull(oResponse.data) && (this.vitisCheck.isObject(oResponse.data) || this.vitisCheck.isObject(JSON.parse(oResponse.data)))) {
      if (!this.vitisCheck.isObject(oResponse.data)) {
        oResponse.data = JSON.parse(oResponse.data);
      }
      // gestions des autres erreurs
      // Message d'erreur si token expiré ou invalide.
      if (this.vitisCheck.isDefAndNotNull(oResponse.data.errorMessage)) {
        sMessage = oResponse.data.errorMessage;
        aTranslationsKey.push(sMessage);
      }
      if (this.vitisCheck.isDefAndNotNull(oResponse.data.errorTitle)) {
        aTranslationsKey.push(oResponse.data.errorTitle);
        sTitle = oResponse.data.errorTitle;
      }

    }

    // affcihage du message d'erreur
    if (oOptions.displayErrorMessage !== false) {
      this.translateService.get(aTranslationsKey).subscribe((res: string) => {
        oModalOptions.title = (this.vitisCheck.isDefAndNotNull(res[sTitle])) ? res[sTitle] : sTitle;
        oModalOptions.text = (this.vitisCheck.isDefAndNotNull(res[sMessage])) ? res[sMessage] : sMessage;
        vitisAlert.alert(oModalOptions).then((resp) => {
          if (sMessage === 'ERROR_SESSION_EXPIRED') {
            this.router.navigate(['login'], { queryParamsHandling: 'merge' });
          }
        });
      });
    }
    // gestion des promises/observables
    if (this.vitisCheck.isFunction(oOptions.error)) {
      oOptions.error(oResponse);
    }
  }

  /**
   * Gère la réponse lors d'un appel à la fonction ajaxRequest en cas d'annulation
   * @param  {any} oRequest
   * @param  {any} oResponse
   * @param  {any} oOptions
   */
  private ajaxRequestHandleAbordedResponse_(oRequest: any, oResponse: any, oOptions: any) {

    //     if (typeof (oOptions['aborded']) == 'function') {
    //       // Si un scope d'Angular est passé en paramètre : évaluation de la fonction dans ce scope.
    //       if (typeof (oOptions['scope']) != 'undefined')
    //         oOptions['scope'].$applyAsync(oOptions['aborded'](oResponse));
    //       else
    //         oOptions['aborded'](oResponse);
    //     }
  }

  /**
   * To do
   */
  private ajaxRequestHandleProgress_(xhr: any, oOptions: any) {

    // //si fonction calcul de progrès
    // if (typeof (oOptions['progress']) == 'function') {
    //   // Si un scope d'Angular est passé en paramètre : évaluation de la fonction dans ce scope.
    //   if (typeof (oOptions['scope']) != 'undefined')
    //     xhr.onprogress = function (e) {
    //       oOptions['scope'].$applyAsync(oOptions['progress'](e));
    //     }
    //   else
    //     xhr.onprogress = function (e) {
    //       oOptions['progress'](e);
    //     }
    // }

    // //si fonction calcul de progrès
    // if (typeof (oOptions['loadStart']) == 'function') {
    //   // Si un scope d'Angular est passé en paramètre : évaluation de la fonction dans ce scope.
    //   if (typeof (oOptions['scope']) != 'undefined')
    //     xhr.onloadstart = function (e) {
    //       oOptions['scope'].$applyAsync(oOptions['loadStart'](e));
    //     }
    //   else
    //     xhr.onloadstart = function (e) {
    //       oOptions['loadStart'](e);
    //     }
    // }

    // //si fonction calcul de progrès
    // if (typeof (oOptions['loadEnd']) == 'function') {
    //   // Si un scope d'Angular est passé en paramètre : évaluation de la fonction dans ce scope.
    //   if (typeof (oOptions['scope']) != 'undefined')
    //     xhr.onloadend = function (e) {
    //       oOptions['scope'].$applyAsync(oOptions['loadEnd'](e));
    //     }
    //   else
    //     xhr.onloadend = function (e) {
    //       oOptions['loadEnd'](e);
    //     }
    // }
  }

  /**
   * To do
   */
  private ajaxRequestHandleLoader_(xhr: any, oOptions: any) {
    // Affiche le loader.
    if (oOptions.ajaxLoader) {
      this.showAjaxLoader();
    }
  }

  /**
   * To do
   */
  private ajaxRequestHandleTimeout_(xhr: any, oOptions: any) {
    // // Timeout
    // if (this.vitisCheck.isNumber(oOptions['timeout'])) {
    //   setTimeout(function () {
    //     if (xhr['readyState'] !== 4 && xhr['aborded'] !== true) {
    //       var iSeconds = oOptions['timeout'] / 1000;
    //       var intervalID, dialog;
    //       var setDialog = function () {
    //         dialog = bootbox.confirm({
    //           'message': 'Requête sans réponse',
    //           'buttons': {
    //             'confirm': {
    //               'label': 'Annuler',
    //               'className': 'btn-success'
    //             },
    //             'cancel': {
    //               'label': 'Attendre',
    //               'className': 'btn-default'
    //             }
    //           },
    //           'callback': function (abord) {
    //             // Annuler ?
    //             if (abord) {
    //               xhr.abort();
    //               console.info('request aborded');
    //               clearInterval(intervalID);
    //             } else {
    //               setTimeout(function () {
    //                 if (xhr['readyState'] !== 4) {
    //                   setDialog();
    //                 }
    //               }, oOptions['timeout']);
    //             }
    //           }
    //         });
    //       };
    //       setDialog();
    //       dialog['init'](function () {
    //         var updateTitle = function () {
    //           var sMessage = '<h4>Requête sans réponse depuis ' + iSeconds + 's...</h4>';
    //           dialog['find']('.bootbox-body').html(sMessage);
    //         }
    //         intervalID = window.setInterval(function () {
    //           if (xhr['readyState'] === 4) {
    //             dialog.modal('hide');
    //             clearInterval(intervalID);
    //           }
    //           if (iSeconds > 100) {
    //             xhr.abort();
    //             console.info('request aborded');
    //             clearInterval(intervalID);
    //           }
    //           updateTitle();
    //           iSeconds++;
    //         }, 1000);
    //         updateTitle();
    //       });
    //     }
    //   }, oOptions['timeout']);
    // }
  }

  /**
   * Abort la requête si la valeur de l'observable est égale à true
   * @param xhr
   * @param oOptions
   */
  private ajaxRequestHandleAbord_(xhr: any, oOptions: any) {
    if (oOptions.abord && oOptions.abord instanceof Observable) {
      if (oOptions.abord.subscribe instanceof Function) {
        oOptions.abord.subscribe(abortStatut => {
          if (abortStatut === true) {
            xhr['aborded'] = true;
            xhr.abort();
          }
        });
      }
    }
  }

  /**
   * Fusionne deux objets
   * @param  {any} oTarget
   * @param  {any} oSource
   * @returns any
   */
  private mergeObject_(oTarget: any, oSource: any): any {
    for (const sKey in oSource) {
      if (oSource.hasOwnProperty(sKey)) {
        oTarget[sKey] = oSource[sKey];
      }
    }
    return oTarget;
  }

  /**
   * Concatène les paramètres dans l'url
   * @param oParams
   * @param sUrl
   */
  private concatParamsWithURL_(oParams: any, sUrl: any) {
    if (oParams) {
      for (const sParamKey in oParams) {
        if (oParams.hasOwnProperty(sParamKey)) {
          if (this.vitisCheck.isObject(oParams[sParamKey])) {
            oParams[sParamKey] = JSON.stringify(oParams[sParamKey]);
          }
          sUrl += '&' + sParamKey + '=' + encodeURIComponent(oParams[sParamKey]);
        }
      }
    }

    return sUrl;
  }

  /**
   * Nettoie les paramètres en enlevant les valeurs undefined et les token
   * @param {object} oObject
   * @returns {object}
   */
  private cleanParams_(oObject: any, sUrl: any): any {

    // Service d'authentification
    const authService = this.injector.get(AuthService);

    // Nettoie les paramètres
    for (const key in oObject) {
      if (oObject.hasOwnProperty(key)) {
        if (this.vitisCheck.isUndefined(oObject[key])) {
          delete oObject.key;
        }
        if (this.vitisCheck.isDefAndNotNull(authService)) {
          if (key === 'token' || (this.vitisCheck.isDefAndNotNull(authService.getToken()) && oObject[key] === authService.getToken())) {
            console.error('Token in params:', oObject, sUrl);
            alert('Attention: token passé en paramètre');
            delete oObject.key;
          }
        }
      }
    }
    return oObject;
  }

  /**
   * Fonction de récupération du cache
   * Si le résultat est déjà présent dans le cache alors retourne le résultat, sinon null
   * @param oOptions
   */
  private ajaxRequestGetFromCache_(oOptions: any): any {
    const oParams = oOptions.params ? oOptions.params : oOptions.data ? oOptions.data : null;
    const sUrl = this.concatParamsWithURL_(oParams, oOptions.url);
    const oResult = this.sessionStorageService.get(sUrl);
    if (oResult) {
      return oResult;
    }
    return null;
  }

  /**
   * Fonction d'écriture du cache
   * Écrit le résultat de la requête dans sessionsStorage
   * @param oOptions
   * @param oResult
   */
  private ajaxRequestWriteToCache_(oOptions: any, oResult: any): any {
    const oParams = oOptions.params ? oOptions.params : oOptions.data ? oOptions.data : null;
    const sUrl = this.concatParamsWithURL_(oParams, oOptions.url);
    try {
      this.sessionStorageService.set(sUrl, oResult);
    } catch (error) {

    }
  }
}
