import { Component, OnInit, ViewChild } from '@angular/core';
import { ActivatedRoute, Router } from '@angular/router';
import { TranslateService } from '@ngx-translate/core';
import { VitisCheck, vitisAlert, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { DataGridComponent } from '../data-grid/data-grid.component';
import { DataGridOptions } from 'src/app/vitis/data-grid/data-grid.model';
import { WorkspaceListOptions } from '../workspacelist/workspacelist.model';
import { environment } from 'src/environments/environment';
import { AuthService } from 'src/app/vitis/core/services/auth/auth.service';
import { CustomTranslationService } from './services/customtranslation.service';
import ResizeObserver from 'resize-observer-polyfill';

@Component({
  selector: 'app-customtranslation',
  templateUrl: './customtranslation.component.html',
  styleUrls: ['./customtranslation.component.scss']
})
export class CustomTranslationComponent implements OnInit {

  @ViewChild(DataGridComponent, { static: false, read: DataGridComponent }) set oDataGrid(oDataGrid: DataGridComponent) {
    setTimeout(() => {
      if (this.vitisCheck.isDefAndNotNull(oDataGrid)) {
        this.initDataGrid(oDataGrid);
        // Sauve le composant de la grille de données.
        this.oDataGrid_ = oDataGrid;
      }
    });
  }
  // Composant de la grille de données.
  public oDataGrid_: any;

  // Options de la grille de données.
  public oDataGridOptions: DataGridOptions = {
    datatableOptions: {
      select: {
        style: 'multiple',
        selector: 'tr'
      },
      processing: true,
      serverSide: false,
      searching: false,
      lengthMenu: [5, 10, 20, 50, 100],  // Tableau choix du nombre d'éléments par page
      pageLength: Number(environment.properties.rows_per_page),  // Nombre d'éléments par page
      order: [4, 'desc'], // Ordre
      displayStart: 0,  // Offset
      scrollY: () => {
        this.calcDataTableHeight(); // Calcul de la hauteur de la grille.
      },
      scrollX: true,
      rowId: '',
      selected: 'selected'
    },
    datagridOptions: WorkspaceListOptions
  };

  // Instance de la grille de données.
  public dataTableInstance: any;

  // Mode en cours
  public oMode: any;

  // Types de filtre disponible
  public oFilterTypes: any = {
    custom: 'VITIS_WORKSPACELIST_FILTER_CUSTOM',
    form: 'VITIS_WORKSPACELIST_FILTER_FORM'
  };

  // Filtre en cours
  public oFilter: any = {
    type: 'custom',
    nodes: []
  };

  constructor(
    private route: ActivatedRoute,
    private router: Router,
    private vitisCheck: VitisCheck,
    private translate: TranslateService,
    private auth: AuthService,
    private vitisRequest: VitisRequest,
    private customTranslationService: CustomTranslationService
  ) { }

  ngOnInit() {
    // Mode en cours d'utilisation
    this.route.data.subscribe(mode => {
      this.oMode = mode;
      // options app.json
      if (mode.options) {
        this.oDataGridOptions.datagridOptions = mode.options;
      }
    });
  }

  /**
  * Initialise la grille de données.
  * @param {DataGridComponent} oDataGrid
  * @returns void
  */
  private initDataGrid(oDataGrid: DataGridComponent) {
    // Paramètres du plugin DataTable.
    this.oDataGridOptions.datatableOptions.columns = this.oMode.options.columns;
    //
    this.oDataGridOptions.datatableOptions.rowId = this.oDataGridOptions.datagridOptions.id_field;
    // Boutons display (Evènement).
    this.oDataGridOptions.displayEvent = (data: any) => {
      this.displayRowForm_(data, 'display');
    };
    // Boutons update (Evènement).
    this.oDataGridOptions.updateEvent = (data: any) => {
      this.displayRowForm_(data, 'update');
    };
    // Colonnes de la liste (définies dans le mode).
    this.oDataGridOptions.datatableOptions.columns = this.getParsedColumns_();
    // Bouton delete
    if (this.oDataGridOptions.datagridOptions.delete === true) {
      this.oDataGridOptions.buttons = [{
        type: 'button',
        class: 'btn-default',
        label: 'LIST_CONFIGURATION_CUSTOM_TRANSLATIONS_DELETE_BTN',
        icon: 'fa fa-trash',
        event: () => {
          this.deleteSelectedRows();
        }
      }];
    }
    // Création de la grille.
    this.dataTableInstance = oDataGrid.initTable(this.oDataGridOptions);
    this.customTranslationService.dataTableInstance = this.dataTableInstance;
    // Evènement à l'affichage de la grille.
    this.dataTableInstance.on('draw.dt', () => {
      // Format les champs avec un template
      // oDataGrid.formatFields(this.oDataGridOptions);
      // Réajuste les dimensions de la table
      this.autoAjustTables(oDataGrid);
    });

    // Rechargement de la grille (bouton refresh).
    this.dataTableInstance.ajax.reload = () => {
      // Recharge les traductions personnalisées.
      this.customTranslationService.loadCustomTranslations()
        .then(() => {
          // Recharge les données de la grille.
          this.setDataGridContent();
        });
    };
    // Recharge les traductions personnalisées.
    this.customTranslationService.loadCustomTranslations()
      .then(() => {
        // Recharge les données de la grille.
        this.setDataGridContent();
      });

    // Réajuste les dimensions de la table lors fermeture/ouverture du dashboard
    const dDashboardContent = document.querySelector('#dashboard-content');
    const resizeObserver = new ResizeObserver(entries => {
      this.autoAjustTables(oDataGrid);
    });
    resizeObserver.observe(dDashboardContent);
  }

  /**
   * Récupère les colonnes définies dans le mode au format BootstrapTable
   * @returns Array
   */
  private getParsedColumns_(): Array<any> {
    const aColumns: any = [{
      targets: 0,
      data: null,
      defaultContent: '',
      orderable: false,
      className: 'select-checkbox',
      width: '40px'
    }];

    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {
      // Colonne boutons
      if (this.oDataGridOptions.datagridOptions.display === true
        || this.oDataGridOptions.datagridOptions.update === true) {

        let defaultContent = '<div style="text-align: center;">';
        if (this.oDataGridOptions.datagridOptions.display === true &&
          (environment.properties.data_grid_update_btn_only === false || this.oDataGridOptions.datagridOptions.update !== true)) {
          defaultContent += '<span class="datatable-btn datatable-btn-display icon-vitis-search"></span>';
        }
        if (this.oDataGridOptions.datagridOptions.update === true) {
          defaultContent += '<span class="datatable-btn datatable-btn-edit icon-vitis-draw"></span>';
        }
        defaultContent += '</div>';

        aColumns.push({
          targets: -1,
          data: null,
          orderable: false,
          defaultContent,
          width: '90px'
        });
      }

      // Colonnes du mode
      if (this.vitisCheck.isArray(this.oDataGridOptions.datagridOptions.columns)) {
        const aUserPrivileges = this.auth.getPrivileges();
        for (let i = 0; i < this.oDataGridOptions.datagridOptions.columns.length; i++) {
          const column = this.oDataGridOptions.datagridOptions.columns[i];
          // Privilège obligatoire pour afficher la colonne.
          if (this.vitisCheck.isDefAndNotNull(column.privilege)) {
            if (aUserPrivileges.indexOf(column.privilege) == -1) {
              continue;
            }
          }
          //
          column.translateTitle = this.translate.instant(column.title);
          aColumns.push({
            title: column.translateTitle,
            data: column.field,
            name: column.field,
            visible: column.visible === false ? false : true,
            type: column.template ? 'html' : 'string',
            render: (data: any, type: any, row: any, meta: any) => {
              if (column.template) {
                data = '<field-render column_number="' + i + '">' + encodeURIComponent(JSON.stringify(row)) + '</field-render>';
              }
              return data;
            },
            orderable: column.sortable,
            template: column.template,
            // Alignement du contenu de la colonne.
            className: 'table-align-' + (column.align ? column.align : 'left')
          });
        }
      }
    }
    return aColumns;
  }

  /**
   * Calcule la hauteur de la table
   */
  private calcDataTableHeight = () => {
    if (this.vitisCheck.isNumber(this.oMode.options.height)) {
      return this.oMode.options.height;
    } else {
      const dashboardHeight = $('#dashboard-content').height();
      const titleHeight = $('.mode-title').height();
      const filterHeight = $('.grid-filter').height();
      const menusHeight = 150;
      const tableHeight = dashboardHeight - (menusHeight + filterHeight + titleHeight);
      return tableHeight;
    }
  }

  /**
   * Réajuste les dimensions de la table
   * @param {DataGridComponent} oDataGrid
   */
  private autoAjustTables = (oDataGrid: DataGridComponent) => {
    if ($('#' + oDataGrid.tableId).length) {
      this.dataTableInstance.columns.adjust();
      $('#' + oDataGrid.tableId + '_wrapper').find('.dataTables_scrollBody').css('height', this.calcDataTableHeight());
    }
  }

  /**
   * Affichage des traductions dans la grille.
   */
  private setDataGridContent = () => {
    // Traductions de la langue sélectionnée.
    const oTranslations: any = this.translate.translations[environment.properties.language];
    const oOriginalTranslations: any = this.customTranslationService.oTranslations[environment.properties.language];
    const aRows: any = [];
    // Formatage des données de traduction pour la grille.
    for (const sKey in oTranslations) {
      if (sKey != '') {
        aRows.push({
          translation_id: sKey,
          original_text: oOriginalTranslations[sKey] ? oOriginalTranslations[sKey] : '',
          text: (this.customTranslationService.aCustomTranslations.indexOf(sKey) !== -1) ? (oTranslations[sKey] ? oTranslations[sKey] : '') : '',
          custom_translation: (this.customTranslationService.aCustomTranslations.indexOf(sKey) !== -1)
        });
      }
    }
    // Application des filtres de la grille.
    let aFilteredRows: any = [];
    if (this.oFilter.nodes.length > 0) {
      let bAddRow: boolean;
      for (let i = 0; i < aRows.length; i++) {
        bAddRow = true;
        for (let j = 0; j < this.oFilter.nodes.length; j++) {
          if (String(aRows[i][this.oFilter.nodes[j].column.field]).toLowerCase().indexOf(
            String(this.oFilter.nodes[j].value).toLowerCase()) === -1) {
            bAddRow = false;
            break;
          }
        }
        if (bAddRow) {
          aFilteredRows.push(aRows[i]);
        }
      }
    } else {
      aFilteredRows = aRows;
    }

    // Vidage de la grille.
    this.dataTableInstance.clear();
    // Affichage des traductions dans la grille.
    this.dataTableInstance.rows.add(aFilteredRows).draw();
    // Réajustement de la taille des colonnes.
    setTimeout(() => {
      this.dataTableInstance.columns.adjust();
    }, 500);
  }

  /**
   * Affiche le formulaire correspondant
   * @param  {any} oRow
   * @param  {String} sFormName
   */
  private displayRowForm_(oRow: any, sFormName: String) {
    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {
      if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.id_field)) {
        if (this.vitisCheck.isDefAndNotNull(oRow)) {
          if (this.vitisCheck.isDefAndNotNull(oRow[this.oDataGridOptions.datagridOptions.id_field])) {
            if (this.vitisCheck.isDefAndNotNull(sFormName)) {
              // Redirection vers le formulaire d'édition.
              const sId = String(oRow[this.oDataGridOptions.datagridOptions.id_field]).toLowerCase();
              const aModeRoute = this.oMode.route.split('/');
              this.router.navigate([sId, sFormName], { relativeTo: this.route.parent });
            }
          }
        }
      }
    }
  }

  /**
   * Suppression des traductions personnalisées sélectionnées.
   */
  private deleteSelectedRows(): void {
    const aTranslationId = this.oDataGrid_.getSelectedIds_();
    // Au moins un enregistrement doit être sélectionné.
    if (aTranslationId.length > 0) {
      // Au moins une traduction personnalisée doit être sélectionné.
      const aCustomTranslation: any = [];
      for (let i = 0; i < aTranslationId.length; i++) {
        if (this.customTranslationService.aCustomTranslations.indexOf(aTranslationId[i]) !== -1) {
          aCustomTranslation.push(aTranslationId[i]);
        }
      }
      if (aCustomTranslation.length > 0) {
        if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {
          if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.ressource)) {
            // Chemin de la ressource REST.
            let sRessourceURL = environment.properties.app_api_url;
            sRessourceURL += '/' + this.oDataGridOptions.datagridOptions.ressource;
            vitisAlert.alert({
              type: 'warning',
              title: this.translate.instant('LIST_DELETE_CONFIRM'),
              showCancelButton: true
            }).then((result: any) => {
              if (result.isConfirmed === true) {
                // Suppression des traductions personnalisées en base.
                this.vitisRequest.ajaxRequestPromise({
                  method: 'DELETE',
                  url: sRessourceURL,
                  data: {
                    idList: aCustomTranslation
                  }
                }).then((resp) => {
                  if (resp.data) {
                    if (resp.data.status === 1) {
                      // Message de succès.
                      vitisAlert.fireToast({
                        icon: 'success',
                        title: this.translate.instant('LIST_DELETE_OK')
                      });
                      // Recharge les traductions personnalisées.
                      this.customTranslationService.loadCustomTranslations()
                        .then(() => {
                          // Recharge les données de la grille.
                          this.setDataGridContent();
                        });
                    }
                  }
                });
              }
            });
          }
        }
      } else {
        vitisAlert.alert({
          type: 'error',
          title: this.translate.instant('LIST_CONFIGURATION_CUSTOM_TRANSLATIONS_DELETE_NO_TRANSLATION'),
        });
      }
    } else {
      vitisAlert.alert({
        type: 'error',
        title: this.translate.instant('LIST_DELETE_NO_SELECTION'),
      });
    }
  }
}