import { Component, OnInit, Input, ElementRef } from '@angular/core';
import { TranslateService } from '@ngx-translate/core';
import { Validators } from '@angular/forms';
import { FormreaderField } from '../../formreader-field';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';
import { ColorPickerService, Hsva } from 'ngx-color-picker';
// use in html : https://www.npmjs.com/package/ngx-color-picker

/**
 * Type color_picker
 *
 * Arguments:
 * type : type de champ
 * name : attribut utilisé par le champ
 * label : libellé utilisé par le champ
 * nb_cols : nombre de colonnes à utiliser (de 1 à 12)
 * id? : id du champ à utiliser
 * required? : champ requis
 * default_value? : valeur par défaut
 * options?
 * options.format? : rgb, rgba, hex, hsl, hsla, cmyk, cmyka, hsv, hsva
 *
 * Exemple:
 * {
 *   "type": "color_picker",
 *   "name": "colorpicker",
 *   "label": "Colorpicker",
 *   "nb_cols": 12,
 *   "required": true,
 *   "options": {
 *       "format": "rgba",
 *   }
 * }
 */
@Component({
  selector: 'app-colorpicker',
  templateUrl: './colorpicker.component.html',
  styleUrls: ['./colorpicker.component.scss']
})
export class ColorpickerComponent extends FormreaderField implements OnInit {

  @Input() sSize: string;
  @Input() oField: any;
  @Input() oFormValues: any;

  public color: any;
  public displayColorPicker: boolean = false;
  public cpOutputFormat: string = 'rgba';
  public cpAlphaChannel: string = 'always';
  public cpCmykEnabled: boolean = false;

  constructor(
    translate: TranslateService,
    vitisCheck: VitisCheck,
    public element: ElementRef,
    private oCPService: ColorPickerService
  ) {
    super(translate, vitisCheck);
  }

  ngOnInit() {

    // set les options de ngx-color-picker
    if (this.oField) {
      if (this.oField.options) {
        this.setOptions_(this.oField.options);
      }
    }

    // Fonction getValue
    this.oField.getValue = () => {
      return this.getValue(this.oFormValues, this.oField);
    };

    // Fonction setValue
    this.oField.setValue = (oValue: any) => {
      if (this.vitisCheck.isDefAndNotNull(oValue) && oValue != '') {
        this.color = oValue;
      }
      return this.setValue(this.oFormValues, this.oField, oValue);
    };

    this.oField.refreshTooltip = () => {
      return this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);
    };

    this.oField.onOptionsChanges = (oNewOptions: any) => {
      this.setOptions_(oNewOptions);

      // Permet de recharger le composant
      this.displayColorPicker = false;
      setTimeout(() => {
        this.displayColorPicker = true;
      });
    };

    // Tooltip
    this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);

    if (this.vitisCheck.isDefAndNotNull(this.oField.getValue()) && this.oField.getValue() != '') {
      this.color = this.oField.getValue();
    }

    // Création d'un validateur spécifique
    this.updateFieldValidator_(this.oField);

    // Affiche le composant
    this.displayColorPicker = true;
  }

  /**
   * Attribut la valeur du colorpicker au controleur lors d'une modification sur le colorpicker
   * @param colorCode
   */
  public onColorPickerChange(colorCode: string): void {
    if (this.cpOutputFormat === 'cmyka') {
      return;
    }

    if (this.cpOutputFormat === 'hsva') {
      colorCode = this.hexToHsva(colorCode);
    }

    this.setColorCode(colorCode);
  }

  /**
   * Transforme le hexa vers du hsv
   * @param {string} sColorCode code hexa type #FFFFFF
   * @returns {string} code hsv/hsva hsv(360,100%,100%,1)
   */
  private hexToHsva (sColorCode: string): string {
    const bAllowAlpha: boolean = (sColorCode.length === 9);
    const aColor: Hsva = this.oCPService.stringToHsva(sColorCode, bAllowAlpha);
    let sReturn = (bAllowAlpha) ? 'hsva' : 'hsv';
    sReturn += '(';
    sReturn += Math.round((aColor.h) * 360) + ',';
    sReturn += Math.round((aColor.s) * 100) + '%,';
    sReturn += Math.round((aColor.v) * 100) + '%';
    if (bAllowAlpha) {
      sReturn += ',' + Math.round(aColor.a * 100) / 100;
    }
    sReturn += ')';
    return sReturn;
  }

  /**
   * Attribut la valeur du colorpicker au controleur lors d'une modification sur le colorpicker
   * uniquement pour le cmyk
   * @param colorCode
   */
  public onColorPickerChangeCmyk(colorCode: string): void {
    if (this.cpOutputFormat !== 'cmyka') {
      return;
    }
    this.setColorCode(colorCode);
  }

  /**
   * possibilité de set la valeur depuis deux fonction différente en fonction de
   * si la couleur est de type cmyk ou non
   * @param colorCode
   */
  private setColorCode(colorCode: string): void {
    this.oField.setValue(colorCode);
    this.oFormValues.controls[this.oField.name].markAsDirty();
    this.oFormValues.controls[this.oField.name].markAsTouched();
  }

  /**
   * Création d'un validateur spécifique
   * @param oField
   */
  private updateFieldValidator_(oField: any) {

    let aValidators = [];

    // Récupère les validators inclus par FormReader
    if (this.vitisCheck.isArray(oField.validators_list)) {
      aValidators = oField.validators_list;
    }

    // Pattern validator
    const aPatternMap: any = {
      'rgb': '^rgb\\((\\d{1,3}),\\s*(\\d{1,3}),\\s*(\\d{1,3})\\)$',
      'rgba': '^rgba\\((\\d{1,3}),\\s*(\\d{1,3}),\\s*(\\d{1,3}),\\s*(\\d*(?:\\.\\d+)?)\\)$',
      'hsl': '^hsl\\((\\d{1,3}),\\s*(\\d{1,3}%),\\s*(\\d{1,3}%)\\)$',
      'hsla': '^hsla\\((\\d{1,3}),\\s*(\\d{1,3}%),\\s*(\\d{1,3}%),\\s*(\\d*(?:\\.\\d+)?)\\)$',
      'hsv': '^hsv\\((\\d{1,3}),\\s*(\\d{1,3}%),\\s*(\\d{1,3}%)\\)$',
      'hsva': '^hsva\\((\\d{1,3}),\\s*(\\d{1,3}%),\\s*(\\d{1,3}%),\\s*(\\d*(?:\\.\\d+)?)\\)$',
      'cmyk': '^cmyk\\((\\d{1,3}),\\s*(\\d{1,3}),\\s*(\\d{1,3}),\\s*(\\d{1,3})\\)$',
      'cmyka': '^cmyka\\((\\d{1,3}),\\s*(\\d{1,3}),\\s*(\\d{1,3}),\\s*(\\d{1,3}),\\s*(\\d*(?:\\.\\d+)?)\\)$',
      'hex': '^#([a-f0-9]{6}|[a-f0-9]{3})\\b$',
      'hex8': '^#([a-f0-9]{8}|[a-f0-9]{4})\\b$'
    };

    let sPattern: any;
    if (oField.options) {
      if (oField.options.format in aPatternMap) {
        sPattern = aPatternMap[oField.options.format];
      }
    }

    aValidators.push(Validators.pattern(sPattern));
    this.oFormValues.controls[oField.name].setValidators(aValidators);

    if (!oField.validation_messages) {
      oField.validation_messages = [];
    }
    oField.validation_messages.push({
      type: 'pattern',
      message: this.translate.instant('FORM_READER_ERROR_COLOR_PICKER')
    });
  }

  /**
   * Met à jour les options de ngx-color-picker
   * @param oOptions
   */
  private setOptions_(oOptions: any) {
    if (oOptions) {
      if (this.vitisCheck.isDefAndNotNull(oOptions.format)) {
        switch (oOptions.format) {
          case 'rgb':
            this.cpOutputFormat = 'rgba';
            break;
          case 'hex8':
            this.cpOutputFormat = 'hex';
            break;
          case 'hsv':
            this.cpOutputFormat = 'hsva';
            break;
          case 'hsl':
            this.cpOutputFormat = 'hsla';
            break;
          case 'cmyk':
          case 'cmyka':
            this.cpOutputFormat = 'cmyka';
            this.cpCmykEnabled = true;
            break;
          default:
            this.cpOutputFormat = oOptions.format;
            break;
        }
        if (oOptions.format === 'rgba' || oOptions.format === 'hsla'
          || oOptions.format === 'cmyka' || oOptions.format === 'hsva'
          || oOptions.format === 'hex8') {
          this.cpAlphaChannel = 'always';
        } else {
          this.cpAlphaChannel = 'disabled';
        }
      }
    }
  }
}
