import { Component, OnInit, Input, ViewChild, ElementRef } from '@angular/core';
import { GlobalFilePickerComponent } from './global-file-picker/global-file-picker.component';
import { FormreaderField } from '../../formreader-field';
import { TranslateService } from '@ngx-translate/core';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';
import { ValidationErrors, ValidatorFn } from '@angular/forms';

@Component({
  selector: 'app-file-picker',
  templateUrl: './file-picker.component.html',
  styleUrls: ['./file-picker.component.scss']
})
export class FilePickerComponent extends FormreaderField implements OnInit {

  @Input() sSize: string;
  @Input() oField: any;
  @Input() oFormValues: any;

  @ViewChild(GlobalFilePickerComponent, { static: false }) oGlobalFilePicker: GlobalFilePickerComponent;

  constructor(
    translate: TranslateService,
    vitisCheck: VitisCheck,
    public element: ElementRef
  ) {
    super(translate, vitisCheck);
  }

  ngOnInit() {

    // Fonction getValue
    this.oField.getValue = () => {
      const oFileList = this.oGlobalFilePicker.getValue();
      return oFileList; // this.getValue(this.oFormValues, this.oField);
    };

    // Fonction setValue
    this.oField.setValue = (oValue: any) => {
      return this.setValue(this.oFormValues, this.oField, oValue);
    };

    this.oField.refreshTooltip = () => {
      return this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);
    };

    this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);


    this.oField.refresh = () => {
      this.oGlobalFilePicker.refresh();
    };

    this.oField.reset = () => {
      this.oGlobalFilePicker.reset();
    };

    // Validateurs
    this.updateFieldValidator_();
  }

  /**
   * Check si un fichier dans la liste du champ upload ne respecte pas la limitation de taille
   * @returns {ValidationErrors | null}
   */
  public filePickerValidator(): ValidationErrors | null {

    let isValid = true;

    if (this.oGlobalFilePicker) {
      const iLimitation: number = (this.oField.max_size) ? this.oField.max_size : -1;

      const oFileList = this.oGlobalFilePicker.getValue();
      for (const oFile of oFileList.files) {
        if (!this.vitisCheck.isUploadable(oFile, iLimitation)) {
          isValid = false;
        }
      }

      // Met à jour le message d'erreur
      if (this.oField &&
        this.oField.validation_messages) {
        for (let i = 0; i < this.oField.validation_messages.length; i++) {
          const element = this.oField.validation_messages[i];
          if (element.type === 'validateFileSize') {
            element.message = this.translate.instant('FORM_UPLOAD_FILE_TOO_BIG_NOT_VALID');
          }
        }
      }
    }

    return isValid ? null : {
      validateFileSize: {
        valid: false,
      }
    };
  }

  /**
   * Création d'un validateur spécifique
   * @param oField
   */
  private updateFieldValidator_(): void {

    let aValidators = [];

    // Récupère les validators inclus par FormReader
    if (this.vitisCheck.isArray(this.oField.validators_list)) {
      aValidators = this.oField.validators_list;
    }

    // Multiple email
    const self = this;
    const validateFileSize: ValidatorFn = () => {
      return self.filePickerValidator();
    };
    aValidators.push(validateFileSize);

    // Ajoute les validateurs
    this.oFormValues.controls[this.oField.name].setValidators(aValidators);

    // Messages
    if (!this.oField.validation_messages) {
      this.oField.validation_messages = [];
    }

    this.oField.validation_messages.push({
      type: 'validateFileSize',
      message: this.translate.instant('FORM_UPLOAD_FILE_TOO_BIG_NOT_VALID') + '.'
    });
  }

  /**
   * Set la valeur du fichier
   * @param oFile
   */
  public setFileValue(oFile): void {
    const oValue = oFile.map(file => {
      if (this.vitisCheck.isDefAndNotNull(file)) {
        if (this.vitisCheck.isDefAndNotNull(file._url)) {
          return file._url;
        }
        return this.oGlobalFilePicker.getFileName(file);
      }
    });

    // Set l'élément comme modifié
    this.oFormValues.get(this.oField.name).markAsTouched();

    this.oField.setValue(oValue);
  }

  /**
   * Récupère la liste des urls pour le global file picker
   */
  public getUrls() {
    const sValue = this.oFormValues.get(this.oField.name).value;
    if (this.vitisCheck.isDefAndNotNull(sValue) && this.vitisCheck.isString(sValue) && (sValue.match(/http/g) || []).length > 1) {
      return sValue.split(',');
    } else {
      return this.oFormValues.get(this.oField.name).value;
    }
  }
}
