import { Component, OnInit, Input, ElementRef } from '@angular/core';
import { TranslateService } from '@ngx-translate/core';
import { FormreaderField } from '../../formreader-field';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';

/**
 * Type liste déroulante
 *
 * Arguments:
 * type : type de champ
 * name : attribut utilisé par le champ
 * label : libellé utilisé par le champ
 * nb_cols : nombre de colonnes à utiliser (de 1 à 12)
 * id? : id du champ à utiliser
 * required? : champ requis
 * default_value? : valeur par défaut
 * editable? : select éditable
 * refresh_btn? : select avec bouton pour rafraichir la source de données
 * datasource? : source de données
 * datasource.datasource_id : identifiant de la source de données (spécifiée dans le formulaire)
 * datasource.id_key : attribut utilisé pour la valeur du champ
 * datasource.label_key : attribut à afficher
 * datasource.attributs : attributs à demander séparés par des "|"
 * datasource.sort_order? : type de sort (ASC, DESC)
 * datasource.order_by? : attribut utilisé pour l'ordre
 * datasource.distinct? : valeurs distinctes (true, false)
 * datasource.parameters? : paramètres supplémentaires à passer dans la requêtre GET
 *
 * Exemple:
 * {
 *   "type": "select",
 *   "name": "Timezone",
 *   "label": "Liste",
 *   "required": true,
 *   "nb_cols": 12,
 *   "datasource": {
 *     "datasource_id": "datasource_timezone_id",
 *     "sort_order": "ASC",
 *     "distinct": "true",
 *     "label_key": "label",
 *     "order_by": "name",
 *     "id_key": "name",
 *     "attributs": "name|label",
 *     "parameters": {
 *       "filter": {"relation":"AND","operators":[{"column":"country","compare_operator":"=","value":"={{country}}"}]}
 *     }
 *   }
 * }
 */
@Component({
  selector: 'app-select',
  templateUrl: './select.component.html',
  styleUrls: ['./select.component.scss']
})
export class SelectComponent extends FormreaderField implements OnInit {

  @Input() sSize: string;
  @Input() oField: any;
  @Input() oFormValues: any;
  @Input() oFormReader: any;

  constructor(
    translate: TranslateService,
    vitisCheck: VitisCheck,
    public element: ElementRef
  ) {
    super(translate, vitisCheck);
  }

  ngOnInit() {

    // Fonction getValue
    this.oField.getValue = () => {
      return this.getValue(this.oFormValues, this.oField);
    };

    // Fonction setValue
    this.oField.setValue = (oValue: any) => {
      return this.setValue(this.oFormValues, this.oField, String(oValue));
    };

    this.oField.refreshTooltip = () => {
      return this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);
    };

    // Tooltip
    this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);

    // Observe l'arrivée des choix (pour utilisation avec datasource)
    if (this.vitisCheck.isDefAndNotNull(this.oField.options)) {
      if (this.vitisCheck.isDefAndNotNull(this.oField.options.observableChoises)) {
        this.oField.options.observableChoises.subscribe((aChoises: any) => {

          // Remplace les valeurs des options de type nombre au format string
          this.makeOptionsAsString_(this.oField.options.choises);

          // Set valeur "" si aucune valeur
          setTimeout(() => {
            if (!this.vitisCheck.isDefAndNotNull(this.oField.getValue()) && !this.oField.disabled) {
              this.oField.setValue('');
            }
          });
        });
      }
    }
  }

  /**
   * Retourne true / false suivant si il faut afficher une option vide
   * @returns boolean
   */
  public hasEmptyOption(): boolean {
    let firstOption = true;

    if (this.vitisCheck.isDefAndNotNull(this.oField)) {
      if (this.oField.required) {
        firstOption = false;
      }

      if (this.vitisCheck.isDefAndNotNull(this.oField.options)) {
        if (this.vitisCheck.isDefAndNotNull(this.oField.options.choises)) {
          if (this.vitisCheck.isDefAndNotNull(this.oField.options.choises[0])) {
            if (this.oField.options.choises[0].value === '') {
              firstOption = false;
            }
          }
        }
      }
    }

    return firstOption;
  }

  /**
   * Remplace les valeurs des options de type nombre au format string
   * @param  {Array<any>} aChoises
   * @returns void
   */
  private makeOptionsAsString_(aChoises: Array<any>): void {

    for (let i = 0; i < aChoises.length; i++) {
      if (this.vitisCheck.isNumber(aChoises[i].value)) {
        aChoises[i].value = String(aChoises[i].value);
      }
    }
  }

  /**
   * Raffraichit les données depuis datasource
   */
  public refreshDatasource() {
    this.oFormReader.reloadDatasource(this.oField);
  }
}
