import { Component, OnInit, Input, OnChanges, Output, EventEmitter, ChangeDetectorRef, ViewChild } from '@angular/core';
import { Validators, FormBuilder, FormGroup, FormControl, FormsModule } from '@angular/forms';
import { TranslateService } from '@ngx-translate/core';
import { Observable, Subject } from 'rxjs';
import { VitisCheck, VitisRequest, VitisClone, vitisAlert } from 'src/app/vitis/core/services/utils/vitis-utils';
import { environment } from 'src/environments/environment';
import { FieldsService } from './services/fields.service';
import { FormreaderService } from './services/formreader.service';
import { ActivatedRoute } from '@angular/router';
import { AuthService } from 'src/app/vitis/core/services/auth/auth.service';
import { FunctionService } from '../core/services/utils/function/function.service';
import { ModalFormreaderOptions } from './formreader.model';

declare var $: any;

@Component({
  selector: 'app-formreader',
  templateUrl: './formreader.component.html',
  styleUrls: ['./formreader.component.scss']
})
export class FormreaderComponent implements OnInit, OnChanges {

  @Input() public formOptions: any;

  @Output() public refreshParentFormValues = new EventEmitter<any>();

  public oFormDefinition: any;
  public sFormDefinitionName: any;
  public oFormValues: FormGroup;
  public showTabs: boolean;
  public iDisplayedTab: number;
  public fieldsService: FieldsService;
  public oParams: any;

  @ViewChild(FormreaderComponent, { static: false }) oModalFormReader: FormreaderComponent;
  public oModalOptions: ModalFormreaderOptions = {
    title: '',
    cancel_button: true,
    valid_button: true,
    initEvent: (this_: FormreaderComponent) => { },
    beforeEvent: (this_: FormreaderComponent) => { },
    afterEvent: (this_: FormreaderComponent) => { },
    event: this.submitModalEvent_
  };

  public isSubmitted: boolean = false;

  private hasSetValues: boolean = false;

  constructor(
    private translate: TranslateService,
    private formBuilder: FormBuilder,
    private vitisRequest: VitisRequest,
    private vitisCheck: VitisCheck,
    private vitisClone: VitisClone,
    private formreaderSrvc: FormreaderService,
    private route: ActivatedRoute,
    private authService: AuthService,
    protected changeDetectorRef: ChangeDetectorRef,
    private functionService: FunctionService,
  ) { }

  ngOnInit() {
    this.route.params.subscribe(oParams => {
      this.oParams = oParams;
    });
  }

  ngOnChanges() {
    if (this.formOptions && this.formOptions.form && this.formOptions.formName) {
      this.oFormValues = null; // Déclenche ngIf pour réinitaliser les champs du formulaire
      // Attend que ngIf ai été activé
      setTimeout(() => {
        this.initForm(this.formOptions.form, this.formOptions.formName, this.formOptions.formValues);

        // Sauve la liste des onglets et les valeurs du formulaire.
        this.formOptions.formTabs = this.formOptions.form[this.formOptions.formName].tabs.list;
        this.formOptions.oFormValues = this.oFormValues;
        // Detecte les changements sur sa vue et ses vues enfants
        // utilisé pour revérifier la validité du ReactiveForm après un changement de valeurs et/ou de validateur
        this.changeDetectorRef.detectChanges();
      });
    }
  }

  /**
   * Initialise le FormReader
   * @param  {any} oFormDefinition
   * @param  {any} sFormDefinitionName
   * @param  {any} oFormValues?
   * @param  {boolean} bLoadDatasource? default true
   * @param  {boolean} bLoadWebserviceGrid? default true
   */
  public initForm(oFormDefinition: any, sFormDefinitionName: any, oFormValues?: any, bLoadDatasource: boolean = true, bLoadWebserviceGrid: boolean = true) {

    this.sFormDefinitionName = sFormDefinitionName;
    this.oFormDefinition = this.parseFormDef_(oFormDefinition, this.sFormDefinitionName, oFormValues, bLoadDatasource, bLoadWebserviceGrid);
    this.oFormValues = this.parseFormValues_(this.oFormDefinition, this.sFormDefinitionName, oFormValues);

    this.showTabs = true; // Toujours vrai pour le moment
    this.iDisplayedTab = 0;
    this.initFormTabs_();
    this.fieldsService = new FieldsService(this.vitisCheck,
      this.oFormValues,
      this.showTabs,
      this.changeDetectorRef,
      this.authService,
      this.functionService);

    // Lance initEvent
    if (this.vitisCheck.isDefAndNotNull(this.oFormDefinition[sFormDefinitionName].initEvent)) {
      const eventContainer = this.oFormDefinition[sFormDefinitionName].eventContainer;
      const initEvent = this.oFormDefinition[sFormDefinitionName].initEvent;
      this.runEvents_([initEvent], oFormValues, [eventContainer, this]);
    }

    this.setFormDynamicValues_();
    this.setFormDynamicEvents();

    setTimeout(() => {
      this.setFormValues(oFormValues);
      setTimeout(() => {
        // Evite les conflits dirty == true suite au setFormValues pour le changement de page (canDeactivate)
        this.oFormValues.markAsPristine();
      }, 500);
    });
  }

  /**
   * Réinitialise les valeurs du formulaire
   */
  public clearValues() {

    // Liste des fileds
    const aFields = this.getFormFields(this.oFormDefinition, this.sFormDefinitionName);

    // Set la valeur à vite en utilisant les setters
    for (let i = 0; i < aFields.length; i++) {
      const oField = aFields[i];
      const oFieldName = oField.name;

      if (this.vitisCheck.isDefAndNotNull(oFieldName)) {
        oField.setValue('');
      }
    }

    // Reset le FormGroup
    this.oFormValues.reset();
  }

  /**
   * Retourne l'élément du formulaire
   * @param  {string} sFieldName
   * @param  {any} oFormDefinition?
   * @param  {any} sFormDefinitionName?
   * @returns Any
   */
  public getFormField(sFieldName: string,
    oFormDefinition: any = this.oFormDefinition,
    sFormDefinitionName: any = this.sFormDefinitionName): any {
    return this.formreaderSrvc.getFormField(sFieldName, oFormDefinition, sFormDefinitionName);
  }

  /**
   * Retourne un tableau contenant les elements du formulaire
   * @param  {any} oFormDefinition?
   * @param  {any} sFormDefinitionName?
   * @returns Array
   */
  public getFormFields(oFormDefinition: any = this.oFormDefinition, sFormDefinitionName: any = this.sFormDefinitionName): Array<any> {
    return this.formreaderSrvc.getFormFields(oFormDefinition, sFormDefinitionName);
  }

  /**
   * Retourne un tableau contenant les lignes du formulaire
   * @param  {any} oFormDefinition?
   * @param  {any} sFormDefinitionName?
   * @returns Array
   */
  public getFormRows(oFormDefinition: any = this.oFormDefinition, sFormDefinitionName: any = this.sFormDefinitionName): Array<any> {
    return this.formreaderSrvc.getFormRows(oFormDefinition, sFormDefinitionName);
  }

  /**
   * Retourne le nombre de colones utilisées pour le formulaire
   * @param  {any} oFormDefinition?
   * @param  {any} sFormDefinitionName?
   * @returns string
   */
  public getFormSize(oFormDefinition: any = this.oFormDefinition, sFormDefinitionName: any = this.sFormDefinitionName): string {
    return this.formreaderSrvc.getFormSize(oFormDefinition, sFormDefinitionName);
  }

  /**
   * Retourne la taille des champs utiliés par le formulaire
   * @param  {any} oFormDefinition?
   * @param  {any} sFormDefinitionName?
   * @returns string
   */
  public getFormInputsSize(oFormDefinition: any = this.oFormDefinition, sFormDefinitionName: any = this.sFormDefinitionName): string {
    return this.formreaderSrvc.getFormInputsSize(oFormDefinition, sFormDefinitionName);
  }

  /**
   * Retourne les valeurs du formulaire sous forme d'objet
   * @param bOnlyFilled si true alors ne renvoi pas les valeurs vides
   * @returns Object
   */
  public getFormValues(bOnlyFilled = false): Object {

    // Valeurs du formulaire
    const oFormValues = {};
    const filesParams = {
      vitis_unchanged_files: {},
      vitis_deleted_files: {},
      vitis_thumbnail_image_files: {}
    };
    let bHaveFile: boolean = false;

    // Liste des fileds
    const aFields = this.getFormFields(this.oFormDefinition, this.sFormDefinitionName);

    // Récupère les valeurs en utilisant la fonction getValue surchargée dans oField par les composants
    for (let i = 0; i < aFields.length; i++) {
      const oField = aFields[i];
      const oFieldName = oField.name;

      if (this.vitisCheck.isDefAndNotNull(oFieldName)) {
        if (oField.getValue) {
          const val = oField.getValue();
          if (this.vitisCheck.isObject(val) && this.vitisCheck.isArray(val.files)) {
            bHaveFile = true;
            if (val.files.length > 1) {
              oFormValues[oFieldName] = [];
              for (let j = 0; j < val.files.length; j++) {
                // Ajoute les fichiers
                oFormValues[oFieldName].push(val.files[j]);
              }
            } else if (val.files.length == 1) {
              oFormValues[oFieldName] = val.files[0];
            }
            filesParams.vitis_unchanged_files[oFieldName] = val.vitis_unchanged_files;
            filesParams.vitis_deleted_files[oFieldName] = val.vitis_deleted_files;
            if (oField.type == 'image_wsdata') {
              filesParams.vitis_thumbnail_image_files[oFieldName] = val.vitis_thumbnail_image_files;
            }
            // Other
          } else {
            if (bOnlyFilled) {
              if (this.vitisCheck.isDefAndNotNull(val) &&
                val !== '') {
                oFormValues[oFieldName] = val;
              }
            } else {
              oFormValues[oFieldName] = val;
            }
          }
        }
      }
    }
    if (bHaveFile === true) {
      oFormValues['vitis_unchanged_files'] = JSON.stringify(filesParams.vitis_unchanged_files);
      oFormValues['vitis_deleted_files'] = JSON.stringify(filesParams.vitis_deleted_files);
      oFormValues['vitis_thumbnail_image_files'] = JSON.stringify(filesParams.vitis_thumbnail_image_files);
    }

    return oFormValues;
  }

  /**
   * Set les valeurs sous forme d'objet au formulaire
   * @param  {Object} oFormValues
   */
  public setFormValues(oFormValues: Object) {

    // Ajoute les valeurs par défaut
    oFormValues = this.useDefaultValues(oFormValues);

    // Vide les valeurs en cours
    this.oFormValues.reset();

    // Liste des fileds
    const aFields = this.getFormFields(this.oFormDefinition, this.sFormDefinitionName);

    // Ajoute les valeurs en utilisant les setters
    for (let i = 0; i < aFields.length; i++) {
      const oField = aFields[i];
      const oFieldName = oField.name;

      if (this.vitisCheck.isDefAndNotNull(oFieldName)) {
        if (this.vitisCheck.isDefAndNotNull(oFormValues[oFieldName])) {
          if (oField.setValue
            && (!this.vitisCheck.isDefAndNotNull(oField.dynamic_visible) || this.hasSetValues === true)) {
            oFormValues[oFieldName] = oField.setValue(oFormValues[oFieldName]);
          } else {
            const fieldValue = oFormValues[oFieldName];
            oField.oninit.subscribe(() => {
              if (oField.setValue) {
                oFormValues[oFieldName] = oField.setValue(fieldValue);
              }
            });
          }
        }
      }
    }

    this.hasSetValues = true;
  }

  /**
   * On met à jour oFormValues avec les valeurs de la base de données
   * puis on parcourt les champs et si il y a une fonction refresh on la lance.
   */
  public refreshFormValues() {

    this.refreshParentFormValues.emit((oFormValues: any) => {
      this.setFormValues(oFormValues);

      setTimeout(() => {
        const aFields = this.getFormFields(this.oFormDefinition, this.sFormDefinitionName);
        for (let i = 0; i < aFields.length; i++) {
          const oField = aFields[i];
          if (this.vitisCheck.isFunction(oField.refresh)) {
            oField.refresh();
          }
        }
      });
    });
  }

  /**
   * Set le nom du formulaire
   * @param sFormDefinitionName
   */
  public setFormDefinitionName(sFormDefinitionName: any) {
    this.sFormDefinitionName = sFormDefinitionName;
  }

  /**
   * Set la définition du formulaire
   * @param oFormDefinition
   */
  public setFormDefinition(oFormDefinition: any) {
    this.oFormDefinition = oFormDefinition;
  }

  /**
   * Fonction appelée au submit du formulaire
   * @param  {any} oFormDefinition?
   * @param  {any} sFormDefinitionName?
   * @param  {any} oFormValues?
   */
  public submitForm(oFormDefinition?: any, sFormDefinitionName?: any, oFormValues?: any) {

    // Paramètres par défaut
    oFormDefinition = this.vitisCheck.isDefAndNotNull(oFormDefinition) ? oFormDefinition : this.oFormDefinition;
    sFormDefinitionName = this.vitisCheck.isDefAndNotNull(sFormDefinitionName) ? sFormDefinitionName : this.sFormDefinitionName;
    oFormValues = this.vitisCheck.isDefAndNotNull(oFormValues) ? oFormValues : this.getFormValues();

    // Tests et execution de la fonction
    if (this.vitisCheck.isDefAndNotNull(oFormDefinition)) {
      if (this.vitisCheck.isDefAndNotNull(sFormDefinitionName)) {
        if (this.vitisCheck.isDefAndNotNull(oFormDefinition[sFormDefinitionName])) {

          const eventContainer = this.oFormDefinition[sFormDefinitionName].eventContainer;
          const aEvents = [];

          // beforeEvent
          if (this.vitisCheck.isDefAndNotNull(this.oFormDefinition[sFormDefinitionName].beforeEvent)) {
            aEvents.push(this.oFormDefinition[sFormDefinitionName].beforeEvent);
          }

          // event
          if (this.vitisCheck.isDefAndNotNull(this.oFormDefinition[sFormDefinitionName].event)) {
            aEvents.push(this.oFormDefinition[sFormDefinitionName].event);
            aEvents.push('setIsFormSubmitted');
          }

          // afterEvent
          if (this.vitisCheck.isDefAndNotNull(this.oFormDefinition[sFormDefinitionName].afterEvent)) {
            aEvents.push(this.oFormDefinition[sFormDefinitionName].afterEvent);
          }

          // Lance les événements en cascade
          this.runEvents_(aEvents, oFormValues, [eventContainer, this]);
        }
      }
    }
  }

  /**
   * Evenement à lancer après le submit pour déclarer que le submit a été fait.
   */
  public setIsFormSubmitted() {
    this.isSubmitted = true;
    return true;
  }

  /**
   * Permet de savoir le formulaire a été submit.
   */
  public getIsSubmittedForm() {
    return this.isSubmitted;
  }

  /**
   * Appelle un event appelé par un bouton
   * @param  {any} oButton
   */
  public callBtnEvent(oButton: any) {
    if (this.vitisCheck.isDefAndNotNull(oButton.event)) {
      this.fieldsService.callFunction(oButton.event, this.getFormValues(), [this.getFormContainer(), this]);
    }
  }

  /**
   * Retourne de container du formulaire
   * @returns any
   */
  public getFormContainer(): any {

    let eventContainer = this;
    if (this.vitisCheck.isDefAndNotNull(this.oFormDefinition)) {
      if (this.vitisCheck.isDefAndNotNull(this.sFormDefinitionName)) {
        if (this.vitisCheck.isDefAndNotNull(this.oFormDefinition[this.sFormDefinitionName])) {
          if (this.vitisCheck.isDefAndNotNull(this.oFormDefinition[this.sFormDefinitionName].eventContainer)) {
            eventContainer = this.oFormDefinition[this.sFormDefinitionName].eventContainer;
          }
        }
      }
    }

    return eventContainer;
  }

  /**
   * Retourne true / false si les valeurs du formulaire sont correctes
   * @returns boolean
   */
  public isFormValid(): boolean {

    // Fait apparaître les messages d'erreurs si besoin
    if (this.vitisCheck.isDefAndNotNull(this.oFormValues.controls)) {
      const aInvalidFields: Array<any> = [];
      for (const key in this.oFormValues.controls) {
        if (this.oFormValues.controls.hasOwnProperty(key)) {
          const control = this.oFormValues.controls[key];
          if (!control.valid) {
            console.error(key + ' not valid');
            aInvalidFields.push(key);
          }
          control.markAsTouched();
        }
      }
      // Affichage des champs de form. en erreur qui sont invisibles (onglets repliés).
      if (aInvalidFields.length > 0) {
        const oFormTabs = this.oFormDefinition[this.sFormDefinitionName].tabs;
        if (this.vitisCheck.isDefAndNotNull(oFormTabs) &&
          this.vitisCheck.isDefAndNotNull(oFormTabs.list) &&
          oFormTabs.type === 'container') {
          for (let i = 0; i < oFormTabs.list.length; i++) {
            if (oFormTabs.list[i].elements.filter((element: string) => aInvalidFields.indexOf(element) !== -1).length > 0) {
              oFormTabs.list[i].collapsed = false;
            }
          }
        }
      }
    }

    return this.oFormValues.valid;
  }

  /**
   * Lance en cascade les événements contenus dans aEvents
   * @param  {Array<any>} aEvents
   * @param  {any} oFormValues
   * @param  {Array<any>} aContainers
   */
  private runEvents_(aEvents: Array<any>, oFormValues: any, aContainers: Array<any>) {

    // lance le premier événement
    if (this.vitisCheck.isDefAndNotNull(aEvents[0])) {

      // Lance la fonction
      const eventResult = this.fieldsService.callFunction(aEvents[0], oFormValues, aContainers);

      // Supprime le premier évent de la liste et relance la fonction
      if (eventResult === true) {
        aEvents.shift();
        this.runEvents_(aEvents, oFormValues, aContainers);
      } else if (this.vitisCheck.isDefAndNotNull(eventResult)) {
        if (this.vitisCheck.isFunction(eventResult.then)) {
          eventResult.then(() => {
            aEvents.shift();
            this.runEvents_(aEvents, oFormValues, aContainers);
          });
        }
      }
    }
  }

  /**
   * Parse oFormDefinition
   * Reorganise les datasources, onglets, etc...
   * @param  {any} oFormDefinition
   * @param  {any} sFormDefinitionName
   * @param  {boolean} bLoadDatasource
   * @param  {boolean} bLoadWebserviceGrid
   * @returns any
   */
  private parseFormDef_(oFormDefinition: any, sFormDefinitionName: any, oFormValues?: any, bLoadDatasource: boolean = true, bLoadWebserviceGrid: boolean = true): any {

    const aFields = this.getFormFields(oFormDefinition, sFormDefinitionName);

    for (let i = 0; i < aFields.length; i++) {
      const oField = aFields[i];

      // id
      oField.id = this.vitisCheck.isDefAndNotNull(oField.id) ? oField.id : this.getUniqueId_();

      // options
      oField.options = this.vitisCheck.isDefAndNotNull(oField.options) ? oField.options : {};

      // observable d'initialisation
      oField.oninit = new Subject<number>();
      oField.onvisible = new Subject<number>();
      oField._visible = false;

      // datasource
      if (oField.datasource && bLoadDatasource) {
        // ne charge pas les datasources si il y a un refresh_btn
        if (oField.datasource.datasource_id && !oField.refresh_btn) {
          const oDatasource = this.vitisClone.clone(this.getDatasourceById_(oField.datasource.datasource_id, oFormDefinition));

          // Parents : définit que les parents ne sont pas écoutés
          if (oField.datasource.parents) {
            oField._parentsListened = false;
          }

          // Remplit oField.options.choises
          this.pullDatasource_(oDatasource, oField);
        }
      }

      // webservice_grid
      if (oField.webservice_grid && bLoadWebserviceGrid) {
        if (oField.webservice_grid.url) {
          // Timeout pour être sur d'avoir les valeurs du formulaire
          // en cas de fonction dans le filtre
          setTimeout(() => {
            const aParams = this.prepareParameters_(oField.webservice_grid.params, oField.webservice_grid.params);
            const oRefreshedFormValues = this.getFormValues();

            if (this.vitisCheck.isDefAndNotNull(oRefreshedFormValues)) {
              aParams.values = oRefreshedFormValues;
            }
            // Récupère et formate la liste des champs à afficher.
            if (!this.vitisCheck.isDefAndNotNull(aParams.attributs)) {
              for (let j = 0; j < oField.options.columns.length; j++) {
                // Ignore les valeurs null
                if (this.vitisCheck.isDefAndNotNull(oField.options.columns[j].data)) {
                  if (!this.vitisCheck.isDefAndNotNull(aParams.attributs)) {
                    aParams.attributs = oField.options.columns[j].data;
                  } else {
                    aParams.attributs += '|' + oField.options.columns[j].data;
                  }
                }
              }
            }

            // Type de colonne pour éventuellement formaté les données dans le php en fonction du type
            aParams.columns_type = {};
            for (let j = 0; j < oField.options.columns.length; j++) {
              aParams.columns_type[oField.options.columns[j].data] = oField.options.columns[j].type;
            }

            this.vitisRequest.ajaxRequestObservable({
              method: 'GET',
              url: environment.properties.app_api_url + oField.webservice_grid.url,
              params: aParams,
              httpMethodOverride: true,
              headers: {
                'Accept': 'application/x-vm-json'
              }
            }).subscribe((resp) => {
              if (this.vitisCheck.isDefAndNotNull(resp.data)) {
                if (resp.data.status === 1) {
                  if (this.vitisCheck.isArray(resp.data.data)) {
                    if (oField.setValue) {
                      oField.setValue(resp.data.data);
                    }
                  }
                }
              }
            });
          }, 50);
        }
      }
    }

    // Patch input_size xxs
    if (oFormDefinition &&
      oFormDefinition[sFormDefinitionName]) {
      if (oFormDefinition[sFormDefinitionName].input_size === 'xxs') {
        oFormDefinition[sFormDefinitionName].input_size = 'xs';
      }
    }

    return oFormDefinition;
  }

  /**
   * Recharge la datasource d'un champ
   * @param  {any} oField
   */
  public reloadDatasource(oField: any) {
    if (oField.datasource) {
      if (oField.datasource.datasource_id) {
        const oDatasource = this.vitisClone.clone(this.getDatasourceById_(oField.datasource.datasource_id, this.oFormDefinition));

        // Parents : définit que les parents ne sont pas écoutés
        if (oField.datasource.parents && oField._parentsListened !== true) {
          oField._parentsListened = false;
        }

        // Remplit oField.options.choises
        oField.options.choises = [];
        oField.options.observableChoises = new Subject<number>();

        if (oDatasource.type === 'web_service') {

          // Datasource de type WebService
          this.pullWebServiceDatasource_(oDatasource, oField);
        } else if (oDatasource.type === 'text' || oDatasource.type === 'object') {

          // Datasource de type text
          this.pullTextDatasource_(oDatasource, oField);
        } else {
          console.error('Datasource type ' + oDatasource.type + ' not yet implemented');
        }
      }
    }
  }

  /**
   * Parse oFormValues
   * Crée un FormGroup spécifiant les valeurs requises etc...
   * @param  {any} oFormDefinition
   * @param  {any} sFormDefinitionName
   * @param  {any} oFormValues?
   * @returns FormGroup
   */
  private parseFormValues_(oFormDefinition: any, sFormDefinitionName: any, oFormValues?: any): FormGroup {

    const tmpValidForm = {};

    // Champs du formulaire
    const aFields = this.getFormFields(oFormDefinition, sFormDefinitionName);

    // Ajoute les champs à tmpValidForm
    for (let i = 0; i < aFields.length; i++) {
      const oField = aFields[i];
      if (oField.name) {
        tmpValidForm[oField.name] = this.getFieldValidator_(oField);
      }
    }

    return this.formBuilder.group(tmpValidForm);
  }

  /**
   * Parse un field et retourne un FormControl ou un FormArray
   * Spécifie les valeurs requises, messages d'erreur etc..
   * @param  {any} oField
   * @returns FormControl | FormArray
   */
  private getFieldValidator_(oField: any) {

    const aValidators = [];
    oField.validation_messages = [];
    oField.validators_list = [];

    // Field required
    if (oField.required === true) {

      // Ajoute à aValidators
      aValidators.push(Validators.required);
      if (!oField.validation_messages) {
        oField.validation_messages = [];
      }
      oField.validation_messages = this.addFieldValidationMsg(oField.validation_messages, {
        type: 'required',
        message: this.translate.instant(oField.label ? oField.label : ' ') + ' ' + this.translate.instant('FORM_FIELD_REQUIRED') + '.'
      });
    }

    // Pattern
    if (this.vitisCheck.isDefAndNotNull(oField.pattern)) {

      // Ajoute à aValidators
      aValidators.push(Validators.pattern(oField.pattern));
      if (!oField.validation_messages) {
        oField.validation_messages = [];
      }
      oField.validation_messages = this.addFieldValidationMsg(oField.validation_messages, {
        type: 'pattern',
        message: this.translate.instant(oField.label ? oField.label : ' ') + ' ' + this.translate.instant('FORM_FIELD_NOT_VALID') + '.'
      });
    }
    // Sauvegarde des fonctions de validation
    oField.validators_list = aValidators;

    const oValidator = new FormControl({ value: '', disabled: oField.disabled }, Validators.compose(aValidators));
    return oValidator;
  }

  /**
   * Trouve un identifiant unique à un élément
   * @returns string
   */
  private getUniqueId_(): string {
    let id = '_' + Math.random().toString(36).substr(2, 9);
    if (document.getElementById(id)) {
      id = this.getUniqueId_();
    }
    return id;
  }

  /**
   * Télécharge les données de la source de données et les place sur oField.options.choises
   * @param  {any} oDatasource
   * @param  {any} oField
   * @param  {boolean} bForcePull permet de pull même si le champ est déjà chargé
   */
  private pullDatasource_(oDatasource: any, oField: any, bForcePull = false) {

    oField.options.choises = [];
    oField.options.observableChoises = new Subject<number>();
    oField._datasourcePulled = false;

    const pullFieldDatasource = () => {
      if (oDatasource.type === 'web_service') {

        // Datasource de type WebService
        if (!this.vitisCheck.isArray(oField.datasource.parents)) {
          this.pullWebServiceDatasource_(oDatasource, oField);
        } else {
          // Timeout pour être sur d'avoir les valeurs du formulaire
          setTimeout(() => {
            this.pullWebServiceDatasource_(oDatasource, oField);
          }, 100);
        }

      } else if (oDatasource.type === 'text' || oDatasource.type === 'object') {

        // Datasource de type text
        this.pullTextDatasource_(oDatasource, oField);
      } else {
        console.error('Datasource type ' + oDatasource.type + ' not yet implemented');
      }

      oField._datasourcePulled = true;
    };

    oField.onvisible.subscribe((bVisible: boolean) => {
      if (bVisible && oField._datasourcePulled !== true) {
        pullFieldDatasource();
      }
    });

    if (bForcePull) {
      pullFieldDatasource();
    }
  }

  /**
   * Télécharge les données de la source de données et les place sur oField.options.choises
   * @param  {any} oDatasource
   * @param  {any} oField
   */
  private pullTextDatasource_(oDatasource: any, oField: any) {

    if (this.vitisCheck.isArray(oDatasource.options)) {
      for (let i = 0; i < oDatasource.options.length; i++) {
        let options = oDatasource.options[i];
        if (this.vitisCheck.isString(options)
          && options.includes('|')) {
          options = options.split('|');
        }
        if (this.vitisCheck.isArray(options)) {
          oField.options.choises.push({
            label: options[0],
            value: options[1]
          });
        } else {
          oField.options.choises.push({
            label: options,
            value: options
          });
        }
      }

      setTimeout(() => {
        oField.options.observableChoises.next(oField.options.choises);
      });
    }
  }

  /**
   * Télécharge les données de la source de données et les place sur oField.options.choises
   * @param  {any} oDatasource
   * @param  {any} oField
   */
  private pullWebServiceDatasource_(oDatasource: any, oField: any) {

    // Écoute des parents
    let bLoadData = true;
    if (this.vitisCheck.isArray(oField.datasource.parents)) {
      bLoadData = this.listenDatasourceParents_(oDatasource, oField);
    }

    if (bLoadData) {
      // Timeout pour être sur d'avoir les valeurs du formulaire
      // en cas de fonction dans le filtre
      setTimeout(() => {
        // Téléchargement des données
        this.getDatasourceWebServiceData_(oDatasource, oField).subscribe((resp: any) => {
          this.setDatasourceResponseInField_(resp, oField, oDatasource);
        });
      }, 500);
    }
  }

  /**
   * Écoute les parents pour faire des listes cascade
   * @param oDatasource
   * @param oField
   */
  private listenDatasourceParents_(oDatasource: any, oField: any): boolean {
    let bLoadData = true;

    // Gestion parents
    if (this.vitisCheck.isArray(oField.datasource.parents)) {

      let bParentsHasValue = false;
      let nbparentsHasValue = 0;

      for (let i = 0; i < oField.datasource.parents.length; i++) {

        // Parent
        const parent = oField.datasource.parents[i];
        if (!parent.name || !parent.filter_attr) {
          console.error('Parent not valid', parent);
          return false;
        }

        // Field du parent
        const parentField: any = this.getFormField(parent.name);
        if (!parentField) {
          console.error('Parent not valid', parent);
          return false;
        }

        // Valeur du parent
        const parentValue = parentField.getValue();

        // Filtre à utiliser
        const parentFilter = this.getParentFilter_(parentField, parent, parentValue);

        if (parentValue && parentFilter) {
          nbparentsHasValue++;

          // Ajoute le filtre au datasource
          oDatasource = this.formreaderSrvc.addDatasourceFilter(oDatasource, parentFilter, true);
        } else {

          // Si on attend pas le parent c'est comme s'il avait une valeur
          if (!parent.wait_for_parent) {
            nbparentsHasValue++;

            // Enlève le filtre au datasource (la valeur et l'opérateur ne sont pas pris en compte)
            oDatasource = this.formreaderSrvc.removeDatasourceFilter(oDatasource, {
              column: parent.filter_attr,
              compare_operator: '=',
              value: ''
            });
          }
        }

        // Écoute le parent
        if (!oField._parentsListened) {

          // Observe les changements de valeur du parent
          this.oFormValues.controls[parent.name].valueChanges.subscribe((value) => {
            // Suppression de la sélection du double select.
            if (oField.type == 'double_select') {
              oField.resetValue();
            }

            // Relance la fonction
            this.pullDatasource_(oDatasource, oField, true);
          });
        }
      }

      // Vérifie que tous les parents aient étés chargés
      if (nbparentsHasValue === oField.datasource.parents.length) {
        bParentsHasValue = true;
      }

      // Enregistre le fait que les parents ont été écoutés
      // permet de ne pas dupliquer l'écoute des parents
      oField._parentsListened = true;

      // Si les parents ne sont pas chargés alors sort de la fonction
      // et n'effectue pas le chargmenet de données
      if (!bParentsHasValue) {
        bLoadData = false;
      }
    }
    return bLoadData;
  }

  /**
   * Retourne le filtre à appliquer pour un usage cascade
   * @param parentField
   * @param parent
   * @param parentValue
   */
  private getParentFilter_(parentField: any, parent: any, parentValue: any): any {

    let oFilter = null;
    const compareOperator = parent.filter_equality ? parent.filter_equality : '=';

    if (parentField.type === 'list' || parentField.type === 'double_select') {
      let aValues = [];
      if (this.vitisCheck.isString(parentValue)) {
        aValues = parentValue.split('|');
      } else if (this.vitisCheck.isArray(parentValue)) {
        aValues = parentValue;
      }
      if (aValues.length > 0) {
        const aOperators = [];
        for (let i = 0; i < aValues.length; i++) {
          const sVal = aValues[i];
          aOperators.push({
            column: parent.filter_attr,
            compare_operator: compareOperator,
            value: sVal
          });
        }
        oFilter = {
          relation: 'OR',
          operators: aOperators
        };
      }
    } else {
      oFilter = {
        column: parent.filter_attr,
        compare_operator: compareOperator,
        value: parentValue
      };
    }

    return oFilter;
  }

  /**
   * Récupère la datasource par son ID
   * @param  {string} sDatasourceId
   * @param  {any} oFormDefinition
   * @returns any
   */
  private getDatasourceById_(sDatasourceId: string, oFormDefinition: any): any {

    if (this.vitisCheck.isDefAndNotNull(oFormDefinition)) {
      if (this.vitisCheck.isDefAndNotNull(oFormDefinition.datasources)) {
        if (this.vitisCheck.isDefAndNotNull(oFormDefinition.datasources[sDatasourceId])) {
          return oFormDefinition.datasources[sDatasourceId];
        }
      }
    }
  }

  /**
   * Télécharge les donées d'une source de données de type web_service
   * @param  {any} oDatasource
   * @param  {any} oField
   * @returns Observable
   */
  private getDatasourceWebServiceData_(oDatasource: any, oField: any): Observable<any> {

    const sUrl: string = environment.properties.app_api_url + '/' + oDatasource.ressource_id;
    let oParams: any = {};

    if (this.vitisCheck.isDefAndNotNull(oField)) {
      if (this.vitisCheck.isDefAndNotNull(oField.datasource)) {

        // Params
        if (this.vitisCheck.isObject(oField.datasource.parameters)) {
          oParams = this.prepareParameters_(oField.datasource.parameters, oParams);
        }
        if (this.vitisCheck.isObject(oDatasource.parameters)) {
          oParams = this.prepareParameters_(oDatasource.parameters, oParams);
        }

        // Attributs
        if (this.vitisCheck.isString(oField.datasource.attributs)) {
          oParams.attributs = oField.datasource.attributs;
        }

        // OrderBy
        if (this.vitisCheck.isString(oField.datasource.order_by)) {
          oParams.order_by = oField.datasource.order_by;
        }

        // SortOrder
        if (this.vitisCheck.isString(oField.datasource.sort_order)) {
          oParams.sort_order = oField.datasource.sort_order;
        }

        // Distinct
        if (this.vitisCheck.isString(oField.datasource.distinct)) {
          oParams.distinct = oField.datasource.distinct;
        }

      }
    }

    return this.vitisRequest.ajaxRequestObservable({
      method: 'GET',
      url: sUrl,
      params: oParams,
      httpMethodOverride: true,
      enableCache: oDatasource.enable_cache,
      headers: {
        'Accept': 'application/x-vm-json'
      }
    });
  }

  /**
   * Prépare les paramètres pour récuperer le webservice
   * @param {any} oParameters
   * @param {any} oParams
   */
  private prepareParameters_(oParameters: any, oParams: any): any {
    for (const key in oParameters) {
      if (oParameters.hasOwnProperty(key)) {

        let sValue = oParameters[key];
        if (key == 'filter') {
          this.prepareFilters_(sValue);
        }

        // Cas où le paramètre soit une fonction
        if (this.fieldsService.isFunctionCall(sValue, true, [this.getFormContainer(), this])) {
          sValue = this.fieldsService.callFunction(sValue, this.getFormValues(), [this.getFormContainer(), this]);
        }

        // Cas où le paramètre bind une autre valeur
        const tmpValue = this.fieldsService.getTernaryStringValue(sValue);
        if (this.vitisCheck.isDefAndNotNull(tmpValue)) {
          sValue = tmpValue;
        }

        oParams[key] = sValue;
      }
    }
    return oParams;
  }

  /**
   * Prépare les filtres des paramètre pour récupérer le webservice
   * @param oFilter
   */
  private prepareFilters_(oFilter: any) {
    if (this.vitisCheck.isDefAndNotNull(oFilter.operators)) {
      for (let i = 0; i < oFilter.operators.length; i++) {
        const sValue = oFilter.operators[i].value;
        const tmpValue = this.fieldsService.getTernaryStringValue(sValue);
        if (this.vitisCheck.isDefAndNotNull(tmpValue)) {
          oFilter.operators[i].value = tmpValue;
        }
      }
    }
  }

  /**
   * Utilise les valeurs par défaut
   * @param {any} oFormValues
   * @param {boolean} bEraseFormValues vrai pour écraser les valeurs de oFormValues
   * @param {string} sElementName Si il est définie, la fonction ne met à jour que l'élement donné.
   */
  public useDefaultValues(oFormValues?: any, bEraseFormValues?: boolean, sElementName?: string): Object {
    const aFields = this.getFormFields(this.oFormDefinition, this.sFormDefinitionName);

    if (!this.vitisCheck.isDefAndNotNull(oFormValues)) {
      oFormValues = this.oFormValues.value;
    }

    if (!this.vitisCheck.isDefAndNotNull(bEraseFormValues)) {
      bEraseFormValues = false;
    }
    for (let i = 0; i < aFields.length; i++) {
      let oField = aFields[i];

      if (oField.name && this.vitisCheck.isDefAndNotNull(oField.default_value)) { // Il existe une valeur par défaut ?

        // sElementName différent du nom de l'élément en cours ?
        if (this.vitisCheck.isDefAndNotNull(sElementName)) {
          if (oField.name !== sElementName) {
            continue;
          }
        }

        // Mise à jour de la valeur par défaut si besoin
        oField = this.updateSpecificDefaultValue(oField);

        // On assigne la valeur par défaut si c'est nécessaire.
        if (this.isNeededDefaultValue_(oFormValues, oField, bEraseFormValues)) {
          oFormValues[oField.name] = oField.default_value;
        }
      }
    }
    return oFormValues;
  }

  /**
   * Met à jour oField.default_value en fonction du type de composant ou si c'est une fonction
   * @param {any} oField
   * @return {object} oField
   */
  private updateSpecificDefaultValue(oField: any): any {

    // La valeur par défaut est un appel de fonction ?
    if (this.vitisCheck.isDefAndNotNull(oField.default_value) && oField.default_value !== '') {
      if (this.fieldsService.isFunctionCall(oField.default_value)) {
        const container = this.getContainer_(oField.default_value);
        oField.default_value = this.fieldsService.callFunction(oField.default_value, null, container);
      }
    }

    // Le composant est une liste ou un select ?
    if (oField.type === 'list' || oField.type === 'double_select' || oField.type === 'editable_list' || oField.type === 'searchable_select') {
      if (this.vitisCheck.isDefAndNotNull(oField.default_value) && oField.default_value !== '') {
        let aDefaultKeys = [];
        if (this.vitisCheck.isString(oField.default_value)) {
          aDefaultKeys = oField.default_value.split('|');
        } else {
          aDefaultKeys = oField.default_value;
        }
        const aDefaults = [];
        for (let iii = 0; iii < aDefaultKeys.length; iii++) {
          aDefaults.push(aDefaultKeys[iii]);
        }
        oField.default_value = aDefaults;
      }
    }

    return oField;
  }

  /**
   * Retourne vrai si le champ a besoin de la valeur par défaut
   * @param {any} oFormValues
   * @param {any} oField
   * @param {boolean} bEraseFormValues
   * @return {boolean} vrai si le champ a besoin de la valeur par défaut
   * @export
   */
  private isNeededDefaultValue_(oFormValues: any, oField: any, bEraseFormValues?: boolean): boolean {
    if ((this.vitisCheck.isDefAndNotNull(oField.default_value) && oField.default_value !== '')) {
      // Vérifie qu'il existe pas déjà une valeur
      if (bEraseFormValues === true) {
        return true;
      } else {
        if (!this.vitisCheck.isDefAndNotNull(oFormValues[oField.name])) {
          return true;
        }
        if (oFormValues[oField.name].length === 0) {
          return true;
        }
      }
    } else {
      return false;
    }
  }

  /**
   * Récupère le container qui permettra d'executer la function
   * @param {string} sString Chaine de caractère.
   * @param {boolean} bCheckFunction? Vérification de l'existence de la fonction (vrai par défaut).
   * @return {any} oFunctionContainer
   */
  private getContainer_(sString: string, bCheckFunction?: boolean): any {
    let oFunctionContainer: any;
    if (!this.fieldsService.isFunctionCall(sString, bCheckFunction)) {
      const sFunctionName = sString.substr(0, sString.indexOf('('));

      const eventContainer = this.oFormDefinition[this.sFormDefinitionName].eventContainer;
      if (this.vitisCheck.isDefAndNotNull(eventContainer) && this.vitisCheck.isFunction(eventContainer[sFunctionName])) {
        oFunctionContainer = eventContainer;
      } else if (this.vitisCheck.isDefAndNotNull(this[sFunctionName])) {
        oFunctionContainer = this;
      }
    }

    return oFunctionContainer;
  }

  /**
   * Traduit la valeur par défaut si c'est une chaine de caractère
   * @param {any} oFormValues
   * @param {any} oField
   * @return {any} oFormValues
   */
  private translateDefaultValue(oFormValues: any, oField: any): any {
    // traduction de la valeur par défaut
    if (typeof (oField.default_value) == 'string') {
      if (/^[A-Z0-9_]+$/.test(oField.default_value)) {
        this.translate.get(oField.default_value).subscribe((translation) => {
          if (oFormValues[oField.name] == oField.default_value) {
            oFormValues[oField.name] = translation[oField.default_value];
          }
        });
      }
    }

    return oFormValues;
  }

  /**
   * Initialise les onglets
   * @export
   */
  private initFormTabs_() {

    if (this.vitisCheck.isDefAndNotNull(this.oFormDefinition)) {
      if (this.vitisCheck.isDefAndNotNull(this.sFormDefinitionName)) {
        if (this.vitisCheck.isDefAndNotNull(this.oFormDefinition[this.sFormDefinitionName])) {

          // Liste des noms de champs
          const aFieldsNames = this.formreaderSrvc.getFormFieldsNames(this.oFormDefinition, this.sFormDefinitionName);

          // Structure des onglets valide
          const bTabsStructOk = this.formreaderSrvc.isTabsValid(this.oFormDefinition[this.sFormDefinitionName].tabs);

          // Si aucun onglet défini, crée un onglet vide
          if (!bTabsStructOk) {
            this.oFormDefinition[this.sFormDefinitionName].tabs = {
              'type': 'tab',
              'list': [{
                'label': 'Onglet',
                'elements': []
              }]
            };
          }

          // Rétrocompatibilité position: top => type: tab
          if (this.vitisCheck.isDefAndNotNull(this.oFormDefinition[this.sFormDefinitionName].tabs.position)) {
            this.oFormDefinition[this.sFormDefinitionName].tabs.type = 'tab';
          }

          // Vérifie que chaque champ appatienne à un onglet, sinon le place sur le premier onglet
          for (let i = 0; i < aFieldsNames.length; i++) {
            const sFieldName = aFieldsNames[i];
            let bIsInTabs = false;

            // Remplit bIsInTabs
            if (this.vitisCheck.isArray(this.oFormDefinition[this.sFormDefinitionName].tabs.list)) {
              for (let ii = 0; ii < this.oFormDefinition[this.sFormDefinitionName].tabs.list.length; ii++) {
                const oTab = this.oFormDefinition[this.sFormDefinitionName].tabs.list[ii];
                if (this.vitisCheck.isArray(oTab.elements)) {
                  if (oTab.elements.indexOf(sFieldName) !== -1) {
                    bIsInTabs = true;
                  }
                }
              }
            }

            // Si non présent alors ajoute sur le premier onglet
            if (!bIsInTabs) {
              this.oFormDefinition[this.sFormDefinitionName].tabs.list[0].elements.push(sFieldName);
            }
          }
        }
      }
    }
  }

  /**
   * Appel le service FieldsService
   * @param {any} oField
   * @param {any} oTab
   * @param {boolean} bCheckButtons
   *
   * @returns {boolean}
   */
  public isFieldPresent(oField: any, oTab?: any, bCheckButtons?: boolean): boolean {
    return this.fieldsService.isFieldPresent(oField, oTab, bCheckButtons);
  }

  /**
   * Si au moins un champ est visible, le container est visible
   * @param {any} oTab
   *
   * @returns {boolean}
   */
  public isHiddenContainer(oTab: any): boolean {
    const rows = this.getFormRows();
    for (let i = 0; i < rows.length; i++) {
      for (let j = 0; j < rows[i].fields.length; j++) {
        const oField = rows[i].fields[j];
        if (oField.type == 'button') {
          for (let k = 0; k < oField.buttons.length; k++) {
            const oButton = oField.buttons[k];
            if ((oTab.elements.indexOf(oField.name) !== -1 || oTab.elements.indexOf(oButton.name) !== -1) &&
              (this.isFieldPresent(oField, oTab, true) || this.isFieldPresent(oButton, oTab, true))) {
              return false;
            }
          }
        } else {
          if (oTab.elements.indexOf(oField.name) !== -1 && this.isFieldPresent(oField, oTab, null) === true) {
            return false;
          }
        }

      }
    }
    return true;
  }

  /**
   * Change la valeur de iDisplayedTab
   * @param {number} index
   */
  public changeDisplayedTab(index: number) {
    this.iDisplayedTab = index;
  }

  /**
   * Initialise les valeurs dynamiques
   */
  private setFormDynamicValues_() {
    const aFields = this.getFormFields(this.oFormDefinition, this.sFormDefinitionName);

    this.fieldsService.setFormDynamicValues(aFields);
  }

  /**
   * Initialise les evenements dynamiques
   */
  public setFormDynamicEvents() {
    const aFields = this.getFormFields(this.oFormDefinition, this.sFormDefinitionName);

    for (let i = 0; i < aFields.length; i++) {
      const oField = aFields[i];

      // Événements dynamiques
      if (this.vitisCheck.isDefAndNotNull(oField.dynamic_event_type) &&
        this.vitisCheck.isDefAndNotNull(oField.dynamic_event_function)) {

        const eventContainer = this.oFormDefinition[this.sFormDefinitionName].eventContainer;

        const sEvent = oField.dynamic_event_type;
        if (sEvent === 'change') {
          this.oFormValues.controls[oField.name].valueChanges.subscribe((value) => {
            this.fieldsService.callbackFormDynamicFunction(oField, [this.getFormContainer(), this], value);
          });
        } else {
          oField.oninit.subscribe(() => {
            if (this.vitisCheck.isDefAndNotNull(oField.element)) {
              oField.element.on(sEvent, () => {
                this.fieldsService.callbackFormDynamicFunction(oField, [this.getFormContainer(), this]);
              });
            }
          });
        }
      }
    }
  }

  /**
   * Set value yes/no traduit pour le formulaire en display
   * @param {string} fieldName
   */
  public initDisplayBoolean(fieldName: string) {
    if (this.sFormDefinitionName === 'display') {
      const oField = this.getFormField(fieldName);

      oField.oninit.subscribe(() => {
        if (oField.setValue) {
          this.translate.get(['FORM_YES', 'FORM_NO']).subscribe((translations) => {
            if (this.oFormValues.get(fieldName).value == true) {
              oField.setValue(translations['FORM_YES']);
            } else {
              oField.setValue(translations['FORM_NO']);
            }
          });
        }
      });
    }
  }

  /**
   * selectFirstOption function.
   * Sélectionne la 1ere option d'un <select>.
   * @param {string} sFieldName Nom du <select>.
   **/
  public selectFirstOption(sFieldName: string) {
    const oField = this.getFormField(sFieldName);
    if (oField.type == 'select' && oField.setValue && this.vitisCheck.isDefAndNotNull(oField.options.choises[0])) {
      oField.setValue(oField.options.choises[0].value);
    }
  }

  /**
   * Ajoute un message de validation à un champ de formulaire.
   * Spécifie les valeurs requises, messages d'erreur etc..
   * @param  {any} aValidationMsgs
   * @param  {any} oValidationMsg
   * @returns any
   */
  private addFieldValidationMsg(aValidationMsgs: any, oValidationMsg: any) {
    let bAddMsg: boolean = true;
    // Vérifie si un message existe déjà pour ce type d'erreur.
    for (let i = 0; i < aValidationMsgs.length; i++) {
      if (aValidationMsgs[i].type == oValidationMsg.type) {
        bAddMsg = false;
        break;
      }
    }
    // Ajoute le message d'erreur.
    if (bAddMsg) {
      aValidationMsgs.push(oValidationMsg);
    }
    return aValidationMsgs;
  }

  /**
   * Formate et met les valeurs de la source de données dans l'objet oField
   * @param resp
   * @param oField
   * @param oDatasource
   */
  private setDatasourceResponseInField_(resp: any, oField: any, oDatasource?: any) {
    // Check et push les données sur oField.options.choises
    if (this.vitisCheck.isDefAndNotNull(resp.data)) {
      if (resp.data.status === 1) {
        if (this.vitisCheck.isArray(resp.data.data)) {
          for (let i = 0; i < resp.data.data.length; i++) {
            if (this.vitisCheck.isDefAndNotNull(resp.data.data[i][oField.datasource.label_key]) &&
              this.vitisCheck.isDefAndNotNull(resp.data.data[i][oField.datasource.id_key])) {
              oField.options.choises.push({
                label: String(resp.data.data[i][oField.datasource.label_key]),
                value: resp.data.data[i][oField.datasource.id_key]
              });
            }
          }
          // Affichage des données si le champ est un double select en cascade (Ne pas supprimer !!!).
          if (oField.type == 'double_select' && this.vitisCheck.isDefAndNotNull(oField.datasource)
            && this.vitisCheck.isDefAndNotNull(oField.datasource.parents)) {
            oField.setValue(null);
          }
        }

        // Si il y a une callback et que la fonction existe on la lance
        if (this.vitisCheck.isDefAndNotNull(oDatasource.callback) &&
          this.fieldsService.isFunctionCall(oDatasource.callback, true, [this.getFormContainer(), this])) {
          this.fieldsService.callFunction(oDatasource.callback, this.getFormValues(), [this.getFormContainer(), this]);
        }
      }
    }

    oField.options.observableChoises.next(oField.options.choises);
  }

  /**
   * Récupère le datasource avec le champ désigné et ajoute les données dans tout les champs de la liste
   * @param aFieldList
   * @param oUsedField
   * @param datasourceId
   */
  public setWebserviceDatasourceInFields(aFieldList: Array<any>, oUsedField: any, datasourceId: string) {
    const oDatasource = this.vitisClone.clone(this.getDatasourceById_(datasourceId, this.oFormDefinition));

    // Timeout pour être sur d'avoir les valeurs du formulaire
    // en cas de fonction dans le filtre
    setTimeout(() => {
      // Téléchargement des données
      this.getDatasourceWebServiceData_(oDatasource, oUsedField).subscribe((resp: any) => {
        for (let i = 0; i < aFieldList.length; i++) {
          const oField = aFieldList[i];

          if (!this.vitisCheck.isDefAndNotNull(oField.datasource)) {
            // Supprime datasource_id avant de l'attribuer au oField sinon au prochain lancement du formulaire la requête est exécuté 2 fois
            delete oUsedField.datasource.datasource_id;
            oField.datasource = oUsedField.datasource;
          }
          // Remplit oField.options.choises
          oField.options.choises = [];
          oField.options.observableChoises = new Subject<number>();
          this.setDatasourceResponseInField_(resp, oField, oDatasource);
        }
      });
    }, 50);
  }

  /**
   * Affiche un formulaire dans une fenêtre modale
   * @param  {string} sFormUrl
   * @param  {string} sFormMode
   * @param  {any} oFormValues
   * @param  {any} oOptions
   * @return {void}
   */
  public displayModalForm(sFormUrl: string, sFormMode: string = 'insert', oFormValues: any = {}, oOptions: any = {}) {
    // Chargement du formulaire.
    this.vitisRequest.ajaxRequestObservable({
      method: 'GET',
      url: sFormUrl,
    }).subscribe(resp => {
      if (resp.status === 200 && resp.data) {
        // Définition et valeurs par défaut du formulaire pour le FormReader.
        const oFormOptions = {
          form: resp.data,
          formName: sFormMode,
          formValues: {...oFormValues}
        };
        // Fusionne les options de la grille de données.
        this.oModalOptions = {
          ...this.oModalOptions,
          ...oOptions,
          formOptions: oFormOptions
        };
        // timeout pour désyncrho et faire passer l'instruction à la queue de la file
        setTimeout(() => {
          // Affichage de la fenêtre modale du formulaire.
          $('#formreader_modal_additional_form').modal('show');

          if (typeof this.oModalOptions.initEvent == 'function') {
            this.oModalOptions.initEvent(this);
          }
        }, 0);

      }
    });
  }

  /**
   * fonction qui va lancer les différents event de validation
   * @returns {void}
   */
  public submitModalForm() {
    if (typeof this.oModalOptions.beforeEvent == 'function') {
      if (this.oModalOptions.beforeEvent(this) === false) {
        return;
      }
    }
    if (typeof this.oModalOptions.event == 'function') {
      if (this.oModalOptions.event(this) === false) {
        return;
      }
    }

    if (typeof this.oModalOptions.beforeEvent == 'function') {
      if (this.oModalOptions.afterEvent(this) === false) {
        return;
      }
    }

    // Ferme la fenêtre modale.
    $('#formreader_modal_additional_form').modal('hide');
  }

  /**
   * Check le formulaire et ferme la modale
   * @param {FormreaderComponent} this_ formreader parent
   * @return {boolean} false pour ne pas appeler l'event suivant
   */
  protected submitModalEvent_(this_: any): boolean {
    const aFields = this_.oModalFormReader.getFormFields(this_.oModalFormReader.oFormDefinition, this_.oModalFormReader.sFormDefinitionName);
    // Set la valeur à vide en utilisant les setters
    const aFormValues: any = {};
    // for (let i = 0; i < aFields.length; i++) {
    for (const oField of aFields) {
      // const oField = aFields[i];
      if (this_.vitisCheck.isDefAndNotNull(oField.getDisplayedValue)) {
        aFormValues[oField.name] = oField.getDisplayedValue();
      } else if (this_.vitisCheck.isDefAndNotNull(oField.getValue)) {
        aFormValues[oField.name] = oField.getValue();

        if (oField.required === true && !aFormValues[oField.name]) {
          vitisAlert.fireToast({
            type: 'error',
            title: this_.translate.instant('DATA_GRID_FORMREADERCOMPONENT_FIELD_REQUIRED') + ' ' +
              this_.translate.instant(oField.label),
          });
          return false;
        }
      }
    }
    return true;
  }
}

