import { Injectable, ChangeDetectorRef } from '@angular/core';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';
import { FormGroup, Validators } from '@angular/forms';
import { environment } from 'src/environments/environment';
import { AuthService } from 'src/app/vitis/core/services/auth/auth.service';
import { FunctionService } from '../../core/services/utils/function/function.service';

/**
 * Service FieldsService
 *
 * Nécéssitant instanciation permet de lancer des opétations
 * complexes sur les fields d'un formulaire
 *
 */
export class FieldsService {

  public oProperties: any = environment.properties;
  public aUserPrivileges: Array<string>;

  constructor(
    private vitisCheck: VitisCheck,
    public oFormValues: FormGroup,
    public showTabs: boolean,
    protected changeDetectorRef: ChangeDetectorRef,
    private authService: AuthService,
    private functionService: FunctionService
  ) {
    this.aUserPrivileges = this.authService.getPrivileges();
  }

  /**
   * Appelle une fonction appartenant à oFunctionContainer
   * @param  {any} sFunction
   * @param  {any} oValues
   * @param  {any} oFunctionContainer? objet ou tableau d'objets dans lequel est possiblement contenu la fonction
   */
  public callFunction(sFunction: any, oValues: any, oFunctionContainer?: any) {

    oFunctionContainer = this.vitisCheck.isDefAndNotNull(oFunctionContainer) ? oFunctionContainer : this;

    return this.functionService.callFunction(sFunction, oValues, oFunctionContainer);
  }

  /**
   * Determine si le champ doit être présent
   * @param {any} oField
   * @param {any} oTab
   * @param {boolean} bCheckButtons
   * @returns {Boolean}
   */
  public isFieldPresent(oField: any, oTab?: any, bCheckButtons?: boolean): boolean {
    bCheckButtons = this.vitisCheck.isDefAndNotNull(bCheckButtons) ? bCheckButtons : false;

    let bReturn = false;

    // Visible
    let isVisible = true;
    if (this.vitisCheck.isDefAndNotNull(oField.visible)) {
      if (oField.visible === false || oField.visible === 'false') {
        isVisible = false;
      }
    }

    // Patch car en 2019 on pouvait mettre des expressions ternaires dans le champ visible
    if (!this.vitisCheck.isDefAndNotNull(oField.dynamic_visible) &&
      this.vitisCheck.isDefAndNotNull(oField.visible) &&
      (this.isFunctionCall(oField.visible) ||
        this.isTernaryString_(oField.visible) ||
        this.isCustomTernaryString_(oField.visible, true))) {
      oField.dynamic_visible = oField.visible;
    }

    // dynamic_visible
    if (this.vitisCheck.isDefAndNotNull(oField.dynamic_visible)) {
      isVisible = this.isVisible_(oField.dynamic_visible);
    }

    // Tabs
    let isOnTab = false;
    if (oTab) {
      if (oTab.elements.indexOf(oField.name) !== -1) {
        isOnTab = true;
      }
      if (oField.visibleAllTabs === true) {
        isOnTab = true;
      }
      if (this.vitisCheck.isDefAndNotNull(oField.buttons) && bCheckButtons === false) {
        isOnTab = true;
      }
      if (!this.showTabs) {
        isOnTab = true;
      }
    } else {
      isOnTab = true;
    }

    // dynamic_disabled
    if (this.vitisCheck.isDefAndNotNull(oField.dynamic_disabled)) {
      oField.disabled = this.isVisible_(oField.dynamic_disabled);
    }

    this.setDynamicRequired(oField, isVisible);

    // TODO Wab
    // if (this['useWab']()) {
    //     var isOnWab = false;
    //     var wabValue = this['getWabField'](oField);
    //     if (wabValue === 'r' || wabValue === 'rw') {
    //         isOnWab = true;
    //     }
    //     if (this.vitisCheck.isDefAndNotNull(oField.buttons)) {
    //         isOnWab = true;
    //     }
    //     if (wabValue === 'r') {
    //         oField.disabled = true;
    //     }
    // }
    const isOnWab = null;

    // Return
    if (this.vitisCheck.isDefAndNotNull(isOnWab)) {
      if (isVisible && isOnTab && isOnWab) {
        bReturn = true;
      } else {
        bReturn = false;
      }
    } else {
      if (isVisible && isOnTab) {
        bReturn = true;
      } else {
        bReturn = false;
      }
    }

    // Event onvisible
    if (oField.onvisible) {
      if (isOnTab) {
        const bTabVisible = (bReturn && oTab.collapsed !== true);
        if (oField._visible !== bTabVisible) {
          oField.onvisible.next(bTabVisible);
        }
        oField._visible = bTabVisible;
      }
    }

    return bReturn;
  }

  /**
   * Vérifie si la chaine de caractère passée est un appel de fonction
   * @param {string} sString Chaine de caractère.
   * @param {boolean} bCheckFunction? Vérification de l'existence de la fonction (vrai par défaut).
   * @param {any} oFunctionContainer? objet ou tableau d'objets dans lequel est possiblement contenu la fonction
   * @return {boolean}
   */
  public isFunctionCall(sString: string, bCheckFunction?: boolean, oFunctionContainer?: any): boolean {
    return this.functionService.isFunctionCall(sString, bCheckFunction, oFunctionContainer);
  }

  /**
   * isCustomTernaryString function.
   * La chaine de caractère passée est une expression ternaire définie par l'administrateur
   * ex : "= {{puissance}} == 500"
   * @param {string} sString Chaine de caractère.
   * @param {boolean} bIsCondition True si il s'agis d'une condition true/false
   * @return {boolean}
   */
  private isCustomTernaryString_(sString: string, bIsCondition: boolean): boolean {
    if (this.vitisCheck.isDefAndNotNull(this.translateCustomTernaryString_(sString, bIsCondition))) {
      return true;
    } else {
      return false;
    }
  }

  /**
   * La chaine de caractère passée est une expression ternaire ?
   * @param {string} sString Chaine de caractère.
   * @return {boolean}
   */
  private isTernaryString_(sString: string): boolean {
    const ter = /([^?]{1,})\?([^:]{1,}):([^;]{1,})/gm;
    if (this.vitisCheck.isString(sString)) {
      return sString.match(ter) ? true : false;
    } else {
      return false;
    }
  }

  /**
   * Vérifie si le champs est visible au travers d'une fonction ou d'un ternaire
   * @param {any} exp expression
   *
   * @returns {boolean}
   */
  private isVisible_(exp: any): boolean {
    // Visible au travers une fonction ou un ternaire
    if (this.isFunctionCall(exp)) {
      return this.callFunction(exp, null);
    } else if (this.isTernaryString_(exp)) {
      try {
        return eval(exp);
      } catch (e) {
        return false;
      }
    } else if (this.isCustomTernaryString_(exp, true)) {
      try {
        const evalString = this.translateCustomTernaryString_(exp, true, this.oFormValues);
        return eval(evalString);
      } catch (e) {
        return false;
      }
    } else if (this.vitisCheck.isBoolean(exp)) {
      return exp;
    } else if (exp == true || exp == false || exp == 'true' || exp == 'false') {
      return exp;
    } else {
      return false;
    }
  }

  /**
   * translateCustomTernaryString function
   * Traduit une chaine ternaire customisée par une chaine ternaire exploitable
   * ex : "= {{puissance}} == 500" devient "oFormValues.value[puissance] == 500 ? true : false"
   * @param  {string} sString
   * @param  {boolean} bIsCondition
   * @param  {object} oFormValues
   * @return {string}
   */
  private translateCustomTernaryString_(sString: string, bIsCondition: boolean, oFormValues?: any): string {

    // Verif type
    if (!this.vitisCheck.isString(sString)) {
      return null;
    }

    // Vérification chaine
    if (sString.substr(0, 1) !== '=') {
      return null;
    }

    // Enlève le premier signe "="
    sString = sString.substr(1);

    // Parse le template
    sString = this.parseCustomTernaryString_(
      sString,
      // S'il y a des symboles d'interpolation
      sString.indexOf('{{') >= 0 && sString.indexOf('}}') >= 0
    );

    // Ajoute le retour true/false
    if (bIsCondition) {
      sString = '(' + sString + ') ? true : false';
    }

    return sString;
  }

  /**
   * parseCustomTernaryString_ function
   * Parse une chaine ternaire customisée par une chaine ternaire exploitable
   * ex : "= oProperties.enableEncryptionForUsers == true" devient "true == true"
   * @param  {string} sString
   * @param  {boolean} bMustache
   * @return {string}
   */
  private parseCustomTernaryString_(sString: string, bMustache: boolean): string {
    // S'il y a des crochets on remplace par oFormValues
    if (bMustache) {
      // Suppression d'espaces entre crochets
      sString = sString.replace(/{{ */g, '{{');
      sString = sString.replace(/ *}}/g, '}}');
      const aKey = sString.match(/{{(\w+(\.\w+)*)}}/g);
      // Parcours tous les attributs trouvé
      aKey.forEach((sKey: string) => {
        // Remplace toutes les occurances trouvées
        sString = sString.replace(RegExp(sKey, 'g'), 'oFormValues.' + sKey.replace(/[{}]+/g, ''));
      });
    }
    // Trouve tout les mots(variables) dans le template
    const aFinded = sString.match(/(\w+(\.\w+)*)/g);
    aFinded.forEach((sFinded: string) => {
      let replacementValue: any = null;
      // Decoupe le mot trouvé
      const aSplited = sFinded.split('.');
      // Remplace la première partie par ça valeur
      switch (aSplited[0]) {
        case 'oFormValues':
          replacementValue = this.oFormValues.getRawValue();
          break;
        case 'oProperties':
          replacementValue = this.oProperties;
          break;
        case 'oPrivileges':
          replacementValue = {};
          for (let i = 0; i < this.aUserPrivileges.length; i++) {
            const sPrivilege = this.aUserPrivileges[i];
            replacementValue[sPrivilege] = true;
          }
          break;
      }
      // si autre chose est bind on bypass
      if (replacementValue !== null && aSplited.length > 1) {
        // Remplace les attributes par leur valeur
        for (let idx = 1; idx < aSplited.length; idx++) {
          if (this.vitisCheck.isDefAndNotNull(replacementValue[aSplited[idx]])) {
            replacementValue = replacementValue[aSplited[idx]];
          } else {
            replacementValue = '';
            break;
          }
        }
        let sValue = (replacementValue === null) ? '""' : replacementValue;

        if (typeof (replacementValue) === 'string') {
          // Echappement des quotes.
          sValue = '\'' + replacementValue.replace(/'/g, '\\\'') + '\'';
        }
        // Remplace la variable trouvé par sa valeur
        sString = sString.replace(sFinded, sValue);
      }
    });
    return sString;
  }

  /**
   * Cherche les attributs dans l'expression ternaire
   *
   * @param  {string} sString
   * @return {array}
   */
  private getCustomTernaryStringAttrs_(sString: string): Array<any> {
    // Cherche les attributs
    const aAttrs = [];
    const attrRegex = RegExp('{{(\\w*?)}}', 'g');
    let attrRegexResult: Array<any>;
    while ((attrRegexResult = attrRegex.exec(sString)) !== null) {
      aAttrs.push(attrRegexResult);
    }

    return aAttrs;
  }

  /**
   * Set the filed value using dynamic_value
   *
   * @param  {any} oField
   */
  private setDynamicValue_(oField: any) {

    if (oField.getValue) {
      let value = oField.getValue();

      if (this.isCustomTernaryString_(oField.dynamic_value, false)) {
        try {
          value = eval(this.translateCustomTernaryString_(oField.dynamic_value, false));

          if (oField.setValue) {
            oField.setValue(value);
          }
        } catch (e) {
          console.error('error with ternary : ' + oField.dynamic_value);
        }
      }
    }
  }

  /**
   * Ajoute les watchers sur les champs présents dans l'expression dynamic_value
   * @param {Array<any>} aFields
   */
  public setFormDynamicValues(aFields: Array<any>) {
    for (let i = 0; i < aFields.length; i++) {
      const oField = aFields[i];

      // Valeurs dynamiques
      if (this.vitisCheck.isDefAndNotNull(oField.dynamic_value)) {

        // Ajoute les watchers sur les champs présents dans l'expression dynamic_value
        if (this.isCustomTernaryString_(oField.dynamic_value, false)) {
          const aAttrs = this.getCustomTernaryStringAttrs_(oField.dynamic_value);
          for (let ii = 0; ii < aAttrs.length; ii++) {
            if (this.vitisCheck.isDefAndNotNull(aAttrs[ii][1])) {
              this.oFormValues.controls[aAttrs[ii][1]].valueChanges.subscribe((value) => {
                setTimeout(() => {
                  this.setDynamicValue_(oField);
                }, 100);
              });
            }
          }
        }
      }
    }
  }

  /**
   * Appel une fonction pour les evenements dynamiques du formulaire
   * @param oField
   * @param oContainer
   * @param value
   */
  public callbackFormDynamicFunction(oField: any, oContainer: any, value?: any) {
    const sFunction = oField.dynamic_event_function;
    if (this.isFunctionCall(sFunction, true, oContainer)) {
      this.callFunction(sFunction, value, oContainer);
    } else if (this.isCustomTernaryString_(sFunction, false)) {
      eval(this.translateCustomTernaryString_(sFunction, true));
    } else {
      try {
        eval(oField.dynamic_event_function);
      } catch (error) {
        console.error('Error while calling event ' + oField.dynamic_event_type, error);
      }
    }
  }

  /**
   * Retire le validateur required si un champ n'est pas visible
   * @param {any} oField
   * @param {boolean} isVisible
   */
  public setDynamicRequired(oField: any, isVisible: boolean) {
    if (oField.required === true) {
      let aValidators = [];
      let bModified = false;

      // Récupère les validators inclus par FormReader
      if (this.vitisCheck.isArray(oField.validators_list)) {
        aValidators = oField.validators_list;
      }

      // Récupère l'index du validator required (pour vérifier son existance)
      const iRequiredValidatorIndex = aValidators.findIndex((validator) => validator == Validators.required);

      if (isVisible === true) {

        // Ajoute le validateur required
        if (iRequiredValidatorIndex === -1) {
          aValidators.push(Validators.required);
          bModified = true;
        }

      } else {

        // Retire le validateur required
        if (iRequiredValidatorIndex !== -1) {
          delete aValidators[iRequiredValidatorIndex];
          bModified = true;
        }
      }

      // Applique la mise à jour
      if (bModified) {
        if (this.oFormValues.controls[oField.name]) {
          this.oFormValues.controls[oField.name].clearValidators();
          this.oFormValues.controls[oField.name].setValidators(aValidators);
          this.oFormValues.controls[oField.name].updateValueAndValidity();
          this.changeDetectorRef.detectChanges();
        }
      }
    }
  }

  /**
   * Récupère la valeur qui correspond à un ternaire
   * @param {string} sString expression
   *
   * @returns {string}
   */
  public getTernaryStringValue(sString: string): string {
    if (this.isCustomTernaryString_(sString, false)) {
      try {
        const evalString = this.translateCustomTernaryString_(sString, false);
        return eval(evalString);
      } catch (e) {
        return null;
      }
    } else {
      return null;
    }
  }
}
