import { Component, OnInit } from '@angular/core';
import { ActivatedRoute, Router } from '@angular/router';
import { TranslateService } from '@ngx-translate/core';
import { VitisCheck, VitisRequest, vitisAlert } from 'src/app/vitis/core/services/utils/vitis-utils';
import { environment } from 'src/environments/environment';

declare var $: any;
declare var monaco: any;

@Component({
  selector: 'app-logs',
  templateUrl: './logs.component.html',
  styleUrls: ['./logs.component.scss']
})
export class LogsComponent implements OnInit {
  public mode: any;

  // Paramètres du Treeview.
  public treeviewConfig = {
    titleAttribute: 'filename',
    nodeAttribute: 'content',
    pathAttribute: 'key',
    dateAttribute: 'lastModification',
    defaultOpenLevel: 0,
    multiSelect: false,
    showDate: false,
    hideEmptyNodes: true,
    hoverButtons: [{
      icon: 'fa fa-download',
      event: (item: any) => {
        this.downloadTreeviewItem(item);
      }
    }, {
      icon: 'fa fa-trash',
      event: (item: any) => {
        this.removeLogFile(item);
      }
    }]
  };
  public treeviewItems: any;
  public treeviewFilter: any;
  public sLogFileName: string;

  // Paramètre de l'éditeur de texte.
  public editorCode: string;
  public editorOptions = {
    theme: 'vs',
    language: 'plaintext',
    readOnly: true,
    fontFamily: 'Arial',
    fontSize: 13,
    automaticLayout: true
  };

  constructor(
    private route: ActivatedRoute,
    private router: Router,
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private translate: TranslateService
  ) { }

  ngOnInit() {
    // Chargement de l'arborescence des logs.
    this.loadTreviewData();
  }

  /**
   * loadTreviewData function
   * @returns void
   */
  public loadTreviewData = function () {

    // Charge l'arborescence des fichiers de logs
    this.vitisRequest.ajaxRequestPromise({
      method: 'GET',
      url: environment.properties.app_api_url + '/vitis/logs'
    }).then((resp: any) => {
      if (resp.data) {
        if (resp.data.status === 1) {
          this.treeviewItems = resp.data.data;
        }
      }
    });
  };

  /**
   * onEditorInit function
   * @param  {any} editor
   * @returns void
   */
  public onEditorInit = function (editor: any) {

    // Création d'un nouveau langage "logs" dans l'éditeur de texte.
    monaco.languages.register({ id: 'logs' });

    // Register a tokens provider for the language
    monaco.languages.setMonarchTokensProvider('logs', {
      tokenizer: {
        root: [
          [/((e|E)rror)/, 'custom-error'], // Erreurs
          [/(notice|(w|W)arning|warn)/i, 'custom-notice'], // Warnings
          [/\[info.*/, 'custom-info'],
          [/^\[?[0-9a-zA-Z- :\/]+\]?/, 'custom-date'] // Date du log
        ]
      }
    });

    // Création d'un nouveau thème "logsTheme" dans l'éditeur de texte.
    monaco.editor.defineTheme('logsTheme', {
      base: 'vs',
      inherit: false,
      rules: [
        { token: 'custom-info', foreground: '808080' },
        { token: 'custom-error', foreground: 'ff0000', fontStyle: 'bold' },
        { token: 'custom-notice', foreground: 'FFA500' },
        { token: 'custom-date', foreground: '008800' },
      ],
      colors: {}
    });

    /**
     * getLogsProvider function
     * @param  {any} monaco_
     * @returns suggestions
     */
    this.getLogsProvider = function (monaco_: any) {
      return {
        provideCompletionItems: () => {
          const suggestions = [{
            label: 'simpleText',
            kind: monaco_.languages.CompletionItemKind.Text,
            insertText: 'simpleText'
          }, {
            label: 'testing',
            kind: monaco_.languages.CompletionItemKind.Keyword,
            insertText: 'testing(${1:condition})',
            insertTextRules: monaco_.languages.CompletionItemInsertTextRule.InsertAsSnippet
          }, {
            label: 'ifelse',
            kind: monaco_.languages.CompletionItemKind.Snippet,
            insertText: [
              'if (${1:condition}) {',
              '\t$0',
              '} else {',
              '\t',
              '}'
            ].join('\n'),
            insertTextRules: monaco_.languages.CompletionItemInsertTextRule.InsertAsSnippet,
            documentation: 'If-Else Statement'
          }];
          return { suggestions: suggestions };
        }
      };
    };

    monaco.languages.registerCompletionItemProvider('logs', this.getLogsProvider(monaco));

    // Change le thème et le langage.
    monaco.editor.setTheme('logsTheme');
    monaco.editor.setModelLanguage(editor.getModel(), 'logs');
  };

  /**
   * onSelectedChange function
   * @param  {any} selectedItems
   * @returns void
   */
  public onSelectedChange = function (selectedItems: any) {
    // Charge le contenu du fichier sélectionné
    if (this.vitisCheck.isArray(selectedItems) && selectedItems.length > 0) {
      this.sLogFileName = selectedItems[0];
      // Si ce n'est pas un fichier log ne l'affiche pas
      if (this.sLogFileName.split('.').pop() !== 'log') {
        return;
      }
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: environment.properties.app_api_url + '/vitis/logs/content',
        params: {
          key: selectedItems.pop()
        }
      }).then((resp: any) => {
        if (resp.data) {
          if (resp.data.status === 1) {
            this.editorCode = resp.data.data.length == 0 ? ' ' : resp.data.data;
          } else {
            // Message d'erreur
            this.translate.get('LOADING_FILE_ERROR_LOGS').subscribe((translation) => {
              const options: any = {
                type: 'error',
                title: translation
              };
              vitisAlert.alert(options);
            });
          }
        }
      });
    }
  };

  /**
   * onFilterChange function
   * @param  {any} event
   * @returns void
   */
  public onFilterChange = function (event: any) { };

  /**
   * removeTreeviewItem function
   * @param  {array} aTree
   * @param  {any} itemValue
   * @returns void
   */
  public removeTreeviewItem = function (aTree: any, itemValue: any) {
    for (let i = 0; i < aTree.length; i++) {
      if (aTree[i].path == itemValue) {
        if (aTree[i].selected) {
          this.editorCode = null;
        }
        aTree.splice(i, 1);
      } else if (this.vitisCheck.isArray(aTree[i].children)) {
        this.removeTreeviewItem(aTree[i].children, itemValue);
      }
    }
  };

  /**
   * removeLogFile function
   * @param  {object} item
   * @returns void
   */
  public removeLogFile = function (item: any) {
    // Modale de confirmation.
    this.translate.get(['DELETE_FILE_CONFIRM_LOGS']).subscribe((translations) => {

      vitisAlert.alert({
        type: 'warning',
        html: translations['DELETE_FILE_CONFIRM_LOGS'] + '<br>(' + item.key + ')',
        showCancelButton: true
      }).then(result => {
        if (result.value) {
          // Suppression du fichier.
          this.vitisRequest.ajaxRequestPromise({
            method: 'DELETE',
            url: environment.properties.app_api_url + '/vitis/logs/delete',
            params: {
              key: item.key
            },
          }).then((resp: any) => {
            if (resp.data) {
              if (resp.data.status === 1) {
                // Suppression de l'item dans l'arborescence du treeview.
                this.removeTreeviewItem(this.treeviewItems, item.key);
                // Message de succès.
                this.translate.get('DELETE_FILE_SUCCESS_LOGS').subscribe((translation) => {
                  vitisAlert.fireToast({
                    icon: 'success',
                    title: translation
                  });
                });
                // Recharge l'arborescence des logs.
                this.editorCode = '';
                this.sLogFileName = '';
                this.loadTreviewData();
              } else {
                // Message d'erreur
                let sErrorMessage = 'DELETE_FILE_ERROR_LOGS';
                if (this.vitisCheck.isDefAndNotNull(resp.data.errorMessage)) {
                  sErrorMessage = resp.data.errorMessage;
                }
                this.translate.get(sErrorMessage).subscribe((translation) => {
                  const options: any = {
                    type: 'error',
                    text: translation
                  };
                  vitisAlert.alert(options);
                });
              }
            }
          });
        }
      });
    });
  };

  /**
   * downloadTreeviewItem function
   * @param  {object} item
   * @returns void
   */
  public downloadTreeviewItem = function (item: any) {
    /* Charge le contenu du fichier sélectionné */
    const aSelectedFilePath = item.key.split('/');
    const sFileName = aSelectedFilePath.pop();
    this.vitisRequest.ajaxRequestPromise({
      method: 'GET',
      url: environment.properties.app_api_url + '/vitis/logs/download',
      params: {
        key: item.key
      },
      responseType: 'blob',
      headers: {
        Accept: 'application/octet-stream'
      }
    }).then((resp: any) => {
      if (resp.data) {
        if (resp.data instanceof Blob) {
          if (typeof (window.navigator.msSaveOrOpenBlob) == 'undefined') {
            // Crée un lien et simule un click dessus.
            const link = document.createElement('a');
            link.href = window.URL.createObjectURL(resp.data);
            link.download = sFileName;
            link.style.display = 'none';
            document.body.appendChild(link);
            link.click();
            document.body.removeChild(link);
          } else {
            // Pour Edge et IE.
            window.navigator.msSaveOrOpenBlob(resp.data, sFileName);
          }
        } else {
          // Message d'erreur
          this.translate.get('LOADING_FILE_ERROR_LOGS').subscribe((translation) => {
            const options: any = {
              type: 'error',
              title: translation
            };
            vitisAlert.alert(options);
          });
        }
      }
    });
  };

  /**
   * removeLogsHistory function
   * @returns void
   */
  public removeLogsHistory = function () {

    // Modale de confirmation.
    this.translate.get(['DELETE_HISTORY_CONFIRM_LOGS']).subscribe((translations) => {
      vitisAlert.alert({
        input: 'number',
        inputValue: '30',
        inputAttributes: {
          min: '0',
          max: '100'
        },
        title: translations['DELETE_HISTORY_CONFIRM_LOGS'],
        showCancelButton: true
      }).then(result => {
        if (typeof (result.value) != 'undefined') {
          // Suppression de l'historique.
          this.vitisRequest.ajaxRequestPromise({
            method: 'DELETE',
            url: environment.properties.app_api_url + '/vitis/logs/purge',
            params: {
              log_lifetime: result.value
            }
          }).then((resp: any) => {
            if (resp.data) {
              if (resp.data.status === 1) {
                // Message de succès.
                this.translate.get('DELETE_FILES_SUCCESS_LOGS').subscribe((translation) => {
                  vitisAlert.fireToast({
                    icon: 'success',
                    title: translation
                  });
                });
                // Recharge l'arborescence des logs.
                this.editorCode = '';
                this.sLogFileName = '';
                this.loadTreviewData();
              } else {
                // Message d'erreur
                let sErrorMessage = 'DELETE_FILES_ERROR_LOGS';
                if (this.vitisCheck.isDefAndNotNull(resp.data.errorMessage)) {
                  sErrorMessage = resp.data.errorMessage;
                }
                this.translate.get(sErrorMessage).subscribe((translation) => {
                  const options: any = {
                    type: 'error',
                    text: translation
                  };
                  vitisAlert.alert(options);
                });
              }
            }
          });
        }
      });
    });
  };

  /**
   * reloadTreeviewData function
   * @returns void
   */
  public reloadTreeviewData = function () {
    this.editorCode = '';
    this.sLogFileName = '';
    this.loadTreviewData();
  };

  /**
   * reloadLogData function
   * @returns void
   */
  public reloadLogData = function () {
    this.vitisRequest.ajaxRequestPromise({
      method: 'GET',
      url: environment.properties.app_api_url + '/vitis/logs/content',
      params: {
        key: this.sLogFileName
      }
    }).then((resp: any) => {
      if (resp.data) {
        if (resp.data.status === 1) {
          this.editorCode = resp.data.data.length == 0 ? ' ' : resp.data.data;
        } else {
          // Message d'erreur
          this.translate.get('LOADING_FILE_ERROR_LOGS').subscribe((translation) => {
            const options: any = {
              type: 'error',
              title: translation
            };
            vitisAlert.alert(options);
          });
        }
      }
    });
  };
}
