import { Injectable, SimpleChanges } from '@angular/core';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';
import { MapCoreManagerService } from '../map-core-manager/map-core-manager.service';

// OpenLayers
import Map from 'node_modules/ol/Map';
import { Subject, Observable, BehaviorSubject } from 'rxjs';
import { MapJSON } from '../map-models.model';
import { filter, map } from 'rxjs/operators';
import { MapCoreUtilsService } from '../map-core-utils/map-core-utils.service';

/**
 * Service permettant d'utiliser la carte dans le formulaires FormReader
 */
@Injectable({
  providedIn: 'root'
})
export class MapCoreFormreaderService {
  private mapFormreaderEvent = new BehaviorSubject<any>(undefined);

  constructor(
    private vitisCheck: VitisCheck,
    private mapCoreManagerService: MapCoreManagerService,
    private mapCoreUtilsService: MapCoreUtilsService,
  ) { }

  /**
   * Initialise la carte pour FormReader
   * @param mapJSON
   * @param targetId
   * @param tileSize
   */
  public initMapFormReader(mapJSON: MapJSON, targetId: string, tileSize: Array<number>): Observable<Map> {
    const oReturn = new Subject<Map>();

    this.getMapDefinition(mapJSON).then((oMapDefinition) => {
      const olMap = this.mapCoreManagerService.initMapFromDef(oMapDefinition, targetId, tileSize);
      if (olMap) {
        // Supprime le zoom au double click.
        const aInteractions = olMap.getInteractions().getArray();
        for (let i = 0; i < aInteractions.length; i++) {
          if (aInteractions[i].constructor.name == 'DoubleClickZoom') {
            olMap.removeInteraction(aInteractions[i]);
            break;
          }
        }
        oReturn.next(olMap);
      }
    });
    return oReturn;
  }

  /**
   * Récupère la définition de la carte en fonction d'une définition JSON
   * @param  {any} mapJSON
   */
  public getMapDefinition(mapJSON: MapJSON): Promise<any> {
    return new Promise((resolve) => {
      // vmap
      if (mapJSON.type == 'vmap') {

        // Initialisation par un arbre
        if (this.vitisCheck.isDefAndNotNull(mapJSON.tree)) {
          resolve(mapJSON.tree);
        } else if (this.vitisCheck.isDefAndNotNull(mapJSON.map_id)) {
          const getMapDefAndInitOl = () => {
            this.mapCoreManagerService.ajaxGetMap(mapJSON.map_id).then((oMapDefinition) => {
              resolve(oMapDefinition);
            });
          };
          this.mapCoreManagerService.generatePrivateWMSService().then(() => {
            getMapDefAndInitOl();
          }).catch(() => {
            getMapDefAndInitOl();
          });
        }
      } else if (mapJSON.type.toLowerCase() === 'osm') {

        const oOSMMapDefinition = this.getOSMDef_(mapJSON);
        resolve(oOSMMapDefinition);

      } else if (mapJSON.type.toLowerCase() == 'bing') {

        const oBingMapDefinition = this.getBingDef_(mapJSON);
        resolve(oBingMapDefinition);

      } else {
        console.error('Map type not allowed: ', mapJSON.type);
      }
    });
  }

  /**
   * Retourne le définition de la vue en fonction de mapJSON
   * @param mapJSON
   */
  private getViewDef_(mapJSON: MapJSON): any {

    const viewDef: any = {
      extent: [-1576070.422, 5031389.978, 2176070.422, 6968610.022],
      projection: 'EPSG:3857'
    };

    if (mapJSON.proj) {
      viewDef.projection = mapJSON.proj;
    }

    if (mapJSON.extent) {
      viewDef.extent = mapJSON.extent;
    }

    if (mapJSON.center) {
      delete viewDef.extent;
      if (mapJSON.center.coord) {
        viewDef.center = mapJSON.center.coord;
      }
      if (mapJSON.center.extent) {
        viewDef.extent = mapJSON.center.extent;
      }
      if (mapJSON.center.scale) {
        viewDef.scale = mapJSON.center.scale;
      }
    }

    return viewDef;
  }

  /**
   * Retourne une définition de carte OSM depuis mapJSON
   * @param mapJSON
   */
  private getOSMDef_(mapJSON: MapJSON): any {

    const oViewDef = this.getViewDef_(mapJSON);

    const oMapDefinition = {
      name: 'Tree',
      children: [{
        view: oViewDef
      }, {
        name: 'Open Street Maps',
        children: [{
          index: 1,
          layerType: 'osm',
          name: 'Open Street Maps',
          visible: true
        }]
      }]
    };

    return oMapDefinition;
  }

  /**
   * Retourne une définition de carte OSM depuis mapJSON
   * @param mapJSON
   */
  private getBingDef_(mapJSON: MapJSON): any {

    const oViewDef = this.getViewDef_(mapJSON);

    const oMapDefinition = {
      name: 'Tree',
      children: [{
        view: oViewDef
      }, {
        name: 'Bing',
        children: [{
          index: 1,
          layerType: 'bing',
          name: mapJSON.source.style,
          key: mapJSON.source.key,
          culture: mapJSON.source.culture,
          imagerySet: mapJSON.source.style,
          visible: true
        }]
      }]
    };

    return oMapDefinition;
  }

  /**
   * Permet d'emettre un evenement pour la carte
   * @param fieldId identifiant du champ formreader
   * @param data
   */
  public emitMapFormreaderEvent(fieldId: any, data: any) {
    this.mapFormreaderEvent.next({ fieldId, data });
  }

  /**
   * Listener
   * @param fieldId identifiant du champ formreader
   */
  public mapFormreaderEventListner(fieldId: any) {
    return this.mapFormreaderEvent.asObservable()
      .pipe(
        filter(e => {
          if (!e) {
            return;
          }
          return e.fieldId === fieldId;
        }),
        map(e => e.data)
      );
  }

  /**
   * Met les valeurs set par le formreader dans la carte
   * @param changements
   * @param olMap
   * @returns aFeatures Tableau des features ajouté à la carte.
   */
  public setValues(changements: SimpleChanges, olMap: Map) {
    if (this.vitisCheck.isDefAndNotNull(changements.sGeom)) {
      if (this.vitisCheck.isDefAndNotNull(changements.sGeom.currentValue) &&
        changements.sGeom.currentValue !== changements.sGeom.previousValue) {
        if (this.vitisCheck.isDefAndNotNull(olMap)) {
          const olDrawLayer = olMap.get('draw_layer');

          let aFeatures = [];
          if (this.vitisCheck.isArray(changements.sGeom.currentValue)) {
            for (let i = 0; i < changements.sGeom.currentValue.length; i++) {
              const sGeom = changements.sGeom.currentValue[i];
              aFeatures = [...aFeatures, ...this.mapCoreUtilsService.getFeaturesFromGeom(sGeom, olMap)];
            }
          } else {
            aFeatures = this.mapCoreUtilsService.getFeaturesFromGeom(changements.sGeom.currentValue, olMap);
          }

          if (this.vitisCheck.isDefAndNotNull(aFeatures) && this.vitisCheck.isDefAndNotNull(olDrawLayer)) {
            // Ajout des features au composant dessin
            olDrawLayer.getSource().clear();
            olDrawLayer.getSource().addFeatures(aFeatures);

            return aFeatures;
          }
        }
      }
    }
  }
}
