import { Injectable } from '@angular/core';
import sha256 from 'crypto-js';

// Vitis
import { VitisCheck, VitisClone } from 'src/app/vitis/core/services/utils/vitis-utils';
import { AuthService } from 'src/app/vitis/core/services/auth/auth.service';
import { environment } from 'src/environments/environment';

// OpenLayers
import View from 'ol/View.js';
import { BingMaps, ImageWMS, TileWMS, Vector as VectorSource, XYZ, OSM, Source } from 'ol/source';
import WMTS, { optionsFromCapabilities } from 'ol/source/WMTS';
import { GeoJSON } from 'ol/format';
import ImageLayer from 'ol/layer/Image';
import VectorLayer from 'ol/layer/Vector';
import TileLayer from 'ol/layer/Tile';
import Layer from 'ol/layer/Layer';
import Map from 'node_modules/ol/Map';
import olms from 'ol-mapbox-style';
import { getLayers as mapBoxGetLayers, apply as mapBoxApply } from 'ol-mapbox-style';

import { containsCoordinate as extentContainsCoordinate } from 'ol/extent';
import { Geometry } from 'ol/geom';
import ImageSource from 'ol/source/Image';
import TileSource from 'ol/source/Tile';

import WMTSTileGrid from 'ol/tilegrid/WMTS';
import { get as getProjection } from 'ol/proj';
import { getWidth } from 'ol/extent';
import { Subject } from 'rxjs';


@Injectable({
  providedIn: 'root'
})
export class MapCoreJSONService {

  constructor(
    private vitisCheck: VitisCheck,
    private vitisClone: VitisClone,
    private auth: AuthService
  ) { }

  /**
   * Retourne une vue OpenLayers depuis la définition de la carte
   * @param {object} oMapDefinition
   * @param {object} opt_options?
   * @param {array} opt_options.size
   * @returns {View}
   */
  public getViewFromDef(oMapDefinition: any, opt_options?: any): View {

    opt_options = this.vitisCheck.isDefAndNotNull(opt_options) ? opt_options : {};

    if (!this.vitisCheck.isDefAndNotNull(oMapDefinition)) {
      console.error('oMapDefinition is not defined: ' + oMapDefinition);
      return null;
    }
    if (!this.vitisCheck.isDefAndNotNull(oMapDefinition.children)) {
      console.error('oMapDefinition.children is not defined: ' + oMapDefinition);
      return null;
    }
    if (!this.vitisCheck.isDefAndNotNull(oMapDefinition.children[0])) {
      console.error('oMapDefinition.children[0] is not defined: ' + oMapDefinition);
      return null;
    }
    if (!this.vitisCheck.isDefAndNotNull(oMapDefinition.children[0].view)) {
      console.error('oMapDefinition.children[0].view is not defined: ' + oMapDefinition);
      return null;
    }

    const oViewDef = oMapDefinition.children[0].view;

    const oViewOptions: any = {};
    oViewOptions.center = this.vitisCheck.isDefAndNotNull(oViewDef.center) ? oViewDef.center : [0, 0];
    oViewOptions.zoom = this.vitisCheck.isDefAndNotNull(oViewDef.zoom) ? oViewDef.zoom : 0;
    oViewOptions.constrainRotation = oViewDef.constrainRotation;
    oViewOptions.enableRotation = this.vitisCheck.isDefAndNotNull(oViewDef.enableRotation) ? oViewDef.enableRotation : false;
    oViewOptions.extent = oViewDef.maxExtent;
    oViewOptions.maxResolution = oViewDef.maxResolution;
    oViewOptions.minResolution = oViewDef.minResolution;
    oViewOptions.maxZoom = oViewDef.maxZoom;
    oViewOptions.minZoom = oViewDef.minZoom;
    oViewOptions.projection = oViewDef.projection;
    oViewOptions.resolution = oViewDef.resolution;
    oViewOptions.resolutions = oViewDef.resolutions;
    oViewOptions.rotation = oViewDef.rotation;
    oViewOptions.zoomFactor = oViewDef.zoomFactor;

    const olView = new View(oViewOptions);

    // Étendue de la vue
    if (this.vitisCheck.isDefAndNotNull(oViewDef.extent)) {
      if (this.vitisCheck.isDefAndNotNull(opt_options.size)) {
        olView.fit(oViewDef.extent, {
          size: opt_options.size,
          nearest: true
        });
      } else {
        console.error('Cannot calc olView.extent if opt_options.size is not defined');
      }
    }

    // Vérifie si le centre de la carte est bien placée dans l'étendu du système de projection
    const onExtent = extentContainsCoordinate(olView.getProjection().getExtent(), olView.getCenter());
    if (this.vitisCheck.isDefAndNotNull(oViewDef.center) && !onExtent) {
      console.error('Map center outside projection extent');
      olView.setCenter([0, 0]);
    }

    return olView;
  }

  /**
   * Retourne un tableau de couches OpenLayers depuis la définition de la carte
   * @param {object} oMapDefinition
   * @param {object} opt_options?
   * @param {array} opt_options.size
   * @param {array} opt_options.tileSize
   * @param {array} opt_options.isOverview
   * @returns {Array<ol.layer>}
   */
  public getLayersFromDef(oMapDefinition: any, opt_options?: any, olMap?: Map, vectortilesLoaded?: Subject<any>): any {

    opt_options = this.vitisCheck.isDefAndNotNull(opt_options) ? opt_options : {};

    if (!this.vitisCheck.isDefAndNotNull(oMapDefinition)) {
      console.error('oMapDefinition is not defined: ' + oMapDefinition);
      return null;
    }
    if (!this.vitisCheck.isDefAndNotNull(oMapDefinition.children)) {
      console.error('oMapDefinition.children is not defined: ' + oMapDefinition);
      return null;
    }
    if (!this.vitisCheck.isDefAndNotNull(opt_options.size)) {
      console.error('opt_options.size not defined');
      return null;
    }
    if (!this.vitisCheck.isDefAndNotNull(opt_options.tileSize)) {
      console.error('opt_options.tileSize not defined');
      return null;
    }

    const aLayersDef = this.getLayersDef_(oMapDefinition);
    const aLayers = [];
    let bHaveVectorTiles = false;

    for (let i = 0; i < aLayersDef.length; i++) {
      if (this.vitisCheck.isDefAndNotNull(aLayersDef[i].url)) {
        if (aLayersDef[i].url.indexOf('[token]') > -1) {
          aLayersDef[i].url = aLayersDef[i].url.replace(/\[token\]/, sha256(this.auth.getToken()));
        }
        if (aLayersDef[i].url.indexOf('[localhost]') > -1) {
          aLayersDef[i].url = aLayersDef[i].url.replace(/\[localhost\]/, environment.properties.web_server_name);
        }
      }
      const oSource = this.getSourceFromLayerDef_(oMapDefinition, aLayersDef[i], opt_options);
      if (this.vitisCheck.isDefAndNotNull(oSource)) {
        const oLayer = this.getLayerFromLayerDef_(oMapDefinition, aLayersDef[i], oSource);
        if (this.vitisCheck.isDefAndNotNull(oLayer)) {
          aLayers.push(oLayer);
        }
      }
      if (olMap && aLayersDef[i].type === 'vectortiles') {
        bHaveVectorTiles = true;
      }
    }

    // Vectortiles non disponibles dans l'overview pour des raisons de performance
    if (bHaveVectorTiles && opt_options.isOverview !== true) {
      this.handleVectorTiles_(olMap, aLayersDef, vectortilesLoaded);
    }

    return aLayers;
  }

  /**
   * Gère l'affectaction des couches vectortiles
   * @param olMap
   * @param aLayersDef
   * @param vectortilesLoaded
   */
  private handleVectorTiles_(olMap: Map, aLayersDef: Array<any>, vectortilesLoaded?: Subject<any>) {
    const aVectorTilesDefs = aLayersDef.filter(layer => layer.type === 'vectortiles');
    for (let i = 0; i < aVectorTilesDefs.length; i++) {
      const oVmapLayer = aVectorTilesDefs[i];
      try {
        const oMapboxStyle = JSON.parse(oVmapLayer.mapbox_gl_style);

        // Charge les couches sur la carte
        const map = mapBoxApply(olMap, oMapboxStyle);
        setTimeout(() => {
          if (oMapboxStyle && oMapboxStyle.sources) {
            for (const sourceName in oMapboxStyle.sources) {
              if (Object.prototype.hasOwnProperty.call(oMapboxStyle.sources, sourceName)) {

                // Récupère les couches chargées (une couche vMap peut correspondre à plusieurs couches vectortiles)
                const aLayers = mapBoxGetLayers(olMap, sourceName);

                // Attend le chargement de la couche
                for (let ii = 0; ii < aLayers.length; ii++) {
                  const olLayer = aLayers[ii];
                  olLayer.once('postrender', () => {

                    // Set les paramètres de la couche (visibilité etc)
                    this.setLayerParameters_(olLayer, oVmapLayer);

                    // Déclenche l'event de chargement ok (pour toutes les couches, d'où le gros timeout)
                    setTimeout(() => {
                      vectortilesLoaded.next(aLayersDef);
                    }, 1000);
                  });
                }
              }
            }
          }
        });
      } catch (error) {
        console.error('MapboxStyle not valid', oVmapLayer);
      }
    }
  }

  /**
   * Retourne la définition JSON de la carte (sans le bind des couches)
   * @param olMap
   * @returns
   */
  public getMapDef(olMap: Map): any {
    const oMapDef = olMap.get('map_definition');
    const oNewMapDef: any = {};

    if (oMapDef.options) {
      oNewMapDef.options = this.vitisClone.clone(oMapDef.options);
    }
    if (oMapDef.name) {
      oNewMapDef.name = this.vitisClone.clone(oMapDef.name);
    }
    if (oMapDef.maptools) {
      oNewMapDef.maptools = this.vitisClone.clone(oMapDef.maptools);
    }
    if (oMapDef.children) {
      oNewMapDef.children = this.cloneMapDefChildrens_(oMapDef.children);
    }

    return oNewMapDef;
  }

  /**
   * Retourne les mesures de la carte (sans le bind des couches)
   * @param olMap
   * @returns
   */
  public getMeasuresInfos(olMap: Map): any {
    const measure_features = olMap.get('measure_features');
    let oNewMeasuresInfos: any = {};

    if (measure_features && measure_features.length > 0) {
      oNewMeasuresInfos = this.vitisClone.clone(measure_features.map(infos => {
        const { popup, ...infosMeasure } = infos;
        return infosMeasure;
      }));
    }

    return oNewMeasuresInfos;
  }

  // /**
  //  * Get the ol tile grid from the map definition
  //  * @param {object} oMapDefinition
  //  * @param {object} opt_options
  //  * @param {array} opt_options.size
  //  * @param {array} opt_options.tileSize
  //  * @returns {TileGrid}
  //  * @export
  //  */
  // public getTileGridFromDef(oMapDefinition: any, opt_options: any): TileGrid {

  //   opt_options = this.vitisCheck.isDefAndNotNull(opt_options) ? opt_options : {};

  //   var olView = this.getViewFromDef(oMapDefinition, opt_options);

  //   if (!this.vitisCheck.isDefAndNotNull(olView)) {
  //     return null;
  //   }
  //   if (!this.vitisCheck.isDefAndNotNull(opt_options.tileSize)) {
  //     console.error('opt_options.tileSize not defined');
  //     return null;
  //   }
  //   if (!this.vitisCheck.isDefAndNotNull(opt_options.size)) {
  //     console.error('opt_options.size not defined');
  //     return null;
  //   }

  //   var extent = olView.calculateExtent(opt_options.size);

  //   var projExtent = olView.getProjection().getExtent();
  //   var startResolution = ol.extent.getWidth(projExtent) / 256;
  //   var resolutions = new Array(22);

  //   for (var i = 0, ii = resolutions.length; i < ii; ++i) {
  //     resolutions[i] = startResolution / Math.pow(2, i);
  //   }

  //   var tileGrid = new TileGrid({
  //     origin: ol.extent.getTopLeft(extent),
  //     extent: projExtent,
  //     resolutions: resolutions,
  //     tileSize: opt_options.tileSize
  //   });

  //   return tileGrid;
  // };

  /**
     * Ajoute le proxy à l'url
     * @param {string} url
     * @returns {String}
     */
  private parseProxyUrl_(url: string): string {

    let parsedProxyUrl = url;
    const proxyUrl = environment.properties.proxy_url;

    if (this.vitisCheck.isDefAndNotNull(proxyUrl)) {
      url = url.replace(/\?/g, '&');
      parsedProxyUrl = proxyUrl + '?url=' + url;
    }

    return parsedProxyUrl;
  }

  /**
   * Ajoute le proxy à l'url
   * @param {array<string>} aUrls
   * @returns {Array<string>}
   */
  private parseProxyUrls_(aUrls: Array<string>): Array<string> {

    if (this.vitisCheck.isArray(aUrls)) {
      for (let i = 0; i < aUrls.length; i++) {
        aUrls[i] = this.parseProxyUrl_(aUrls[i]);
      }
    }

    return aUrls;
  }

  /**
   * Retourne la défintion des couches depuis la définition de la carte
   * @param {object} oMapDefinition
   * @returns {Array}
   */
  private getLayersDef_(oMapDefinition: any): Array<any> {

    let aLayersDef = [];
    const aChildrens = oMapDefinition.children;

    // Ne prend pas le premier element (il s'agit de la vue)
    for (let i = 1; i < aChildrens.length; i++) {
      this.recursiveGetLayersDef_(aChildrens[i], aLayersDef);
    }

    // Ordonne les couches suivant leur index
    aLayersDef = this.sortLayersDef_(aLayersDef);

    return aLayersDef;
  }

  /**
   * Cherche les couches de l'arbre et remplit aLayersDef
   * @param oChildren
   * @param aLayersDef
   */
  private recursiveGetLayersDef_(oChildren: any, aLayersDef: any): Array<any> {
    if (oChildren.children) {
      for (let i = 0; i < oChildren.children.length; i++) {
        this.recursiveGetLayersDef_(oChildren.children[i], aLayersDef);
      }
    } else {
      aLayersDef.push(oChildren);
    }

    return aLayersDef;
  }

  /**
   * Ordonne les couches par index
   * @param {array} aLayersDef
   * @returns {array}
   */
  private sortLayersDef_(aLayersDef: any): Array<any> {
    function compare(a: any, b: any) {
      if (a.index < b.index) {
        return -1;
      }
      if (a.index > b.index) {
        return 1;
      }
      return 0;
    }
    aLayersDef.sort(compare);
    return aLayersDef;
  }

  /**
   * Retourne une source OpenLayers depuis une definition de couche
   * @param {object} oMapDefinition
   * @param {object} oLayerDef
   * @param {object} opt_options
   * @param {array} opt_options.size
   * @param {array} opt_options.tileSize
   * @returns {any}
   */
  private getSourceFromLayerDef_(oMapDefinition: any, oLayerDef: any, opt_options: any): any {

    opt_options = this.vitisCheck.isDefAndNotNull(opt_options) ? opt_options : {};

    if (!this.vitisCheck.isDefAndNotNull(opt_options.tileSize)) {
      console.error('opt_options.tileSize not defined');
      return null;
    }
    if (!this.vitisCheck.isDefAndNotNull(opt_options.size)) {
      console.error('opt_options.size not defined');
      return null;
    }

    let source: any;
    const type = oLayerDef.layerType;

    // Ajout d'un timestamp afin de maitriser les caches lors du rafraichissement
    // L'ajout de ce paramètre a également pour effet d'annuler les caches lors du chargement des couches
    if (this.vitisCheck.isDefAndNotNull(oLayerDef.params)) {
      if (oLayerDef.is_dynamic === true) {
        oLayerDef.params.TIMESTAMP = Date.now();
      }
    }

    // Options du service
    const oServiceOptions = this.getServiceOptions_(oLayerDef);

    // Options de la couche
    const oLayerOptions = this.getlayerOptions_(oLayerDef);

    // Styles des couches
    oLayerDef = this.setLayerDefStyle_(oLayerDef, oLayerOptions);

    // Authentification
    oLayerDef = this.setLayerDefAuth_(oLayerDef, oServiceOptions);

    // Bing
    if (type === 'bing') {
      source = this.getBingSource_(oLayerDef);
    } else if (type === 'tilewms' || type === 'imagewms') {
      source = this.getWMSSource_(oLayerDef, oMapDefinition, opt_options);
    } else if (type === 'imagevector') {
      source = this.getVectorSource_(oLayerDef);
    } else if (type === 'wmts') {
      source = this.getWMTSSource_(oLayerDef, oMapDefinition);
    } else if (type === 'xyz') {
      source = this.getXYZSource_(oLayerDef);
    } else if (type === 'osm') {
      source = this.getOSMSource_(oLayerDef);
    } else if (type === 'vectortiles') {
      // Pas besoin de définir de source car ol-mapbox-style va en créer plusieurs
    } else if (type === 'vector') {
      // Pas besoin de définir de source car elle est définie dans la fonction addLayerVector de MapCoreManager
    } else {
      console.error('Error: layerType (' + type + ') is not supported');
      source = null;
    }

    return source;
  }

  /**
   * Get the ol layer from the layer definition
   * @param {object} oMapDefinition
   * @param {object} oLayerDef
   * @param {object} oSource
   * @returns {ol.layer.Tile|ol.layer.Image}
   */
  private getLayerFromLayerDef_(oMapDefinition: any, oLayerDef: any, oSource: any): any {

    let layer: any;
    const type = oLayerDef.layerType;

    if (type === 'imagewms') {
      layer = this.getImageLayer_(oSource);
    } else if (type === 'imagevector') {
      layer = this.getVectorLayer_(oLayerDef, oSource);
    } else {
      layer = this.getTileLayer_(oSource);
    }

    // Set les paramètres additionnels pour le bon fonctionnement dans l'outil
    return this.setLayerParameters_(layer, oLayerDef);
  }

  /**
   * Set les paramètres additionnels pour le bon fonctionnement dans l'outil
   * @param olLayer Couche OpenLayers
   * @param oLayerDef
   * @returns
   */
  private setLayerParameters_(olLayer: any, oLayerDef: any): Layer<any> {

    const type = oLayerDef.layerType;

    // Type
    olLayer.set('type', type); // Type OL3
    olLayer.set('layerType', oLayerDef.type); // Type vMap

    olLayer.set('attribution', oLayerDef.source);

    // Name
    olLayer.set('name', oLayerDef.name);

    // Layer_id
    olLayer.set('layer_id', oLayerDef.layer_id);

    // layer_grid_id pour identifier la couche dans une grille
    olLayer.set('layer_grid_id', oLayerDef.layer_grid_id);

    // Legend
    olLayer.set('legend', true);

    // definit si la légende est selectionnable ou non
    // queryable: la couche peut être ou pas interrogeable
    // select: la couche est interrogeable par défaut
    olLayer.set('queryable', true);
    if ((type !== 'imagewms') && (type !== 'tilewms')) {
      olLayer.set('queryable', false);
    }

    // Select
    oLayerDef.select = true;
    olLayer.set('select', true);

    // Auth
    if (this.vitisCheck.isDefAndNotNull(oLayerDef.service_login)) {
      olLayer.set('service_login', oLayerDef.service_login);
    }
    if (this.vitisCheck.isDefAndNotNull(oLayerDef.service_password)) {
      olLayer.set('service_password', oLayerDef.service_password);
    }

    // Events
    if (this.vitisCheck.isDefAndNotNull(oLayerDef.events)) {
      olLayer.set('events', oLayerDef.events);
    }

    // Params
    if (this.vitisCheck.isDefAndNotNull(oLayerDef.params)) {
      if (this.vitisCheck.isDefAndNotNull(oLayerDef.params.LAYERS)) {
        const aLayers = oLayerDef.params.LAYERS.split(',');
        if (this.vitisCheck.isArray(aLayers)) {

          // sublayers
          olLayer.set('sublayers', aLayers);

          // activeSublayers
          olLayer.set('activeSublayers', aLayers);

          // queryableSublayers
          if (this.vitisCheck.isDefAndNotNull(oLayerDef.is_queryable_getfeatureinfo)) {
            olLayer.set('queryableSublayers', aLayers);
          }
        }
      }
    }

    // GetFeatureInfo
    if (this.vitisCheck.isDefAndNotNull(oLayerDef.is_queryable_getfeatureinfo)) {
      olLayer.set('is_queryable_getfeatureinfo', oLayerDef.is_queryable_getfeatureinfo);
      olLayer.set('wms_queryable', true);
    }

    // Dynamic
    if (this.vitisCheck.isDefAndNotNull(oLayerDef.is_dynamic)) {
      olLayer.set('is_dynamic', oLayerDef.is_dynamic);
    }

    // Dynamic Legend
    if (this.vitisCheck.isDefAndNotNull(oLayerDef.is_dynamic_legend)) {
      olLayer.set('is_dynamic_legend', oLayerDef.is_dynamic_legend);
    }

    // Filtre
    if (this.vitisCheck.isDefAndNotNull(oLayerDef.is_filtered)) {
      if (this.vitisCheck.isDefAndNotNull(oLayerDef.filter_form)) {
        if ((this.vitisCheck.isString(oLayerDef.filter_form) && oLayerDef.filter_form.length !== 0) ||
          this.vitisCheck.isObject(oLayerDef.filter_form)) {
          if (this.vitisCheck.isString(oLayerDef.filter_form)) {
            oLayerDef.filter_form = JSON.parse(oLayerDef.filter_form);
          }
          oLayerDef.filter_values = this.getFilterFormValues_(oLayerDef.filter_form);

          // filter_form
          olLayer.set('filter_form', oLayerDef.filter_form);

          // filter_form_embedjs
          olLayer.set('filter_form_embedjs', oLayerDef.filter_form_embedjs);

          // is_filtered
          olLayer.set('is_filtered', oLayerDef.is_filtered);

          // is_bo_filtered
          olLayer.set('is_bo_filtered', oLayerDef.is_bo_filtered);

          // filter_values
          olLayer.set('filter_values', oLayerDef.filter_values);

          // Fonction applyFilter
          const this_ = this;
          olLayer.applyFilter = function () {
            this_.setFilterOnLayer_(this);
          };

          // Applique le filtre
          olLayer.applyFilter();
        }
      }
    }

    // layer_options
    if (this.vitisCheck.isDefAndNotNull(oLayerDef.layer_options)) {
      olLayer.set('layer_options', oLayerDef.layer_options);
    }

    // Visibilité de la couche
    if (oLayerDef.is_map_hidden === 'true' || oLayerDef.is_map_hidden === true) {
      oLayerDef.visible = false;
    } else if (oLayerDef.visible === 'true' || oLayerDef.visible === true) {
      oLayerDef.visible = true;
    } else {
      oLayerDef.visible = false;
    }
    olLayer.setVisible(oLayerDef.visible);

    // Opacité de la couche
    if (this.vitisCheck.isDefAndNotNull(oLayerDef.opacity)) {
      olLayer.setOpacity(oLayerDef.opacity);
    }

    // Mémorise la couche dans la definition
    oLayerDef.olLayer = olLayer;

    // export à la compilation pour les values
    olLayer.values = olLayer.values_;

    // Set ordre (utile pour les vectortiles qui sont chargés en asynchrone)
    olLayer.setZIndex(oLayerDef.index);

    return olLayer;
  }

  /**
   * Change the layer source URL to apply the filter_values
   * @param {Layer} olLayer
   */
  private setFilterOnLayer_(olLayer: Layer<any>) {

    if (!this.vitisCheck.isDefAndNotNull(olLayer.get('filter_values'))) {
      return null;
    }
    if (!this.vitisCheck.isDefAndNotNull(olLayer.get('filter_values').search)) {
      return null;
    }

    const oValues = this.getBOValuesFromFormValues_(olLayer.get('filter_values').search);
    olLayer.set('filter_values_cleared', oValues);

    // if (typeof oVmap !== "undefined") { // Obligatoire lors d'une impression
    //   if (this.vitisCheck.isDefAndNotNull(oVmap)) {
    //     if (this.vitisCheck.isDefAndNotNull(oVmap['scope'])) {
    //       oVmap['scope'].$broadcast('layersChanged');
    //     }
    //   }
    // }

    if (
      olLayer.getSource() instanceof ImageWMS ||
      olLayer.getSource() instanceof TileWMS
    ) {
      const params = olLayer.getSource().getParams();

      for (const key in oValues) {
        if (oValues.hasOwnProperty(key)) {
          if (this.vitisCheck.isArray(oValues[key])) {
            // en cas de tableau, affiche les valeurs séparées par des virgules
            params[key] = oValues[key].join('|');
          } else {
            params[key] = oValues[key];
          }

          if (this.vitisCheck.isDefAndNotNull(params[key])) {
            if (this.vitisCheck.isArray(params[key]) && this.vitisCheck.isObject(params[key])) {
              delete params[key];
              continue;
            }
            if (this.vitisCheck.isString(params[key])) {
              if (params[key] === '') {
                delete params[key];
                continue;
              }
            }
          }
        }
      }

      olLayer.getSource().updateParams(params);
      olLayer.getSource().refresh();
    }

  }

  /**
   * Get the values from oFormValues
   * @param {any} oFormValues
   * @returns {any}
   */
  private getBOValuesFromFormValues_(oFormValues: any): any {

    let oFormValuesValues: any;
    try {
      oFormValuesValues = { ...oFormValues };
    } catch (e) {
      return {};
    }

    for (const key in oFormValuesValues) {
      if (this.vitisCheck.isObject(oFormValuesValues[key])) {
        if (this.vitisCheck.isDefAndNotNull(oFormValuesValues[key].selectedOption)) {
          if (this.vitisCheck.isDefAndNotNull(oFormValuesValues[key].selectedOption.value)) {
            oFormValuesValues[key] = oFormValuesValues[key].selectedOption.value;
          } else
            // Cas de liste à choix multiples
            if (this.vitisCheck.isArray(oFormValuesValues[key].selectedOption)) {
              const aValues = [];
              for (let i = 0; i < oFormValuesValues[key].selectedOption.length; i++) {
                if (this.vitisCheck.isDefAndNotNull(oFormValuesValues[key].selectedOption[i].value)) {
                  aValues.push(oFormValuesValues[key].selectedOption[i].value);
                }
              }
              oFormValuesValues[key] = [...aValues];
            } else {
              delete oFormValuesValues[key];
            }
        }
      } else if (!this.vitisCheck.isDefAndNotNull(oFormValuesValues[key])) {
        delete oFormValuesValues[key];
      }
    }

    return oFormValuesValues;
  }

  /**
   * Retourne les options du service depuis la defintion de la couche
   * @param  {any} oLayerDef
   * @returns any
   */
  private getServiceOptions_(oLayerDef: any): any {
    let oServiceOptions = {};
    if (this.vitisCheck.isString(oLayerDef.service_options)) {
      try {
        oServiceOptions = JSON.parse(oLayerDef.service_options);
      } catch (e) {
        oServiceOptions = {};
      }
    } else if (this.vitisCheck.isObject(oLayerDef.service_options)) {
      oServiceOptions = oLayerDef.service_options;
    }
    return oServiceOptions;
  }

  /**
   * Retourne les options de la couche depuis la defintion de la couche
   * @param  {any} oLayerDef
   * @returns any
   */
  private getlayerOptions_(oLayerDef: any): any {
    let oLayerOptions = {};
    if (this.vitisCheck.isString(oLayerDef.layer_options)) {
      try {
        oLayerOptions = JSON.parse(oLayerDef.layer_options);
      } catch (e) {
        oLayerOptions = {};
      }
    } else if (this.vitisCheck.isObject(oLayerDef.layer_options)) {
      oLayerOptions = oLayerDef.layer_options;
    }
    return oLayerOptions;
  }

  /**
   * Set le style définit dans les options dans la définition de la couche
   * @param  {any} oLayerDef
   * @param  {any} oLayerOptions
   * @returns any
   */
  private setLayerDefStyle_(oLayerDef: any, oLayerOptions: any): any {

    const type = oLayerDef.layerType;

    if (type === 'tilewms' || type === 'imagewms') {

      // Styles des couches
      if (this.vitisCheck.isDefAndNotNull(oLayerOptions)) {
        if (this.vitisCheck.isDefAndNotNull(oLayerOptions.layer_style)) {
          if (this.vitisCheck.isDefAndNotNull(oLayerDef.params.LAYERS)) {
            const aStyles = [];
            const aLayers = oLayerDef.params.LAYERS.split(',');
            for (let i = 0; i < aLayers.length; i++) {
              if (this.vitisCheck.isDefAndNotNull(oLayerOptions.layer_style[aLayers[i]])) {
                aStyles.push(oLayerOptions.layer_style[aLayers[i]].Name);
              } else {
                aStyles.push('');
              }
            }
            oLayerDef.params.STYLES = aStyles.join(',');
          }
        }
      }
      if (this.vitisCheck.isDefAndNotNull(oLayerDef.params) &&
        !this.vitisCheck.isDefAndNotNull(oLayerDef.params.STYLES) &&
        !this.vitisCheck.isDefAndNotNull(oLayerDef.params.styles)) {
        oLayerDef.params.STYLES = '';
      }
      if (!this.vitisCheck.isDefAndNotNull(oLayerDef.tileLoadFunction)) {
        oLayerDef.tileLoadFunction = function (imageTile: any, src: any) {
          src = src.replace('STYLES&', 'STYLES=&');
          imageTile.getImage().src = src;
        };
      }
    }

    return oLayerDef;
  }

  /**
   * Set le login/password définit dans les options dans la définition de la couche
   * @param  {any} oLayerDef
   * @param  {any} oServiceOptions
   * @returns any
   */
  private setLayerDefAuth_(oLayerDef: any, oServiceOptions: any): any {

    const type = oLayerDef.layerType;

    if (type === 'tilewms' || type === 'imagewms') {

      // Retrocompatibilité: authentification des couches
      if (this.vitisCheck.isDefAndNotNull(oServiceOptions)) {
        if (this.vitisCheck.isDefAndNotNull(oServiceOptions.login)) {
          oLayerDef.service_login = oServiceOptions.login;
        }
        if (this.vitisCheck.isDefAndNotNull(oServiceOptions.password)) {
          oLayerDef.service_password = oServiceOptions.password;
        }
      }
    }


    if (this.vitisCheck.isDefAndNotNull(oLayerDef.service_login)
      && this.vitisCheck.isDefAndNotNull(oLayerDef.service_password)) {

      // Si login et password sont renseignés, alors effectue une requête "fantome" pour logguer l'ip client
      const newLoadFunction = function (image, src) {

        // Styles pour le WMS
        src = src.replace('STYLES&', 'STYLES=&');

        // Authentification
        const xhr = new XMLHttpRequest();
        xhr.responseType = 'blob';
        xhr.onreadystatechange = function () {
          if (xhr.readyState === XMLHttpRequest.DONE && xhr.status === 200) {
            image.getImage().src = URL.createObjectURL(xhr.response);
          }
        };
        xhr.open('GET', src, true);
        xhr.setRequestHeader('Authorization', 'Basic ' + btoa(oLayerDef.service_login + ':' + oLayerDef.service_password));
        xhr.send();
      };
      oLayerDef.tileLoadFunction = newLoadFunction;
      oLayerDef.imageLoadFunction = newLoadFunction;
    }

    return oLayerDef;
  }

  /**
   * Retourne une source BingMaps OpenLayers depuis une definition de couche
   * @param  {any} oLayerDef
   * @returns BingMaps
   */
  private getBingSource_(oLayerDef: any): BingMaps {
    const source = new BingMaps({
      key: oLayerDef.key,
      wrapX: oLayerDef.wrapX,
      culture: oLayerDef.culture,
      maxZoom: oLayerDef.maxZoom,
      tileLoadFunction: oLayerDef.tileLoadFunction,
      imagerySet: oLayerDef.imagerySet
    });
    return source;
  }

  /**
   * Retourne une source TileWMS ou ImageWMS OpenLayers depuis une definition de couche
   * @param  {any} oLayerDef
   * @returns any
   */
  private getWMSSource_(oLayerDef: any, oMapDefinition: any, opt_options: any): any {

    let source: any;
    const type = oLayerDef.layerType;

    if (type === 'tilewms') {
      source = new TileWMS({
        url: (environment.properties.use_proxy_for_tiles) ? this.parseProxyUrl_(oLayerDef.url) : oLayerDef.url,
        urls: (environment.properties.use_proxy_for_tiles) ? this.parseProxyUrls_(oLayerDef.urls) : oLayerDef.urls,
        hidpi: oLayerDef.hidpi,
        wrapX: oLayerDef.wrapX,
        gutter: oLayerDef.gutter,
        // tileGrid: this.getTileGridFromDef(oMapDefinition, opt_options),
        projection: oLayerDef.projection,
        serverType: oLayerDef.serverType,
        crossOrigin: oLayerDef.crossOrigin,
        tileLoadFunction: oLayerDef.tileLoadFunction,
        params: oLayerDef.params
      });
    }

    if (type === 'imagewms') {
      source = new ImageWMS({
        url: (environment.properties.use_proxy_for_tiles) ? this.parseProxyUrl_(oLayerDef.url) : oLayerDef.url,
        hidpi: oLayerDef.hidpi,
        ratio: oLayerDef.ratio,
        projection: oLayerDef.projection,
        serverType: oLayerDef.serverType,
        resolutions: oLayerDef.resolutions,
        crossOrigin: oLayerDef.crossOrigin,
        imageLoadFunction: oLayerDef.imageLoadFunction,
        params: oLayerDef.params
      });
    }

    return source;
  }

  /**
   * Retourne une source Vector OpenLayers depuis une definition de couche
   * @param  {any} oLayerDef
   * @returns VectorSource
   */
  private getVectorSource_(oLayerDef: any): VectorSource<Geometry> {

    let source: any;
    if (oLayerDef.features) {
      source = new VectorSource({
        features: oLayerDef.features
      });
    } else if (oLayerDef.url) {
      source = new VectorSource({
        format: new GeoJSON(),
        url: (environment.properties.use_proxy_for_tiles) ? this.parseProxyUrl_(oLayerDef.url) : oLayerDef.url
      });
    } else {
      console.error('Error: veuillez renseinger une url ou une feature');
    }
    return source;
  }

  /**
   * Retourne une source WMTS OpenLayers depuis une definition de couche
   * @param  {any} oLayerDef
   * @returns WMTS
   */
  private getWMTSSource_(oLayerDef: any, oMapDefinition: any): WMTS {

    let oOptions: any;
    let source: any;

    if (!this.vitisCheck.isDefAndNotNull(oLayerDef.service_options)) {
      console.error('oLayerDef.service_options undefined');
      return null;
    }

    try {
      const oCapabilities = JSON.parse(oLayerDef.service_options);
      oOptions = optionsFromCapabilities(oCapabilities, {
        layer: oLayerDef.layer,
        matrixSet: oLayerDef.matrixSet,
        requestEncoding: oLayerDef.requestEncoding,
        style: oLayerDef.style,
        format: oLayerDef.format
      });

      // Des fois optionsFromCapabilities est ronchon et ne donne pas le bon style
      if (this.vitisCheck.isDefAndNotNull(oLayerDef.style)) {
        oOptions.style = oLayerDef.style;
      }
      if (this.vitisCheck.isDefAndNotNull(oLayerDef.tileLoadFunction)) {
        oOptions.tileLoadFunction = oLayerDef.tileLoadFunction;
      }

      // Génération auto de la tileGrid car celle parsée depuis les capabilities n'est souvent pas bonne
      try {
        const resolutions = [];
        const matrixIds = [];
        const mapProj = getProjection(oMapDefinition.children[0].view.projection);
        const maxResolution = getWidth(mapProj.getExtent()) / 256;

        for (let i = 0; i < 20; i++) {
          matrixIds[i] = i.toString();
          resolutions[i] = maxResolution / Math.pow(2, i);
        }

        const tileGrid = new WMTSTileGrid({
          origin: [-20037508, 20037508],
          resolutions: resolutions,
          matrixIds: matrixIds,
        });

        oOptions.tileGrid = tileGrid;
      } catch (error) { }

      source = new WMTS(oOptions);
    } catch (e) {
      console.error('error while parsing WMTS options');
    }

    return source;
  }

  /**
   * Retourne une source XYZ OpenLayers depuis une definition de couche
   * @param  {any} oLayerDef
   * @returns XYZ
   */
  private getXYZSource_(oLayerDef: any): XYZ {
    const source = new XYZ({
      url: oLayerDef.url,
      tileLoadFunction: oLayerDef.tileLoadFunction
    });
    return source;
  }

  /**
   * Retourne une source OSM OpenLayers depuis une definition de couche
   * @param  {any} oLayerDef
   * @returns OSM
   */
  private getOSMSource_(oLayerDef: any): OSM {

    if (!this.vitisCheck.isDefAndNotNull(oLayerDef.url) || oLayerDef.url === '') {
      oLayerDef.url = 'https://{a-c}.tile.openstreetmap.org/{z}/{x}/{y}.png';
    }

    const source = new OSM({
      url: oLayerDef.url
    });
    return source;
  }

  /**
   * Retourne une couche OpenLayers depuis une source
   * @param  {any} oLayerDef
   * @returns OSM
   */
  private getImageLayer_(oSource: any): ImageLayer<ImageSource> {
    const layer = new ImageLayer({
      source: oSource
    });
    return layer;
  }

  /**
   * Retourne une couche OpenLayers depuis une definition de couche
   * @param  {any} oLayerDef
   * @param  {any} oSource
   * @returns OSM
   */
  private getVectorLayer_(oLayerDef: any, oSource: any): VectorLayer<VectorSource<Geometry>> {
    const layer = new VectorLayer({
      // renderMode: 'image',
      style: oLayerDef.style,
      source: oSource,
      zIndex: 1000,
    });
    return layer;
  }

  /**
   * Retourne une couche OpenLayers depuis une source
   * @param  {any} oLayerDef
   * @returns OSM
   */
  private getTileLayer_(oSource: any): TileLayer<TileSource> {
    const layer = new TileLayer({
      source: oSource
    });
    return layer;
  }

  /**
   * Retourne les valeurs du formulaire de filtre
   * @param {object} filterForm
   * @returns {object} filter form values
   */
  private getFilterFormValues_(filterForm: any): any {

    const filterValues = {};
    for (const key in filterForm) {
      if (key !== 'datasources') {
        filterValues[key] = {};

        for (let i = 0; i < filterForm[key].rows.length; i++) {
          for (let ii = 0; ii < filterForm[key].rows[i].fields.length; ii++) {
            const name = filterForm[key].rows[i].fields[ii].name;
            const defaultValue = filterForm[key].rows[i].fields[ii].default_value;

            // Ajoute la valeur si elle n'existe pas déjà ou qu'elle est vide
            if (!this.vitisCheck.isDefAndNotNull(filterValues[key][name])) {
              filterValues[key][name] = '';
              if (this.vitisCheck.isDefAndNotNull(defaultValue)) {
                filterValues[key][name] = defaultValue;
              }
            } else {
              if (this.vitisCheck.isDefAndNotNull(defaultValue) && filterValues[key][name] === '') {
                filterValues[key][name] = defaultValue;
              }
            }
          }
        }
      }
    }
    return filterValues;
  }

  /**
   * Clone les childrens d'une def
   * @param aChildrens
   * @returns
   */
  private cloneMapDefChildrens_(aChildrens: Array<any>) {
    const aReturn = [];

    for (let i = 0; i < aChildrens.length; i++) {
      const oNewChildren: any = {};
      const oChildren = aChildrens[i];
      if (oChildren.children) {
        oNewChildren.children = this.cloneMapDefChildrens_(oChildren.children);
      }

      for (const key in oChildren) {
        if (key !== 'olLayer') {
          if (Object.prototype.hasOwnProperty.call(oChildren, key)) {
            try {
              oNewChildren[key] = JSON.parse(JSON.stringify(oChildren[key]));
            } catch (error) { }
          }
        } else {
          oNewChildren.visible = oChildren.olLayer.get('visible') === true ? true : false;
        }
      }

      aReturn.push(oNewChildren);
    }

    return aReturn;
  }
}