import { Component, OnInit, Input, forwardRef, ViewChild, AfterViewInit, Injectable, Output, EventEmitter } from '@angular/core';
import { NgbDateStruct, NgbPopoverConfig, NgbPopover, NgbDatepickerI18n } from '@ng-bootstrap/ng-bootstrap';
import { NG_VALUE_ACCESSOR } from '@angular/forms';
import { DatePipe } from '@angular/common';
import { TranslateService } from '@ngx-translate/core';
import moment from 'moment-timezone';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';
import { AuthService } from 'src/app/vitis/core/services/auth/auth.service';

@Injectable()
export class CustomDatepickerI18n extends NgbDatepickerI18n {

  constructor(private translate: TranslateService) {
    super();
  }

  getWeekdayShortName(weekday: number): string {
    let value = '';
    this.translate.get('DATE_TIME_PICKER_WEEKDAYS').subscribe((res: string) => {
      value = res[weekday - 1];
    });
    return value;
  }

  getMonthShortName(month: number): string {
    let value = '';
    this.translate.get('DATE_TIME_PICKER_MONTHS').subscribe((res: string) => {
      value = res[month - 1];
    });
    return value;
  }

  getMonthFullName(month: number): string {
    return this.getMonthShortName(month);
  }

  getDayAriaLabel(date: NgbDateStruct): string {
    return `${date.day}/${date.month}/${date.year}`;
  }
}

@Component({
  selector: 'app-datetimepicker',
  templateUrl: './datetimepicker.component.html',
  styleUrls: ['./datetimepicker.component.scss'],
  providers: [
    DatePipe,
    {
      provide: NG_VALUE_ACCESSOR,
      useExisting: forwardRef(() => DatetimepickerComponent),
      multi: true
    },
    { provide: NgbDatepickerI18n, useClass: CustomDatepickerI18n }
  ]
})
export class DatetimepickerComponent implements OnInit, AfterViewInit {

  @Input() oField: any;
  @Input() oFormValues?: any;
  @Input() showInputField: any = true;
  @Input() showDateTimePicker: any = false;
  @Input() showCancelBtn: any = true;

  @Output() submit = new EventEmitter<number | string>();
  @Output() valueChange = new EventEmitter<number | string>();

  // affiche / cache la selection du temps
  public showTimePickerToggle = false;

  // Format d'affichage de la date
  public sDateFormat: string;

  // Date affichée
  public sDisplayDate: string;

  // Date selectionnée
  public oSelectedDate: any = {};

  // Heure selectionnée
  public oSelectedTime: any = {
    hour: 0,
    minute: 0,
    second: 0
  };

  // Timezone de l'utilisateur
  private sUserTimezone_: string;

  // Composant qui contiendra la selection de l'heure
  @ViewChild(NgbPopover, { static: false }) popover: NgbPopover;

  constructor(
    private config: NgbPopoverConfig,
    private vitisCheck: VitisCheck,
    private authService: AuthService
  ) {
    config.autoClose = 'outside';
    config.placement = 'auto';
  }

  ngOnInit(): void {

    // Format d'affichage de la date
    this.sDateFormat = this.getDateFormat();

    // Fonction getValue
    this.oField.getValue = () => {
      return this.getValue();
    };

    // Fonction setValue
    this.oField.setValue = (oValue: any) => {
      return this.setValue(oValue);
    };

    // Fonction getDisplayedValue
    this.oField.getDisplayedValue = () => {
      return this.getDisplayedValue();
    };

    // Timezone de l'utilisateur
    const oUser = this.authService.getUserInfos();
    this.sUserTimezone_ = oUser.timezone_id;

    if (this.sUserTimezone_) {
      moment.tz.setDefault(this.sUserTimezone_);
    }
  }

  ngAfterViewInit(): void {

    // Si la min_date vaut 'now' on selectionne la date du jour
    if (this.vitisCheck.isDefAndNotNull(this.oField) && this.oField.min_date == 'now') {
      this.oField.min_date = {
        year: new Date().getFullYear(),
        month: new Date().getMonth() + 1,
        day: new Date().getDate()
      };
    }

    // Si la max_date vaut 'now' on selectionne la date du jour
    if (this.vitisCheck.isDefAndNotNull(this.oField) && this.oField.max_date == 'now') {
      this.oField.max_date = {
        year: new Date().getFullYear(),
        month: new Date().getMonth() + 1,
        day: new Date().getDate()
      };
    }

    // Ferme la selection des heures lorsqu'on ferme le composant
    if (this.showInputField === false) {
      this.popover.hidden.subscribe(() => {
        this.showTimePickerToggle = false;
      });
    }

    this.popover.shown.subscribe(() => {
      if (this.vitisCheck.isEmptyObject(this.oSelectedDate)) {
        this.setValue(moment().unix());
      }
    });
  }

  /**
   * Ferme la popover
   */
  public closePopover(bEmptyVal = false) {
    if (bEmptyVal) {
      this.setValue('');
    } else {
      this.submit.emit(this.getValue());
    }
    if (this.showInputField !== false) {
      this.popover.close();
    } else {
      this.showDateTimePicker = false;
    }
  }

  /**
   * Retourne le format de date à afficher
   * @returns string
   */
  public getDateFormat(): string {

    // Valeur par défaut
    let sDateFormat = 'DD/MM/YYYY';

    // Valeur de l'utilisateur en cours
    const oUser = this.authService.getUserInfos();
    if (this.vitisCheck.isDefAndNotNull(oUser)) {
      if (this.vitisCheck.isDefAndNotNull(oUser.formatdate_id)) {
        sDateFormat = oUser.formatdate_id;
      }
    }

    if (this.oField.type === 'datetime') {
      sDateFormat += ' HH:mm';
    }

    return sDateFormat;
  }

  /**
   * Met à jour this.sDisplayDate en fonction de oSelectedDate et oSelectedTime
   */
  public refreshValue() {

    const date = this.getMomentDate();

    // Date affichée dans le composant
    if (this.vitisCheck.isDefAndNotNull(date)) {
      this.sDisplayDate = date.format(this.sDateFormat);
      // Supprime l'espace entre le "T" et l'heure (ISO 8601).
      this.sDisplayDate = String(this.sDisplayDate).replace('T ', 'T');
    } else {
      this.sDisplayDate = '';
    }

    // Date utilisée dans le formulaire
    this.setFormValue(this.getValue());
  }

  /**
   * Retourne la date en cours au format moment en fonction de oSelectedDate et oSelectedTime
   * @returns any
   */
  public getMomentDate(): any {

    // Date au format momentjs
    let date: any;

    if (this.vitisCheck.isDefAndNotNull(this.oSelectedDate)) {

      const year = this.oSelectedDate.year;
      // Patch car momentjs utilise 0 our le mois de janvier
      const month = this.oSelectedDate.month - 1;
      const day = this.oSelectedDate.day;

      if (this.vitisCheck.isDefAndNotNull(year) &&
        this.vitisCheck.isDefAndNotNull(month) &&
        this.vitisCheck.isDefAndNotNull(day)) {

        date = moment().year(year).month(month).date(day);

        if (this.vitisCheck.isDefAndNotNull(this.oSelectedTime)) {

          const hour = this.oSelectedTime.hour;
          const minute = this.oSelectedTime.minute;

          date = date.hour(hour).minute(minute).second(0);

        }
      }
    }

    return date;
  }

  /**
   * Permet le switch entre les calendriers date et temps.
   */
  public toggleDateTimeState() {
    this.showTimePickerToggle = !this.showTimePickerToggle;
  }

  /**
   * Getter
   * @returns String
   */
  public getValue(): number | string {
    const date = this.getMomentDate();
    if (this.vitisCheck.isDefAndNotNull(date)) {
      return date.unix();
    }
    return '';
  }

  /**
   * Setter
   * @param  {number} sTimestamp
   */
  public setValue(sTimestamp: any) {

    this.oSelectedDate = {};
    if (this.vitisCheck.isDefAndNotNull(sTimestamp) &&
      sTimestamp !== '') {

      // Transforme la chaine de caractère en entier car moment() n'accepte pas une chaine de caractère
      if (this.vitisCheck.isString(sTimestamp)) {
        sTimestamp = parseInt(sTimestamp, 10);
      }

      const date = moment.unix(sTimestamp);
      const year = date.year();
      let month = date.month();
      const day = date.date();
      let hour: number = 0;
      let minute: number = 0;
      // Si le champ est de type "datetime" -> récupération de l'heure.
      if (this.oField.type == 'datetime') {
        hour = date.hour();
        minute = date.minute();
      }

      if (this.vitisCheck.isNumber(year) &&
        this.vitisCheck.isNumber(month) &&
        this.vitisCheck.isNumber(day)) {
        month += 1;

        this.oSelectedDate = {
          year,
          month,
          day
        };

        if (this.vitisCheck.isNumber(hour) &&
          this.vitisCheck.isNumber(minute)) {

          this.oSelectedTime = {
            hour,
            minute
          };
        }
      }
    }

    setTimeout(() => {
      this.refreshValue();
    });
  }

  /**
   * Retourne la date affichée.
   * @returns String
   */
  public getDisplayedValue(): String {
    return this.sDisplayDate;
  }

  /**
   * Set la valeur dans this.oFormValues
   * @param  {number} iValue
   */
  public setFormValue(iValue: any) {

    if (this.vitisCheck.isDefAndNotNull(this.oFormValues)) {
      this.oFormValues.controls[this.oField.name].setValue(iValue);
      this.oFormValues.controls[this.oField.name].markAsTouched();
    }

    // Event valueChange
    this.valueChange.emit(iValue);
  }

  // Événements

  /**
   * Événement touched (obligatoire pour le fonctionnement de datepicker)
   */
  public onTouched() { }

  /**
   * Événement change (obligatoire pour le fonctionnement de datepicker)
   * @param  {any} oVal
   */
  public onChange(oVal: any) { }

  /**
   * Événement onDateChange (obligatoire pour le fonctionnement de datepicker)
   * @param  {any} oVal
   */
  public onDateChange(oVal: any) {
    if (this.vitisCheck.isDefAndNotNull(oVal)) {
      this.oSelectedDate = oVal;
      this.refreshValue();
    }
  }

  /**
   * Événement onTimeChange (obligatoire pour le fonctionnement de datepicker)
   * @param  {any} oVal
   */
  public onTimeChange(oVal: any) {
    this.oSelectedTime = oVal;
    this.refreshValue();
  }

  /**
   * Événement lors du changement de la date depuis le champ texte
   * @param  {any} $event
   */
  public onInputChange($event: any) {

    const value = $event.target.value;
    const date = moment(value, this.sDateFormat);
    const sTimestamp = date.unix();

    this.setValue(sTimestamp);
  }

  /**
   * Événement blur (obligatoire pour le fonctionnement de datepicker)
   */
  public inputBlur(event: any) { }
}