import { OnInit, Component, Input, EventEmitter, Output } from '@angular/core';
import { ActivatedRoute } from '@angular/router';
import { Subject } from 'rxjs';
import { VitisCheck, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { environment } from 'src/environments/environment';

@Component({
  selector: 'app-tabs-filter',
  templateUrl: './tabs-filter.component.html',
  styleUrls: ['./tabs-filter.component.scss']
})
export class TabsFilterComponent implements OnInit {

  public mode: any;

  @Input() aSections: Array<any>;
  @Input() oTabSubject: Subject<any>;
  @Input() sSelectLabel: String;
  @Input() iDefaultTab: number;
  @Input() applyFilterEvent: EventEmitter<any>;

  @Output() tabChanged = new EventEmitter<number>();

  public tabs: Array<any> = [];

  public order_index: number;
  public order_by: string;
  public sort_order: string;
  public operator: any;

  private iActivetedTab: number;

  constructor(
    private route: ActivatedRoute,
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest
  ) { }

  async ngOnInit() {

    // récupération du mode
    this.route.data.subscribe(mode => {
      this.mode = mode;
    });

    for (const section of this.aSections) {
      switch (section.type) {
        case 'custom':
          this.tabs.push({
            active: false,
            section: section
          });
          break;

        case 'ressource':
          await this.fetchRessourceSections(section);
          break;

        default:
          break;
      }
    }

    // get all active tabs
    const activeTabs = this.tabs.filter((tab) => tab.active);

    // if there is no active tab set, activate the first
    if (activeTabs.length === 0) {
      this.selectTab(this.iDefaultTab ? this.iDefaultTab : 0);
    }

    // Changement de filtre
    this.applyFilterEvent.subscribe((oFilter: any) => {

      // Si on est sur Nos favorits, passe sur l'onglet "tout"
      if (this.iActivetedTab === 0 &&
        oFilter && oFilter.nodes && oFilter.nodes.length > 0) {
        this.activateTab_(1);
      }
    });
  }

  /**
   * Émet un event sur oTabSubject
   */
  public emitData() {
    this.oTabSubject.next({
      order_by: this.order_by,
      sort_order: this.sort_order,
      operator: this.operator
    });
  }

  /**
   * Sélectionne un onglet
   * @param tabIndex
   */
  public selectTab(tabIndex: number) {

    // Active l'onglet
    this.activateTab_(tabIndex);

    const section = this.tabs[tabIndex].section;
    const node = this.getNodeFromSection(section);

    this.order_by = section.order_by;
    this.sort_order = section.sort_order;

    if (section.field && section.value) {
      this.operator = this.getOperatorFromNode(node);
    } else {
      this.operator = null;
    }

    const options = this.mode.options.sort_options;

    if (options) {
      for (let i = 0; i < options.length; i++) {
        if (this.order_by == options[i].order_by && this.sort_order == options[i].sort_order) {
          this.order_index = i;
        }
      }
    }

    this.emitData();
    this.tabChanged.next(tabIndex);
    this.iActivetedTab = tabIndex;
  }

  /**
   * Permet de trier
   * @param index
   */
  public sortBy(index) {
    this.order_index = index;

    this.order_by = this.mode.options.sort_options[index].order_by;
    this.sort_order = this.mode.options.sort_options[index].sort_order;

    this.emitData();
  }

  /**
   * Recherche les sections de la ressource
   * @param section
   */
  public fetchRessourceSections(section): Promise<any> {
    return new Promise((resolve, reject) => {
      const oParams = {
        order_by: section.order_by ?? section.id_field,
        sort_order: section.sort_order ?? 'ASC'
      };

      this.vitisRequest.ajaxRequest({
        method: 'GET',
        url: environment.properties.app_api_url + '/' + section.ressource,
        params: oParams,
        success: (resp: any) => {
          const data = resp.data.data;

          for (const element of data) {
            this.tabs.push({
              active: false,
              section: {
                type: 'ressource',
                title: element.title,
                translateTitle: element.title,
                field: section.id_field,
                compareOperator: '~',
                value: element[section.id_field],
                order_by: oParams.order_by,
                sort_order: oParams.sort_order
              }
            });
          }

          resolve(true);
        }
      });
    });
  }

  /**
   * Retourne l'opérateur du noeud passé en paramètre
   * @param oNode
   * @returns
   */
  public getOperatorFromNode(oNode) {
    const oNodeFilter: any = {};

    if (this.vitisCheck.isDefAndNotNull(oNode.column.field) &&
      this.vitisCheck.isDefAndNotNull(oNode.compareOperator) &&
      this.vitisCheck.isDefAndNotNull(oNode.value)) {
      oNodeFilter.column = oNode.column.field;
      oNodeFilter.compare_operator = oNode.compareOperator;
      oNodeFilter.value = oNode.value;

      // Type de filtre
      oNodeFilter.type = 'tab';

      if (oNode.compareOperator === '~') {
        oNodeFilter.compare_operator = 'LIKE';
        oNodeFilter.compare_operator_options = {
          'case_insensitive': true
        };
        oNodeFilter.value = '%' + oNodeFilter.value + '%';

      }
      // modification de la colonne à utiliser pour filtrer
      if (this.vitisCheck.isDefAndNotNull(oNode.column.searchColumn)) {
        oNodeFilter.column = oNode.column.searchColumn;
      }
    }

    return oNodeFilter;
  }

  /**
   * Retourne le noeud de la section passé en paramètre
   * @param oSection
   * @returns
   */
  public getNodeFromSection(oSection) {
    const oNode = {
      column: {
        title: oSection.title,
        translateTitle: oSection.translateTitle,
        field: oSection.field,
        searchable: false,
        searchType: 'string'
      },
      compareOperator: oSection.compareOperator,
      displayValue: oSection.value,
      value: oSection.value
    };

    return oNode;
  }

  /**
   * Active un onglet dans l'interface
   * @param iTabIndex
   */
  private activateTab_(iTabIndex: number) {

    // deactivate all tabs
    this.tabs.forEach(tab => tab.active = false);

    // activate the tab the user has clicked on.
    this.tabs[iTabIndex].active = true;
  }
}
