import { Component, OnInit, OnDestroy, ViewChild, Input, Injector } from '@angular/core';
import { ActivatedRoute, Router } from '@angular/router';
import { TranslateService } from '@ngx-translate/core';
import { VitisCheck, VitisRequest, vitisAlert, LocalStorage } from 'src/app/vitis/core/services/utils/vitis-utils';
import { AuthService } from '../core/services/auth/auth.service';
import { FormreaderComponent } from '../formreader/formreader.component';
import { environment } from 'src/environments/environment';
import { FormreaderService } from '../formreader/services/formreader.service';
import { CanComponentDeactivate } from 'src/app/vitis/core/services/deactivate/can-deactivate.guard';
import { Observable } from 'rxjs';
import { InterfaceService } from '../core/services/utils/interface/interface.service';
import { SessionStorageService } from '../core/services/utils/sessionstorage/session-storage.service';
import { LocalStorageService } from '../core/services/utils/localstorage/local-storage.service';

@Component({
  selector: 'app-simpleform',
  templateUrl: './simpleform.component.html',
  styleUrls: ['./simpleform.component.scss']
})
export class SimpleformComponent implements OnInit, OnDestroy, CanComponentDeactivate {

  @ViewChild(FormreaderComponent, { static: false }) oFormReader: FormreaderComponent;
  public oMode: any;
  public oParams: any;
  public sFormName: string = 'update';
  public insertedId: any;
  private sStarificationChar_ = '****';

  // Permet de passer oInputMode en argument du template
  @Input() set oInputMode(oMode: any) {
    if (oMode.options) {
      this.initForm(oMode);
    }
  }

  constructor(
    private route: ActivatedRoute,
    private router: Router,
    private vitisRequest: VitisRequest,
    private vitisCheck: VitisCheck,
    private auth: AuthService,
    private translate: TranslateService,
    private formreaderSrvc: FormreaderService,
    private interfaceService: InterfaceService,
    private sessionStorageService: SessionStorageService,
    private localStorageService: LocalStorageService,
    private localStorage: LocalStorage,
    private injector: Injector,
  ) { }

  /**
   * Initialisation du composant
   */
  ngOnInit() {
    this.route.data.subscribe(oMode => {

      if (oMode.type === 'sectionSimpleform' && oMode.options.default_form) {
        const sFirstsectionRoute = '../' + oMode.sections[0].route.replace(`:form`, oMode.options.default_form);
        this.router.navigate([sFirstsectionRoute], { relativeTo: this.route.parent});
        return;
      }
      // Le timeout permet d'attendre que this.oParams soit rempli
      setTimeout(() => {

        // Initialise le formulaire
        this.initForm(oMode);
      });
    });
    this.route.params.subscribe(oParams => {
      this.oParams = oParams;
    });
  }

  /**
   * Destruction du composant
   */
  ngOnDestroy() {
    if ((<any>window).constructor_form) {
      (<any>window).constructor_form = () => { };
    }
  }

  /**
   * Vérifie si le formulaire a été modifié
   */
  canDeactivate(): Observable<boolean> | Promise<boolean> | boolean {
    const bIsModified = this.oFormReader.oFormValues && this.oFormReader.oFormValues.dirty && !this.oFormReader.getIsSubmittedForm();

    if (bIsModified) {
      return new Promise((resolve) => {
        this.translate.get(['FORM_MODIFIED_WARNING_TITLE',
          'FORM_MODIFIED_WARNING',
          'CONFIRM_CANCEL',
          'CONFIRM_CONFIRM']
        ).subscribe((translations) => {
          vitisAlert.alert({
            title: translations['FORM_MODIFIED_WARNING_TITLE'],
            text: translations['FORM_MODIFIED_WARNING'],
            type: 'warning',
            showCancelButton: true,
            confirmButtonText: translations['CONFIRM_CONFIRM'],
            cancelButtonText: translations['CONFIRM_CANCEL'],
          }).then((result) => {
            resolve(result.value);
          });
        });
      });
    }

    return true;
  }

  /**
   * Initialise le formulaire
   * @param  {any} oMode
   */
  public initForm(oMode: any) {

    this.oMode = oMode;

    // Type de formulaire
    if (this.vitisCheck.isDefAndNotNull(this.oParams)) {
      if (this.vitisCheck.isString(this.oParams.form)) {
        if (this.oMode.options[this.oParams.form] === true) {
          this.sFormName = this.oParams.form;
        } else {
          console.error('Form type not allowed');
          return null;
        }
      }
    }

    // Charge les données du formulaire
    this.loadFormValues_(oMode).then((bValuesLoaded) => {

      // Charge le JS associé
      this.loadFormJavaScript_(oMode).then((bJsLoaded) => {

        // Charge le formulaire
        this.loadJSONForm_(oMode);

        // Services intégrés par défaut
        const oServices = {
          vitisRequest: this.vitisRequest,
          vitisCheck: this.vitisCheck,
          vitisAlert: vitisAlert,
          auth: this.auth,
          translate: this.translate,
          formreaderSrvc: this.formreaderSrvc,
          route: this.route,
          router: this.router,
          SimpleformComponent: this,
          interfaceService: this.interfaceService,
          sessionStorageService: this.sessionStorageService,
          localStorageService: this.localStorageService,
          oFormReader: this.oFormReader,
        };

        // Import des services
        if (oMode.options && oMode.options.js_services) {
          this.importCustomServices(oMode.options.js_services).then((oInjectedServices: any) => {

            // Lance le JS associé
            if ((<any>window).constructor_form) {
              (<any>window).constructor_form(this.oFormReader, environment, { ...oServices, ...oInjectedServices });
            }
          });

        } else {
          // Lance le JS associé
          if ((<any>window).constructor_form) {
            (<any>window).constructor_form(this.oFormReader, environment, oServices);
          }
        }
      });
    });
  }

  /**
   * Charge les valeurs du formulaire en fonction de l'identifiant passé dans l'URL
   * @param  {any} oMode
   * @returns Promise
   */
  private loadFormValues_(oMode: any): Promise<boolean> {
    return new Promise((resolve, reject) => {
      let bValuesLoading = false;

      if (this.vitisCheck.isDefAndNotNull(oMode.options)) {
        if (this.vitisCheck.isDefAndNotNull(oMode.options.ressource)) {

          const sRessourcePath = oMode.options.ressource;

          if (this.vitisCheck.isString(sRessourcePath)) {

            const sUrl = this.router.url;
            if (this.vitisCheck.isString(sUrl)) {

              // Récupère l'id dans les paramètres de la route
              let sRouteId: any;
              if (this.vitisCheck.isDefAndNotNull(this.localStorage.get('idClone'))) {
                sRouteId = this.localStorage.get('idClone');
                this.localStorage.set('idClone', null);
              } else if (this.vitisCheck.isDefAndNotNull(this.oParams)) {
                if (this.vitisCheck.isString(this.oParams.id) && this.sFormName !== 'insert') {
                  sRouteId = this.oParams.id;
                }
              }

              // Recupère les données
              if (this.vitisCheck.isDefAndNotNull(sRouteId)) {
                bValuesLoading = true;
                this.vitisRequest.ajaxRequestPromise({
                  method: 'GET',
                  url: this.vitisRequest.getApiRessourceUrl(sRessourcePath, sRouteId)
                }).then((resp) => {
                  if (resp.data) {
                    if (resp.data.status === 1) {
                      if (resp.data.data) {
                        if (resp.data.data) {
                          oMode.options.values = resp.data.data;
                        }
                      }
                    }
                  }
                  resolve(true);
                });
              } else {
                oMode.options.values = {};
              }
            }
          }
        }
      }

      if (!bValuesLoading) {
        resolve(false);
      }
    });
  }

  /**
   * Charge le JS associé au mode
   * @param  {any} oMode
   * @returns Promise
   */
  private loadFormJavaScript_(oMode: any): Promise<boolean> {
    return new Promise((resolve, reject) => {

      // Si le mode contient un JS associé
      if (this.vitisCheck.isDefAndNotNull(oMode.options)) {
        if (this.vitisCheck.isDefAndNotNull(oMode.options.js_path)) {

          // Charge le JS du mode
          this.vitisRequest.loadExternalJs(oMode.options.js_path).then(() => {
            resolve(true);
          });
        } else {
          resolve(false);
        }
      } else {
        resolve(false);
      }
    });
  }

  /**
   * Charge le formulaire
   * @param  {any} oMode
   * @returns Promise
   */
  private loadJSONForm_(oMode: any): Promise<void> {
    return new Promise((resolve, reject) => {
      if (this.vitisCheck.isDefAndNotNull(oMode.options)) {
        if (this.vitisCheck.isDefAndNotNull(oMode.options.form_path)) {

          // Récupère le formulaire
          this.getJSONForm_(oMode).then((oFormDefinition) => {
            if (this.vitisCheck.isDefAndNotNull(oFormDefinition)) {
              if (this.vitisCheck.isDefAndNotNull(oFormDefinition[this.sFormName])) {

                // AfterEvent par défaut
                if (!this.vitisCheck.isDefAndNotNull(oFormDefinition[this.sFormName].afterEvent)) {
                  oFormDefinition[this.sFormName].afterEvent = `setMode('search')`;
                }

                // Ouvre le formulaire dans FormReader
                oFormDefinition[this.sFormName].eventContainer = this;
                this.oFormReader.initForm(oFormDefinition, this.sFormName, oMode.options.values);
              }
            }
          });
        } else {
          resolve();
        }
      } else {
        resolve();
      }
    });
  }

  /**
   * Récupère la définition JSON d'un formulaire
   * @param  {any} oMode
   * @returns Promise
   */
  private getJSONForm_(oMode: any): Promise<any> {
    return new Promise((resolve, reject) => {

      let url: string = '';
      // si c'est un string ou un objet
      if (this.vitisCheck.isString(oMode.options.form_path)) {
        url = oMode.options.form_path;
      } else if (this.vitisCheck.isString(oMode.options.map_field)) {
        if (this.vitisCheck.isDefAndNotNull(oMode.options.values)) {
          const sKey = oMode.options.values[oMode.options.map_field];
          // Attribut de mappage dans les valeurs de l'objet
          if (this.vitisCheck.isDefAndNotNull(sKey) &&
            this.vitisCheck.isDefAndNotNull(oMode.options.form_path[sKey])) {
            url = oMode.options.form_path[sKey];
          } else {
            // si impossibilité de récupérer la valeur, on récupère le premier formulaire présent dans form_path OBJECT
            url = oMode.options.form_path[Object.keys(oMode.options.form_path)[0]];
          }
        } else { // Si pas d'attribut de mappage on récupère le premier formulaire présent dans form_path OBJECT
          url = oMode.options.form_path[Object.keys(oMode.options.form_path)[0]];
        }
      } else if (oMode.section_parent_mode &&
        this.vitisCheck.isString(oMode.options.parent_map_field)) {

        // Récupération des valeurs du parent
        if (this.route.snapshot &&
          this.route.snapshot.params &&
          this.route.snapshot.params.id_parent) {
          const idParent = this.route.snapshot.params.id_parent;
          this.vitisRequest.ajaxRequestPromise({
            method: 'GET',
            url: this.vitisRequest.getApiRessourceUrl(oMode.section_parent_mode.options.ressource, idParent)
          }).then((resp) => {
            if (resp.data) {
              if (resp.data.status === 1) {
                if (resp.data.data) {

                  // Attribut de mappage dans les valeurs du parent
                  oMode.options.parent_values = resp.data.data;
                  const sKey = oMode.options.parent_values[oMode.options.parent_map_field];
                  if (this.vitisCheck.isDefAndNotNull(sKey) &&
                    this.vitisCheck.isDefAndNotNull(oMode.options.form_path[sKey])) {
                    url = oMode.options.form_path[sKey];
                  }

                  // Chargement du formulaire
                  if (url) {
                    this.getJSONFormFromUrl_(url).then((data) => {
                      resolve(data);
                    }).catch(() => {
                      reject(false);
                    });
                  }
                }
              }
            }
          });
        }
      }

      if (url) {
        this.getJSONFormFromUrl_(url).then((data) => {
          resolve(data);
        }).catch(() => {
          reject(false);
        });
      }
    });
  }

  /**
   * Récupère la définition JSON d'un formulaire depuis son URL
   * @param  {string} sUrl
   * @returns Promise
   */
  private getJSONFormFromUrl_(sUrl: string): Promise<any> {
    return new Promise((resolve, reject) => {
      if (sUrl) {

        // Récupère le formulaire
        this.vitisRequest.ajaxRequestPromise({
          method: 'GET',
          url: sUrl
        }).then((resp) => {
          if (this.vitisCheck.isDefAndNotNull(resp.data)) {
            resolve(resp.data);
          } else {
            console.error('Error getting JSON form');
            reject(false);
          }
        });
      }
    });
  }

  /**
   * Change le mode en update
   */
  public editSectionForm() {
    this.setMode('update');
  }

  /**
   * Valide l'envoi d'un formulaire
   * @param  {any} oValues
   * @return {false|Promise<any>}
   */
  public sendSimpleForm(oValues: any): any {
    if (this.oFormReader.isFormValid()) {
      if (this.vitisCheck.isDefAndNotNull(oValues)) {
        if (this.vitisCheck.isDefAndNotNull(this.oMode)) {
          if (this.vitisCheck.isDefAndNotNull(this.oMode.options)) {
            if (this.vitisCheck.isDefAndNotNull(this.oMode.options.ressource)) {
              if (this.sFormName === 'update') {

                // Détection si un champ de type password a été modifié
                const bHasPassChanged = this.hasPasswordChanged_(oValues);
                if (bHasPassChanged) {
                  return new Promise<void>((resolve, reject) => {
                    this.showPassChangedMessage_().then(() => {
                      this.updateRessource(this.oMode.options.ressource, oValues).then(() => {
                        resolve();
                      }).catch(() => {
                        reject();
                      });
                    }).catch(() => {
                      reject();
                    });
                  });
                } else {
                  return this.updateRessource(this.oMode.options.ressource, oValues);
                }
              }

              if (this.sFormName === 'insert') {
                return this.insertRessource(this.oMode.options.ressource, oValues);
              }
            }
          }
        }
      }
    }
    return false;
  }

  /**
   * Met à jour l'enregistrement
   * @param  {any} oValues
   * @return {false|Promise<any>}
   */
  public updateRessource(sRessourceURL: string, oValues: any): any {

    // Attribut identifiant
    if (this.vitisCheck.isDefAndNotNull(this.oMode.options.id_field)) {
      if (this.vitisCheck.isDefAndNotNull(oValues[this.oMode.options.id_field])) {
        sRessourceURL = this.vitisRequest.getApiRessourceUrl(sRessourceURL, oValues[this.oMode.options.id_field]);
      } else if (this.vitisCheck.isDefAndNotNull(this.oParams.id)) {
        sRessourceURL = this.vitisRequest.getApiRessourceUrl(sRessourceURL, this.oParams.id);
      } else {
        console.error('Id field not valid');

        this.translate.get(['VITIS_ERROR', 'VITIS_SIMPLEFORM_FIELD_NOT_VALID']).subscribe((translations) => {
          translations['VITIS_SIMPLEFORM_FIELD_NOT_VALID'] = translations['VITIS_SIMPLEFORM_FIELD_NOT_VALID']
            .replace('{{::id_field}}', this.oMode.options.id_field);

          vitisAlert.alert({
            type: 'error',
            title: translations['VITIS_ERROR'],
            text: translations['VITIS_SIMPLEFORM_FIELD_NOT_VALID']
          });
        });
        return false;
      }
    }

    return new Promise<void>((resolve, reject) => {

      this.formreaderSrvc.manageLargesFiles(oValues).then((fileValues: any) => {
        // Lancement de la requête PUT
        this.vitisRequest.ajaxRequestPromise({
          method: 'PUT',
          url: sRessourceURL,
          data: this.formreaderSrvc.transformToDataForm(fileValues)
        }).then((resp) => {
          if (resp.data) {
            if (resp.data.status === 1) {
              this.translate.get('LIST_EDIT_OK').subscribe((translation) => {
                vitisAlert.fireToast({
                  icon: 'success',
                  title: translation
                });
                resolve();
              });
            } else {
              reject();
            }
          }
        });
      });
    });
  }

  /**
   * Ajoute l'enregistrement
   * @param  {any} oValues
   * @return {Promise<void>}
   */
  public insertRessource(sRessourceURL: string, oValues: any): Promise<void> {
    return new Promise((resolve, reject) => {

      delete this.insertedId;
      sRessourceURL = this.vitisRequest.getApiRessourceUrl(sRessourceURL);
      this.formreaderSrvc.manageLargesFiles(oValues).then((fileValues: any) => {
        // Lancement de la requête PUT
        this.vitisRequest.ajaxRequestPromise({
          method: 'POST',
          url: sRessourceURL,
          data: this.formreaderSrvc.transformToDataForm(fileValues)
        }).then((resp) => {
          if (resp.data) {
            if (resp.data.status === 1) {

              // Message succès
              this.translate.get('LIST_ADD_OK').subscribe((translation) => {
                vitisAlert.fireToast({
                  icon: 'success',
                  title: translation
                });

                // Sauvegarde l'identifiant précédament inserré
                if (this.vitisCheck.isDefAndNotNull(this.oMode.options)) {
                  if (this.vitisCheck.isDefAndNotNull(this.oMode.options.id_field)) {
                    const idField = this.oMode.options.id_field;
                    if (this.vitisCheck.isDefAndNotNull(resp.data[idField])) {
                      this.insertedId = resp.data.data[idField];
                    }
                    if (this.vitisCheck.isDefAndNotNull(resp.data.data) &&
                      this.vitisCheck.isDefAndNotNull(resp.data.data[idField])) {
                      this.insertedId = resp.data.data[idField];
                    }
                  }
                }

                if (!this.vitisCheck.isDefAndNotNull(this.insertedId)) {
                  console.error('Insert id could not be founded');
                }

                resolve();
              });
            } else {
              reject();
            }
          }
        });
      });
    });
  }

  /**
   * Change de mode (update, insert...)
   * @param  {string} sMode
   * @returns void
   */
  public setMode(sMode: string): void {

    // Identifiant (url ou précédament inserré)
    let sId: any;
    if (this.vitisCheck.isDefAndNotNull(this.insertedId)) {
      sId = this.insertedId;
    }
    if (this.vitisCheck.isDefAndNotNull(this.oParams)) {
      if (this.vitisCheck.isDefAndNotNull(this.oParams.id)) {
        sId = this.oParams.id;
      }
    }
    if (sMode === 'insert') {
      sId = '';
    }

    // Fin de l'url (après :form/).
    let endUrl: string = '', startUrl: string = '';
    if (this.oMode.route.indexOf(':form/') !== -1) {
      endUrl = '/' + this.oMode.route.split(':form/').pop();
      startUrl = '../';
    }

    if (sMode === 'search') {

      // Ouvre la reoute demandée
      if (this.vitisCheck.isDefAndNotNull(this.oParams)) {

        let relativeRoute = this.route.parent;

        // Cas de Workspacelist dans une section
        if (this.oMode.route.indexOf(':form/') !== -1) {
          if (this.route.snapshot &&
            this.route.snapshot.params &&
            this.route.snapshot.params.id_parent &&
            this.route.snapshot.params.form_parent) {
            relativeRoute = this.route;
          }
        }

        if (this.vitisCheck.isDefAndNotNull(this.oParams.id) && this.oParams.form !== 'insert') {
          this.router.navigate([startUrl + '../../'], { relativeTo: relativeRoute });
        } else if (this.vitisCheck.isDefAndNotNull(this.oParams.form)) {
          this.router.navigate([startUrl + '../'], { relativeTo: relativeRoute });
        }
      }
    } else {

      // Récupère l'id dans les paramètres de la route
      if (this.vitisCheck.isDefAndNotNull(sId)) {

        // Ouvre la reoute demandée
        if (this.vitisCheck.isDefAndNotNull(this.oParams.id)) {
          this.router.navigate([startUrl + '../../' + sId + '/' + sMode + endUrl], { relativeTo: this.route.parent });
        } else {
          this.router.navigate([startUrl + '../' + sId + '/' + sMode + endUrl], { relativeTo: this.route.parent });
        }

        // Déclenche l'événement
        const routeData: any = this.route.data;
        routeData.next(this.oMode);
      }
    }
  }

  /**
   * Récupère les données de la base et lance la callback
   * @param {Function} callback
   */
  public refreshFormValues(callback: Function) {
    // Charge les données du formulaire
    this.loadFormValues_(this.oMode).then((bValuesLoaded) => {
      if (this.vitisCheck.isFunction(callback)) {
        callback(this.oMode.options.values);
      }
    });
  }

  /**
   * Retourne true si un des champs de type password a été modifié
   * @param oValues
   */
  private hasPasswordChanged_(oValues: any): boolean {
    const oFormDefinition = this.oFormReader.oFormDefinition;
    const sFormDefinitionName = this.oFormReader.sFormDefinitionName;
    const aFieds = this.formreaderSrvc.getFormFields(oFormDefinition, sFormDefinitionName);

    for (let i = 0; i < aFieds.length; i++) {
      const oField = aFieds[i];
      if (oField.type === 'password') {
        const sValue = oValues[oField.name];
        if (this.vitisCheck.isDefAndNotNull(sValue) &&
          sValue !== this.sStarificationChar_) {
          return true;
        }
      }
    }

    return false;
  }

  /**
   * Affiche la modale de changement de mot de passe
   */
  private showPassChangedMessage_(): Promise<void> {
    return new Promise((resolve, reject) => {
      this.translate.get(['FORM_PASS_MODIFIED_WARNING_TITLE',
        'FORM_PASS_MODIFIED_WARNING',
        'CONFIRM_CANCEL',
        'CONFIRM_CONFIRM']
      ).subscribe((translations) => {
        vitisAlert.alert({
          title: translations['FORM_PASS_MODIFIED_WARNING_TITLE'],
          text: translations['FORM_PASS_MODIFIED_WARNING'],
          type: 'warning',
          showCancelButton: true,
          confirmButtonText: translations['CONFIRM_CONFIRM'],
          cancelButtonText: translations['CONFIRM_CANCEL'],
        }).then((result: any) => {
          if (result &&
            result.isConfirmed === true) {
            resolve();
          } else {
            reject();
          }
        });
      });
    });
  }

  /**
   * Utilise l'injector pour instancier des services et les retourne dans un objet
   * @param aImportServices
   * @returns
   */
  private importCustomServices(aImportServices: any) {
    return new Promise((resolve, reject) => {

      const oRenderedServices = {};

      let toLoad = 0;
      let loaded = 0;
      for (const key in aImportServices) {
        if (Object.prototype.hasOwnProperty.call(aImportServices, key)) {
          toLoad++;
          aImportServices[key].then((importedClass: any) => {
            try {
              oRenderedServices[key] = this.injector.get(importedClass[key]);
            } catch (error) {
              console.error(`fail to load service ${key}`, error);
            }
            loaded++;
            if (toLoad === loaded) {
              resolve(oRenderedServices);
            }
          });
        }
      }
    });
  }
}