/**
 * constructor_form
 * Fonction lancée au chargement du formulaire
 * @param {object} scope Formreader
 * @param {object} env environment
 * @param {object} services
 * @returns {undefined} 
 */
var constructor_form = function (scope, env, services) {

  const themeAndColorsFieldsName = [
    'app_theme',
    'color_primary',
    'color_primary_light',
    'color_primary_dark',
    'color_primary_text',
    'color_secondary',
    'color_secondary_active',
    'color_secondary_text',
    'color_secondary_text_active',
    'color_background',
    'color_background_dark5',
    'color_background_dark15',
    'color_background_text'
  ];

  /**
   * Boolean déterminant si le test des couleurs est en cours
   * Permet d'éviter que la fonction cancelTestColors ne s'execute si il n'y a pas de test en cours
   */
  scope.isBeingTested = false;

  /**
   * Remplit le formulaire avec les valeurs des properties
   */
  scope.loadProperties = function () {
    let sRessourceURL = env.properties.app_api_url + '/vitis/properties';

    services.vitisRequest.ajaxRequestPromise({
      method: 'GET',
      url: sRessourceURL,
    }).then((resp) => {
      if (resp.data) {
        if (resp.data.status === 1) {
          services.route.data.subscribe((oConfig) => {
            scope.oConfig = oConfig;
            const oSessionStorageValues = services.sessionStorageService.get(scope.oConfig.name);
            let oValues = resp.data.data;

            // Si il y a des valeurs dans la session, Lit les couleurs du storage
            if (services.vitisCheck.isDefAndNotNull(oSessionStorageValues) && services.vitisCheck.isDefAndNotNull(oSessionStorageValues.theme)) {
              const oThemeStorage = oSessionStorageValues.theme;
              oThemeStorage.app_theme = oThemeStorage.theme_id; // Récupère le theme_id pour set les valeurs
              oValues = { ...resp.data.data, ...oThemeStorage };

              // Si il y a la session c'est que le test est en cours
              scope.isBeingTested = true;
              scope.updateVisibleBtn(false, true);
            } else {
              const oTheme = scope.getPersonalizeTheme(resp.data.data.themes);
              oValues = { ...resp.data.data, ...oTheme };
            }

            scope.setFormValues(oValues);

            // On attend un peu pour ne pas prendre en compte la modification de mise en place (ci-dessus)
            setTimeout(() => {
              const aObserveControls = [];
              for (let i = 0; i < themeAndColorsFieldsName.length; i++) {
                aObserveControls.push(scope.oFormValues.controls[themeAndColorsFieldsName[i]])
              }
              
              scope.controlsSubscribed = [];
              // Si on modifie une valeur du theme ou des couleurs personnalisé, on annule le test en cours
              for (let i = 0; i < aObserveControls.length; i++) {
                const oControl = aObserveControls[i];
                scope.controlsSubscribed.push(oControl.valueChanges.subscribe((value) => {
                  scope.cancelTestColors();
                }));
              }
            }, 500);
          });


        }
      }
    });
  }

  /**
   * Fonction de mise à jour des éléments de l'interface, met à jour le thème personalisé si besoin puis met à jour les properties
   */
  scope.updateInterface = function (oFormValues) {
    return new Promise((resolve, reject) => {
      let oDataCss = {};
      let oDataColors = {};
      let oDataProperties = {};
      for (const key in oFormValues) {
        if (services.vitisCheck.isDefAndNotNull(oFormValues[key])) {
          if (key === 'custom_css') {
            oDataCss[key] = oFormValues[key]
          } 
          if (themeAndColorsFieldsName.indexOf(key) !== -1) {
            oDataColors[key] = oFormValues[key];
          } 
          if (key === 'app_theme' || key === 'app_description') {
            oDataProperties[key] = oFormValues[key];
          }
        }
      }

      scope.updateCss(oDataCss).then(() => {
        scope.updateTheme(oDataColors).then(() => {
          scope.updateProperties(oDataProperties).then(() => {
            resolve();
          });
        });
      });
    });
  }

  /**
   * Fonction de mise à jour des properties
   */
  scope.updateProperties = function (oData) {

    return new Promise((resolve, reject) => {
      let sRessourceURL = env.properties.app_api_url + '/' + scope.oConfig.options.ressource;
      
      services.vitisRequest.ajaxRequestPromise({
        method: 'PUT',
        url: sRessourceURL,
        data: oData
      }).then((resp) => {
        if (resp.data) {
          if (resp.data.status === 1) {
            services.translate.get('FORM_PROPERTIES_EDIT_OK').subscribe((translation) => {

              // Supprimer les couleurs du storage
              services.sessionStorageService.container.removeItem(scope.oConfig.name);
              scope.updateVisibleBtn(true, false);

              services.vitisAlert.fireToast({
                type: 'success',
                title: translation
              });
              resolve();
            });
          }
        }
      });
    })
  }

  /**
   * Fonction de mise à jour le theme personnalisé
   */
  scope.updateTheme = function (oData) {
    return new Promise((resolve, reject) => {
      if (oData.app_theme === env.properties.personalize_theme_id) {
        // Sauvegarde les modifications du thème personnalisé
        services.vitisRequest.ajaxRequestPromise({
          method: 'PUT',
          url: env.properties.app_api_url + '/vitis/theme/' + oData.app_theme,
          data: oData
        }).then((resp) => {
          if (resp.data) {
            if (resp.data.status === 1) {
              resolve();
            }
          }
        });
        // Applique visuellement le theme
        services.interfaceService.toggleTheme(oData.app_theme, [services.interfaceService.getPersonalizeTheme(oData)]);
      } else {
        // Si c'est un thème par défaut on applique le thème
        services.interfaceService.toggleTheme(oData.app_theme);
        resolve();
      }
    });
  }

  /**
   * Fonction de mise à jour du css
   */
  scope.updateCss = function (oData) {
    return new Promise((resolve, reject) => {
      // Si custom_css a été modifié on fait la requête pour le mettre à jour
      if (scope.oFormValues.controls['custom_css'].dirty) {
        services.vitisRequest.ajaxRequestPromise({
          method: 'PUT',
          url: env.properties.app_api_url + '/vitis/css/' + env.properties.custom_css_id,
          data: oData
        }).then((resp) => {
          if (resp.data) {
            if (resp.data.status === 1) {
              services.interfaceService.setDynamicStyle(oData.custom_css);
              resolve();
            }
          }
        });
      } else {
        resolve();
      }
    });
  }

  /**
   * Action du bouton pour tester les couleurs
   */
  scope.testColors = function () {
    scope.isBeingTested = true;

    let oData = scope.getFormValues();
    const personalizeTheme = services.interfaceService.getPersonalizeTheme(oData);

    // Change l'affichage du theme
    if (oData.app_theme == env.properties.personalize_theme_id) {
      services.interfaceService.toggleTheme(oData.app_theme, [personalizeTheme]);
    } else {
      services.interfaceService.toggleTheme(oData.app_theme);
    }

    // Sauvegarde les couleurs dans le storage
    services.sessionStorageService.add(scope.oConfig.name, {
      theme: personalizeTheme
    });

    // Modifie le bouton
    scope.updateVisibleBtn(false, true);
  }
  
  /**
   * Action du bouton d'annulation du test de couleur
   */
  scope.cancelTestColors = function () {
    if (scope.isBeingTested) {
      scope.isBeingTested = false;
      // Change le theme avec le contenu des properties
      services.interfaceService.toggleTheme();
  
      // Refresh les valeurs du formulaire ?
  
      // Supprime les couleurs dans le storage
      services.sessionStorageService.container.removeItem(scope.oConfig.name);
  
      // Modifie le bouton
      scope.updateVisibleBtn(true, false);
    }
  }

  /**
   * Met à jour la visibilité des boutons test et annulation du test
   * @param {boolean} setBtnTestVisible 
   * @param {boolean} setBtnCancelTestVisible
   */
  scope.updateVisibleBtn = function (setBtnTestVisible, setBtnCancelTestVisible) {
    let oBtnTest = scope.getFormField('button_test');
    let oBtnCancelTest = scope.getFormField('button_test_cancel');
    oBtnTest.visible = setBtnTestVisible;
    oBtnCancelTest.visible = setBtnCancelTestVisible;
  }

  /**
   * Surcharge de la fonction canDeactivate
   */
  services.SimpleformComponent.canDeactivate = function () {
    const component = services.SimpleformComponent
    let bIsModified = false;

    const oSessionStorageValues = services.sessionStorageService.get(scope.oConfig.name);

    // Si il y a des valeurs dans la session, on vérifie que les champs autre que les couleurs
    if (scope.isBeingTested &&
      services.vitisCheck.isDefAndNotNull(oSessionStorageValues) &&
      services.vitisCheck.isDefAndNotNull(oSessionStorageValues.theme)) {
      for (const key in scope.oFormValues.controls) {
        if (scope.oFormValues.controls.hasOwnProperty(key)) {
          // On vérifie si un champ a été modifié
          if (themeAndColorsFieldsName.indexOf(key) === -1) {
            const oControl = scope.oFormValues.controls[key];
            if (oControl.dirty) {
              bIsModified = oControl.dirty;
            }

          }
        }
      }
    } else { // Sinon on vérifie tout le formulaire
      bIsModified = scope.oFormValues.dirty;
    }

    if (bIsModified && !component.oFormReader.getIsSubmittedForm()) {
      return new Promise((resolve) => {
        services.translate.get(['FORM_MODIFIED_WARNING_TITLE',
          'FORM_MODIFIED_WARNING',
          'CONFIRM_CANCEL',
          'CONFIRM_CONFIRM']
        ).subscribe((translations) => {
          services.vitisAlert.alert({
            title: translations['FORM_MODIFIED_WARNING_TITLE'],
            text: translations['FORM_MODIFIED_WARNING'],
            type: 'warning',
            showCancelButton: true,
            confirmButtonText: translations['CONFIRM_CONFIRM'],
            cancelButtonText: translations['CONFIRM_CANCEL'],
          }).then((result) => {
            if (result.value) {
              scope.unsubscribeColors();
            }
            resolve(result.value);
          });
        });
      });
    }

    scope.unsubscribeColors();

    return true;
  }

  /**
   * annule les subscribes qui surveille les changements de valeurs du formulaire
   */
  scope.unsubscribeColors = function () {
    for (let i = 0; i < scope.controlsSubscribed.length; i++) {
      scope.controlsSubscribed[i].unsubscribe();
    }
  }

  /**
   * Récupère le thème personnalisé à partir de la liste des thèmes et set les valeurs du theme qui sont null avec les valeurs courrantes du css.
   */
  scope.getPersonalizeTheme = function(aThemes) {
    let oTheme;
    // Récupère les valeurs du thème personnalisé dans les properpeties
    for (let i = 0; i < aThemes.length; i++) {
      if (aThemes[i].theme_id == env.properties.personalize_theme_id) {
        oTheme = aThemes[i];
        for (const themeKey in oTheme) {
          if (oTheme.hasOwnProperty(themeKey)) {
            const varStyleValue = getComputedStyle(document.documentElement).getPropertyValue('--application-' + themeKey.replace(/_/g, '-'));
            // Si la valeur n'existe pas dans le theme et que la variable de css a une valeur, on attribut la valeur de la variable css au theme
            if ((!services.vitisCheck.isDefAndNotNull(oTheme[themeKey]) || oTheme[themeKey] == '')
              && services.vitisCheck.isDefAndNotNull(varStyleValue) && varStyleValue != '') {
              oTheme[themeKey] = varStyleValue;
            }
          }
        }
      }
    }

    return oTheme;
  }
}
