import fs from 'fs';

/**
 * Un PDF se décompose en plusieurs partie que l'on peut considérer comme un bloc 
 * Les modifications peuvent de trouver réparti en plusieurs bloques (comme un système de versionning intégré)
 * Pour signer un document, il faut prévoir un espace pret à recevoir la signature cette espace est en fait un bloc
 * 
 * 
 *  ex de PDF signé  :
 *  _________________________________________________
 * |%PDF-version                                     |
 * |%^^^^                                            |
 * |_________________________________________________|
 * |                                                 |
 * |                     Header                      |
 * |_________________________________________________|
 * |                                                 |
 * |                      Body                       |
 * |_________________________________________________|
 * |                                                 |
 * |                     Trailer                     |
 * |_________________________________________________|
 * |%%EOF                                            |
 * |_________________________________________________|
 *   ////////////////////////////////////////////////
 *  //////////////////////////////////////////////// 
 * |                      BODY 2                     |
 * |Espace signature                                 |
 * |Widget signature                                 |
 * |Champ signature                                  |
 * |Surcharge Catalog/Root                           |
 * |Intégration Widget dans la dernière page         |
 * |_________________________________________________|
 * |                                                 |
 * |                     Trailer 2                   |
 * |_________________________________________________|
 * |%%EOF                                            |
 * |_________________________________________________|
 * 
 * 
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */


export interface PlaceHolderOptions {
  reason: string;
  contactInfo: string;
  name: string;
  location: string;
  signatureLength: number;
  subFilter: string; // doit être une des 4 constante de placeHolder
  date: Date;
}

interface PDFInfo {
  xref: any,
  rootRef: string,
  root: string,
  infoRef?: string,
  trailerStart: number,
  previousXrefs: any[],
  xRefPosition: number,
}

interface PDFBuffer {
  index: number,
  ref: string,
  buffer: Buffer,
}

export class PlaceHolder {

  public static DEFAULT_SIGNATURE_LENGTH: number = 8192;
  public static SUBFILTER_ADOBE_PKCS7_DETACHED = 'adbe.pkcs7.detached';
  public static SUBFILTER_ADOBE_PKCS7_SHA1 = 'adbe.pkcs7.sha1';
  public static SUBFILTER_ADOBE_X509_SHA1 = 'adbe.x509.rsa.sha1';
  public static SUBFILTER_ETSI_CADES_DETACHED = 'ETSI.CAdES.detached';

  private pdfBuffer: Buffer;

  private signature?: PDFBuffer;
  private widget?: PDFBuffer;
  private form?: PDFBuffer;
  private overWritedRoot?: PDFBuffer;
  private overWritedPage?: PDFBuffer;

  private byteRange: number[] = [
    0,
    0,
    0,
    0
  ];

  private options: PlaceHolderOptions = {
    reason: 'default reason',
    contactInfo: 'contact@cert.fr',
    name: 'certificator',
    location: 'somewhere',
    signatureLength: PlaceHolder.DEFAULT_SIGNATURE_LENGTH,
    subFilter: PlaceHolder.SUBFILTER_ADOBE_PKCS7_DETACHED,
    date: new Date()
  }
  /**
   * info du Vuffer initial
   */
  private info: PDFInfo;
  /**
   * Référence de l'objet page à éditer dans le pdf
   */
  private pageRef: string;
  private pageIndex: number;

  private addedReferences: Map<number, number>;

  constructor(sPdfPath: string, options: PlaceHolderOptions) {
    this.pdfBuffer = fs.readFileSync(sPdfPath);
    this.options = {
      ...this.options,
      ...options
    }
    this.removeTrailingNewLine();
    this.info = this.readPdf();
    this.pageRef = this.getPageRef();
    this.pageIndex = this.getIndexFromRef(this.info.xref, this.pageRef);
    this.addedReferences = new Map();
    this.generatePlaceHolder();
  }

  /**
   * Retourne le fichier avec son placeholder utilisable
   * @returns {Buffer}
   */
  public getBuffer(): Buffer {
    return this.concatElement();
  }

  /**
   * concatène tous les éléments et regénère récursivement la signature jusqu'à avoir quelques chose de cohérent
   * @returns {Buffer}
   */
  private concatElement(): Buffer {
    let pdf: Buffer = Buffer.concat([
      this.pdfBuffer,
      // comme on trim toujours les espaces en fin de fichier on en remet un de rab pour la lisibilité mais un seul suffirai
      //Buffer.from('\n'),
      Buffer.from('\n'),
    ]);

    if (!this.signature) {
      throw new Error('you have to call generateSignature before')
    }

    this.addedReferences.set(this.signature.index, pdf.length);
    pdf = Buffer.concat([
      pdf,
      this.signature.buffer,
      Buffer.from('\n'),
    ]);

    if (!this.widget) {
      throw new Error('you have to call generateWidget before')
    }

    this.addedReferences.set(this.widget.index, pdf.length);
    pdf = Buffer.concat([
      pdf,
      this.widget.buffer,
      Buffer.from('\n'),
    ]);

    if (!this.form) {
      throw new Error('you have to call generateForm before')
    }

    this.addedReferences.set(this.form.index, pdf.length);
    pdf = Buffer.concat([
      pdf,
      this.form.buffer,
      Buffer.from('\n'),
    ]);

    // la surcharge de l'élement root ne se mais que si form est nouveau dans le fichier
    // si un acroform existait déjà alors il ne sert à rien de le surcharger
    if (this.overWritedRoot) {
      this.addedReferences.set(this.overWritedRoot.index, pdf.length);
      pdf = Buffer.concat([
        pdf,
        this.overWritedRoot.buffer,
        Buffer.from('\n'),
      ]);
    }

    // surcharge de la page qui prendre en charge le widget comme annotation
    if (!this.overWritedPage) {
      throw new Error('you have to call generateOverwritingPageWithAnnotation before')
    }

    this.addedReferences.set(this.overWritedPage.index, pdf.length);
    pdf = Buffer.concat([
      pdf,
      this.overWritedPage.buffer,
      Buffer.from('\n'),
    ]);

    // on ajoute le trailer
    pdf = Buffer.concat([
      pdf,
      this.createBufferTrailer(pdf),
    ]);

    // si el byterange n'est psa bon on regènère la signature et on relance une concaténation
    if (!this.checkArrayByteRange(pdf)) {
      this.generateSignature();
      return this.concatElement();
    }

    return pdf;
  }

  /**
   * détermine si le byteRange dans le fichier est cohérent ou s'il faut regénérer tout le buffer
   * @param {Buffer} pdf buffer du pdf ave cle placeholder  
   * @returns {boolean} si false il faut regénérer la signature
   */
  private checkArrayByteRange(pdf: Buffer): boolean {

    // la signature sera tujours calculé depuis le début
    // ByteRange 0 dans le cas d'une signature de première version toujours 0
    if (this.byteRange[0] !== 0) {
      return false;
    }
    // cherche le byterange que l'on a ajouté
    const byteRangePos = pdf.indexOf('/ByteRange', this.pdfBuffer.length);

    if (byteRangePos === -1) {
      throw new Error(`Could not find ByteRange placeholder`);
    }
    // cherche la ligne contenant la signature aprés byteRange
    const contentsTagPos = pdf.indexOf('/Contents ', byteRangePos);
    // cherche la ligne contenant l'ouverture de signature
    const placeholderPos = pdf.indexOf('<', contentsTagPos);
    // cherche la ligne contenant la fin de signature
    const placeholderEnd = pdf.indexOf('>', placeholderPos);

    // Byterange 1 Position d'ouverture de signature <
    if (this.byteRange[1] !== placeholderPos) {
      this.byteRange[1] = placeholderPos;
      // comme byterange 2 est la fin de l'emplacement autant anticipé on s'économisera une itération
      // le chevron est les délimiteur du placeholder mais il n'est ni dans le hash, ni dans le byteRange (enfin entre 1 et 2)
      this.byteRange[2] = placeholderPos + this.options.signatureLength + 2;
      // la longueur de fin de fichier à peu de chnage de bouger aussi dans le doute on le tente
      this.byteRange[3] = (pdf.length) - this.byteRange[2];
      return false;
    }

    if (this.byteRange[2] != placeholderEnd + 1) {
      // erreur de calcul précédent on retape une itération
      // console.log('calcul à revoir')
      this.byteRange[2] = placeholderEnd + 1;
      this.byteRange[3] = (pdf.length) - this.byteRange[2];
      return false;
    }

    // longueur restant jusqu'à la fin du fichier
    if (this.byteRange[3] !== (pdf.length) - this.byteRange[2]) {
      this.byteRange[3] = (pdf.length) - this.byteRange[2];
      return false;
    }

    return true;
  }

  private generatePlaceHolder() {
    // Check if pdf already contains acroform field
    const acroFormPosition = this.pdfBuffer.lastIndexOf('/Type /AcroForm');
    const isAcroFormExists = (acroFormPosition !== -1);
    let fieldIds: string[] = [];
    let acroFormId;

    if (isAcroFormExists) {
      let acroFormStart = acroFormPosition;
      // 10 is the distance between "/Type /AcroForm" and AcroFrom ID
      const charsUntilIdEnd = 10;
      const acroFormIdEnd = acroFormPosition - charsUntilIdEnd;
      // Let's find AcroForm ID by trying to find the "\n" before the ID
      // 12 is a enough space to find the "\n"
      // (generally it's 2 or 3, but I'm giving a big space though)
      const maxAcroFormIdLength = 12;
      let foundAcroFormId = '';
      let index = charsUntilIdEnd + 1;
      for (index; index < charsUntilIdEnd + maxAcroFormIdLength; index += 1) {
        const acroFormIdString = this.pdfBuffer
          .slice(acroFormPosition - index, acroFormIdEnd).toString();

        if (acroFormIdString[0] === '\n') {
          break;
        }

        foundAcroFormId = acroFormIdString;
        acroFormStart = acroFormPosition - index;
      }

      const pdfSlice = this.pdfBuffer.slice(acroFormStart);
      const acroForm = pdfSlice.slice(0, pdfSlice.indexOf('endobj')).toString();
      acroFormId = parseInt(foundAcroFormId);

      const acroFormFields = acroForm.slice(acroForm.indexOf('/Fields [') + 9, acroForm.indexOf(']'));
      fieldIds = acroFormFields
        .split(' ')
        .filter((element, i) => i % 3 === 0)
        .map((fieldId) => this.intToRef(parseInt(fieldId)));
    }


    /* eslint-disable no-underscore-dangle,no-param-reassign */
    // Generate the signature placeholder
    this.generateSignature();

    this.generateWidget(fieldIds);

    if (acroFormId) {
      this.form = {
        index: acroFormId,
        ref: this.intToRef(acroFormId),
        buffer: Buffer.from('')
      }
    }

    // Include the widget in a page
    this.generateForm(fieldIds);
    if (!isAcroFormExists) {
      // Create a form (with the widget) and link in the _root
      // this.addedReferences.set(rootIndex, this.pdfBuffer.length + 1);
      this.generateOverwritingRootWithAcroform();
    }

    // this.addedReferences.set(this.pageIndex, pdf.length + 1);
    this.generateOverwritingPageWithAnnotation();

    // return Buffer.concat([
    //   pdf,
    //   Buffer.from('\n'),
    //   this.createBufferTrailer(pdf, info, addedReferences),
    // ]);
  }

  private getNewObjectIndex(): number {
    this.info.xref.maxIndex += 1;
    return this.info.xref.maxIndex;
  }

  /**
   * Génère le buffer pour la signature et stock sa référence
   */
  private generateSignature() {
    let index: number;
    if (!this.signature) {
      index = this.getNewObjectIndex();
    } else {
      index = this.signature.index;
    }
    const date: Date = this.options.date;
    const sYear = date.getUTCFullYear();
    const sMonth = (date.getUTCMonth() + 1 < 10) ? '0' + (date.getUTCMonth() + 1) : date.getUTCMonth() + 1;
    const sDay = (date.getUTCDate() < 10) ? '0' + date.getUTCDate() : date.getUTCDate();
    const sHour = (date.getUTCHours() < 10) ? '0' + date.getUTCHours() : date.getUTCHours();
    const sMinute = (date.getUTCMinutes() < 10) ? '0' + date.getUTCMinutes() : date.getUTCMinutes();
    const sSecond = (date.getUTCSeconds() < 10) ? '0' + date.getUTCSeconds() : date.getUTCSeconds();
    const sDateM: string = `${sYear}${sMonth}${sDay}${sHour}${sMinute}${sSecond}Z`;

    const ref: string = this.intToRef(index);
    let buffer: Buffer = Buffer.concat([
      Buffer.from(`${index} 0 obj\n`),
      Buffer.from('<<\n'),
      Buffer.from(`/Type /Sig\n`),
      Buffer.from(`/Filter /Adobe.PPKLite\n`),
      Buffer.from(`/SubFilter /${this.options.subFilter}\n`),
      Buffer.from(`/ByteRange [${this.byteRange.join(' ')}]\n`),
      Buffer.from(`/Contents <`),
      Buffer.from(Buffer.from(String.fromCharCode(0).repeat(this.options.signatureLength)).toString('hex')),
      Buffer.from(`>\n`),
      Buffer.from(`/Reason (${this.prepareString(this.options.reason)})\n`),
      Buffer.from(`/M (D:${sDateM})\n`),
      Buffer.from(`/ContactInfo (${this.prepareString(this.options.contactInfo)})\n`),
      Buffer.from(`/Name (${this.prepareString(this.options.name)})\n`),
      Buffer.from(`/Location (${this.prepareString(this.options.location)})\n`),
      Buffer.from('>>\nendobj\n'),
    ]);

    this.signature = {
      index,
      ref,
      buffer
    }
  }

  /**
   * génère le widget qui stockera la signature
   * @param fieldIds 
   */
  private generateWidget(fieldIds: any[]) {

    if (!this.signature) {
      throw new Error('you have to call generateSignature before')
    }

    let index: number;
    if (!this.widget) {
      index = this.getNewObjectIndex();
    } else {
      index = this.widget.index;
    }

    const signatureName: string = 'Signature';
    const ref: string = this.intToRef(index);

    const sT: string = new String(signatureName + (fieldIds.length + 1)).toString();

    // Generate signature annotation widget
    const buffer: Buffer = Buffer.concat([
      Buffer.from(`${index} 0 obj\n`),
      Buffer.from('<<\n'),
      Buffer.from(`/Type /Annot\n`),
      Buffer.from(`/Subtype /Widget\n`),
      Buffer.from(`/FT /Sig\n`),
      Buffer.from(`/Rect [0 0 0 0]\n`),
      Buffer.from(`/V ${this.signature.ref}\n`),
      Buffer.from(`/T (${sT})\n`),
      Buffer.from(`/F 4\n`),
      Buffer.from(`/P ${this.pageIndex} 0 R\n`),
      Buffer.from('>>\nendobj\n'),
    ]);

    this.widget = {
      index,
      ref,
      buffer
    }
  }

  /**
   * 
   * @param fieldIds 
   * @param index 
   */
  private generateForm(fieldIds: string[]) {

    if (!this.widget) {
      throw new Error('you have to call generateWidget before')
    }

    let index: number;
    if (!this.form) {
      index = this.getNewObjectIndex();
    } else {
      index = this.form.index;
    }

    const ref: string = this.intToRef(index);;

    fieldIds.push(this.widget.ref)

    const buffer: Buffer = Buffer.concat([
      Buffer.from(`${index} 0 obj\n`),
      Buffer.from('<<\n'),
      Buffer.from(`/Type /AcroForm\n`),
      Buffer.from(`/SigFlags 3\n`),
      Buffer.from(`/Fields [${fieldIds.join(' ')}]\n`),
      Buffer.from('>>\nendobj\n'),
    ]);

    this.form = {
      index,
      ref,
      buffer
    }
  }

  private generateOverwritingRootWithAcroform() {

    if (!this.form) {
      throw new Error('you have to call generateForm before')
    }

    const rootIndex = this.getIndexFromRef(this.info.xref, this.info.rootRef);

    const buffer: Buffer = Buffer.concat([
      Buffer.from(`${rootIndex} 0 obj\n`),
      Buffer.from('<<\n'),
      Buffer.from(`${this.info.root}\n`),
      Buffer.from(`/AcroForm ${this.form.ref}\n`),
      Buffer.from('>>\nendobj\n'),
    ]);

    this.overWritedRoot = {
      index: rootIndex,
      ref: this.intToRef(rootIndex),
      buffer
    }
  };

  /**
   *  A revoir
   */
  private generateOverwritingPageWithAnnotation() {
    if (!this.widget) {
      throw new Error('you have to call generateWidget before')
    }
    const pagesDictionary = this.findObject(this.info.xref, this.pageRef).toString();


    // Extend page dictionary with newly created annotations
    let annotsStart: number = pagesDictionary.indexOf('/Annots');
    let annotsEnd;
    let annots;
    if (annotsStart > -1) {
      annotsEnd = pagesDictionary.indexOf(']', annotsStart);
      // annots = pagesDictionary.substr(annotsStart, annotsEnd + 1 - annotsStart);
      annots = pagesDictionary.substring(annotsStart, annotsEnd + 1);
      // annots = annots.substr(0, annots.length - 1); // remove the trailing ]
      annots = annots.substring(0, annots.length - 1); // remove the trailing ]
    } else {
      annotsStart = pagesDictionary.length;
      annotsEnd = pagesDictionary.length;
      annots = '/Annots [';
    }

    const pagesDictionaryIndex = this.getIndexFromRef(this.info.xref, this.pageRef);

    annots = `${annots} ${this.widget.ref}]`; // add the trailing ] back

    const preAnnots = pagesDictionary.substring(0, annotsStart + 1);
    let postAnnots = '';
    if (pagesDictionary.length > annotsEnd) {
      postAnnots = pagesDictionary.substring(annotsEnd + 1);
    }

    const buffer: Buffer = Buffer.concat([
      Buffer.from(`${pagesDictionaryIndex} 0 obj\n`),
      Buffer.from('<<\n'),
      Buffer.from(`${preAnnots + annots + postAnnots}\n`),
      Buffer.from('>>\nendobj\n'),
    ]);

    this.overWritedPage = {
      index: pagesDictionaryIndex,
      ref: this.intToRef(pagesDictionaryIndex),
      buffer
    }
  };

  private createBufferTrailer(pdf: Buffer): Buffer {
    let rows = [];
    rows[0] = '0000000000 65535 f '; // info.xref.tableRows[0];

    this.addedReferences.forEach((offset, index) => {
      const paddedOffset = (`0000000000${offset}`).slice(-10);
      rows[index + 1] = `${index} 1\n${paddedOffset} 00000 n `;
    });
    rows = rows.filter((row) => row !== undefined);

    return Buffer.concat([
      Buffer.from('xref\n'),
      Buffer.from(`${this.info.xref.startingIndex} 1\n`),
      Buffer.from(rows.join('\n')),
      Buffer.from('\ntrailer\n'),
      Buffer.from('<<\n'),
      Buffer.from(`/Size ${this.info.xref.maxIndex + 1}\n`),
      Buffer.from(`/Root ${this.info.rootRef}\n`),
      Buffer.from(this.info.infoRef ? `/Info ${this.info.infoRef}\n` : ''),
      Buffer.from(`/Prev ${this.info.xRefPosition}\n`),
      Buffer.from('>>\n'),
      Buffer.from('startxref\n'),
      Buffer.from(`${pdf.length}\n`),
      Buffer.from('%%EOF'),
    ]);
  };

  /**
   * Nettoie la fin du fichier en cas de retour chariot en trop, ...
   * Et vérifie que la fin du pdf existant est correct
   * @throws {Error} si la fin du fichier n'est pas conforme
   * @returns {void} 
   */
  private removeTrailingNewLine(): void {
    let output: Buffer = this.pdfBuffer;

    output = this.sliceLastChar(output, '\n');
    output = this.sliceLastChar(output, '\r');

    const lastLine = output.slice(output.length - 6).toString();
    if (lastLine !== '\n%%EOF') {
      throw new Error('A PDF file must end with an EOF line.');
    }

    this.pdfBuffer = output;
  }

  /**
   * Extrait les informations présentes dans le PDF
   * @return {PDFInfo}
   */
  private readPdf(): PDFInfo {
    const pdfBuffer: Buffer = this.pdfBuffer;
    // Extraction du trailer du PDF d'origine.
    const trailerStart: number = pdfBuffer.lastIndexOf('trailer');
    // The trailer is followed by xref. Then an EOF. EOF's length is 6 characters.
    const trailer: Buffer = pdfBuffer.slice(trailerStart, pdfBuffer.length - 6);

    const xRefPosition: string = trailer.slice(trailer.lastIndexOf('startxref') + 10).toString();

    const iRefPosition: number = parseInt(xRefPosition);
    const refTable = this.readRefTable();

    const rootRef = this.getValue(trailer, '/Root');

    if (typeof rootRef !== 'string') {
      throw Error('rootRef is not a string')
    }
    const root = this.findObject(refTable, rootRef).toString();

    const infoRef = this.getValue(trailer, '/Info');

    return {
      xref: refTable,
      rootRef,
      root,
      infoRef,
      trailerStart,
      previousXrefs: [],
      xRefPosition: iRefPosition,
    };
  };


  /**
   * 
   * @returns {number} valeur du xrefposition du dernier trailer dans le PDF
   */
  private getLastTrailerPosition(): number {
    const trailerStart = this.pdfBuffer.lastIndexOf('trailer');
    const trailer = this.pdfBuffer.slice(trailerStart, this.pdfBuffer.length - 6);

    const xRefPosition = trailer
      .slice(trailer.lastIndexOf('startxref') + 10)
      .toString();

    return parseInt(xRefPosition);
  };

  /**
   * 
   * @param {number} position 
   * @returns {any} 
   * {
   *  size : Nombre d'objet présent dans le pdf
   *  prev : Index du prev si présent dans le dernier trailer (je crois)
   *  xRefContent : chaine contenant le xref (je crois)
   * }
   */
  private getXref(position: number): any {
    let refTable = this.pdfBuffer.slice(position);

    refTable = refTable.slice(4);
    refTable = refTable.slice(refTable.indexOf('\n') + 1);
    const size = refTable.toString().split('/Size')[1];
    const [objects, infos] = refTable.toString().split('trailer');

    const isContainingPrev = infos.split('/Prev')[1] != null;

    let prev;
    let reducer;

    if (isContainingPrev) {
      const pagesRefRegex = /Prev (\d+)/g;
      const match = pagesRefRegex.exec(infos);
      if (!match) {
        throw new Error('Cannot find Prev tag')
      }
      const [, prevPosition] = match;
      prev = prevPosition;
      reducer = (prev: any, curr: string) => {
        const isObjectId = curr.split(' ').length === 2;

        if (isObjectId) {
          const [id] = curr.split(' ');
          return { ...prev, [id]: undefined };
        }

        const [offset] = curr.split(' ');
        const prevId = Object.keys(prev).find((id) => prev[id] === undefined);

        return { ...prev, [(prevId) ? prevId : '']: parseInt(offset) };
      };
    } else {
      reducer = (prev: any, l: any, i: any) => {
        const element = l.split(' ')[0];
        const isPageObject = parseInt(element) === 0 && element.length > 3;

        if (isPageObject) {
          return { ...prev, 0: 0 };
        }

        let [offset] = l.split(' ');
        offset = parseInt(offset);

        return { ...prev, [i - 1]: offset };
      };
    }

    const xRefContent = objects
      .split('\n')
      .filter((l) => l !== '')
      .reduce(reducer, {});

    return {
      size,
      prev,
      xRefContent,
    };
  };

  /**
   * 
   * @param trailer 
   * @param key 
   * @returns 
   */
  private getValue(trailer: Buffer, key: string): string | undefined {
    let KeyPos: number = trailer.indexOf(key);

    if (KeyPos === -1) {
      return;
    }

    // ne tien pas compte de ce qu'il y a avant la clé
    const slice: Buffer = trailer.slice(KeyPos);
    let endPos: number;

    // détection de la clé suivante
    endPos = slice.indexOf('/', 1);

    // s'il n'y pas d'autre clé par la suite
    // on cherche la fin de l'objet courant
    if (endPos == -1) {
      endPos = slice.indexOf('>', 1);
    }

    // détection des sous objet à la place des références
    let subObject = slice.indexOf('<<');
    if (subObject > -1 && subObject < endPos) {
      endPos = slice.indexOf('>>', 1) + 3; // fin de sous objet et retour chariot
    }

    return slice.slice(key.length + 1, endPos).toString().trim(); // key + at least one space
  };

  /**
   * Trouve la référence de la page pour y intégrer le widget par la suite
   * @returns {string}
   */
  private getPageRef(): string {
    const pagesRef = this.getPagesDictionaryRef();
    const pagesDictionary = this.findObject(this.info.xref, pagesRef);
    const kidsPosition = pagesDictionary.indexOf('/Kids');
    const kidsStart = pagesDictionary.indexOf('[', kidsPosition) + 1;
    const kidsEnd = pagesDictionary.indexOf(']', kidsPosition);
    const pages = pagesDictionary.slice(kidsStart, kidsEnd).toString();
    const split = pages.trim().split(' ', 3);
    return `${split[0]} ${split[1]} ${split[2]}`;
  }

  /**
   * 
   * @returns {any}
   * {
   *  maxOffset : ,number
   *  startingIndex : 0
   *  maxIndex : Nombre d'entré dans le trailer (je crois)
   *  offsets : Map offset
   * }
   */
  private readRefTable(): any {
    const offsetsMap = new Map();
    const fullXrefTable = this.getFullXrefTable(this.pdfBuffer);

    const startingIndex = 0;

    let maxOffset = 0;
    const maxIndex = Object.keys(fullXrefTable).length - 1;

    Object.keys(fullXrefTable).forEach((id) => {
      const offset = parseInt(fullXrefTable[id]);
      maxOffset = Math.max(maxOffset, offset);
      offsetsMap.set(parseInt(id), offset);
    });

    return {
      maxOffset,
      startingIndex,
      maxIndex,
      offsets: offsetsMap,
    };
  };

  /**
   * fonction récursive qui va aanalyser les différent trailer du buffer
   * @param pdf : Buffer
   * @returns {any} 
   */
  private getFullXrefTable(pdf: Buffer): any {
    const lastTrailerPosition = this.getLastTrailerPosition();
    const lastXrefTable = this.getXref(lastTrailerPosition);

    if (lastXrefTable.prev === undefined) {
      return lastXrefTable.xRefContent;
    }
    const pdfWithoutLastTrailer = pdf.slice(0, lastTrailerPosition);
    const partOfXrefTable = this.getFullXrefTable(pdfWithoutLastTrailer);

    return {
      ...partOfXrefTable,
      ...lastXrefTable.xRefContent,
    };
  }

  /**
   * 
   * @returns {string} référence du root
   */
  private getPagesDictionaryRef(): string {
    const pagesRefRegex = /\/Pages\s+(\d+\s+\d+\s+R)/g;
    const match = pagesRefRegex.exec(this.info.root);
    if (match === null) {
      throw new Error('Failed to find the pages descriptor. This is probably a problem in node-signpdf.',);
    }

    return match[1];
  }

  /**
   * conveti une référence d'objet en offset pour manipuler les buffer
   * @param {any} refTable 
   * @param {string} ref 
   * @returns 
   */
  private getIndexFromRef(refTable: any, ref: string): number {
    let [index] = ref.split(' ');
    const iIndex = parseInt(index);

    if (!refTable.offsets.has(iIndex)) {
      throw new Error(`Failed to locate object "${ref}".`,);
    }
    return iIndex;
  };

  /**
   * Trouve un objet dans le pdf à partir d'une reférence
   * problème si l'objet est un stream
   * @param {any} refTable 
   * @param {string} ref 
   * @returns {Buffer}
   */
  private findObject(refTable: any, ref: string): Buffer {
    const index = this.getIndexFromRef(refTable, ref);

    const offset = refTable.offsets.get(index);
    let slice = this.pdfBuffer.slice(offset);
    slice = slice.slice(0, slice.indexOf('endobj'));

    // FIXME: What if it is a stream?
    slice = slice.slice(slice.indexOf('<<') + 2);
    slice = slice.slice(0, slice.lastIndexOf('>>'));
    return slice;
  };

  /**
   * Supprime un caractère à la fin d'un buffer s'il est présent
   * @param pdf 
   * @param character 
   * @returns 
   */
  private sliceLastChar(pdf: Buffer, character: string): Buffer {
    const lastChar = pdf.slice(pdf.length - 1).toString();
    if (lastChar === character) {
      return pdf.slice(0, pdf.length - 1);
    }

    return pdf;
  }

  /**
   * génére une référence basique à partir d'un nombre 
   * @param {number} index 
   * @returns {string}
   */
  private intToRef(index: number): string {
    return `${index} 0 R`;
  }


  private prepareString(string: string): string {
    const escapableRe = /[\n\r\t\b\f()\\]/g;
    const escapable: any = {
      '\n': '\\n',
      '\r': '\\r',
      '\t': '\\t',
      '\b': '\\b',
      '\f': '\\f',
      '\\': '\\\\',
      '(': '\\(',
      ')': '\\)',
    };

    // // Detect if this is a unicode string
    // for (let i = 0, end = string.length; i < end; i += 1) {
    //   if (string.charCodeAt(i) > 0x7f) {
    //     string = string.substring(0, i) + String.fromCharCode(0) + string.substring(i + 1)
    //     // If so, encode it as big endian UTF-16
    //     // stringBuffer = Buffer.from(`\ufeff${string}`, 'utf16le').swap16();
    //     // break;
    //   }
    // }
    let stringBuffer: Buffer = Buffer.from(string, 'ascii');

    // Escape characters as required by the spec
    return stringBuffer.toString('binary').replace(escapableRe, (c: string) => escapable[c]);
  }
}


