<?php

namespace App\Module\Vitis\WebService\Controller;

use App\Module\Vitis\WebService\Controller\UserController;
use App\Module\Vitis\WebService\Entity\Privilege;
use App\Module\Vitis\WebService\Entity\User;
use App\Module\Vitis\WebService\Entity\UserRequest;
use App\Module\Vitis\WebService\Repository\GenericRepository;
use App\Module\Vitis\WebService\Repository\PrivilegeRepository;
use App\Module\Vitis\WebService\Repository\UserRepository;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\DataAccessor\DatabaseConnection;
use App\Module\Vitis\WebService\Service\FileAccessor\FileManager;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\Mail\Context\AccountContext;
use App\Module\Vitis\WebService\Service\Mail\TemplateRenderer;
use App\Module\Vitis\WebService\Service\Token\JweTokenManager;
use App\Module\Vitis\WebService\Service\VitisRequest;
use Ramsey\Uuid\Uuid;
use Symfony\Component\HttpClient\HttpClient;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller pour gérer un compte utilisateur.
 *
 * @author Ludivine Premont <ludivine.premont@veremes.com>
 */
class AccountController extends UserController {
  /**
   * @var string
   */
  private const FORGOTTEN_PWD_URL = '@prop(app_client_url)/login/forgotten-pwd?datakey=';
  /**
   * @var string
   */
  private const SIGN_UP_URL = '@prop(app_client_url)/login/sign-up?datakey=';
  /**
   * @var string
   */
  private const API_RECAPTCHA_URL = 'https://www.google.com/recaptcha/api/siteverify';
  /**
   * @var JweTokenManager
   */
  private $oJweTokenManager;

  /**
   * @var TemplateRenderer
   */
  private $oTplRenderer;

  /**
   * Initialise le controler et instancie les services requis
   * @param Properties $oPropertiesSrvc Service d'accès aux properties
   * @param VitisRequest $oVitisRequest Service d'abstraction de la requête Symfony
   * @param Route $oRouteConfig Service d'accès à la configuration de la route
   * @param LoggerManager $oLoggerManager Service d'utilsation des logs
   * @param DatabaseConnection $oDatabaseConnection Service d'accès à la base de données
   * @param FileManager $oFileManager Service de gestion des fichiers
   * @param JweTokenManager $oJweTokenManager Service de manipulation des token JWE
   * @param TemplateRenderer $oTplRenderer Service d'évaluation des template de mail
   */
  public function __construct(Properties $oPropertiesSrvc, VitisRequest $oVitisRequest, Route $oRouteConfig, LoggerManager $oLoggerManager, DatabaseConnection $oDatabaseConnection, FileManager $oFileManager, JweTokenManager $oJweTokenManager, TemplateRenderer $oTplRenderer) {
    parent::__construct($oPropertiesSrvc, $oVitisRequest, $oRouteConfig, $oLoggerManager, $oDatabaseConnection, $oFileManager);
    $this->oJweTokenManager = $oJweTokenManager;
    $this->oTplRenderer = $oTplRenderer;
  }

  /**
   * Récupère l'utilisateur, créer un token et l'envoi par email
   * 1er étape: l'email
   * Si l'email est unique dans la bdd => email envoyé
   * Si l'email est trouvé plusieurs fois dans la bdd => étape 2
   * 2eme étape: le login
   * @return Response
   */
  public function forgottenPassword(): Response {
    if ($this->oProperties->get('advanced_login_page') === false) {
      return $this->formatGenericFeatureNotEnabled();
    }

    $aData = $this->requestToArray();

    $user = $this->checkUserInfos($aData);

    if ($user === null) {
      return $this->formatGenericError('LOGINNOTKNOWN', '', 'Can\'t get user', Response::HTTP_BAD_REQUEST);
    }

    // Si il y a un problème avec la récupération de l'utilisateur on renvoit un message d'erreur
    if (count($user) > 1) {
      return $this->formatGenericError('SEVERAL_USER_FOUNDED', '', 'Several user found', Response::HTTP_BAD_REQUEST);
    }

    $sUrl = $this->oProperties->replacePropertiesInString(self::FORGOTTEN_PWD_URL) . $this->getJWEToken($user[0]);

    if (!$this->prepareAndSendMail(AccountContext::FORGOTTEN_PASSWORD, $user[0], $sUrl)) {
      return $this->formatGenericErrorCannotSendMail();
    }

    return $this->formatResponse(['status' => 1, 'data' => $user[0]]);
  }

  /**
   * Vérifie l'email ou le login de l'utilisateur
   * @param array $aData
   * @return array|null
   */
  private function checkUserInfos(array $aData): ?array{
    $oEm = $this->oDatabaseConnection->getAdminEntityManager();
    try {
      $userRepository = $oEm->getRepository(User::class);

      if (!$userRepository instanceof UserRepository) {
        throw new \Exception('Repository is not an UserRepository');
      }

      if (array_key_exists('fpwd_email', $aData)) {
        $user = $userRepository->findByEmail($aData['fpwd_email']);
      } else if (array_key_exists('fpwd_login', $aData)) {
        $user = $userRepository->findByLogin($aData['fpwd_login']);
      } else {
        return null;
      }
    } catch (\Exception $e) {
      $this->oLoggerManager->get('SQL_ERROR')->error('User : can\'t get user with email or login: ' . $e->getMessage());
      return null;
    }

    if (!empty($user)) {
      return $user;
    }

    return null;
  }

  /**
   * Undocumented function
   *
   * @param  string $sLogin
   * @return bool|null
   */
  private function checkSignUpRessourceForLogin(string $sLogin): ?bool{
    $oEm = $this->oDatabaseConnection->getAdminEntityManager();
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEm);

    // check les role postgres
    if ($oQueryAccessor->ifPgRoleExist($sLogin)) {
      return true;
    }
    if ($oQueryAccessor->ifUserExist($sLogin)) {
      return true;
    }
    // check les demandes d'inscription
    $userRepository = $oEm->getRepository(UserRequest::class);

    if (!$userRepository instanceof GenericRepository) {
      throw new \Exception('Repository is not an UserRepository');
    }

    try {
      $user = $userRepository->findById($sLogin);
    } catch (\Exception $e) {
      $this->oLoggerManager->get('SQL_ERROR')->error('User : can\'t get user\'s request with login: ' . $e->getMessage());
      return null;
    }

    if (!empty($user)) {
      return true;
    }

    return false;
  }

  /**
   * Undocumented function
   *
   * @param  string $sLogin
   * @param  int    $iStatus
   * @return bool
   */
  private function updateStatusUserRequest(string $sLogin, int $iStatus): bool {
    $oEm = $this->oDatabaseConnection->getAdminEntityManager();

    try {
      $oEntity = $oEm->find(UserRequest::class, $sLogin);

      $oEntity->setStatus($iStatus);
      $oEntity->setDatakey(null);
      // Sauvegarde en base de donnée
      $oEm->persist($oEntity);
      $oEm->flush();

    } catch (\Exception $e) {
      $this->oLoggerManager->get('SQL_ERROR')->error('User : can\'t update user\'s request status with login: ' . $e->getMessage());
      return false;
    }

    return true;
  }

  /**
   * Génére un token JWE à partir d'un payload porvenant de l'entité User
   * @param array $aUserData Payload provenant de l'entité User
   * @return string Token JWE
   */
  private function getJWEToken(array $aUserData): string {
    // The payload we want to encrypt. It MUST be a string.
    $iIssuedAt = time();
    $iExp = $iIssuedAt + $this->oProperties->get('token_email_fpwd');

    $sIssuer = $this->oProperties->get('app_identifier');
    $sIpAudience = $this->oVitisRequest->getClientIp();
    $sUrlAudience = str_replace('/', '', $this->oVitisRequest->getRequest()->getBaseUrl());
    $sJti = Uuid::uuid1()->toString(); //ex : e4eaaaf2-d142-11e1-b3e4-080027620cdd
    $payload = [
      'iat' => $iIssuedAt,
      'nbf' => $iIssuedAt,
      'exp' => $iExp,
      'iss' => $sIssuer,
      'aud' => $sIpAudience . ':' . $sUrlAudience,
      'sub' => 'Issued by ' . $sIssuer . ' from ' . $sIpAudience . ' on ' . $sUrlAudience . ' at ' . $iIssuedAt . ' for user ' . $aUserData['login'],
      'jti' => $sJti,
      'data' => $aUserData,
    ];

    return $this->oJweTokenManager->encryptToken($payload);
  }

  /**
   * Vérifie que le token soit valide, utilisateur qui existe et l'expiration
   * @return Response
   */
  public function checkToken(): Response {
    if ($this->oProperties->get('advanced_login_page') === false) {
      return $this->formatGenericFeatureNotEnabled();
    }

    $aData = $this->requestToArray();
    $payload = $this->oJweTokenManager->decryptData($aData['datakey']);

    $userRepository = $this->oDatabaseConnection->getAdminEntityManager()->getRepository(User::class);
    $user = $userRepository->find($payload['data']['user_id']);

    // Vérifie la présence de l'utilisateur dans le token
    if (!$user) {
      return $this->formatGenericError('NO_VALID_USER_IN_TOKEN', '', '', Response::HTTP_BAD_REQUEST);
    }

    // Vérifie que le token ne soit pas expiré
    if ($payload['exp'] < time()) {
      return $this->formatGenericError('EXPIRED_TOKEN', '', '', Response::HTTP_BAD_REQUEST);
    }
    return $this->formatResponse(['status' => 1, 'data' => $payload['data']]);
  }

  /**
   * Modifie le mot de passe
   * @return Response
   */
  public function changePassword(): Response {
    // Vérifie que la properties advanced_login_page est valide
    if ($this->oProperties->get('advanced_login_page') === false) {
      return $this->formatGenericFeatureNotEnabled();
    }

    $aData = $this->requestToArray();
    if (!isset($aData['datakey']) || empty($aData['datakey'])) {
      return $this->formatErrorInvalidToken('Missing token');
    }
    $payload = $this->oJweTokenManager->decryptData($aData['datakey']);
    // Vérifie qu'un utilisateur est présent dans le token
    if (!isset($payload['data']) || !isset($payload['data']['user_id'])) {
      return $this->formatErrorInvalidToken();
    }
    // Vérifie que l'utilisateur présent dans le token existe
    $userRepository = $this->oDatabaseConnection->getAdminEntityManager()->getRepository(User::class);
    $user = $userRepository->find($payload['data']['user_id']);
    if (!$user) {
      return $this->formatErrorInvalidToken();
    }

    // Vérifie les données du formulaire (mdp & confirmation mdp)
    if (!$this->checkPassword($aData)) {
      return $this->formatGenericError('PASSWORD_AND_CONFIRM_NO_EQUALS', '', 'Password and confirm password not equal', Response::HTTP_BAD_REQUEST);
    }

    // Récupère la connexion à la base de données
    $this->oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($this->oDatabaseConnection->getAdminEntityManager());
    // Si la connexion à la base de données n'a pas pu être récupéré.
    if (empty($this->oQueryAccessor)) {
      return $this->formatPostgresBadConnection();
    }

    // protège le mot de passe d'après la configuraton
    $sProtectedPassword = $this->oDatabaseConnection->getHashedPassword($payload['data']['login'], $aData['password']);
    // Lance la modification du mot de passe
    $response = $this->oQueryAccessor->updateRolePassword($payload['data']['login'], $sProtectedPassword);

    // Si la requête n'a pas réussi à s'exécuter
    if (!$response) {
      return $this->formatGenericError('ERROR_POSTGRES_USER_PASSWORD_UPDATE', '',
        'Cannot set password for Postgres User', Response::HTTP_UNAUTHORIZED);
    }

    return $this->formatResponse(['status' => 1]);
  }

  /**
   * Vérifie que le mot de passe et la confirmation de mot de passe soient présent et égaux.
   * @param array $aData
   * @return bool
   */
  private function checkPassword(array $aData): bool {
    if (!isset($aData['password']) || !isset($aData['password_confirm'])) {
      return false;
    }

    if (empty($aData['password']) || empty($aData['password_confirm'])) {
      return false;
    }

    if ($aData['password'] !== $aData['password_confirm']) {
      return false;
    }

    return true;
  }

  /**
   * Point d'entré pour la demande d'inscription
   * Fait les vérifications sur le formulaire et envoi le mail pour la suite de la procédure
   * @return Response
   */
  public function signUp(): Response {
    // Vérifie que la properties advanced_login_page est valide
    if ($this->oProperties->get('advanced_login_page') === false) {
      return $this->formatGenericFeatureNotEnabled();
    }

    $aData = $this->requestToArray();

    // Vérifie les champs obligatoires du formualire
    if (!$this->checkSignUpRequiredField($aData)) {
      return $this->formatGenericError('MISSING_REQUIRED_DATA', '', 'Missing required data', Response::HTTP_BAD_REQUEST);
    }

    // Vérifie le captcha
    $captchaData = [
      'secret' => $this->oProperties->get('google_private_captcha'),
      'response' => $aData['captcha'],
    ];

    // Configuration du serveur proxy (http(s)://user:password@ip:port).
    $aHttpClientParams = [];
    $sProxyUrl = $this->oProperties->get('proxy_server_url');
    $sProxyPort = $this->oProperties->get('proxy_server_port');
    if (!empty($sProxyUrl) && !empty($sProxyPort)) {
      $aProxyUrlParams = parse_url($sProxyUrl);
      $sProxy = $aProxyUrlParams['host'] . ':' . $sProxyPort;
      $sProxyLogin = $this->oProperties->get('proxy_server_login');
      $sProxyPassword = $this->oProperties->get('proxy_server_password');
      if (!empty($sProxyLogin) && !empty($sProxyPassword)) {
        $sProxy = $sProxyLogin . ':' . $sProxyPassword . '@' . $sProxy;
      }
      $sProxy = $aProxyUrlParams['scheme'] . '://' . $sProxy;
      $aHttpClientParams['proxy'] = $sProxy;
    }

    // Lance la requête pour vérifier le captcha
    $client = HttpClient::create($aHttpClientParams);
    $response = $client->request('POST', self::API_RECAPTCHA_URL, ['body' => $captchaData]);
    $responseContent = json_decode($response->getContent(), true);

    // Vérifie le success
    if ($response->getStatusCode() != 200 || $responseContent['success'] === false) {
      $sMessage = $responseContent['error-codes'] ?? 'Captcha error';
      return $this->formatGenericError('INVALID_CAPTCHA', '', $sMessage, Response::HTTP_UNAUTHORIZED);
    }

    // Vérifie l'unicité du login
    $user = $this->checkUserInfos(['fpwd_login' => $aData['login']]);
    if ($user !== null) {
      return $this->formatGenericError('LOGINNOTAVAILABLE', '', 'user already exist', Response::HTTP_UNAUTHORIZED);
    }
    if ($this->checkSignUpRessourceForLogin($aData['login'])) {
      return $this->formatGenericError('LOGINNOTAVAILABLE', '', 'user already exist', Response::HTTP_UNAUTHORIZED);
    }

    // Vérifie les domaines de mail interdit
    if (!$this->checkDomainMail($aData['email'])) {
      return $this->formatGenericError('MAILDOMAINFORBIDDEN', '', 'Domain forbidden', Response::HTTP_UNAUTHORIZED);
    }

    // Préparation des données
    $aData['protected_password'] = $this->oDatabaseConnection->getHashedPassword($aData['login'], $aData['password']);
    $aData['password'] = '';
    $aData['password_confirm'] = '';
    $sDataKey = $this->getJWEToken($aData);

    $sUrl = $this->oProperties->replacePropertiesInString(self::SIGN_UP_URL) . $sDataKey;

    // chargement de la connexion à la base de donnée
    $oEntityManager = $this->oDatabaseConnection->getAdminEntityManager();
    $oDataAccessor = $this->getDataAccessor($oEntityManager);

    $aData['datakey'] = $sDataKey;
    $sData = json_encode($aData);
    $aReturn = $oDataAccessor->insertEntity($sData, false);

    if ($aReturn['status'] !== 1) {
      return $this->formatGenericResponseForErrorOnInsert('Cannot store data for admin validation');
    }

    // Cache le nom complet si l'admin doit valider
    if ($this->oProperties->get('automated_sign_up') == false) {
      $aData['name'] = '';
    }

    if (!$this->prepareAndSendMail(AccountContext::SIGN_UP, $aData, $sUrl, true)) {
      return $this->formatGenericErrorCannotSendMail();
    }

    return $this->formatResponse(['status' => 1]);
  }

  /**
   * Vérifie que les paramètres obligatoire pour s'inscrire sont correctement rempli.
   * @param array $aData Données à vérifier
   * @param bool $bCheckPassword Vérification ou non de la confirmation du mot de passe
   * @return boolean
   */
  private function checkSignUpRequiredField(array $aData, bool $bCheckPassword = true): bool {
    if (!isset($aData['login']) || empty($aData['login'])) {
      return false;
    }
    if ($bCheckPassword && !$this->checkPassword($aData)) {
      return false;
    }
    if (!$bCheckPassword && (!isset($aData['protected_password']) || empty($aData['protected_password']))) {
      return false;
    }
    if (!isset($aData['email']) || empty($aData['email'])) {
      return false;
    }
    if (!isset($aData['name']) || empty($aData['name'])) {
      return false;
    }
    if (!isset($aData['captcha']) || empty($aData['captcha'])) {
      return false;
    }

    return true;
  }

  /**
   * Vérifie si le domaine respecte les règles d'obligation et d'interdiction
   * @param string $sEmail
   * @return bool
   */
  private function checkDomainMail(string $sEmail): bool {
    $explodeEmail = explode('@', $sEmail);
    $sDomain = end($explodeEmail);
    // check if granted
    $domainGranted = $this->oProperties->get('sign_up_domain_granted');
    if (!empty($domainGranted)) {
      for ($i = 0; $i < count($domainGranted); $i++) {
        if ($domainGranted[$i] === $sDomain) {
          return true;
        }
      }
      return false;
    }
    // check if forbidden
    $domainForbidden = $this->oProperties->get('sign_up_domain_forbidden');
    if (!empty($domainForbidden)) {
      for ($i = 0; $i < count($domainForbidden); $i++) {
        if ($domainForbidden[$i] === $sDomain) {
          return false;
        }
      }
    }
    return true;
  }

  /**
   * Vérifie le token et lance l'ajout de l'utilisateur.
   * @return Response
   */
  public function confirmSignUp(): Response {
    $payload = $this->checkForConfirmSignUp();
    // en cas d'erreur on retourne la réponse pré-formatée
    if ($payload === null) {
      return $this->oResponse;
    }
    
    // Sauvegarde le nouvel utilisateur
    $oData = $this->createUser($payload['data']);
    
    // en cas d'erreur on retourne la réponse pré-formatée
    if ($oData === null) {
      return $this->oResponse;
    }

    return $this->formatResponse($oData, Response::HTTP_CREATED);
  }

  /**
   * check tous les élément avant de confirmer l'inscription
   *
   * @return array|null donnée pour créer l'utilisateur ou null en cas d'erreur (un réponse HTTP est pré-formatté en cas d'erreur)
   */
  protected function checkForConfirmSignUp(): ?array{
    if ($this->oProperties->get('advanced_login_page') === false) {
      $this->formatGenericFeatureNotEnabled();
      return null;
    }

    $aData = $this->requestToArray();

    if (!isset($aData['datakey']) || empty($aData['datakey'])) {
      $this->formatErrorInvalidToken('Missing token');
      return null;
    }

    $payload = $this->oJweTokenManager->decryptData($aData['datakey']);
    // Vérifie la présence des valeurs requise dans le token
    if (!$this->checkSignUpRequiredField($payload['data'], false)) {
      $this->formatErrorInvalidToken();
      return null;
    }

    // Vérifie que l'utilisateur n'existe pas déjà
    $user = $this->checkUserInfos(['fpwd_login' => $payload['data']['login']]);
    if ($user !== null) {
      $this->formatGenericError('USER_ALREADY_CREATE', '', 'user already exists', Response::HTTP_UNAUTHORIZED);
      return null;
    }

    return $payload;
  }

  /**
   * Undocumented function
   *
   * @param  string $id
   * @return Response
   */
  public function abortSignUp($id): Response {
    return $this->formatResponse(['status' => $this->updateStatusUserRequest($id, 3) ? 1 : 0]);
  }

  /**
   * Créé un utilisateur
   * @param array $aData Données de l'utilisateur à ajouter
   * @return array|null donnée inséré ou null en cas d'erreur (une réponse sera préformaté en cas d'erreur)
   */
  protected function createUser(array $aData): ?array{
    // récupération d'un queryAccessor natif pour les opérations sur les roles
    $this->oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($this->oDatabaseConnection->getAdminEntityManager());
    if (empty($this->oQueryAccessor)) {
      $this->formatPostgresBadConnection();
      return null;
    }
    // création de l'utilisateur dans Postgres
    if (!$this->oQueryAccessor->createUser($aData['login'])) {
      $this->formatGenericError('ERROR_POSTGRES_USER_CREATION', '', 'Cannot create Postgres User with this login', Response::HTTP_UNAUTHORIZED);
      return null;
    }
    // mise à jour du mot de passe
    if (!$this->oQueryAccessor->updateRolePassword($aData['login'], $aData['protected_password'])) {
      $this->formatGenericError('ERROR_POSTGRES_USER_PASSWORD_UPDATE', '', 'Cannot set password for Postgres User', Response::HTTP_UNAUTHORIZED);
      return null;
    }

    // insertion de l'utilisateur dans la table user
    // chargement de la connexion à la base de donnée
    $oDataAccessor = $this->getDataAccessor($this->oDatabaseConnection->getAdminEntityManager());

    // Sauvegarde en base de donnée
    $oResponse = $oDataAccessor->insertEntity(json_encode($aData));
    if ($oResponse['status'] === 0) {
      $this->formatGenericError('ERROR_VITIS_USER_INSERT', '', 'Cannot create Vitis User with this login', Response::HTTP_UNAUTHORIZED);
      return null;
    }

    if ($oResponse['data'] instanceof User) {
      $oUser = $oResponse['data'];
      // ajout des privilèges
      $this->updateUserPrivileges($oUser);
      // ajout des groupes par défaut.
      $this->updateUserGroups($oUser);
    }

    $this->updateStatusUserRequest($aData['login'], 2);

    // Si confirmation par l'admin, on lui envoie un mail pour confirmer la réussite de la création.
    $this->sendConfirmationMail($aData);

    return $oResponse;
  }

  /**
   * Met à jour les privilèges de l'utilisateur en fonction de la properties sign_up_automated_roles + vitis_user
   * @param User $oUser l'entité User à mettre à jour
   * @return bool
   */
  private function updateUserPrivileges(User $oUser): bool {
    $privileges = $this->oProperties->get('sign_up_automated_roles');

    if (!isset($privileges)) {
      return false;
    }
    // Récuperation de PrivilegeRepository
    $privilegeRepository = $this->oDatabaseConnection->getAdminEntityManager()->getRepository(Privilege::class);

    if (!$privilegeRepository instanceof PrivilegeRepository) {
      return false;
    }

    // récupère les nouveaux privilèges
    $aNewPrivileges = $privileges;
    if (!is_array($aNewPrivileges)) {
      $aNewPrivileges = explode('|', $aNewPrivileges);
    }
    // On ajoute toujours le role vitis_user
    if (!in_array('vitis_user', $aNewPrivileges)) {
      $aNewPrivileges[] = 'vitis_user';
    }

    $aRolesToDelete = array_diff($oUser->getPrivileges(), $aNewPrivileges);
    // Supprime des privileges
    if (!$this->removePrivilegesToUser($oUser->getLogin(), $aRolesToDelete)) {
      return false;
    }
    // Passage du tableau de string en tableau d'entité existante
    $aPrivileges = $privilegeRepository->findByRolname($aNewPrivileges);
    if ($this->addPrivilegesToUser($oUser, $aPrivileges)) {
      return false;
    }
    return true;
  }

  /**
   * Met à jour les groupes de l'utilisateur en fonction du paramètre
   * sign_up_automated_groups dans les properties.
   * @param User $oUser l'entité User à mettre à jour
   * @return bool
   */
  private function updateUserGroups(User $oUser): bool {
    $aUserGroups = $this->oProperties->get('sign_up_automated_groups');
    if (!empty($aUserGroups) && is_array($aUserGroups)) {
      // lowercase de la liste de groupe à appliquer à l'utilisateur
      $aLowerUserGroups = array_map(
        function (string $value): string {
          return strtolower($value);
        }
        , $aUserGroups);
      // Récupère les id des groupes à associer à l'utilisataeur.
      $oEm = $this->oDatabaseConnection->getAdminEntityManager();
      $qb = $this->oDatabaseConnection->getQueryAccessor($oEm)->createQueryBuilder();

      $aParamsSelect = [
        'SCHEMA' => ['value' => $this->oProperties->get('schema_vitis'), 'type' => 'schema_name'],
        'aGroupList' => ['value' => $aLowerUserGroups, 'type' => 'array'],
      ];

      $qb->select('a.group_id')
        ->from('[SCHEMA].group', 'a')
        ->where('lower(a.name) IN ([aGroupList])');
      try {
        $aGroups = $qb->executeWithParams($aParamsSelect)->fetchAll(\PDO::FETCH_COLUMN);
      } catch (\Exception $e) {
        $this->oLoggerManager->get('SQL_ERROR')->error('Cannot get groups ' . $e->getMessage());
        return false;
      }
      // Insertion de l'association du groupe à l'utilisateur.
      if (!empty($aGroups)) {
        $aParamsInsert = [
          'SCHEMA' => ['value' => $this->oProperties->get('schema_vitis'), 'type' => 'schema_name'],
        ];

        foreach ($aGroups as $iGroupId) {
          $qb = $this->oDatabaseConnection->getQueryAccessor($oEm)->createQueryBuilder();
          $qb->insert('[SCHEMA].user_group')
            ->setValue('user_id', '[USER_ID]')
            ->setValue('group_id', '[GROUP_ID]');

          $aParamsInsert['USER_ID'] = ['value' => $oUser->getUserId(), 'type' => 'integer'];
          $aParamsInsert['GROUP_ID'] = ['value' => $iGroupId, 'type' => 'integer'];
          try {
            $qb->executeWithParams($aParamsInsert);
          } catch (\Exception $e) {
            $this->oLoggerManager->get('SQL_ERROR')->error('Cannot associate group to user ' . $e->getMessage());
            return false;
          }
        }
      }
    }
    return true;
  }

  /**
   * Envoi un email à l'utilisateur pour confirmé que son compte a été créé
   * @param array $aData
   * @return void
   */
  private function sendConfirmationMail(array $aData) {
    // Si confirmation par l'admin, on lui envoie un mail pour confirmer la réussite de la création.
    if ($this->oProperties->get('automated_sign_up') === false) {
      $sUrl = $this->oProperties->get('app_client_url') . '/login';
      $this->prepareAndSendMail(AccountContext::CONFIRM_SIGN_UP, $aData, $sUrl, true);
    }
  }

  /**
   * Prépare et envoi un email en fonction des paramètres
   * @param string $templateTitle
   * @param array $user user.name, user.login, user.company, user.email
   * @param string $sUrl url ou ''
   * @param bool $addCC default false
   * @return bool
   */
  private function prepareAndSendMail(string $templateTitle, array $user, string $sUrl, bool $addCC = false): bool {
    $oEntity = new AccountContext($this->oLoggerManager, $this->oDatabaseConnection, $this->oProperties, $user, $sUrl);
    // récupèration du template de mail
    $oMailTemplate = $oEntity->getMailTemplate($templateTitle);
    if ($oMailTemplate !== null) {
      // Eval du template avec phantomjs
      $oMail = $this->oTplRenderer->evalTemplateWithContext($oMailTemplate, $oEntity);
      // récupération du mail et envoi
      if ($oMail !== null) {
        if ($addCC === true) {
          // Ajoute le/les cc en properties si nécessaire
          if ($this->oProperties->get('sign_up_CC') !== null) {
            if (is_array($this->oProperties->get('sign_up_CC'))) {
              foreach ($this->oProperties->get('sign_up_CC') as $cc) {
                $oMail->addCc($cc);
              }
            } else if (is_string($this->oProperties->get('sign_up_CC'))) {
              $oMail->addCc($this->oProperties->get('sign_up_CC'));
            }
          }
        }
        return $oMail->send();
      } else {
        $this->oLoggerManager->get('API_ERROR')->error('Email can\'t be eval');
        return false;
      }
    }
    $this->oLoggerManager->get('API_ERROR')->error('Can\'t get email template');
    return false;
  }

  /**
   * Désactivation du compte d'un utilisateur.
   * @param mixed $id id de l'utilisateur
   * @return Response réponse à retourner par l'API
   */
  public function unsubscribe($id): Response {
    $aResponse = [
      'status' => 1,
    ];

    // chargement de la connexion à la base de donnée (superuser).
    $oAdminEntityManager = $this->oDatabaseConnection->getAdminEntityManager();

    // Récupération d'un queryAccessor natif pour les opérations sur les roles.
    $this->oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oAdminEntityManager);
    if (empty($this->oQueryAccessor)) {
      return $this->formatPostgresBadConnection();
    }

    // Récupère le login de l'utilisateur.
    $oRepository = $this->oDatabaseConnection->getUserEntityManager()->getRepository(User::class);

    $aEntities = [];
    if ($oRepository instanceof UserRepository) {
      $aEntities = $oRepository->findByIdInRange([$id]);
    }

    if (empty($aEntities)) {
      return $this->formatGenericError('ERROR_UNSUBSCRIBE_USER_NOT_FOUND', '',
        "Cannot unsubscribe the user id $id : not found in database.", Response::HTTP_UNAUTHORIZED);
    }

    // Un compte administrateur ne peut pas être désactivé.
    $oUser = $aEntities[0];
    $sUserLogin = $oUser->getLogin();
    if ($oUser->isAdmin()) {
      return $this->formatGenericError('ERROR_UNSUBSCRIBE_ADMINISTRATOR_ACCOUNT', '',
        "An administrator's account cannot be disabled ('$sUserLogin').", Response::HTTP_UNAUTHORIZED);
    }

    // Révoque le privilège "vitis_user" pour l'utilisateur.
    if ($this->removePrivilegesToUser($sUserLogin, ['vitis_user']) === false) {
      return $this->formatGenericError('ERROR_UNSUBSCRIBE_USER_REVOKE', '',
        "Cannot revoke privilege 'vitis_user' to user '$sUserLogin'.", Response::HTTP_UNAUTHORIZED);
    }

    return $this->formatResponse($aResponse);
  }

  /**
   * @param  string   $sMessage
   * @return Response
   */
  protected function formatErrorInvalidToken(string $sMessage = 'Invalid token (datakey)'): Response {
    return $this->formatGenericError('NO_VALID_USER_IN_TOKEN', '', $sMessage, Response::HTTP_BAD_REQUEST);
  }
}
