<?php

namespace App\Module\Vitis\WebService\Controller;

use App\Module\Vitis\WebService\Controller\GenericController;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\DataAccessor\DatabaseConnection;
use App\Module\Vitis\WebService\Service\FileAccessor\FileManager;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\VitisRequest;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Serializer\Normalizer\AbstractNormalizer;

/**
 * Controller pour les traductions
 */
class CustomTranslationController extends GenericController {
  /**
   * Initialise le controler et instancie les services requis
   * @param Properties $oPropertiesSrvc Service d'accès aux properties
   * @param VitisRequest $oVitisRequest Service d'abstraction de la requête Symfony
   * @param Route $oRouteConfig Service d'accès à la configuration de la route
   * @param LoggerManager $oLoggerManager Service d'utilsation des logs
   * @param DatabaseConnection $oDatabaseConnection Service d'accès à la base de données
   * @param FileManager $oFileManager Service de gestion des fichiers
   */
  public function __construct(Properties $oPropertiesSrvc, VitisRequest $oVitisRequest, Route $oRouteConfig, LoggerManager $oLoggerManager, DatabaseConnection $oDatabaseConnection, FileManager $oFileManager) {
    parent::__construct($oPropertiesSrvc, $oVitisRequest, $oRouteConfig, $oLoggerManager, $oDatabaseConnection, $oFileManager);
  }

  /**
   * Récupère toutes les entités pour une route de type GET Module/Objet.
   * @return Response réponse à retourner par l'API
   */
  public function getTranslations(): Response {
    $aData = $this->requestToArray();

    $oEntityManager = $this->oDatabaseConnection->getAdminEntityManager();

    // chargement de la connexion à la base de donnée
    $oDataAccessor = $this->getDataAccessor($oEntityManager);
    $aEntities = $oDataAccessor->getEntities($aData);

    if ($aEntities['status'] == 0) {
      return $this->formatGenericResponseForErrorOnSelect();
    }

    $aSerializerContext = $this->getSerializerContext();

    // transforme le tableau de réponse dans le résultat souhaité
    return $this->formatResponse($aEntities, Response::HTTP_OK, $aSerializerContext);
  }

  /**
   * Récupère une entité pour une route de type GET Module/Objet/{{id}}.
   * @param mixed $id id de l'entité à trouver
   * @return Response réponse à retourner par l'API
   */
  public function getData($id): Response {
    $aReturn = [];
    // chargement de la connexion à la base de donnée
    $oDataAccessor = $this->getDataAccessor();
    // Si $id = numérique => custom_translation_id (sinon => translation_id).
    if (is_numeric($id)) {
      $aData = $this->requestToArray();
      $oEntity = $oDataAccessor->getEntity($id, $aData);
      if ($oEntity['status'] == 0) {
        $oEntity['message'] = 'Error during get';
        $oEntity['errorMessage'] = 'ERROR_GENERIC_CONTROLLER_GET';
        return $this->formatResponse($oEntity, Response::HTTP_UNAUTHORIZED);
      }
      //
      $aSerializerContext = [];
      if (isset($aData['attributs'])) {
        $aSerializerContext[AbstractNormalizer::ATTRIBUTES] = explode('|', $aData['attributs']);
      }
      return $this->formatResponse($oEntity, Response::HTTP_OK, $aSerializerContext);
    } else {
      // Données de la traduction personnalisée.
      $id = strtoupper($id);
      $qb = $this->oDatabaseConnection->getUserEntityManager()->getConnection()->createQueryBuilder();
      $qb->select('*')
        ->from($this->oProperties->get('schema_vitis') . '.custom_translation', 'ct')
        ->where($qb->expr()->eq('ct.translation_id', ':translation_id'))
        ->setParameter('translation_id', $id);
      try {
        $aResult = $qb->execute()->fetchAllAssociative();
        $aReturn = [
          'status' => 1,
          'data' => [
            'translation_id' => $id,
          ],
        ];
        // Récupère les données de la traduction personnalisée si existante.
        if (!empty($aResult)) {
          $aReturn['data']['custom_translation_id'] = $aResult[0]['custom_translation_id'];
          foreach ($aResult as $aRow) {
            $aReturn['data']['text_' . $aRow['lang']] = $aRow['text'];
          }
        }
        return $this->formatResponse($aReturn);
      } catch (\Exception $e) {
        $this->oLoggerManager->get('SQL_ERROR')->error('Error getting s_vitis.custom_translation rows' . $e->getMessage());
        $aReturn['status'] = 0;
        $aReturn['message'] = 'Error selecting custom translation';
        $aReturn['errorMessage'] = 'REQUEST_ERROR';
        return $this->formatResponse($aReturn, Response::HTTP_INTERNAL_SERVER_ERROR);
      }
    }
  }

  /**
   * Met à jour une entité pour une route de type PUT Module/Objet/{{id}}.
   * @param mixed $id id de l'entité à trouver
   * @return Response réponse à retourner par l'API
   */
  public function updateData($id): Response {
    $aReturn = [];
    $aData = $this->requestToArray();
    // Suppression de la traduction personnalisée dans toutes les langues.
    $qb = $this->oDatabaseConnection->getUserEntityManager()->getConnection()->createQueryBuilder();
    $qb->delete($this->oProperties->get('schema_vitis') . '.custom_translation', 'ct')
      ->where($qb->expr()->eq('ct.translation_id', ':translation_id'))
      ->setParameter('translation_id', $aData['translation_id']);
    try {
      $qb->execute();
      // Insertion de la traduction personnalisée dans toutes les langues.
      $aLang = ['fr', 'en'];
      foreach ($aLang as $sLang) {
        $qb = $this->oDatabaseConnection->getUserEntityManager()->getConnection()->createQueryBuilder();
        $qb->insert($this->oProperties->get('schema_vitis') . '.custom_translation')
          ->setValue('translation_id', '?')
          ->setValue('text', '?')
          ->setValue('lang', '?')
          ->setParameter(0, $aData['translation_id'])
          ->setParameter(1, $aData["text_$sLang"])
          ->setParameter(2, $sLang);
        try {
          $qb->execute();
        } catch (\Exception $e) {
          $this->oLoggerManager->get('SQL_ERROR')->error('Error inserting s_vitis.custom_translation row' . $e->getMessage());
          $aReturn['status'] = 0;
          $aReturn['message'] = 'Error inserting custom translation';
          $aReturn['errorMessage'] = 'REQUEST_ERROR';
          return $this->formatResponse($aReturn, Response::HTTP_INTERNAL_SERVER_ERROR);
        }
      }
    } catch (\Exception $e) {
      $this->oLoggerManager->get('SQL_ERROR')->error('Error deleting s_vitis.custom_translation rows' . $e->getMessage());
      $aReturn['status'] = 0;
      $aReturn['message'] = 'Error deleting custom translations';
      $aReturn['errorMessage'] = 'REQUEST_ERROR';
      return $this->formatResponse($aReturn, Response::HTTP_INTERNAL_SERVER_ERROR);
    }
    return $this->formatResponse(['status' => 1]);
  }

  /**
   * Supprime les entités d'une liste pour une route de type POST Module/Objet.
   * @return Response réponse à retourner par l'API
   */
  public function deleteDatas(): Response {
    $aReturn = [];
    $aData = $this->massiveRequestGetIdList();
    // Id de traduction entre quotes.
    $aIdList = array_map([$this, 'array_map_callback'], $aData['idList']);
    // Suppression des traductions dans toutes les langues.
    $qb = $this->oDatabaseConnection->getUserEntityManager()->getConnection()->createQueryBuilder();
    $qb->delete($this->oProperties->get('schema_vitis') . '.custom_translation', 'ct')
      ->where($qb->expr()->in('ct.translation_id', $aIdList));
    try {
      $qb->execute();
    } catch (\Exception $e) {
      $this->oLoggerManager->get('SQL_ERROR')->error('Error deleting s_vitis.custom_translation rows' . $e->getMessage());
      $aReturn['status'] = 0;
      $aReturn['message'] = 'Error deleting custom translations';
      $aReturn['errorMessage'] = 'REQUEST_ERROR';
      return $this->formatResponse($aReturn, Response::HTTP_INTERNAL_SERVER_ERROR);
    }
    return $this->formatResponse(['status' => 1]);
  }

  /**
   * ajoute des slashes sur l'élément
   *
   * @param  string|integer $n
   * @return string
   */
  private function array_map_callback($n): string {
    return "'$n'";
  }
}
