<?php

namespace App\Module\Vitis\WebService\Controller;

use Doctrine\ORM\EntityManager;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller Générique de Vitis.
 *
 * @author Ludivine Premont <ludivine.premont@veremes.com>
 */
class GenericQueriesController extends GenericController {

  /**
   * Permet de récupérer les base de données locales
   * @return Response
   */
  public function getDatabases(): Response {
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor();

    // Si la connexion à la base de données n'a pas pu être récupéré.
    if (empty($oQueryAccessor)) {
      return $this->formatPostgresBadConnection();
    }

    $aData = $this->requestToArray();
    // Récupération des bases de données
    $oDatabases = $oQueryAccessor->getDatabases($aData);
    if (!isset($oDatabases)) {
      return $this->formatResponse(['status' => 0]);
    }

    $response = ['status' => 1, 'data' => []];
    foreach ($oDatabases as $oDatabase) {
      $response['data'][] = $oDatabase['database'];
    }

    return $this->formatResponse($response);
  }

  /**
   * Permet de récupérer les schémas d'une base de données locale
   * @param mixed $databaseName Nom de la base de données pour laquelle on veut récupérer les schémas
   * @return Response réponse à retourner par l'API
   */
  public function getSchemas($databaseName) {
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($this->oDatabaseConnection->createExternalEntityManager($databaseName));

    // Si la connexion à la base de données n'a pas pu être récupéré.
    if (empty($oQueryAccessor)) {
      return $this->formatPostgresBadConnection();
    }

    $aData = $this->requestToArray();
    // Récupération des schemas
    $oSchemas = $oQueryAccessor->getSchemas($aData);
    if (!isset($oSchemas)) {
      return $this->formatResponse(['status' => 0]);
    }

    $response = ['status' => 1, 'data' => []];
    foreach ($oSchemas as $oSchema) {
      $response['data'][] = $oSchema['schema_name'];
    }

    return $this->formatResponse($response);
  }

  /**
   * Permet de récupérer les tables d'un schema d'une base de données locale
   * @return Response réponse à retourner par l'API
   */
  public function getTables() {
    $aData = $this->requestToArray();

    if (!isset($aData['database']) || empty($aData['database']) || !isset($aData['schema']) || empty($aData['schema'])) {
      return $this->formatMissingParametersError();
    }

    $oEntityManager = $this->getEntityManagerFromData($aData);
    // Si la connexion à la base de données n'a pas pu être récupéré.
    if (!isset($oEntityManager) || empty($oEntityManager)) {
      return $this->formatPostgresBadConnection();
    }

    if (isset($aData['sgbd']) && !empty($aData['sgbd']) && $aData['sgbd'] == 'pdo_oci') {
      $oQueryAccessor = $this->oDatabaseConnection->getOracleAccessor($oEntityManager);
    } else {
      $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEntityManager);
    }
    // Si la connexion à la base de données n'a pas pu être récupéré.
    if (empty($oQueryAccessor)) {
      return $this->formatPostgresBadConnection();
    }

    // Récupération des tables
    $oTables = $oQueryAccessor->getTables($aData);
    if (!isset($oTables)) {
      return $this->formatResponse(['status' => 0]);
    }

    $response = ['status' => 1, 'data' => []];
    foreach ($oTables as $oTable) {
      $response['data'][] = $oTable['table_name'];
    }

    return $this->formatResponse($response);
  }

  /**
   * Permet de récupérer des données de n'importe quel base de données
   * @param array|null $aData
   * @return Response réponse à retourner par l'API
   */
  public function getGeneric(?array $aData = null) : Response {

    if (!isset($aData)) {
      $aData = $this->requestToArray();
    }

    if (!isset($aData['database']) || empty($aData['database'])) {
      $aData['database'] = $this->oProperties->get('db_name');
    }

    if (!isset($aData['table']) || empty($aData['table']) ||
      !isset($aData['database']) || empty($aData['database']) ||
      !isset($aData['schema']) || empty($aData['schema'])) {
      return $this->formatMissingParametersError();
    }

    $oEntityManager = $this->getEntityManagerFromData($aData);
    // Si la connexion à la base de données n'a pas pu être récupéré.
    if (empty($oEntityManager)) {
      return $this->formatPostgresBadConnection();
    }

    if (isset($aData['sgbd']) && !empty($aData['sgbd']) && $aData['sgbd'] == 'pdo_oci') {
      $oQueryAccessor = $this->oDatabaseConnection->getOracleAccessor($oEntityManager);
    } else {
      $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEntityManager);
    }
    // Si la connexion à la base de données n'a pas pu être récupéré.
    if (empty($oQueryAccessor)) {
      return $this->formatPostgresBadConnection();
    }

    $aFields = $oQueryAccessor->getFieldsFromTable($aData);
    if ($aFields['status'] == 0) {
      return $this->formatResponse($aFields);
    }

    // Si columns_type défini, gère les données spécifiques
    if (isset($aData['columns_type'])) {
      for ($i=0; $i < count($aFields['data']); $i++) {
        $sName = $aFields['data'][$i]['name'];
        if (isset($aData['columns_type'][$sName])) {
          switch ($aData['columns_type'][$sName]) {
            case 'image':
              if (isset($aData['file_path'])) {
                // Modifie le type
                $aFields['data'][$i]['type'] = 'file';
                $aFields['data'][$i]['image'] = true;
                $aFields['data'][$i]['multiple'] = true;
                // Ajoute l'url
                $aFields['data'][$i]['path'] = $aData['file_path'];
              }
              break;
            case 'document':
              if (isset($aData['file_path'])) {
                // Modifie le type
                $aFields['data'][$i]['type'] = 'file';
                $aFields['data'][$i]['image'] = false;
                $aFields['data'][$i]['multiple'] = true;
                // Ajoute l'url
                $aFields['data'][$i]['path'] = $aData['file_path'];
              }
              break;
            case 'geom':
              $aFields['data'][$i]['type'] = 'geom';
              break;
            default:
              break;
          }
        }
      }
    }

    $options = [
      'schema' => $aData['schema'],
      'table' => $aData['table'],
      'fields' => $aFields['data'],
    ];

    if (isset($aData['sgbd']) && !empty($aData['sgbd']) && $aData['sgbd'] == 'pdo_oci') {
      $options['field_id'] = $aFields['data'][0]['name'];
    }

    // Si on connait le champs identifiant on le passe à la config et on l'ajoute au attributs
    if (isset($aData['field_id'])) {
      $options['field_id'] = $aData['field_id'];
      if (isset($aData['attributs'])) {
        $aData['attributs'] .= '|' . $aData['field_id'];
      }
    }

    $this->oRouteConfig->setOptionsParameter($options);

    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEntityManager);
    // Si la connexion à la base de données n'a pas pu être récupéré.
    if (empty($oQueryAccessor)) {
      return $this->formatPostgresBadConnection();
    }

    // Ajout du filtre si besoin
    if (isset($aData['filter_value']) && isset($aData['filter_attr']) && isset($aData['values'])) {

      if (!isset($aData['values'][$aData['filter_value']]) || $aData['values'][$aData['filter_value']] == '') {
        return $this->formatMissingParametersError('La valeur de filter_value ('. $aData['filter_value'] .') n\'est pas présente dans les valeurs');

      }

      // Ajouter le filtre
      $aData['filter'] = json_encode([
        'relation' => 'AND',
        'operators' => [[
          'column' => $aData['filter_attr'],
          'compare_operator' => 'IN',
          'value' => $aData['values'][$aData['filter_value']],
        ]],
      ]);
    }

    // Récupération des tables
    $response = $oQueryAccessor->getEntities($aData);

    return $this->formatResponse($response);
  }

  /**
   * Permet de récupérer les données de n'importe quel column de n'importe quel bdd
   * @param array|null $aData
   * @return Response réponse à retourner par l'API
   */
  public function getDatasColumn(?array $aData = null) : Response {

    if (!isset($aData)) {
      $aData = $this->requestToArray();
    }

    if (!isset($aData['database']) || empty($aData['database'])) {
      $aData['database'] = $this->oProperties->get('db_name');
    }

    if (!isset($aData['table']) || empty($aData['table']) ||
        !isset($aData['database']) || empty($aData['database']) || 
        !isset($aData['schema']) || empty($aData['schema'])) {
      return $this->formatMissingParametersError();

    }

    $oEntityManager = $this->getEntityManagerFromData($aData);
    // Si la connexion à la base de données n'a pas pu être récupéré.
    if (empty($oEntityManager)) {
      return $this->formatPostgresBadConnection();
    }

    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEntityManager);
    // Si la connexion à la base de données n'a pas pu être récupéré.
    if (empty($oQueryAccessor)) {
      return $this->formatPostgresBadConnection();

    }
    
    $sDatas = $oQueryAccessor->getDatasFromColumn($aData);

    if ($sDatas['status'] == 0) {
      return $this->formatResponse($sDatas);
    }

    return $this->formatResponse($sDatas);
  }

  /**
   * Récupère l'entity manager nécessaire en fonction des paramètres.
   *
   * @param mixed $aData
   * @return EntityManager|null EntityManager
   */
  protected function getEntityManagerFromData($aData): ?EntityManager {
    // si la clé n'existe pas on la set à null pour le reste
    if (!array_key_exists('database', $aData)){
      $aData['database'] = null;
    }

    if (!empty($aData['login']) && !empty($aData['password']) &&
        !empty($aData['server']) && !empty($aData['port']) &&
        !empty($aData['sgbd'])) {

      $oEntityManager = $this->oDatabaseConnection->createExternalEntityManager(
        $aData['database'],
        $aData['login'],
        $aData['password'],
        $aData['sgbd'],
        $aData['server'],
        $aData['port'],
        $aData['charset'] ?? null
      );
    } else {
      $oEntityManager = $this->oDatabaseConnection->createExternalEntityManager($aData['database']);
    }

    return $oEntityManager;
  }

  /**
   * @param string $sMessage message de log
   * @return Response
   */
  protected function formatMissingParametersError(string $sMessage = ''): Response {
    return $this->formatGenericError('MISSING_PARAMS', '', $sMessage, Response::HTTP_OK);
  }
}
