<?php

namespace App\Module\Vitis\WebService\Controller;

use App\Module\Vitis\WebService\Controller\GenericController;
use App\Module\Vitis\WebService\Entity\Privilege;
use App\Module\Vitis\WebService\Entity\User;
use App\Module\Vitis\WebService\Repository\PrivilegeRepository;
use App\Module\Vitis\WebService\Repository\UserRepository;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller Générique de Vitis.
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 * @author Alexandre Mironenkov <alexandre.mironenkov@veremes.com>
 */
class PrivilegeController extends GenericController {
  /**
   * @var \App\Module\Vitis\WebService\Service\DataAccessor\QueryAccessor
   */
  private $oQueryAccessor;

  /**
   * Créé un utilisateur et y associe des utilisateurs
   * @return Response réponse à retourner par l'API
   */
  public function insertPrivilege(): Response {
    // récupération du contenu de la requête
    $aData = $this->requestToArray();
    // récupération d'un queryAccessor natif pour les opérations sur les roles
    $this->oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor();
    if (empty($this->oQueryAccessor)) {
      return $this->formatGenericError('ERROR_POSTGRES_PRIVILEGE_BAD_CONNECTION', '', 'Cannot access to database', Response::HTTP_UNAUTHORIZED);
    }
    // création de l'utilisateur dans Postgres
    if (!$this->oQueryAccessor->createRole($aData['rolname'])) {
      return $this->formatGenericError('ERROR_POSTGRES_PRIVILEGE_CREATION', '',
        'Cannot create Postgres User with this login', Response::HTTP_UNAUTHORIZED);
    }
    // insertion de l'utilisateur dans la table user
    // chargement de la connexion à la base de donnée
    $oDataAccessor = $this->getDataAccessor();
    // Sauvegarde en base de donnée
    $oResponse = $oDataAccessor->insertEntity($this->requestToJson());
    if ($oResponse['status'] === 0) {
      return $this->formatGenericError('ERROR_VITIS_PRIVILEGE_INSERT', '',
        'Cannot create Vitis Role with this login', Response::HTTP_UNAUTHORIZED);
    }
    // si des utilisateurs sont passés on lance les grants
    if (isset($aData['users'])) {
      // si c'est un string on explode
      $aUsers = (is_array($aData['users'])) ? $aData['users'] : explode('|', $aData['users']);
      $oResponse['usersGranted'] = $this->grantUsers($aData['rolname'], $aUsers);
    }

    return $this->formatResponse($oResponse, Response::HTTP_CREATED);
  }

  /**
   * Met à jour un privileges et les utilisateurs qui y sont associés
   * @param string $id nom du role à mettre à jour
   * @return Response
   */
  public function updatePrivilege(string $id): Response {
    // récupération du contenu de la requête
    $aData = $this->requestToArray();
    // récupération d'un queryAccessor natif pour les opérations sur les roles
    $this->oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor();
    if (empty($this->oQueryAccessor)) {
      return $this->formatPostgresBadConnection();
    }
    // mise à jour de la table dans vitis
    $oDataAccessor = $this->getDataAccessor();
    // Sauvegarde en base de donnée
    $oResponse = $oDataAccessor->updateEntity($id, $this->requestToJson());
    if ($oResponse['status'] === 0) {
      $sMessage = 'Cannot update Vitis Role ' . $id;
      return $this->formatGenericError('ERROR_VITIS_PRIVILEGE_UPDATE', '', $sMessage, Response::HTTP_UNAUTHORIZED);
    }

    // si des utilisateurs sont passés on lance les grants/revoke
    if (isset($aData['users'])) {
      // revoke sur tout le monde
      $aUserToRevoke = $oResponse['data']->getUsers();
      if (in_array($id, ['vitis_user', 'vitis_admin'])) {
        // retrait de l'utilisateur courant pour vitis_user et vitis_admin
        // il doit toujours y avoir un admin capable de se connecter
        $iCurrentUserId = $this->getCurrentUserEntity()->getUserId();
        $aUserToRevoke = array_diff($aUserToRevoke, [$iCurrentUserId]);
      }
      $oResponse['usersRevoked'] = $this->revokeUsers($id, $aUserToRevoke);
      // si c'est un string on explode
      $aUsers = (is_array($aData['users'])) ? $aData['users'] : explode('|', $aData['users']);
      $oResponse['usersGranted'] = $this->grantUsers($id, $aUsers);
    }

    return $this->formatResponse($oResponse);
  }

  /**
   * Supprime des privileges et revoke les utilisateurs qui y sont associés
   * @return Response
   */
  public function deletePrivileges(): Response {
    // récupération du contenu de la requête
    $aData = $this->massiveRequestGetIdList();

    $aRoleList = ($aData === null) ? [] : $aData['idList'];

    return $this->revokeAndDeletePrivilege($aRoleList);
  }

  /**
   * Supprime un privilege et revoke les utilisateurs qui y sont associés
   * @param string $id nom du role à mettre à jour
   * @return Response
   */
  public function deletePrivilege(string $id): Response {
    // passage format array
    $aRoleList = [$id];

    return $this->revokeAndDeletePrivilege($aRoleList);
  }

  /**
   * Supprime une liste de privileges et revoke les utilisateurs qui y sont associés
   * @param string[] $aRoleList nom du role à mettre à jour
   * @return Response
   */
  private function revokeAndDeletePrivilege(array $aRoleList): Response {
    if (empty($aRoleList)) {
      return $this->formatGenericError('ERROR_VITIS_PRIVILEGE_DELETE_EMPTY_ID_LIST', '',
        'Cannot find Vitis Role(s)', Response::HTTP_NOT_FOUND);
    }

    // récupération d'un queryAccessor natif pour les opérations sur les roles
    $this->oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor();
    if (empty($this->oQueryAccessor)) {
      return $this->formatPostgresBadConnection();
    }

    // chargement de la connexion à la base de donnée
    $oRepository = $this->oDatabaseConnection->getUserEntityManager()->getRepository(Privilege::class);
    if (!$oRepository instanceof PrivilegeRepository) {
      return $this->formatGenericError('ERROR_VITIS_PRIVILEGE_DELETE_NOT_FOUND', '',
        'Cannot find Vitis Roles, bad repository', Response::HTTP_NOT_FOUND);
    }

    // find users
    $aEntities = $oRepository->findByIdInRange($aRoleList);
    if (empty($aEntities)) {
      return $this->formatGenericError('ERROR_VITIS_PRIVILEGE_DELETE_NOT_FOUND', '',
        'Cannot find Vitis Roles', Response::HTTP_NOT_FOUND);
    }
    // gestion des roles dans postgres
    $aLockedRoles = [];
    $aFailedDrop = [];
    foreach ($aEntities as $oPrivilege) {
      $sRoleName = $oPrivilege->getRolname();
      // on n'efface pas les roles vérrouiller
      if ($oPrivilege->getLocked()) {
        $aLockedRoles[] = $sRoleName;
        continue;
      }
      // on revoke tous les utilisateurs
      $this->revokeUsers($sRoleName, $oPrivilege->getUsers());
      // on drop le role postgres
      // ne fonctionnera pas si le role est utilisé ailleurs
      if (!$this->oQueryAccessor->dropRole($sRoleName)) {
        $aFailedDrop[] = $sRoleName;
      }
    }
    // on retire les roles protégés pour ne pas les effacer
    $aRoleList = array_diff($aRoleList, $aLockedRoles);
    // on retire les roles qui n'ont pas pû être drop
    $aRoleList = array_diff($aRoleList, $aFailedDrop);
    // nettoyage de la table vitis
    $oDataAccessor = $this->getDataAccessor();
    $aEntities = $oDataAccessor->deleteEntities($aRoleList);

    if (!empty($aFailedDrop)) {
      $sMessage = 'Cannot drop Postgres Roles : ' . implode(',', $aFailedDrop);
      return $this->formatGenericError('ERROR_VITIS_PRIVILEGE_DELETE_DROP_ROLE', '', $sMessage, Response::HTTP_UNAUTHORIZED);
    }

    if (!empty($aLockedRoles)) {
      $sMessage = 'Cannot drop locked Roles : ' . implode(',', $aLockedRoles);
      return $this->formatGenericError('ERROR_VITIS_PRIVILEGE_DELETE_DROP_LOCKED_ROLE', '', 
        $sMessage, Response::HTTP_UNAUTHORIZED);
    }

    return $this->formatResponse($aEntities);
  }

  /**
   * @param string $sRolname nom du role
   * @param int[] $aUsers Liste des id utilisateur à grant
   * @return bool true si mise à jour réussie false sinon
   */
  private function grantUsers(string $sRolname, array $aUsers): bool {
    $aUserLogin = $this->getArrayLogins($aUsers);

    if (empty($aUserLogin)) {
      return false;
    }

    // call query Accessor
    return $this->oQueryAccessor->grantRoleToRoles($sRolname, $aUserLogin);
  }

  /**
   * @param string $sRolname nom du role
   * @param int[] $aUsers Liste des id utilisateur à grant
   * @return bool true si mise à jour réussie false sinon
   */
  private function revokeUsers(string $sRolname, array $aUsers): bool {
    $aUserLogin = $this->getArrayLogins($aUsers);

    if (empty($aUserLogin)) {
      return false;
    }

    // call query Accessor
    return $this->oQueryAccessor->revokeRoleFromRoles($sRolname, $aUserLogin);
  }

  /**
   * @param int[] $aUsers Liste des id utilisateur à grant
   * @return string[] array de login des utilsiateurs
   */
  private function getArrayLogins(array $aUsers): array{
    // récupération du repository
    $oRepository = $this->oDatabaseConnection->getUserEntityManager()->getRepository(User::class);
    if (!$oRepository instanceof UserRepository) {
      return [];
    }

    // find users
    $aEntities = $oRepository->findByIdInRange($aUsers);

    if (empty($aEntities)) {
      return [];
    }

    // génère un tableau de login
    $aUserLogin = [];
    foreach ($aEntities as $oUser) {
      $aUserLogin[] = $oUser->getLogin();
    }
    return $aUserLogin;
  }

}
