<?php

namespace App\Module\Vitis\WebService\Controller;

use App\Module\Vitis\WebService\Controller\AbstractVitisController;
use App\Module\Vitis\WebService\Entity\User;
use App\Module\Vitis\WebService\Repository\UserRepository;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\DataAccessor\DatabaseConnection;
use App\Module\Vitis\WebService\Service\FileAccessor\FileManager;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\Token\JweTokenManager;
use App\Module\Vitis\WebService\Service\VitisRequest;
use Ramsey\Uuid\Uuid;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller permettant de gérer les token de connexion à l'application.
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 * @author Armand Bahi <armand.bahi@veremes.com>
 */
class TokenController extends AbstractVitisController {
  /**
   * @var JweTokenManager
   */
  protected $oJweTokenManager;

  /**
   * Initialise le controler et instancie les services requis
   * @param Properties $oPropertiesSrvc Service d'accès aux properties
   * @param VitisRequest $oVitisRequest Service d'abstraction de la requête Symfony
   * @param Route $oRouteConfig Service d'accès à la configuration de la route
   * @param LoggerManager $oLoggerManager Service d'utilsation des logs
   * @param DatabaseConnection $oDatabaseConnection Service d'accès à la base de données
   * @param FileManager $oFileManager Service de gestion des fichiers
   * @param JweTokenManager $oJweTokenManager Service de manipulation des token JWE
   */
  public function __construct(Properties $oPropertiesSrvc, VitisRequest $oVitisRequest, Route $oRouteConfig, LoggerManager $oLoggerManager, DatabaseConnection $oDatabaseConnection, FileManager $oFileManager, JweTokenManager $oJweTokenManager) {
    parent::__construct($oPropertiesSrvc, $oVitisRequest, $oRouteConfig, $oLoggerManager, $oDatabaseConnection, $oFileManager);
    $this->oJweTokenManager = $oJweTokenManager;
  }

  /**
   * Génère un token, utiliser par la route vitis_create_token.
   * exemple de corp de la requête :
   *      {
   *          "login" : "string",
   *          "password" : "string"
   *      }
   * @return Response réponse à retourner par l'API
   */
  public function generateToken(): Response {
    // preset de la strcuture de réponse
    $aReturn = ['status' => 0];

    // récupération login password
    $oData = $this->requestToArray();

    // chargement de la connexion à la base de donnée
    $oEntityManager = $this->oDatabaseConnection->getAdminEntityManager();

    // vérification de l'existence de l'utilisateur
    $oUserRepository = $oEntityManager->getRepository(User::class);

    $aUsers = [];
    if ($oUserRepository instanceof UserRepository) {
      $aUsers = $oUserRepository->findByLoginWithoutCase($oData['user']);
    }

    if (count($aUsers) == 1) {
      // retransformation en objet
      $aUser = $oUserRepository->findOneBy(['login' => $aUsers[0]['login']]);
      // controle de la restriction d'IP
      if (!$this->checkIpConstraint($aUser->getIpConstraint())) {
        $sClientIp = $this->oVitisRequest->getClientIp();
        $this->oLoggerManager->get('API_ERROR')->error('Error with token generation for ' . $aUser->getLogin() . ', this user try to connect from ' . $sClientIp, [$aUser]);
        // L'utilisateur ne peut pas se connecter avec cette ip
        $sMessage = 'Connection forbidden from ip :' . $sClientIp;
        return $this->formatGenericError('ERROR_USER_IP_UNAUTHORIZED', '', $sMessage, Response::HTTP_UNAUTHORIZED);
      }

      // Initialisation de la structure de l'utilisateur
      $aPayload = $aUser->getTokenPayload();

      // récupère le mot de passe pour se conncter à postgres
      // à la connexion il n'est jamais retouché
      if (empty($oData['password'])) {
        $this->oLoggerManager->get('API_ERROR')->error('Password Hash failed for ' . $aUser->getLogin());
        // Le mot de passe n'a pas pu être hash
        return $this->formatGenericError('BAD_HASH_METHOD_FOR_DB_PASSWORD', '',
          'Your password cannot be hashed, contact application\'s administrator', Response::HTTP_NOT_IMPLEMENTED);
      } else {
        // Création du token
        $aTokenStruct = $this->getJWEToken($this->getJWETokenPayload([
          'user' => $aPayload,
          'password' => $oData['password'],
        ]));

        // Test de connexion avec le token
        $aUserTest = $this->getCurrentUserEntity($aTokenStruct['token']);
        if (empty($aUserTest)) {
          // L'utilisateur n'a pas pu être trouvé
          return $this->formatGenericError('ERROR_USER_NOT_EXIST', '', 'This user does not exist', Response::HTTP_UNAUTHORIZED);
        }

        // Insertion des information du token dans les données de l'utilisateur
        $aPayload = array_merge($aPayload, $aTokenStruct);
        if (empty($aPayload['token'])) {
          // Le token n'a pas pu être généré
          $this->oLoggerManager->get('API_ERROR')->error('Token generation failed for ' . $aUser->getLogin(), $aPayload);
          return $this->formatGenericError('ERROR_EMPTY_TOKEN_PAYLOAD', '', 'Problem with Token Payload', Response::HTTP_SERVICE_UNAVAILABLE);

        } else {
          $this->oLoggerManager->get('CONNEXION_INFO')->info('Token generation for user ' . $aUser->getLogin(), $aPayload);
          // mise à jour de la date de dernière connexion
          $aUser->updateLastConnection();
          $oEntityManager->persist($aUser);
          $oEntityManager->flush();
          // Ajout du status et on retourne la requête
          $aReturn['data'] = $aPayload;
          $aReturn['status'] = 1;
          return $this->formatResponse($aReturn);
        }
      }
    }
    $this->oLoggerManager->get('API_ERROR')->error('Error with token generation for ' . $oData['user'] . ', this user doesn\'t exist');
    // L'utilisateur n'a pas pu être trouvé
    return $this->formatGenericError('ERROR_USER_NOT_EXIST', '', 'This user does not exist', Response::HTTP_UNAUTHORIZED);
  }

  /**
   * Retourne les infos du token (Utilisateur, date de fin etc...)
   * @return Response réponse à retourner par l'API
   */
  public function getToken(): Response {
    $aReturn = ['status' => 0];

    // Token
    $sToken = $this->oVitisRequest->getToken();

    $tokenUpdated = false;
    if (Uuid::isValid($sToken)) {
      // Si la chaine token est un uuid alors on récupère le token
      $oEm = $this->oDatabaseConnection->getAdminEntityManager();
      $qb = $this->oDatabaseConnection->getQueryAccessor($oEm)->createQueryBuilder();
      $qb->select('a.token')
        ->from('[SCHEMA].login_token', 'a')
        ->where($qb->expr()->eq('a.token_id', '[TOKEN]'));

      $aParams = [
        'SCHEMA' => ['value' => $this->oProperties->get('schema_vitis'), 'type' => 'schema_name'],
        'TOKEN' => ['value' => $sToken, 'type' => 'string'],
      ];

      $oToken = $qb->executeWithParams($aParams)->fetchAll();

      $tokenUpdated = true;
      $sToken = $oToken[0]['token'];

      $aTokenJWEDecrypt = $this->oJweTokenManager->decryptToken($sToken);
      if (empty($aTokenJWEDecrypt) ||
        (isset($aTokenJWEDecrypt['data']['ip_restriction']) &&
          !empty($aTokenJWEDecrypt['data']['ip_restriction']) &&
          !$this->checkIpConstraint($aTokenJWEDecrypt['data']['ip_restriction']))) {
        $sClientIp = $this->oVitisRequest->getClientIp();
        // L'utilisateur ne peut pas se connecter avec cette ip
        $sMessage = 'Connection forbidden from ip :' . $sClientIp;
        return $this->formatGenericError('ERROR_USER_IP_UNAUTHORIZED', '', $sMessage, Response::HTTP_UNAUTHORIZED);
      }
      
    }
    
    if (empty($sToken)) {
      return $this->formatGenericError('ERROR_MISSING_AUTHORIZATION_HEADER', '', 'Authorization header is missing', Response::HTTP_UNAUTHORIZED);
    }

    // Token décrypté
    $aTokenJWEDecrypt = $this->oJweTokenManager->decryptToken($sToken);

    // Roles si l'utilisateur est connecté
    $oUser = $this->getCurrentUserEntity($sToken);
    if ($oUser !== null) {
      $aReturn['data'] = [
        'privileges' => $oUser->getRoles(),
        'user' => $oUser->getLogin(),
        'user_id' => $oUser->getUserId(),
        'validity_date' => $aTokenJWEDecrypt['exp'],
      ];

      if ($tokenUpdated) {
        $aReturn['data']['token'] = $sToken;
      }

      $aReturn['status'] = 1;
      // log en info
      $this->oLoggerManager->get('CONNEXION_INFO')->info('Get Token for user ' . $aReturn['data']['user'], $aReturn);
      // chargement de la connexion à la base de donnée admin
      // pour passer en update toute les opérations doivent être faite avec le même EntityManager
      // or getcurrentUserEntity utilise le UserEntityManager
      // d'où la double requête
      $oEntityManager = $this->oDatabaseConnection->getAdminEntityManager();
      // vérification de l'existence de l'utilisateur
      $aUser = $oEntityManager->getRepository(User::class)->findOneBy(['login' => $oUser->getLogin()]);
      // mise à jour de la date de dernière connexion
      $aUser = $aUser->updateLastConnection();
      $oEntityManager->persist($aUser);
      $oEntityManager->flush();
      // Retour
      $this->formatResponse($aReturn);
      return $this->oResponse;
    } else {
      $this->oLoggerManager->get('API_ERROR')->error('Get Token user not found : ' . $sToken);
      // session expiré ou structure du payload non conforme
      return $this->formatGenericErrorSessionExpired();
    }
  }

  /**
   * Génére un token JWE à partir d'un payload porvenant de l'entité User
   * @param array $payload Payload pour le token jwe
   * @return array struct :
   *                  - return.token : Token JWE
   *                  - return.exp : Timestamp d'expiration en ms
   *                  - return.iat : Timestamp de création en ms
   */
  protected function getJWEToken(array $payload): array{
    return [
      'token' => $this->oJweTokenManager->encryptToken($payload), // token JWE
      'exp' => ($payload['exp'] * 1000), // timestamp d'expiration en ms
      'iat' => ($payload['iat'] * 1000), // timestamp de création en ms
    ];
  }

  /**
   * Génére payload pour jwetoken à partir d'un payload porvenant de l'entité User
   * @param array $aUserData Payload provenant de l'entité User
   * @return array struct :
   *                  - return.token : Token JWE
   *                  - return.exp : Timestamp d'expiration en ms
   *                  - return.iat : Timestamp de création en ms
   */
  protected function getJWETokenPayload(array $aUserData) {
    // The payload we want to encrypt. It MUST be a string.
    $iIssuedAt = time();
    $iExp = $iIssuedAt + $this->oProperties->get('token_ttl');
    $sIssuer = $this->oProperties->get('app_identifier');
    $sIpAudience = $this->oVitisRequest->getClientIp();
    $sUrlAudience = str_replace('/', '', $this->oVitisRequest->getRequest()->getBaseUrl());
    $sJti = Uuid::uuid1()->toString(); //ex : e4eaaaf2-d142-11e1-b3e4-080027620cdd
    $payload = [
      'iat' => $iIssuedAt,
      'nbf' => $iIssuedAt,
      'exp' => $iExp,
      'iss' => $sIssuer,
      'aud' => $sIpAudience . ':' . $sUrlAudience,
      'sub' => 'Issued by ' . $sIssuer . ' from ' . $sIpAudience . ' on ' . $sUrlAudience . ' at ' . $iIssuedAt . ' for user ' . $aUserData['user']['login'],
      'jti' => $sJti,
      'data' => $aUserData,
    ];

    return $payload;
  }

}
