<?php

namespace App\Module\Vitis\WebService\Controller;

use App\Module\Vitis\WebService\Controller\AbstractVitisController;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\DataAccessor\DatabaseConnection;
use App\Module\Vitis\WebService\Service\FileAccessor\FileManager;
use App\Module\Vitis\WebService\Service\FileAccessor\LfsService;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\VitisRequest;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller qui permet de récupérer et d'intéragir avec un treeview.
 *
 * @author Ludivine Premont <ludivine.premont@veremes.com>
 *
 * La route devra contenir id et form
 * id : identifiant de l'enregistrement
 *
 * Exemple :
 * "route": "/v2/gtf/workspaces/{id}/delete",
 *
 * Dans les options l'attribut dir_path est requis
 * dir_path peut contenir
 * - @prop(nom_properties) => remplacer par un properties
 * - @value(idenfiant_valeur) => remplacer par une valeur du formdata de la requête
 * Dans les options l'attribut keep_only_ext n'est pas requis
 *
 * Exemple :
 * "dir_path": "@prop(workspace_dir)/@id/fme"
 *
 * Exemple:
 * "dir_path": "@value(path)"
 * "keep_only_ext": {
 *   "properties": "datasource_authorized_files"
 *   "value": "authorized_files"
 * }
 *
 */
class TreeController extends AbstractVitisController {
  /**
   * Liste des caractères à échapé pour l'ajout d'un dossier
   * @var array
   */
  private const LIST_ESCAPE_CARACTERE = ['\\', '../', '/..', './'];

  /**
   * Clé utilisée pour l'id dans les chemins
   * @var string
   */
  private const ID_PATH_KEY = '@id';

  /**
   * @var LfsService
   */
  protected $oLfsService;

  /**
   * Initialise le controler et instancie les services requis
   * @param Properties $oPropertiesSrvc Service d'accès aux properties
   * @param VitisRequest $oVitisRequest Service d'abstraction de la requête Symfony
   * @param Route $oRouteConfig Service d'accès à la configuration de la route
   * @param LoggerManager $oLoggerManager Service d'utilsation des logs
   * @param DatabaseConnection $oDatabaseConnection Service d'accès à la base de données
   * @param FileManager $oFileManager Service de gestion des fichiers
   */
  public function __construct(Properties $oPropertiesSrvc, VitisRequest $oVitisRequest, Route $oRouteConfig, LoggerManager $oLoggerManager, DatabaseConnection $oDatabaseConnection, FileManager $oFileManager) {
    parent::__construct($oPropertiesSrvc, $oVitisRequest, $oRouteConfig, $oLoggerManager, $oDatabaseConnection, $oFileManager);
    $this->oLfsService = $oFileManager->getLfsService();
  }

  /**
   * get workspace directory tree
   * @param mixed $id identifiant du projet fme
   * @return Response
   */
  public function getTree($id = null): Response {
    $sDir = $this->getDirPath($id);
    $aKeepOnlyExt = $this->getKeepOnlyExt();
    $aData = $this->requestToArray();

    $iDepth = -1; // Par défaut toute la profondeur de l'arborescence
    if (isset($aData['depth']) && !empty($aData['depth'])) {
      $iDepth = $aData['depth'];
    }

    $aTree = $this->oFileManager->getAccessorForPath($sDir)->getFolderInfos($sDir, $aKeepOnlyExt, [], false, $iDepth);
    $aReturn = [];
    if ($aTree === null) {
      return $this->formatGenericError('TREE_VIEW_ERROR_LOADING_FILES', '', '', Response::HTTP_NOT_FOUND);
    }
    $aReturn = [
      'status' => 1,
      'data' => $aTree,
    ];

    return $this->formatResponse($aReturn);
  }

  /**
   * Supprime un fichier ou un dossier du répertoire du projet
   * @param mixed $id identifiant du projet fme
   * @param array $aForbiddenToDelete
   * @return Response
   */
  public function deleteTreeDirectoryOrFile($id = null, $aForbiddenToDelete = []): Response {
    $aData = $this->requestToArray();

    $sDir = $this->getDirPath($id);

    if (!empty($aData['idList'])) {
      $aPathsToDelete = is_array($aData['idList']) ? $aData['idList'] : explode('|', $aData['idList']);
      foreach ($aPathsToDelete as $sPath) {
        // Vérifie si le chemin/fichier ne fait pas partie de la liste des interdits
        if (in_array('$sPath', $aForbiddenToDelete)) {
          return $this->formatTreeViewErrorDeletingFile('You don\'t have right to delete this file');
        }

        // suppression de tous les charactères permettant de remonter l'arborecence
        foreach (self::LIST_ESCAPE_CARACTERE as $value) {
          $sDir = str_replace($value, '/', $sDir);
          $sPath = str_replace($value, '/', $sPath);
        }

        $sTruePath = $sDir . $sPath;
        // realpath sur shared dir pour éviter tout conflit de / \ entre windows et linux
        if (strpos($sTruePath, $sDir) !== 0) {
          return $this->formatTreeViewErrorDeletingFile('you don\'t have right to delete files outside of directory');

        }

        $oFileAccessor = $this->getFileAccessor($sTruePath);
        if ($oFileAccessor->file_exists($sTruePath) && $sTruePath !== $sDir) {
          if ($oFileAccessor->is_dir($sTruePath)) {
            $oFileAccessor->clearDir($sTruePath);
          } else {
            $oFileAccessor->unlink($sTruePath);
          }
        } else {
          return $this->formatTreeViewErrorDeletingFile('you don\'t have right to delete files outside of directory');
        }
      }
      return $this->formatResponse(['status' => 1], Response::HTTP_OK);
    }
    return $this->formatGenericResponseForMissingIdList();
  }

  /**
   * Création d'un répertoire dans le dossier du projet FME.
   * @param mixed $id id de l'entité à trouver
   * @return Response
   */
  public function createTreeFolder($id = null): Response {
    $aData = $this->requestToArray();

    // Test si le chemin est vide
    if (empty($aData['directory_name'])) {
      $sMessage = 'createDirectoryFolder : directory name is missing';
      return $this->formatTreeViewErroDirCreation($sMessage, Response::HTTP_BAD_REQUEST);
    }
    // suppression de tous les charactères permettant de remonter l'arborecence
    foreach (self::LIST_ESCAPE_CARACTERE as $value) {
      $aData['directory_name'] = str_replace($value, '/', $aData['directory_name']);
      $aData['directory_path'] = str_replace($value, '/', $aData['directory_path']);
    }

    $sDirPath = $this->getDirPath($id) . '/' . $aData['directory_path'];
    $sDirPath .= '/' . $aData['directory_name'];
    $oFileAccessor = $this->getFileAccessor($sDirPath);
    // Test si le dossier existe
    if ($oFileAccessor->file_exists($sDirPath)) {
      $sMessage = 'Directory already exists : ' . $sDirPath;
      return $this->formatTreeViewErroDirCreation($sMessage, Response::HTTP_INTERNAL_SERVER_ERROR);
    }

    // Créer le dossier
    if (!$oFileAccessor->mkdir($sDirPath, 0777, true)) {
      $sMessage = 'createDirectoryFolder : directory creation error ' . $sDirPath;
      return $this->formatTreeViewErroDirCreation($sMessage, Response::HTTP_NOT_FOUND);
    }

    return $this->formatResponse(['status' => 1]);
  }

  /**
   * @param mixed $id
   * @return Response
   */
  public function uploadTreeFile($id = null): Response {
    $aData = $this->requestToArray();
    $aRequestOptions = $this->getRequestParameters();
    $sDirPath = $this->getDirPath($id) . '/' . $aData['directory_path'];
    $sDirPath = str_replace('//', '/', $sDirPath);
    $iLimitation = -1;
    if (array_key_exists('file_size_limitation', $aRequestOptions['options'])){
      $iLimitation = (is_string($aRequestOptions['options']['file_size_limitation'])) ? $this->oProperties->replacePropertiesInString($aRequestOptions['options']['file_size_limitation']) : $aRequestOptions['options']['file_size_limitation'];
    }

    error_log($iLimitation);
    error_log(print_r($aRequestOptions, true));

    $sFileTmpName = $_FILES['file']['tmp_name'];
    $sFileName = $_FILES["file"]["name"];

    // taille max autorisé
    if (!$this->getLocalFileAccessor()->checkFileSize($sFileTmpName, $iLimitation)) {
      return $this->formatGenericError('ERROR_FILE_UPLOAD_SIZE_GREATER_THAN_LIMIT', '', '', Response::HTTP_INTERNAL_SERVER_ERROR);
    }

    // extension non autorisé
    if (!$this->getLocalFileAccessor()->checkExtension($sDirPath . '/' . $sFileName)) {
      return $this->formatGenericError('ERROR_FILE_UPLOAD_FORBIDDEN_EXTENSION', '', '', Response::HTTP_INTERNAL_SERVER_ERROR);
    }

    $oFileAccessor = $this->getFileAccessor($sDirPath);
    if (!$oFileAccessor->copyLocalToDistant($sFileTmpName, $sDirPath . '/' . $sFileName)) {
      return $this->formatGenericError('ERROR_FILE_UPLOAD', '', '', Response::HTTP_INTERNAL_SERVER_ERROR);
    } else {
      if (isset($aData['unzip']) && $aData['unzip'] == 'true' && strtolower(pathinfo($sFileName, PATHINFO_EXTENSION)) == "zip") {
        $oFileAccessor->unZip($sDirPath . '/' . $sFileName, $sDirPath . '/');
        // On supprime le zip après
        $oFileAccessor->unlink($sDirPath . '/' . $sFileName);
      }

      if (array_key_exists('transaction', $_FILES['file'])) {
        // nettoyage de la transaction
        $this->oLfsService->clearTransaction($_FILES['file']['transaction']);
      }

      return $this->formatResponse(['status' => 1]);
    }
  }

  /**
   * Retourne le chemin du dossier
   * @param mixed $id
   * @return string
   */
  protected function getDirPath($id): string {
    $aRequestOptions = $this->getRequestParameters();

    $sPath = $aRequestOptions['options']['dir_path'];
    // Remplacement des properties
    $sPath = $this->oProperties->replacePropertiesInString($sPath);
    $sPath = $this->replaceValueInString($sPath);

    if (isset($id)) {
      // Remplacement de l'id
      $sPath = str_replace(self::ID_PATH_KEY, $id, $sPath);
    }

    return $sPath;
  }

  /**
   * Retourne une valeur en fonction de sa clé
   * @param string $sString Chaine à évaluer pour remplacement des bind
   * @return string chaine avec valeur remplacé
   */
  private function replaceValueInString(string $sString, array $aParentReplacer = []): string {
    $sRegexCapture = '/@value\(([^)]*)\)/';
    $aMatches = [];
    preg_match_all($sRegexCapture, $sString, $aMatches);
    $aData = $this->requestToArray();

    foreach ($aMatches[1] as $sDataKey) {
      $sString = str_replace('@value(' . $sDataKey . ')', $aData[$sDataKey], $sString);
    }

    return $sString;
  }

  /**
   * Récupère par une properties et/ou par une valeur dans la requete la liste des extensions autorisé
   *
   * @return array tableau des extensions à garder
   */
  protected function getKeepOnlyExt(): array{
    $aRequestOptions = $this->getRequestParameters();
    if (!isset($aRequestOptions['options'])) {
      return [];
    }

    // Par défaut: toutes les extensions
    $aKeepOnlyExt = [];
    $aParamsKeepOnlyExt = $aRequestOptions['options']['keep_only_ext'] ?? [];
    if (isset($aParamsKeepOnlyExt) && !empty($aParamsKeepOnlyExt)) {
      // Si une properties
      if (isset($aParamsKeepOnlyExt['properties']) && !empty($aParamsKeepOnlyExt['properties'])) {
        if ($this->oProperties->get($aParamsKeepOnlyExt['properties']) !== null && is_array($this->oProperties->get($aParamsKeepOnlyExt['properties']))) {
          $aKeepOnlyExt = array_merge($aKeepOnlyExt, $this->oProperties->get($aParamsKeepOnlyExt['properties']));
        }
      }
      // Si une valeur dans data
      if (isset($aParamsKeepOnlyExt['value']) && !empty($aParamsKeepOnlyExt['value'])) {
        $aData = $this->requestToArray();
        if (isset($aData[$aParamsKeepOnlyExt['value']]) && is_array($aData[$aParamsKeepOnlyExt['value']])) {
          $aKeepOnlyExt = array_merge($aKeepOnlyExt, $aData[$aParamsKeepOnlyExt['value']]);
        }
      }
    }

    return $aKeepOnlyExt;
  }

  /**
   * @param string $sMessage
   * @return Response
   */
  protected function formatTreeViewErrorDeletingFile(string $sMessage = ''): Response {
    return $this->formatGenericError('TREE_VIEW_ERROR_LOADING_FILES', '', $sMessage, Response::HTTP_FORBIDDEN);
  }

  /**
   * @param string $sMessage
   * @return Response
   */
  protected function formatTreeViewErroDirCreation(string $sMessage = '', int $iHttpCode): Response {
    $this->oLoggerManager->get('API_ERROR')->error($sMessage);
    return $this->formatGenericError('ERROR_DIRECTORY_CREATION', '', $sMessage, $iHttpCode);
  }

}
