<?php

namespace App\Module\Vitis\WebService\Repository;

use App\Module\Vitis\WebService\Entity\Privilege;
use App\Module\Vitis\WebService\Entity\User;
use Doctrine\ORM\EntityManagerInterface;
use Doctrine\ORM\Mapping\ClassMetadata;

/**
 * Repository pour les utilisateurs de Vitis.
 *
 * @method User|null find($id, $lockMode = null, $lockVersion = null)
 * @method User|null findOneBy(array $criteria, array $orderBy = null)
 * @method User[]    findAll()
 * @method User[]    findBy(array $criteria, array $orderBy = null, $limit = null, $offset = null)
 *
 * @author Alexandre Mironenkov   <alexandre.mironenkov@veremes.com>
 */
class UserRepository extends GenericRepository {
  /**
   * {@inheritdoc}
   */
  public function __construct(EntityManagerInterface $em, ClassMetadata $class) {
    parent::__construct($em, $class);
  }

  /**
   * Retourne les Utilisateurs sous forme d'un array et non objet pour le role domandé
   * @param Privilege $oPrivilege l'utilisateur
   * @return array
   * @throws \Exception
   */
  public function findByPrivilege(Privilege $oPrivilege): array{
    $oMetaData = $this->getClassMetadata();
    $sUserTable = $oMetaData->getSchemaName() . '.' . $oMetaData->getTableName();
    $conn = $this->_em->getConnection();
    /* @var $em EntityManager */
    $qb1 = $conn->createQueryBuilder();
    $qb1->select('a.user_id')
      ->from($sUserTable, 'a')
      ->join('a', 'pg_user', 'b', $qb1->expr()->eq('a.login', 'b.usename'))
      ->join('b', 'pg_group', 'c', $qb1->expr()->eq('b.usesysid', 'any(c.grolist)'))
      ->where($qb1->expr()->eq('c.groname', $conn->quote($oPrivilege->getRolname())))
      ->groupBy('a.user_id, a.login');

    $aResults = $qb1->execute()->fetchAll(\PDO::FETCH_COLUMN);

    return $aResults;
  }

  /**
   * Retourne l'utilisateur récupéré avec son login
   * @param string $sLogin Login de l'utilisateur
   * @return array
   * @throws \Exception
   */
  public function findByLogin(string $sLogin): array{
    $oMetaData = $this->getClassMetadata();
    $sUserTable = $oMetaData->getSchemaName() . '.' . $oMetaData->getTableName();
    $conn = $this->_em->getConnection();
    /* @var $em EntityManager */
    $qb1 = $conn->createQueryBuilder();
    $qb1->select('a.*')
      ->from($sUserTable, 'a')
      ->where('a.login = :login')
      ->setParameter('login', $sLogin);

    $aResults = $qb1->execute()->fetchAll();

    return $aResults;
  }

  /**
   * vrai recherche par login
   *
   * @param  string    $sLogin
   * @return User|null
   */
  public function findByLoginEntity(string $sLogin): ?User{
    $qb = $this->createQueryBuilder('a');
    $qb->where($qb->expr()->eq('a.login', ':sLogin'))
      ->setParameter('sLogin', $sLogin);

    $oResult = $qb->getQuery()->getResult();

    if (empty($oResult)) {
      return null;
    }

    if (!$oResult[0] instanceof User) {
      return null;
    }

    return $oResult[0];

  }

  /**
   * Retourne l'utilisateur récupéré avec son login sans tenir conmpte de la casse
   * @param string $sLogin Login de l'utilisateur
   * @return array
   * @throws \Exception
   */
  public function findByLoginWithoutCase(string $sLogin): array{
    $oMetaData = $this->getClassMetadata();
    $sUserTable = $oMetaData->getSchemaName() . '.' . $oMetaData->getTableName();
    $conn = $this->_em->getConnection();
    /* @var $em EntityManager */
    $qb1 = $conn->createQueryBuilder();
    $qb1->select('a.*')
      ->from($sUserTable, 'a')
      ->where('LOWER(a.login) = LOWER(:login)')
      ->setParameter('login', $sLogin);

    $aResults = $qb1->execute()->fetchAll();

    return $aResults;
  }

  /**
   * Retourne l'utilisateur récupéré avec son email
   * @param string $sEmail Email de l'utilisateur
   * @return array
   * @throws \Exception
   */
  public function findByEmail(string $sEmail): array{
    $oMetaData = $this->getClassMetadata();
    $sUserTable = $oMetaData->getSchemaName() . '.' . $oMetaData->getTableName();
    $conn = $this->_em->getConnection();
    /* @var $em EntityManager */
    $qb1 = $conn->createQueryBuilder();
    $qb1->select('a.*')
      ->from($sUserTable, 'a')
      ->where('a.email = :email')
      ->setParameter('email', $sEmail);

    $aResults = $qb1->execute()->fetchAll();

    return $aResults;
  }

  /**
   * trouve l'utilisateur logger s'il existe dans la liste des utilisateurs
   *
   * @return User|null
   */
  public function findCurrentUser(): ?User{
    $conn = $this->_em->getConnection();
    /* @var $em EntityManager */
    $qb1 = $conn->createQueryBuilder();
    $qb1->select('"current_user"()');

    $sLogin = $qb1->execute()->fetchOne();

    if (empty($sLogin)) {
      return null;
    }

    $oUser = $this->findByLoginEntity($sLogin);

    if ($oUser instanceof User) {
      return $oUser;
    }

    return null;
  }
}
