<?php

namespace App\Module\Vitis\WebService\Service\DataAccessor\Fields;

use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\Crypto\JweCrypto;
use App\Module\Vitis\WebService\Service\DataAccessor\Fields\AbstractCustomField;
use App\Module\Vitis\WebService\Service\DataAccessor\Fields\FieldInterface;
use App\Module\Vitis\WebService\Service\FileAccessor\FileManager;
use App\Module\Vitis\WebService\Service\FileAccessor\LocalAccessor;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;

/**
 * Classe pour surcharger les champs de type fichier.
 * remplace les fichiers par l'url d'accès, gère l'upload, ...
 *
 *  Configuration spécifique du champ
 *    "type"      (string) : prend toujours la valeur 'file' pour ce type
 *    "name"      (string) : Nom du champ
 *    "path"      (string) : Chemin à utiliser pour le fichier @prop permet d'utiliser une properties, @id permet d'utiliser la colonne identifiant de la table, @value permet d'utiliser la valeur du champ, @column permet d'utiliser le nom du champ
 *    "multiple"  (bool)   : défini si le champ peut contenir plusieurs fichiers
 *    "image"     (bool)   : défini si le champ est une image
 *    "width"     (int)    : si le champ est une image définit la largeur pour redimensionner l'image
 *    "height"    (int)    : si le champ est une image définit la hauteur pour redimensionner l'image
 *    "thumbnail" (bool)   : si c'est une image défini s'il faut générer une miniature
 *
 *  exemple :
 *    {
 *      "type": "file",
 *      "name": "my_file",
 *      "path": "@prop(ws_data_dir)/module/objet/@id/field/@value",
 *      "multiple": false,
 *      "image": true,
 *      "width": 500,
 *      "height": 500,
 *      "thumbnail": true
 *    }
 *
 * @author Anthony Borghi   <anthony.borghi@veremes.com>
 */
class FileField extends AbstractCustomField implements FieldInterface {
  /**
   * @var string
   * Valeur du champs type dans route.json
   */
  public const TYPE = 'file';
  /**
   * nom de l'attribut contenant la liste des fichiers à effacer pour les champs de type filepicker
   * @var string
   */
  private const VITIS_MULTIFILE_RMLIST_KEY = 'vitis_deleted_files';
  /**
   * nom de l'attribut contenant la liste des fichiers à conserver pour les champs de type filepicker
   * @var string
   */
  private const VITIS_MULTIFILE_UNCHANGELIST_KEY = 'vitis_unchanged_files';
  /**
   * nom de l'attribut contenant la liste des fichiers dont les miniatures sont à générer pour les champs de type filepicker
   * @var string
   */
  private const VITIS_MULTIFILE_THUMBNAILLIST_KEY = 'vitis_thumbnail_image_files';
  /**
   * Clé utilisée pour l'id dans les chemins
   * @var string
   */
  private const ID_PATH_KEY = '@id';
  /**
   * Clé utilisée pour la valeur du champ dans les chemins
   * @var string
   */
  private const VALUE_PATH_KEY = '@value';
  /**
   * Clé utilisée pour la valeur du champ dans les chemins
   * @var string
   */
  private const COLUMN_PATH_KEY = '@column';
  /**
   * @var array
   * Permet de forcer une valeur sur le champs s'il n'est pas défini
   * pour les fonctions utilisé comme clé
   */
  protected const FORCE_ATTRIBUTE = [
    'deleteFieldValue' => true,
  ];
  /**
   * @var string
   * valeur à utiliser pour forcer le champs si besoin
   */
  protected const FORCE_VALUE = 'fake_file_name';

  /**
   * @var FileManager
   */
  private $oFileManager;

  /**
   * @var LocalAccessor
   */
  private $oLocalFileInterface;

  /**
   * {@inheritdoc}
   */
  public function __construct(Properties $oPropertiesSrvc, Route $oRoute, LoggerManager $oLoggerManager, JweCrypto $oJweCrypto) {
    parent::__construct($oPropertiesSrvc, $oRoute, $oLoggerManager);
    // récupération de l'interface fichier
    $this->oFileManager = new FileManager($this->oProperties, $this->oLoggerManager, $oJweCrypto);
    $this->oLocalFileInterface = $this->oFileManager->getLocalFileAccessor();
  }

  /**
   * {@inheritdoc}
   */
  public function getFieldValue(array $aEntity, array $aRequestData, array $aFieldOptions): array{
    $oLogger = $this->oLoggerManager->get('API_ERROR');
    $bIsFileMultiple = (isset($aFieldOptions['multiple']) ? $aFieldOptions['multiple'] : false);

    if (!isset($aEntity[$aFieldOptions['name']])) {
      return $aEntity;
    }

    // génération de la liste de fichier à traiter
    $aFiles = ($bIsFileMultiple) ? explode('|', $aEntity[$aFieldOptions['name']]) : [$aEntity[$aFieldOptions['name']]];
    $aFieldPath = [];
    foreach ($aFiles as $sFile) {
      // calcul du chemin
      $sPath = $this->pathReplacer($aEntity, $aFieldOptions, $sFile);
      // si le chemin à pu être généré
      if (!empty($sPath)) {
        // changement de la valeur du champ
        $sUrl = $this->oFileManager->getAccessorForPath($sPath)->getProxyPassUrl($sPath);
        if (!empty($sUrl)) {
          $aFieldPath[] = $sUrl;
        }
      } else {
        $oLogger->warning('cannot generate download url for field : ' . $aFieldOptions['name']);
        $aFieldPath = [];
        break;
      }
    }

    $aEntity[$aFieldOptions['name']] = (empty($aFieldPath)) ? null : implode(',', $aFieldPath);

    return $aEntity;
  }

  /**
   * {@inheritdoc}
   */
  public function insertFieldValue(array $aEntity, array $aRequestData, array $aFieldOptions): array{
    $bIsFileMultiple = (isset($aFieldOptions['multiple']) ? $aFieldOptions['multiple'] : false);
    $bIsImage = (isset($aFieldOptions['image']) ? $aFieldOptions['image'] : false);

    $aFieldValues = $this->getUnchangedFiles($aRequestData, $aFieldOptions['name']);

    if (isset($_FILES[$aFieldOptions['name']])) {
      // si les fichier multiple sont acceptés et que le champ contient plusieurs fichiers
      if ($bIsFileMultiple && isset($_FILES[$aFieldOptions['name']][0])) {
        // on ajoute les fichier à upload
        foreach ($_FILES[$aFieldOptions['name']] as $aFileStruct) {
          $sFileName = $aFileStruct['name'];
          if ($bIsImage && isset($aFieldOptions['width']) && isset($aFieldOptions['height'])) {
            if (strpos($sFileName, ".") > -1) {
              $aPath = explode(".", $sFileName);
              $aPath[count($aPath) - 1] = "jpg";
              $sFileName = implode(".", $aPath);
            }
          }
          $aFieldValues[] = $sFileName;
        }
      } else {
        $sFileName = $_FILES[$aFieldOptions['name']]['name'];
        if ($bIsImage && isset($aFieldOptions['width']) && isset($aFieldOptions['height'])) {
          if (strpos($sFileName, ".") > -1) {
            $aPath = explode(".", $sFileName);
            $aPath[count($aPath) - 1] = "jpg";
            $sFileName = implode(".", $aPath);
          }
        }
        $aFieldValues[] = $sFileName;
      }
    }

    $aEntity[$aFieldOptions['name']] = (empty($aFieldValues)) ? null : implode('|', $aFieldValues);

    return $aEntity;
  }

  /**
   * {@inheritdoc}
   */
  public function afterInsertFieldValue(array $aEntity, array $aRequestData, array $aFieldOptions): array{
    $oLogger = $this->oLoggerManager->get('API_ERROR');

    $bIsFileMultiple = (isset($aFieldOptions['multiple']) ? $aFieldOptions['multiple'] : false);
    // récupération des fichiers inchangés
    $aFieldValues = $this->getUnchangedFiles($aRequestData, $aFieldOptions['name']);

    if (isset($_FILES[$aFieldOptions['name']])) {
      // si les fichier multiple sont acceptés et que le champ contient plusieurs fichiers
      if ($bIsFileMultiple && isset($_FILES[$aFieldOptions['name']][0])) {
        // on ajoute les fichier à upload
        foreach ($_FILES[$aFieldOptions['name']] as $aFileStruct) {
          $sPath = $this->uploadFile($aEntity, $aFieldOptions, $aFileStruct, $aRequestData);
          if (!empty($sPath)) {
            $aFieldValues[] = $this->oLocalFileInterface->getFileName($sPath);
          } else {
            $oLogger->error('cannot write file (' . $aFileStruct['name'] . ') for field ' . $aFieldOptions['name']);
          }
        }
      } else {
        $sPath = $this->uploadFile($aEntity, $aFieldOptions, $_FILES[$aFieldOptions['name']], $aRequestData);
        if (!empty($sPath)) {
          $aFieldValues[] = $this->oLocalFileInterface->getFileName($sPath);
        }
      }
    }

    $aEntity[$aFieldOptions['name']] = (empty($aFieldValues)) ? null : implode('|', $aFieldValues);

    return $aEntity;
  }

  /**
   * {@inheritdoc}
   */
  public function updateFieldValue(array $aEntity, array $aRequestData, array $aFieldOptions): array{

    // Si la colonne est fournie dans les paramètres
    if ($this->fieldIsSetInData($aEntity, $aFieldOptions, $aRequestData)) {
      $oLogger = $this->oLoggerManager->get('API_ERROR');
      // suppression des vieux fichiers
      if (!$this->deleteFiles($aEntity, $aRequestData, $aFieldOptions)) {
        $oLogger->error('Cannot delete files before update field : ' . $aFieldOptions['name'], [$aEntity, $aRequestData, $_FILES]);
      }
      return $this->afterInsertFieldValue($aEntity, $aRequestData, $aFieldOptions);
    } else {
      return $aEntity;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function updateMassiveFieldValue(array $aEntity, array $aRequestData, array $aFieldOptions): array{
    $oLogger = $this->oLoggerManager->get('API_ERROR');
    // suppression des vieux fichiers
    $aFieldOptionsCopy = [];
    foreach ($aFieldOptions as $sKey => $mValue) {
      $aFieldOptionsCopy[$sKey] = $mValue;
    }
    // on force multiple à false pour écraser l'existant
    $aFieldOptionsCopy['multiple'] = false;
    if (!$this->deleteFiles($aEntity, $aRequestData, $aFieldOptionsCopy)) {
      $oLogger->error('Cannot delete files before update field : ' . $aFieldOptions['name'], [$aEntity, $aRequestData, $_FILES]);
    }
    return $this->afterInsertFieldValue($aEntity, $aRequestData, $aFieldOptions);
  }

  /**
   * {@inheritdoc}
   */
  public function deleteFieldValue(array $aEntity, array $aRequestData, array $aFieldOptions): array{
    $oLogger = $this->oLoggerManager->get('API_ERROR');
    // on efface le dossier au niveau de l'id
    if ($this->deleteEntityDirectory($aEntity, $aFieldOptions)) {
      // on efface la valeur du champ
      $aEntity[$aFieldOptions['name']] = null;
    } else {
      $oLogger->error('Cannot delete directory before remove field : ' . $aFieldOptions['name'], [$aEntity, $aFieldOptions]);
    }

    return $aEntity;
  }

  /**
   * @param array $aEntity Retour de la base de données
   * @param array $aFieldOptions configuration du champ
   * @return string|null chemin vers le fichier sans bind ou null si la fonction rnecontre une erreur
   */
  private function pathReplacer(array $aEntity, array $aFieldOptions, string $sFileName = ''): ?string{
    $aRequestOptions = $this->oRouteConfig->getRequestParameters();

    $sFileName = (empty($sFileName)) ? $aEntity[$aFieldOptions['name']] : $sFileName;

    if (isset($aEntity[$aRequestOptions['options']['field_id']])) {
      // remplacement des properties
      $sPath = $this->oProperties->replacePropertiesInString($aFieldOptions['path']);
      // remplacement de l'id
      $sPath = str_replace(self::ID_PATH_KEY, $aEntity[$aRequestOptions['options']['field_id']], $sPath);
      // remplacement des champs rensigné
      $sPath = str_replace(self::VALUE_PATH_KEY, $sFileName, $sPath);
      // remplacement du nom du champs
      $sPath = str_replace(self::COLUMN_PATH_KEY, $aFieldOptions['name'], $sPath);

      return $sPath;
    }
    return null;
  }

  /**
   * Supprime le dossier d'une entité à partir de la clé @id si elle est défini, le fichier sinon
   * @param array $aEntity Retour de la base de données
   * @param array $aFieldOptions configuration du champ
   * @return bool true si le dossier a pu être supprimé false sinon
   */
  private function deleteEntityDirectory(array $aEntity, array $aFieldOptions): bool {
    // On retouche le chemin fichier pour isoler au niveau de l'id
    $aPath = explode('/', $aFieldOptions['path']);
    $aFinalPath = [];
    $i = 0;

    while ($i < count($aPath)) {
      $aFinalPath[] = $aPath[$i];
      if ($aPath[$i] === self::ID_PATH_KEY) {
        break;
      }
      $i++;
    }
    // insertion du chemin modifié
    $aFieldOptions['path'] = implode('/', $aFinalPath);
    // remplacement du chemin
    $sDirPath = $this->pathReplacer($aEntity, $aFieldOptions);
    // si le fichier ou le dossier existe on supprime
    $oFileInterface = $this->oFileManager->getAccessorForPath($sDirPath);
    if ($oFileInterface->file_exists($sDirPath)) {
      if (!$oFileInterface->clearDir($sDirPath)) {
        return false;
      }
    }

    return true;
  }
  /**
   * Supprime les fichiers à supprimer lors de la mise à jour du champ
   * @param array $aEntity Retour de la base de données
   * @param array $aRequestData Données dans la requête HTTP
   * @param array $aFieldOptions configuration du champ
   * @return bool true si les fichiers ont pu être supprimés false sinon
   */
  private function deleteFiles(array $aEntity, array $aRequestData, array $aFieldOptions): bool {
    $bIsFileMultiple = (isset($aFieldOptions['multiple']) ? $aFieldOptions['multiple'] : false);
    $bUseThumbnail = (isset($aFieldOptions['thumbnail']) ? $aFieldOptions['thumbnail'] : false);
    $bIsImage = (isset($aFieldOptions['image']) ? $aFieldOptions['image'] : false);

    //récupération de la liste des fichiers à supprimer
    $aFilesToDelete = $this->getDeletedFiles($aRequestData, $aFieldOptions['name']);

    $sPath = $this->pathReplacer($aEntity, $aFieldOptions, self::FORCE_VALUE);
    $oFileInterface = $this->oFileManager->getAccessorForPath($sPath);
    // on récupère le dossier parent
    $sDirPath = dirname($sPath);
    // gestion des champs multiple
    if ($bIsFileMultiple) {
      // la liste des champs à effacer est déjà dans $aFilesToDelete
      if ($bIsImage && $bUseThumbnail) {
        // ajout de la suppression des thumbnails pour les fichiers
        $aThumbnailsToDelete = [];
        foreach ($aFilesToDelete as $sFileName) {
          $aThumbnailsToDelete[] = LocalAccessor::thumbnailPathGenerator($sFileName);
        }
        $aFilesToDelete = array_merge($aFilesToDelete, $aThumbnailsToDelete);
      }
    } else {
      // on considère qu'il faut nettoyer tout le dossier si le fichier existe et qu'un nouveau fichier a été uploadé
      if ($oFileInterface->file_exists($sDirPath) && isset($_FILES[$aFieldOptions['name']])) {
        if ($oFileInterface->is_dir($sDirPath)) {
          $aFilesToDelete = array_diff($oFileInterface->scandir($sDirPath), ['.', '..']);
        } else {
          // Cas merdique où routes.json a été mal configuré et que $sDirPath est un fichier
          $aFilesToDelete[] = $sDirPath;
        }
      }
    }

    foreach ($aFilesToDelete as $sFileName) {
      if ($sFileName === $sDirPath) {
        $sFilePath = $sDirPath;
      } else {
        $sFilePath = $sDirPath . '/' . $sFileName;
      }
      if ($oFileInterface->file_exists($sFilePath)) {
        if (!$oFileInterface->unlink($sFilePath)) {
          return false;
        }
      }
    }

    return true;
  }
  /**
   * upload un fichier sur le serveur
   * @param array $aEntity Retour de la base de données
   * @param array $aFieldOptions configuration du champ
   * @param array $aFileStruct Structure contenant les infos du fichier à uploader
   * @param array $aRequestData Données dans la requête HTTP
   * @return string|null nom du fichier après upload, null en cas d'erreur
   */
  private function uploadFile(array $aEntity, array $aFieldOptions, array $aFileStruct, array $aRequestData): ?string{
    $bUseThumbnail = (isset($aFieldOptions['thumbnail']) ? $aFieldOptions['thumbnail'] : false);
    $bIsImage = (isset($aFieldOptions['image']) ? $aFieldOptions['image'] : false);

    $sPath = $this->pathReplacer($aEntity, $aFieldOptions, $aFileStruct['name']);
    $oFileInterface = $this->oFileManager->getAccessorForPath($sPath);
    $iLimitation = -1;

    if (array_key_exists('file_size_limitation', $aFieldOptions)) {
      $iLimitation = (is_string($aFieldOptions['file_size_limitation'])) ? $this->oProperties->replacePropertiesInString($aFieldOptions['file_size_limitation']) : $aFieldOptions['file_size_limitation'];
    }

    // extension non autorisé
    if (!$oFileInterface->checkExtension($sPath)) {
      return null;
    }
    // taille max autorisé
    if (!$this->oLocalFileInterface->checkFileSize($aFileStruct['tmp_name'], $iLimitation)) {
      return null;
    }

    if (!$oFileInterface->copyLocalToDistant($aFileStruct['tmp_name'], $sPath)) {
      return null;
    }

    // resample
    if ($bIsImage && isset($aFieldOptions['width']) && isset($aFieldOptions['height'])) {
      $sPath = $oFileInterface->pictureResampler($sPath, $aFieldOptions['width'], $aFieldOptions['height']);
    }

    // miniature
    if ($bIsImage && $bUseThumbnail) {
      $aFilesToThumbnails = $this->getThumbnailsFiles($aRequestData, $aFieldOptions['name']);
      if (\in_array($aFileStruct['name'], $aFilesToThumbnails)) {
        $oFileInterface->thumbnailResampler($sPath);
      }
    }

    return $sPath;
  }
  /**
   * Récupère la liste des fichier inchangés
   * @param array $aRequestData Données dans la requête HTTP
   * @param string $sFieldName nom du champs en cours de traitement
   * @return array liste des nom de fichiers pour cette clé
   */
  private static function getUnchangedFiles(array $aRequestData, string $sFieldName): array{
    return self::getFilesListForKey($aRequestData, $sFieldName, self::VITIS_MULTIFILE_UNCHANGELIST_KEY);
  }
  /**
   * Récupère la liste des fichier à supprimer
   * @param array $aRequestData Données dans la requête HTTP
   * @param string $sFieldName nom du champs en cours de traitement
   * @return array liste des nom de fichiers pour cette clé
   */
  private static function getDeletedFiles(array $aRequestData, string $sFieldName): array{
    return self::getFilesListForKey($aRequestData, $sFieldName, self::VITIS_MULTIFILE_RMLIST_KEY);
  }
  /**
   * Récupère la liste des fichier pour lesquels il faut une miniature
   * @param array $aRequestData Données dans la requête HTTP
   * @param string $sFieldName nom du champs en cours de traitement
   * @return array liste des nom de fichiers pour cette clé
   */
  private static function getThumbnailsFiles(array $aRequestData, string $sFieldName): array{
    return self::getFilesListForKey($aRequestData, $sFieldName, self::VITIS_MULTIFILE_THUMBNAILLIST_KEY);
  }
  /**
   * Récupère la liste des fichier d'après une clé définie
   * @param array $aRequestData Données dans la requête HTTP
   * @param string $sFieldName nom du champs en cours de traitement
   * @param string $sKey clé à rechercher dans $arequestData
   * @return array liste des nom de fichiers pour cette clé
   */
  private static function getFilesListForKey(array $aRequestData, string $sFieldName, string $sKey): array{
    $aFiles = [];
    if (!empty($aRequestData[$sKey])) {
      $aUnchangedFiles = json_decode($aRequestData[$sKey], true);
      if (isset($aUnchangedFiles[$sFieldName])) {
        $aFiles = $aUnchangedFiles[$sFieldName];
      }
    }
    return $aFiles;
  }
  /**
   * {@inheritdoc}
   */
  public function fieldIsSet(array $aEntity, array $aFieldOptions): bool {

    // Valeur dans l'entity
    if (isset($aEntity[$aFieldOptions['name']])) {
      return true;
    }

    // valeur dans vitis_unchanged_files, vitis_deleted_files, vitis_thumbnail_image_files
    foreach (['vitis_unchanged_files', 'vitis_deleted_files', 'vitis_thumbnail_image_files'] as $key) {
      if (isset($aEntity[$key])) {
        $aFiles = is_string($aEntity[$key]) ? json_decode($aEntity[$key], true) : $aEntity[$key];
        if (isset($aFiles[$aFieldOptions['name']])) {
          return true;
        }
      }
    }

    // Valeur dans $_FILES
    if (!empty($_FILES) && isset($_FILES[$aFieldOptions['name']])) {
      return true;
    }

    return false;
  }

  /**
   * Vérifie qu'il y ait mention du fichier dans les paramètres envoyés à la requête ou dans $_FILES
   */
  public function fieldIsSetInData(array $aEntity, array $aFieldOptions, array $aRequestData): bool {
    // Valeur dans la data envoyée
    if (isset($aRequestData[$aFieldOptions['name']])) {
      return true;
    }

    // valeur dans vitis_unchanged_files, vitis_deleted_files, vitis_thumbnail_image_files
    foreach (['vitis_unchanged_files', 'vitis_deleted_files', 'vitis_thumbnail_image_files'] as $key) {
      if (isset($aRequestData[$key])) {
        $aFiles = is_string($aRequestData[$key]) ? json_decode($aRequestData[$key], true) : $aRequestData[$key];
        if (isset($aFiles[$aFieldOptions['name']])) {
          return true;
        }
      }
    }

    // Valeur dans $_FILES
    if (!empty($_FILES) && isset($_FILES[$aFieldOptions['name']])) {
      return true;
    }

    return false;
  }
}
