<?php

namespace App\Module\Vitis\WebService\Service\DataAccessor\Fields;

use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\DataAccessor\Fields\AbstractCustomField;
use App\Module\Vitis\WebService\Service\DataAccessor\Fields\FieldInterface;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use Doctrine\ORM\EntityManager;

/**
 * Classe abstraite pour gérer un champ de type geom.
 * ne fait que des return sans modification.
 *
 *  Configuration du champ
 *    "type"      (string) : vaut toujours date pour ce champ
 *    "name"      (string) : Nom du champ
 *    "mode"      (string) : valeurs possible : unchanged, nullify (default : nullify)
 *  exemple :
 *    {
 *      "type": "geom",
 *      "name": "my_field"
 *      "srid": "2154"
 *      "geometry_type": "POLYGON" // Ne sert à rien pour l'instant
 *      "coord_dimension": "2" // Ne sert à rien pour l'instant
 *    }
 *
 * @author Armand Bahi   <armand.bahi@veremes.com>
 */

class GeomField extends AbstractCustomField implements FieldInterface {
  /**
   * @var string
   * Valeur du champs type dans route.json
   */
  public const TYPE = 'geom';

  /**
   * @var array
   */
  private $oRouteParameters;

  /**
   * @var EntityManager
   */
  private $oEntityManager;

  /**
   * {@inheritdoc}
   */
  public function __construct(Properties $oPropertiesSrvc, Route $oRoute, LoggerManager $oLoggerManager, EntityManager $oEntityManager) {
    $this->oRouteParameters = $oRoute->getRequestParameters();
    $this->oEntityManager = $oEntityManager;
    parent::__construct($oPropertiesSrvc, $oRoute, $oLoggerManager);
  }

  /**
   * Retourne l'expression utilisée en tant que colonne dans le requête GET
   * {@inheritdoc}
   */
  public function getFieldColumn(string $sSchema, string $sField): string {
    $aFieldOptions = $this->getFieldOptions($sField);
    $sColumn = "\"$sSchema\".\"$sField\"";
    if (!empty($aFieldOptions['result_srid'])) {
      $sColumn = "ST_Transform($sColumn, {$aFieldOptions['result_srid']})";
    }
    return "ST_AsEWKT($sColumn) as $sField";
  }

  /**
   * Retourne la fonction utilisée en tant que colonne dans le requête PUT ou insert
   * {@inheritdoc}
   */
  public function getFieldColumnFunction(string $sField): ?string {
    $aFieldOptions = $this->getFieldOptions($sField);
    $sSRID = isset($aFieldOptions['srid']) ? $aFieldOptions['srid'] : $this->getColumnSRID($this->oRouteParameters['options']['schema'], $this->oRouteParameters['options']['table'], $sField);
    $sReturn = "ST_Transform(ST_GeomFromEWKT(:$sField), $sSRID)";
    return $sReturn;
  }

  /**
   * Retourne un EWKT
   * {@inheritdoc}
   */
  public function getFieldValue(array $aEntity, array $aRequestData, array $aFieldOptions): array{
    return $aEntity;
  }

  /**
   * Récupère un timestamp pour écrire une date
   * {@inheritdoc}
   */
  public function insertFieldValue(array $aEntity, array $aRequestData, array $aFieldOptions): array{
    return $aEntity;
  }

  /**
   * {@inheritdoc}
   */
  public function afterInsertFieldValue(array $aEntity, array $aRequestData, array $aFieldOptions): array{
    return $this->insertFieldValue($aEntity, $aRequestData, $aFieldOptions);
  }

  /**
   * {@inheritdoc}
   */
  public function updateFieldValue(array $aEntity, array $aRequestData, array $aFieldOptions): array{
    return $this->insertFieldValue($aEntity, $aRequestData, $aFieldOptions);
  }

  /**
   * {@inheritdoc}
   */
  public function updateMassiveFieldValue(array $aEntity, array $aRequestData, array $aFieldOptions): array{
    return $this->insertFieldValue($aEntity, $aRequestData, $aFieldOptions);
  }

  /**
   * Retrouve LE SRID d'une colonne
   */
  public function getColumnSRID(string $sSchema, string $sTable, string $sColumn): ?int{

    // inisitalisation du logger
    $oLogger = $this->oLoggerManager->get('SQL_ERROR');
    //initialisation de la connexion à la base
    $oConnexion = $this->oEntityManager->getConnection();

    if (!empty($sSchema) && !empty($sTable)) {

      try {
        $statement = $oConnexion->prepare('SELECT ST_SRID("' . addslashes($sColumn) . '") as column_srid FROM "' . addslashes($sSchema) . '"."' . addslashes($sTable) . '" WHERE "' . addslashes($sColumn) . '" is not null AND ST_SRID("' . addslashes($sColumn) . '") != 0 LIMIT 1');
        $statement->execute();
        $aResult = $statement->fetch();

        if (is_int($aResult['column_srid']) && $aResult['column_srid'] !== 0) {
          return $aResult['column_srid'];
        } else {
          $oLogger->error('Cannot get SRID from table for field : ' . $sSchema . '.' . $sTable . '.' . $sColumn, [$sSchema, $sTable, $sColumn]);
          return null;
        }
      } catch (\Exception $e) {
        $oLogger->error('Cannot get SRID from table for field : ' . $sSchema . '.' . $sTable . '.' . $sColumn . ' (' . $e->getMessage() . ')', [$sSchema, $sTable, $sColumn, $e]);
        return null;
      }
    }
    $oLogger->error('Cannot get SRID with empty schema or table', [$sSchema, $sTable, $sColumn]);
    return null;
  }

  /**
   * Retourne les options du champ
   */
  private function getFieldOptions(string $sField): ?array{
    if (!empty($this->oRouteParameters['options']['fields'])) {
      foreach ($this->oRouteParameters['options']['fields'] as $key => $value) {
        if ($value['name'] === $sField) {
          return $value;
        }
      }
    }
    return null;
  }
}