<?php

namespace App\Module\Vitis\WebService\Service\Mail;

use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use PHPMailer\PHPMailer\PHPMailer;
use PHPMailer\PHPMailer\SMTP;

//use PHPMailer\PHPMailer\Exception;

/**
 * Class permettant l'envoi de mail via un serveur SMTP
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */

class SMTPConnexion {
  /**
   * @var int[] SMTP::Const
   */
  private const DEBUG_LEVEL = [
    SMTP::DEBUG_OFF, //0
    SMTP::DEBUG_CLIENT, //1
    SMTP::DEBUG_SERVER, //2
    SMTP::DEBUG_CONNECTION, //3
    SMTP::DEBUG_LOWLEVEL, //4
  ];
  /**
   * @var array protocol
   */
  private const PROTOCOL_MAP = [
    'tls' => PHPMailer::ENCRYPTION_STARTTLS,
    'ssl' => PHPMailer::ENCRYPTION_SMTPS,
  ];
  /**
   * logs désactivés par défaut
   * @var int
   */
  private const DEFAULT_DEBUG_LEVEL = SMTP::DEBUG_OFF;
  /**
   * UTF-8
   * @var string
   */
  private const DEFAULT_CHARSET = PHPMailer::CHARSET_UTF8;
  /**
   * @var string
   */
  private const DEFAULT_AUTH_TYPE = 'LOGIN';
  /**
   * @array
   */
  private const DEFAULT_UNSECURE_CONFIGURATION = array(
    'ssl' => array(
      'verify_peer' => false,
      'verify_peer_name' => false,
      'allow_self_signed' => true,
    ),
  );
  /**
   * @var Properties
   */
  private $oProperties;
  /**
   * @var \Monolog\Logger
   */
  private $oErrorLogger;
  /**
   * @var \Monolog\Logger
   */
  private $oDebugLogger;
  /**
   * @var PHPMailer
   */
  private $oMailer;
  /**
   * @var array
   */
  private $aConfiguration = [];
  
  /**
   * @param Properties $properties service d'accés aux properties
   * @param LoggerManager $oLogger Service d'accès aux logs
   */
  public function __construct(Properties $properties, LoggerManager $oLogger) {
    $this->oErrorLogger = $oLogger->get('API_ERROR');
    $this->oDebugLogger = $oLogger->get('NOTIFY_DEBUG');
    $this->oProperties = $properties;
    // charge la configuration stocké en properties
    $this->loadConfiguration();
  }

  /**
   * Configure la classe PHP_MAILER
   * @return bool true si la cofniguration se passe bien false sinon
   */
  private function configureSMTP(): bool {
    try {
      // true permet d'activer les exceptions
      $this->oMailer = new PHPMailer(true);
      $this->oMailer->isSMTP(); // Send using SMTP
      $this->oMailer->SMTPDebug = $this->aConfiguration['smtp_debug_level'];
      // on renvoi les log dans le logger d'error
      $this->oMailer->Debugoutput = $this->oDebugLogger;
      $this->oMailer->Host = $this->aConfiguration['smtp_host'];
      $this->oMailer->Port = $this->aConfiguration['smtp_port'];
      // valeur possible '', 'ssl', 'tls'
      $this->oMailer->SMTPSecure = $this->aConfiguration['smtp_protocol'];

      $this->oMailer->SMTPAuth = $this->aConfiguration['smtp_authentification'];
      // if authentication add parameter to connect SMTP
      if ($this->aConfiguration['smtp_authentification']) {
        $this->oMailer->AuthType = self::DEFAULT_AUTH_TYPE;
        $this->oMailer->Username = $this->aConfiguration['smtp_login'];
        $this->oMailer->Password = $this->aConfiguration['smtp_password'];
      }
      // chargement de la configuration pour ne pas controler le certificat
      $this->oMailer->SMTPOptions = self::DEFAULT_UNSECURE_CONFIGURATION;
      
      // PHP Mailer configuré
    } catch (\Exception $e) {
      $this->oErrorLogger->error('Mailer cannot be initialized : ' . $e->getMessage());
      return false;
    }
    $this->oErrorLogger->debug('Mailer initialized');
    return true;
  }

  /**
   * Envoi un mail
   * @return PHPMailer Prepare la conenction SMTP et retoune le mailer
   */
  public function connect(): ?PHPMailer {
    if (!$this->testConnection()) {
      return null;
    }
    // on force le charset voir (self::DEFAULT_CHARSET)
    $this->oMailer->CharSet = self::DEFAULT_CHARSET;
    // définition de l'origine
    $this->oMailer->setFrom($this->aConfiguration['mail_sender'], $this->aConfiguration['nickname_sender']);
    $this->oMailer->addReplyTo($this->aConfiguration['mail_sender'], $this->aConfiguration['nickname_sender']);

    return $this->oMailer;
  }
  /**
   * test la connection au SMTP
   *
   * @return bool true si le SMTP autorise la connection false sinon
   */
  public function testConnection(): bool {
    try {
      if (!$this->configureSMTP()) {
        return false;
      }
      return $this->oMailer->smtpConnect();
    } catch (\Exception $e) {
      $this->oErrorLogger->error('SMTP Connection failed : ' . $e->getMessage());
      return false;
    }
  }
  /**
   * charge la configuration depuis les properties ou depuis un tableau externe
   * @param array $aData données pour se connecter au SMTP (optionnel)
   *
   * @return array tableau contenant la configuration
   */
  private function loadConfiguration(array $aData = []): array{
    // chargement de la configuration simple
    $this->aConfiguration = [
      'smtp_host' => $aData['smtp_host'] ?? $this->oProperties->get('smtp_host'),
      'smtp_port' => $aData['smtp_port'] ?? $this->oProperties->get('smtp_port'),
      'smtp_login' => $aData['smtp_login'] ?? $this->oProperties->get('smtp_login'),
      'smtp_password' => $aData['smtp_password'] ?? $this->oProperties->get('smtp_password'),
      'nickname_sender' => $aData['nickname_sender'] ?? $this->oProperties->get('nickname_sender'),
      'mail_sender' => $aData['mail_sender'] ?? $this->oProperties->get('mail_sender'),
    ];

    // passage en booléen pur
    $bAuth = ($aData['smtp_authentification'] ?? $this->oProperties->get('smtp_authentification'));
    $this->aConfiguration['smtp_authentification'] = ($bAuth === true);

    $bCheckSSL = ($aData['check_ssl'] ?? $this->oProperties->get('proxy_check_ssl'));
    $this->aConfiguration['check_ssl'] = ($bCheckSSL === true);
    // définition du protocol
    $sProtocol = ($aData['smtp_protocol'] ?? $this->oProperties->get('smtp_protocol'));
    switch ($sProtocol) {
      case 'tls':
        $this->aConfiguration['smtp_protocol'] = PHPMailer::ENCRYPTION_STARTTLS;
        break;
      case 'ssl':
        $this->aConfiguration['smtp_protocol'] = PHPMailer::ENCRYPTION_SMTPS;
        break;
      default:
        $this->aConfiguration['smtp_protocol'] = '';
        break;
    }
    // gestion du niveau de debugage
    $iDebugLevel = $aData['smtp_debug_level'] ?? $this->oProperties->get('smtp_debug_level');
    $this->aConfiguration['smtp_debug_level'] = (in_array($iDebugLevel, self::DEBUG_LEVEL)) ? $iDebugLevel : self::DEFAULT_DEBUG_LEVEL;

    return $this->aConfiguration;
  }
  /**
   * surcharge la configuration SMTP avec des données externe
   * @param array $aData données pour se connecter au SMTP
   *
   * @return self
   */
  public function setConfiguration(array $aData): self {
    $this->loadConfiguration($aData);
    return $this;
  }

}