<?php
namespace App\Module\Vitis\WebService\Service\Normalizer;

use App\Module\Vitis\WebService\EventListener\VitisEventArgs;
use App\Module\Vitis\WebService\EventListener\VitisEvents;
use Doctrine\Common\EventManager;
use Symfony\Component\Serializer\Exception\InvalidArgumentException;
use Symfony\Component\Serializer\Exception\LogicException;
use Symfony\Component\Serializer\Normalizer\AbstractNormalizer;
use Symfony\Component\Serializer\Normalizer\DenormalizerInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;

/**
 * A normalizer that uses an objects own JsonSerializable implementation.
 *
 * @author Alexandre Mironenkov <alexandre.mironenkov@veremes.com>
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */
class VitisNormalizer extends AbstractNormalizer {

  /**
   * @var string
   */
  public const DISTINCT = 'distinct';

  /**
   * @var EventManager
   */
  private $oEventManager;

  /**
   * {@inheritdoc}
   */
  public function normalize($object, $format = null, array $context = []) {
    if ($this->isCircularReference($object, $context)) {
      return $this->handleCircularReference($object);
    }

    if (!$object instanceof \JsonSerializable) {
      throw new InvalidArgumentException(sprintf('The object must implement "%s".', \JsonSerializable::class));
    }

    if (!$this->serializer instanceof NormalizerInterface) {
      throw new LogicException('Cannot normalize object because injected serializer is not a normalizer');
    }
    if (!$this->serializer instanceof DenormalizerInterface) {
      throw new LogicException('Cannot normalize object because injected serializer is not a normalizer');
    }

    return $this->serializer->normalize($this->attributesFiltration($object->jsonSerialize(), $context), $format, $context);
  }

  /**
   * permet le filtrage des attributs
   * @param array $aData Données provenant de JsonSerialize
   * @param array $context context de normalisation
   */
  public static function attributesFiltration(array $aData, array $context = []): array{
    if (empty($context)) {
      return $aData;
    }

    if (!isset($context[AbstractNormalizer::ATTRIBUTES]) && empty($context[AbstractNormalizer::ATTRIBUTES])) {
      return $aData;
    }

    // Filtrage des champs
    $aFiltered = [];
    foreach ($aData as $sKey => $attribute) {
      if (in_array($sKey, $context[AbstractNormalizer::ATTRIBUTES])) {
        $aFiltered[$sKey] = $attribute;
      }
    }

    return $aFiltered;
  }

  /**
   * {@inheritdoc}
   */
  public function supportsNormalization($data, $format = null) {
    return $data instanceof \JsonSerializable;
  }

  /**
   * {@inheritdoc}
   */
  public function supportsDenormalization($data, $type, $format = null) {
    // si l'entité existe
    if (class_exists($type)) {
      $reflectionClass = new \ReflectionClass($type);
      $aInterfaceName = $reflectionClass->getInterfaceNames();
      // si l'entité implémente l'interface de vitis
      if (in_array(VitisDeserializerInterface::class, $aInterfaceName)) {
        return true;
      }
    }
    return false;
  }

  /**
   * {@inheritdoc}
   */
  public function denormalize($data, $class, $format = null, array $context = []) {

    $allowedAttributes = $this->getAllowedAttributes($class, $context, true);
    // récupération des information de la classe
    $reflectionClass = new \ReflectionClass($class);
    // création de l'objet et le rempli si besoin
    $object = parent::instantiateObject($data, $class, $context, $reflectionClass, $allowedAttributes, $format);
    // emission de l'event prédénomalization
    if (isset($this->oEventManager)) {
      $this->oEventManager->dispatchEvent(VitisEvents::preDenormalization, new VitisEventArgs($object));
    }
    //Utilisation de l'interface VitisDeserializerInterface pour récupérer la map de déserilization
    $aMapClass = $object->getDeserializeMap();
    // remplissage de l'objet
    $aKeys = array_keys($data);
    foreach ($aMapClass as $sDataKey => $sSetterName) {
      // Si la clé est dans les données et que le setter existe
      if ($reflectionClass->hasMethod($sSetterName) && in_array($sDataKey, $aKeys)) {
        $oReflectionMethod = $reflectionClass->getMethod($sSetterName);
        // Si le setter est public on l'utilise
        if ($oReflectionMethod->isPublic()) {
          // analyse du paramètre pour savoir ce qu'on doit passer
          $aParameters = $oReflectionMethod->getParameters();
          if (count($aParameters) === 1) {
            $oReflectionParameter = $aParameters[0];
            // si un type est défini on l'analyse pour comparer avec la donnée et ne pas envoyer n'importe quoi
            if ($oReflectionParameter->hasType()) {
              $oReflectionType = $oReflectionParameter->getType();
              if ($data[$sDataKey] !== null) {
                switch ($oReflectionType->__toString()) {
                  case 'int':
                  case 'integer':
                    if ($data[$sDataKey] != '') {
                      $data[$sDataKey] = (is_int($data[$sDataKey])) ? $data[$sDataKey] : intval($data[$sDataKey]);
                    } else {
                      $data[$sDataKey] = null;
                    }
                    break;
                  case 'double':
                  case 'float':
                    if ($data[$sDataKey] != '') {
                      $data[$sDataKey] = (is_float($data[$sDataKey])) ? $data[$sDataKey] : floatval($data[$sDataKey]);
                    } else {
                      $data[$sDataKey] = null;
                    }
                    break;
                  case 'bool':
                  case 'boolean':
                    $data[$sDataKey] = (is_bool($data[$sDataKey])) ? $data[$sDataKey] : boolval($data[$sDataKey]);
                    break;
                  default:
                    $data[$sDataKey] = ($oReflectionType->__toString() == gettype($data[$sDataKey])) ? $data[$sDataKey] : null;
                    break;
                }
              }
              if ($data[$sDataKey] === null && !$oReflectionType->allowsNull()) {
                continue;
              }
            }
            $object->$sSetterName($data[$sDataKey]);
          } else {
            throw new LogicException('Cannot set a field with more than one parameter');
          }
        }
      }
    }

    return $object;
  }

  /**
   * {@inheritdoc}
   */
  public function hasCacheableSupportsMethod(): bool {
    return __CLASS__ === \get_class($this);
  }
  /**
   * @param EventManager $oEventManager EventManager doctrine
   * @return void
   */
  public function setEventManager(EventManager $oEventManager): void {
    $this->oEventManager = $oEventManager;
  }
}
