<?php

namespace App\Module\Vitis\WebService\Service\PrintService;

class PdfSigner extends AbstractPrinter {
  /**
   * prepare et signe un PDF
   * @param string $sInputFile le PDF à signé
   * @param string $sOutputFile emplacement du PDF final
   * @param string $sCertFile chemin vers le certificat p12
   * @param int $iCertLength longueur de la clé du certificat
   * @param string $sPassPhrase mot d epasse de la clé si requis
   * @param string $sReason raison de la signature
   * @param string $sName nom du signataire
   * @param string $sLocation lieu de signature
   * @param string $sContact mail du signataire
   * @throws \Exception si le pdf ne peut pas être signé
   * @return bool|null true en cas de succés null sinon
   */
  public function sign(string $sInputFile, string $sOutputFile, string $sCertFile, int $iCertLength, string $sPassPhrase = '', string $sReason = '', string $sName = '', string $sLocation = '', string $sContact = ''): ?bool {
    try {
      // génération du tableau des arguments
      $aArguments = $this->prepareArguments($sInputFile, $sOutputFile, $sCertFile, $iCertLength, $sPassPhrase, $sReason, $sName, $sLocation, $sContact);
      //exécution de la commande
      $sReturn = $this->callPrinter($aArguments);
      $this->oDebugLogger->debug($sReturn);
    } catch (\Exception $e) {
      $this->oErrorLogger->error('Impossible to sign PDF : ' . $e->getMessage());
      return null;
    }
    return true;
  }

  /**
   * Prépare les arguments pour la CLI
   * @param string $sInputFile le PDF à signé
   * @param string $sOutputFile emplacement du PDF final
   * @param string $sCertFile chemin vers le certificat p12
   * @param int $iCertLength longueur de la clé du certificat
   * @param string $sPassPhrase mot d epasse de la clé si requis
   * @param string $sReason raison de la signature
   * @param string $sName nom du signataire
   * @param string $sLocation lieu de signature
   * @param string $sContact mail du signataire
   * @return array tableau d'argument à utiliser pour appeler phantomjs
   */
  private function prepareArguments(string $sInputFile, string $sOutputFile, string $sCertFile, int $iCertLength, string $sPassPhrase = '', string $sReason = '', string $sName = '', string $sLocation = '', string $sContact = ''): array{
    // génération d'un dossier temporaire
    $aArguments = ['signPdf'];
    // chemin vers les ficheir à concaténer
    $aArguments[] = '--inputFile="' . $sInputFile . '"';
    // chemin vers la sortie PDF
    $aArguments[] = '--outputFile="' . $sOutputFile . '"';
    // chemin vers le certificat P12
    $aArguments[] = '--certFile="' . $sCertFile . '"';
    // taille du certificat P12
    $aArguments[] = '--certLength="' . $iCertLength . '"';
    // pasphrase vers le certificat P12
    if (!empty($sPassPhrase)) {
      $aArguments[] = '--certPass="' . $sPassPhrase . '"';
    }
    // autres informations pour la signature
    if (!empty($sReason)) {
      $aArguments[] = '--reason="' . $sReason . '"';
    }
    if (!empty($sContact)) {
      $aArguments[] = '--contact="' . $sContact . '"';
    }
    if (!empty($sName)) {
      $aArguments[] = '--name="' . $sName . '"';
    }
    if (!empty($sLocation)) {
      $aArguments[] = '--location="' . $sLocation . '"';
    }

    return $aArguments;
  }

}
