# Fichier de configuration des routes de l'API

Le fichier routes.json de chaque Module permet de définir les routes du module.

Il est découpé en plusieurs parties :
- le tableau associatif des **routes**
- le tableau associatif des **configurations**.
- Le tableau **deny**

## Routes

Cette partie permet de définir le point de chute de chaque requête HTTP au niveau du VAS.

ex :
```json
"vitis_mail_template_update_v2": {
  "route": "/v2/vitis/emailtemplates/{id}",
  "controller": "App\\Module\\Vitis\\WebService\\Controller\\GenericController",
  "function": "updateData",
  "methods": ["PUT"],
  "type": "orm",
  "role_minimum": "vitis_admin",
  "options": "orm_vitis_mail_template"
}
```

Un route est identifiée par un clé unique (tout module confondu) et contient les champs :
- **route** : Route relative à l'alias rest
- **controller** : Controller (classe php) qui servira de point d'entrée pour traiter les requêtes sur cette route
- **function** : Fonction du controller qui sera appelée comme point d'entrée
- **methods** : Liste de méthodes HTTP acceptées par la fonction du controller (habituellement une seule)
- **type** : Défini si la route va exploité la base, une entité intermédiaire (ORM), ou faire une action plus spécifique (query, orm, custom)
- **role_minimum** : privilège de l'utilisateur requis pour utiliser la route
- **check_user** : booléen permettant de contourner les vérifications sur l'utilisateur (ip, droits, ...)
- **check_origin** : Permet de définir un regexp pour controller l'origine des requêtes sur une route (utilisé principalement pour engine)
- **options** : clé qui sert de lien vers une configuration
- **filter** : permet de définir un filtre spécifique pour la route (ou pour un rôle), les filtres ont la même structures que les filtres envoyés par le client
- **options_by_role** : Permet de définir une configuration relativement au rôle de l'utilisateur


Une modification des paramètres route, controller, function, methods nécessiteront un redémarrage du serveur symfony pour être pris en compte.


## Configurations

Les configuration permettent de piloter des controllers génériques.

La configuration va varier en fonction du type de routes. Pour une route custom vous pouvez créer votre propre type de configuration si le controller tend à être réutilisable.

### Query

La configuration pour un type query se présentera toujours comme l'exemple ci-dessous.

ex:
```json
    "query_vitis_mail_context": {
      "schema": "@prop(schema_vitis)",
      "table": "rt_emailcontext",
      "field_id": "rt_emailcontext_id",
      "fields": [{
        "type": "string",
        "name": "rt_emailcontext_id"
      }, {
        "type": "string",
        "name": "context_label"
      }]
    },
```

A minima, on trouvera une clé unique comme pour les routes et il faudra renseigner :
- **schema** : schéma de la table a interroger
- **table** : nom de la table à interroger
- **field_id** : Colonne à utiliser comme identifiant
- **fields** : Tableau contenant la description des colonnes pour le retour d'API



### ORM

A minima, on trouvera une clé unique comme pour les routes et il faudra renseigner :
- **entity** : Namespace de l'entité à utiliser
- **field_id** : Colonne à utiliser comme identifiant
- **fields** : Tableau contenant la description des colonnes pour le retour d'API

A la différence d'une route de type Query est que chaque field pourra avoir des champs supplémentaires :
- **orm_get**: nom du getter pour récupérer le champ au moment de la sérialisation
- **orm_set**: nom du setter pour la déserialisation
- **orm_col**: nom de la variable dans l'entité si différent du nom que l'on veut au moment de la sérialisation

### Fields

un champ aura à minima deux champs :
- **type**: type de la donnée,
- **name**: nom du champ à retourner au moment de la sérialisation

ci-dessous plus de détail sur chaque type, tout type non détaillé ci-dessous équivaut au type string.
#### String

Type le plus simple aucun traitement spécifique n'est fait.

#### Number

Type très simple, on fait juste quelques contrôle supplémentaire pour les valeurs nulles.

#### Boolean

Gestion de plusieurs type de valeurs pour true et false

#### Date

Force la date en timestamp sur UTC pour normaliser les dates pour le formatage coté client.

#### Password

il contient un champ supplémentaire **mode** qui peut prendre plusieurs valeur :
- unchanged : L'API retourne le mot de passe en clair (comme un string)
- nullify : L'API remplace le mot de passe par une valeur nulle
- star : L'API remplace le mot de passe par la chaîne '\*\*\*\*' (un contrôle sera fait au moment de la mise à jour pour voir si la chaîne est différente de '\*\*\*\*')

ex:
```json
  {
    "type": "password",
    "name": "password",
    "mode": "star"
  }
```

#### File

Le champs path est requis, il permet de définir le chemin de sotckage du fichier, il peut référencer des properties. La clé @id référence l'identifiant de l'enregistrement ne base, la clé @value la valeur du champ fichier (habituellement juste le nom du fichier).

Tous les autres champs sont optionnels :
- **multiple** : détermine si le champ peut contenir plusieurs fichiers
- **image** : détermine si le fichier est une image
- **width** : permet de définir la largeur pour resampler/redimensionner l'image
- **height** : permet de définir la hauteur pour resampler/redimensionner l'image
- **thumbnail** : permet de définir s'il faut générer une miniature (taille définie dans les properties)

La gestion du fichier sur le disque/S3 se fera automatiquement.

ex :
```json
{
  "type": "file",
  "name": "my_file",
  "path": "@prop(ws_data_dir)/module/objet/@id/field/@value",
  "multiple": false,
  "image": true,
  "width": 500,
  "height": 500,
  "thumbnail": true
}
```
#### Geom

Traite les géométries pour utiliser les formats de stockage PostGIS tout en reprojettant la donnée en entrée et en sortie

ex :
```json
{
  "type": "geom",
  "name": "my_field",
  "srid": "2154",
  "geometry_type": "POLYGON", // Ne sert à rien pour l'instant
  "coord_dimension": "2", // Ne sert à rien pour l'instant
}
```

#### Array

Permet de gérer un table de jointure intermédiaire par exemple le lien entre groupe et user

ex :
```json
{
  "type": "array",
  "name" : "users",
  "join" : {
    "schema" : "s_vitis",
    "table" : "user_group",
    "field_id" : "group_id",
    "joined_id" : "user_id"
  }
}
```
### Tree

Permet de déployer des routes pour piloter une arborescence type shared_dir.

le champ **dirpath** est obligatoire mais peut être composé  d'une clé @id.

ex:
```json
"query_dirpath_fmeproject": {
  "dir_path": "@prop(workspace_dir)/@id/fme",
  "schema": "@prop(schema_gtf)",
  "table": "v_workspace",
  "field_id": "workspace_id",
  "fields": [{
    "type": "number",
    "name": "workspace_id"
  }, {
    "type": "string",
    "name": "fmw_file"
  }]
},
```

### Properties

ex :
```json
    "vitis_configuration": {
      "module_name": "Vitis"
    },
```
## Deny

Ce tableau permet à un module de bloquer des routes de son propre modules ou d'autres modules.
C'est simplement un tableau de clé qui est utilisé pour filtrer l'aggrégats des configurations.

L'intéret principal est de pouvoir désactiver des features pour certaines applications ou les surcharger en utilisant les même routes.

exemple de configuration du routes.json
```json
  "deny": [
    "vitis_account_signup_confirm_v2",
    "vitis_users_request_confirm_v2"
  ],
  "routes":[
    "extraction_overwrite_users_request_confirm_v2": {
      "route": "/v2/vitis/user_requests/{id}/confirm",
      "controller": "App\\Module\\Extraction\\WebService\\Controller\\AccountController",
      "function": "confirmSignUp",
      "methods": ["PUT"],
      "type": "orm",
      "role_minimum": "vitis_admin",
      "options": "orm_users"
    },
    "extraction_overwrite_account_signup_confirm_v2": {
      "route": "/v2/vitis/accounts/signup/confirm",
      "controller": "App\\Module\\Extraction\\WebService\\Controller\\AccountController",
      "function": "confirmSignUp",
      "check_user": false,
      "methods": ["POST"],
      "type": "orm",
      "options": "orm_users"
    }
  ]
```

Dans l'exemple ci-dessus je désactive les routes permettant de confirmer un compte, pour pouvoir les surcharger et intégrer une jointure spécifique entre l'utilsiateur et un sous-objet de mon module
## Exemples

### Configuration en fonction des droits

Le premier niveau de clé est l'ordre dans lequel il faut vérifier les droits, ici un admin doit voir plus d'éléments qu'un auteur, on peut pas laisser l'ordre au hasard.

```json
    "gtf_orders_list_v2": {
      "route": "/v2/gtf/orders",
      "controller": "App\\Module\\Gtf\\WebService\\Controller\\OrderController",
      "function": "getDatas",
      "methods": ["GET"],
      "type": "query",
      "role_minimum": ["gtf_admin", "gtf_author"],
      "options_by_role": {
        "0": {
          "gtf_admin": "gtf_orders"
        },
        "1": {
          "gtf_author": "gtf_orders_author"
        }
      }
    },
```