import fs from 'fs';
import Path from 'path';
import { Logger } from "vitis-engine-lib";
import { Db } from "vitis-engine-lib";
import { LocalFS } from "vitis-engine-lib";
import { propertiesSrvc } from "vitis-engine-lib";
import { FileConfigSrvc } from "../../fileconfigSrvc";
import { FileConfig, dirNode, sqlNode } from "../../models";

/**
 * Classe Cleaner : permet la gestion des demandes
 */
export default class Cleaner {

  // properties
  private properties_: any;

  // Base de données
  private db_: Db;

  // Service de logs
  private logger_: Logger;

  private config_: Array<dirNode | sqlNode> = [];

  /**
 * Constructeur
 * @param {any} properties
 * @param {any} db
 * @param {any} licence
 */
  constructor(properties: any, db: Db, logger: Logger, oLocalFS: LocalFS) {
    this.properties_ = properties;
    this.db_ = db;
    logger.setLogFilePath(properties.vitis_cleaner_log_path);
    this.logger_ = logger;
    this.loadConfiguration();
  }

  /**
   * Charge la configuration pour lancer le nettoyage
   * @returns 
   */
  private loadConfiguration(): void {
    const oFileConfigSrvc = new FileConfigSrvc(this.logger_);
    // Fichier de config des jobs
    const oConfig = oFileConfigSrvc.getFileConfig('cleaner');

    if (!oConfig || !oConfig.nodes) {
      return;
    }

    this.config_ = this.config_.concat(oConfig.nodes);
  }

  /**
   * Fonction de lancement du cleaner gloable
   * @param {boolean} bDelete 
   */
  private runCleaner(bDelete: boolean) : void {
    let iTotalFreeSpace : number = 0;
    const sMessage: string = (bDelete) ? 'active' : 'inactive';
    this.logger_.log('Lancement du script, la suppression est ' + sMessage + ' (' + this.config_.length + ' noeud(s) à traiter)', null, 'INFO');
    // pour chaque noeud appelle une fonction spécifique
    for (const oNode of this.config_) {
      if ('path' in oNode) {
        iTotalFreeSpace += this.processDirNode(oNode, bDelete);
      }
    }
    this.logger_.log('Libération totale : ' + this.humanizeSize(iTotalFreeSpace), null, 'INFO');
    this.logger_.log('Fin du script', null, 'INFO');
  }

  /**
   * traite un noeud de nettoyage de type Dir
   * @param {dirNode} oNode 
   * @param {boolean} bDelete 
   * @returns 
   */
  private processDirNode(oNode: dirNode, bDelete: boolean) : number {

    const sPath: string = propertiesSrvc.replacePropertiesInString(oNode.path, this.properties_);
    const iExpiration: number = (typeof oNode.expiration == 'string') ? parseInt(propertiesSrvc.replacePropertiesInString(oNode.expiration, this.properties_)) : oNode.expiration;
    // test existence du dossier
    if (!fs.existsSync(sPath)) {
      this.logger_.log('Le chemin du noeud n\'existe pas :' + sPath, null, 'INFO');
      return 0;
    }
    this.logger_.log('Le chemin pincipal du noeud existe :' + sPath, null, 'INFO');
    this.logger_.log('Lancement du traitement du noeud (expiration : ' + iExpiration + ' jour(s))', null, 'INFO');

    // analyse récursive taille fichier/dossier
    const iTotalFreeSpace: number = this.processPath_(sPath, iExpiration, bDelete);

    const sComplement = (bDelete) ? '' : 'estimée'
    this.logger_.log('Libération totale ' + sComplement + ' pour ce noeud: ' + this.humanizeSize(iTotalFreeSpace), null, 'INFO');

    return iTotalFreeSpace;
  }

  /**
   * Fonction récursive d enettoyage avec récupération de l'espace libérée
   * @param {string} sPath chemin principal
   * @param {number} expiration nombre d ejour pour l'expiration
   * @param {boolean} bDelete détermine s'il faut lancer la suppresison ou non
   * @returns 
   */
  private processPath_(sPath: string, expiration: number, bDelete: boolean): number {
    let fileSizeInBytes: number = 0;
    const now: Date = new Date();
    const expiration_ms: number = expiration * 24 * 3600 * 1000;
    //scan du dossier pour chaque élément on poursuit
    fs.readdirSync(sPath).forEach((element, index) => {
      const curPath: string = Path.join(sPath, element);
      if (fs.lstatSync(curPath).isDirectory()) {
        // Récursive
        fileSizeInBytes += this.processPath_(curPath, expiration, bDelete);
        if (fs.readdirSync(curPath).length === 0 && bDelete) {
          // Suppression du dossier vide
          this.logger_.log(curPath + ' est maintenant vide (suppression)', null, 'DEBUG');
          fs.rmdirSync(curPath);
        }
      } else {
        // analyse taille fichier
        const stats = fs.statSync(curPath)
        // si le ficheir est trop vieux
        if ((now.getTime() - stats.mtime.getTime()) > expiration_ms) {
          fileSizeInBytes += stats.size;
          this.logger_.log(curPath + ' (' + this.humanizeSize(stats.size) + ')', null, 'DEBUG');
          if (bDelete) {
            // Suppression du fichier
            fs.unlinkSync(curPath);
          }
        }
      }
    });

    return fileSizeInBytes;
  }

  /**
   * Convertit une taille en octet en taille générique
   * @param {number} size taille en octet
   * @returns 
   */
  private humanizeSize(size: number): string {
    let sResult: string = size + ' octets';
    // si plus grand que 1Ko
    if (size > 1024) {
      // si plus grand que 1 Mo
      if (size > 1024 * 1024) {
        // si plus grand que 1Go
        if (size > 1024 * 1024 * 1024) {
          // Si plus grand que 1 To (on en reparlera)
          if (size > 1024 * 1024 * 1024 * 1024) {
            sResult = '-';
          } else {
            sResult = (Math.ceil(size / (1024 * 1024 * 1024))) + 'Go';
          }
        } else {
          sResult = (Math.ceil(size / (1024 * 1024))) + 'Mo';
        }
      } else {
        sResult = (Math.ceil(size / 1024)) + 'Ko';
      }
    }
    return sResult;
  }

  /**
   * Lance le cleaner en mode suppression
   */
  public cleanFiles() {
    this.runCleaner(true);
  };
  /**
   * Lance le cleaner en mode analyse
   */
  public analyzeFiles() {
    this.runCleaner(false);
  };
}


